@@ -6,28 +6,37 @@ typealias AbstractMatrix{T} AbstractArray{T,2}
 ## Basic functions ##
 
 size{T,n}(t::AbstractArray{T,n}, d) = (d>n ? 1 : size(t)[d])
+eltype(x) = Any
 eltype{T,n}(::AbstractArray{T,n}) = T
 eltype{T,n}(::Type{AbstractArray{T,n}}) = T
 eltype{T<:AbstractArray}(::Type{T}) = eltype(super(T))
+iseltype(x,T) = eltype(x) <: T
+isinteger(x::AbstractArray) = all(isinteger,x)
+isinteger{T<:Integer,n}(x::AbstractArray{T,n}) = true
+isreal(x::AbstractArray) = all(isreal,x)
+isreal{T<:Real,n}(x::AbstractArray{T,n}) = true
 ndims{T,n}(::AbstractArray{T,n}) = n
 ndims{T,n}(::Type{AbstractArray{T,n}}) = n
 ndims{T<:AbstractArray}(::Type{T}) = ndims(super(T))
-length(t::AbstractArray) = prod(size(t))
+length(t::AbstractArray) = prod(size(t))::Int
 endof(a::AbstractArray) = length(a)
 first(a::AbstractArray) = a[1]
-last(a::AbstractArray) = a[end]
+first(a) = next(a,start(a))[1]
+last(a) = a[end]
+
+function stride(a::AbstractArray, i::Integer)
+    if i > ndims(a)
+        return length(a)
+    end
+    s = 1
+    for n=1:(i-1)
+        s *= size(a, n)
+    end
+    return s
+end
 
 strides(a::AbstractArray) = ntuple(ndims(a), i->stride(a,i))::Dims
 
-isinteger{T<:Integer}(::AbstractArray{T}) = true
-isinteger(::AbstractArray) = false
-isreal{T<:Real}(::AbstractArray{T}) = true
-isreal(::AbstractArray) = false
-iscomplex{T<:Complex}(::AbstractArray{T}) = true
-iscomplex(::AbstractArray) = false
-isbool(::AbstractArray{Bool}) = true
-isbool(::AbstractArray) = false
-
 function isassigned(a::AbstractArray, i::Int...)
     # TODO
     try
@@ -48,26 +57,26 @@ function trailingsize(A, n)
 end
 
 ## Bounds checking ##
-function check_bounds(sz::Int, I::Real)
+function checkbounds(sz::Int, I::Real)
     I = to_index(I)
     if I < 1 || I > sz
         throw(BoundsError())
     end
 end
 
-function check_bounds(sz::Int, I::AbstractVector{Bool})
+function checkbounds(sz::Int, I::AbstractVector{Bool})
     if length(I) > sz
         throw(BoundsError())
     end
 end
 
-function check_bounds{T<:Integer}(sz::Int, I::Ranges{T})
-    if min(I) < 1 || max(I) > sz
+function checkbounds{T<:Integer}(sz::Int, I::Ranges{T})
+    if !isempty(I) && (minimum(I) < 1 || maximum(I) > sz)
         throw(BoundsError())
     end
 end
 
-function check_bounds{T <: Real}(sz::Int, I::AbstractVector{T})
+function checkbounds{T <: Real}(sz::Int, I::AbstractArray{T})
     for i in I
         i = to_index(i)
         if i < 1 || i > sz
@@ -76,37 +85,37 @@ function check_bounds{T <: Real}(sz::Int, I::AbstractVector{T})
     end
 end
 
-function check_bounds(A::AbstractArray, I::AbstractArray{Bool})
+function checkbounds(A::AbstractArray, I::AbstractArray{Bool})
     if !isequal(size(A), size(I)) throw(BoundsError()) end
 end
 
-check_bounds(A::AbstractArray, I) = check_bounds(length(A), I)
+checkbounds(A::AbstractArray, I) = checkbounds(length(A), I)
 
-function check_bounds(A::AbstractMatrix, I, J)
-    check_bounds(size(A,1), I)
-    check_bounds(size(A,2), J)
+function checkbounds(A::AbstractMatrix, I, J)
+    checkbounds(size(A,1), I)
+    checkbounds(size(A,2), J)
 end
 
-function check_bounds(A::AbstractArray, I, J)
-    check_bounds(size(A,1), I)
+function checkbounds(A::AbstractArray, I, J)
+    checkbounds(size(A,1), I)
     sz = size(A,2)
     for i = 3:ndims(A)
         sz *= size(A, i) # TODO: sync. with decision on issue #1030
     end
-    check_bounds(sz, J)
+    checkbounds(sz, J)
 end
 
-function check_bounds(A::AbstractArray, I::Union(Real,AbstractArray)...)
+function checkbounds(A::AbstractArray, I::Union(Real,AbstractArray)...)
     n = length(I)
     if n > 0
         for dim = 1:(n-1)
-            check_bounds(size(A,dim), I[dim])
+            checkbounds(size(A,dim), I[dim])
         end
         sz = size(A,n)
         for i = n+1:ndims(A)
             sz *= size(A,i)     # TODO: sync. with decision on issue #1030
         end
-        check_bounds(sz, I[n])
+        checkbounds(sz, I[n])
     end
 end
 
@@ -144,11 +153,12 @@ end
 reshape(a::AbstractArray, dims::Int...) = reshape(a, dims)
 
 vec(a::AbstractArray) = reshape(a,length(a))
+vec(a::AbstractVector) = a
 
 function squeeze(A::AbstractArray, dims)
     d = ()
     for i in 1:ndims(A)
-        if contains(dims,i)
+        if in(i,dims)
             if size(A,i) != 1
                 error("squeezed dims must all be size 1")
             end
@@ -175,6 +185,53 @@ function copy!(dest::AbstractArray, src)
     return dest
 end
 
+# copy with minimal requirements on src
+# if src is not an AbstractArray, moving to the offset might be O(n)
+function copy!(dest::AbstractArray, doffs::Integer, src, soffs::Integer=1)
+    st = start(src)
+    for j = 1:(soffs-1)
+        _, st = next(src, st)
+    end
+    i = doffs
+    while !done(src,st)
+        val, st = next(src, st)
+        dest[i] = val
+        i += 1
+    end
+    return dest
+end
+
+# NOTE: this is to avoid ambiguity with the deprecation of
+#   copy!(dest::AbstractArray, src, doffs::Integer)
+# Remove this when that deprecation is removed.
+function copy!(dest::AbstractArray, doffs::Integer, src::Integer)
+    dest[doffs] = src
+    return dest
+end
+
+# this method must be separate from the above since src might not have a length
+function copy!(dest::AbstractArray, doffs::Integer, src, soffs::Integer, n::Integer)
+    n == 0 && return dest
+    st = start(src)
+    for j = 1:(soffs-1)
+        _, st = next(src, st)
+    end
+    for i = doffs:(doffs+n-1)
+        done(src,st) && throw(BoundsError())
+        val, st = next(src, st)
+        dest[i] = val
+    end
+    return dest
+end
+
+# if src is an AbstractArray and a source offset is passed, use indexing
+function copy!(dest::AbstractArray, doffs::Integer, src::AbstractArray, soffs::Integer, n::Integer=length(src))
+    for i = 0:(n-1)
+        dest[doffs+i] = src[soffs+i]
+    end
+    return dest
+end
+
 copy(a::AbstractArray) = copy!(similar(a), a)
 copy(a::AbstractArray{None}) = a # cannot be assigned into so is immutable
 
@@ -189,33 +246,56 @@ isempty(a::AbstractArray) = (length(a) == 0)
 
 ## Conversions ##
 
-function iround_to{T}(dest::AbstractArray{T}, src)
-    i = 1
-    for x in src
-        dest[i] = iround(T,x)
-        i += 1
-    end
-    return dest
-end
-
 for (f,t) in ((:char,   Char),
               (:int,    Int),
               (:int8,   Int8),
               (:int16,  Int16),
               (:int32,  Int32),
               (:int64,  Int64),
+              (:int128, Int128),
+              (:uint,   Uint),
               (:uint8,  Uint8),
               (:uint16, Uint16),
               (:uint32, Uint32),
-              (:uint64, Uint64))
-    @eval ($f)(x::AbstractArray{$t}) = x
-    @eval ($f)(x::AbstractArray) = iround_to(similar(x,$t), x)
+              (:uint64, Uint64),
+              (:uint128,Uint128))
+    @eval begin
+        ($f)(x::AbstractArray{$t}) = x
+
+        function ($f)(x::AbstractArray)
+            y = similar(x,$t)
+            i = 1
+            for e in x
+                y[i] = ($f)(e)
+                i += 1
+            end
+            y
+        end
+    end
+end
+
+for (f,t) in ((:integer, Integer),
+              (:unsigned, Unsigned))
+    @eval begin
+        ($f){T<:$t}(x::AbstractArray{T}) = x
+
+        function ($f)(x::AbstractArray)
+            y = similar(x,typeof(($f)(one(eltype(x)))))
+            i = 1
+            for e in x
+                y[i] = ($f)(e)
+                i += 1
+            end
+            y
+        end
+    end
 end
 
 bool(x::AbstractArray{Bool}) = x
 bool(x::AbstractArray) = copy!(similar(x,Bool), x)
 
-for (f,t) in ((:float32,    Float32),
+for (f,t) in ((:float16,    Float16),
+              (:float32,    Float32),
               (:float64,    Float64),
               (:complex64,  Complex64),
               (:complex128, Complex128))
@@ -223,25 +303,32 @@ for (f,t) in ((:float32,    Float32),
     @eval ($f)(x::AbstractArray) = copy!(similar(x,$t), x)
 end
 
-integer{T<:Integer}(x::AbstractArray{T}) = x
-unsigned{T<:Unsigned}(x::AbstractArray{T}) = x
 float{T<:FloatingPoint}(x::AbstractArray{T}) = x
 complex{T<:Complex}(x::AbstractArray{T}) = x
 
-integer (x::AbstractArray) = iround_to(similar(x,typeof(integer(one(eltype(x))))), x)
-unsigned(x::AbstractArray) = iround_to(similar(x,typeof(unsigned(one(eltype(x))))), x)
 float   (x::AbstractArray) = copy!(similar(x,typeof(float(one(eltype(x))))), x)
 complex (x::AbstractArray) = copy!(similar(x,typeof(complex(one(eltype(x))))), x)
 
 dense(x::AbstractArray) = x
 full(x::AbstractArray) = x
 
+## range conversions ##
+
+for fn in _numeric_conversion_func_names
+    @eval begin
+        $fn(r::Range ) = Range($fn(r.start), $fn(r.step), r.len)
+        $fn(r::Range1) = Range1($fn(r.start), r.len)
+    end
+end
+
 ## Unary operators ##
 
 conj{T<:Real}(x::AbstractArray{T}) = x
 conj!{T<:Real}(x::AbstractArray{T}) = x
 
+real{T<:Real}(x::AbstractVector{T}) = x
 real{T<:Real}(x::AbstractArray{T}) = x
+imag{T<:Real}(x::AbstractVector{T}) = zero(x)
 imag{T<:Real}(x::AbstractArray{T}) = zero(x)
 
 +{T<:Number}(x::AbstractArray{T}) = x
@@ -305,8 +392,11 @@ end
 ## specifies what it wants to do before and after each loop.
 ## If genbodies creates an array it must of length N.
 function gen_cartesian_map(cache, genbodies, ranges, exargnames, exargs...)
+    if ranges === ()
+        ranges = (1,)
+    end
     N = length(ranges)
-    if !has(cache,N)
+    if !haskey(cache,N)
         if isdefined(genbodies,:code)
             mod = genbodies.code.module
         else
@@ -432,7 +522,7 @@ end
 #     turn are used for computing the linear index)
 function gen_array_index_map(cache, genbody, ranges, exargnames, exargs...)
     N = length(ranges)
-    if !has(cache,N)
+    if !haskey(cache,N)
         dimargnames = { symbol(string("_d",i)) for i=1:N }
         loopvars = { symbol(string("_l",i)) for i=1:N }
         offsetvars = { symbol(string("_offs",i)) for i=1:N }
@@ -457,15 +547,44 @@ end
 
 
 
-## Indexing: ref ##
+## Indexing: getindex ##
 
-ref(t::AbstractArray, i::Real) = error("indexing not defined for ", typeof(t))
+getindex(t::AbstractArray, i::Real) = error("indexing not defined for ", typeof(t))
+
+# linear indexing with a single multi-dimensional index
+function getindex(A::AbstractArray, I::AbstractArray)
+    x = similar(A, size(I))
+    for i=1:length(I)
+        x[i] = A[I[i]]
+    end
+    return x
+end
 
 # index A[:,:,...,i,:,:,...] where "i" is in dimension "d"
 # TODO: more optimized special cases
 slicedim(A::AbstractArray, d::Integer, i) =
     A[[ n==d ? i : (1:size(A,n)) for n in 1:ndims(A) ]...]
 
+function reverse(A::AbstractVector, s=1, n=length(A))
+    B = similar(A)
+    for i = 1:s-1
+        B[i] = A[i]
+    end
+    for i = s:n
+        B[i] = A[n+s-i]
+    end
+    for i = n+1:length(A)
+        B[i] = A[i]
+    end
+    B
+end
+
+function flipdim(A::AbstractVector, d::Integer)
+    d > 0 || error("dimension out of range")
+    d == 1 || return copy(A)
+    reverse(A)
+end
+
 function flipdim(A::AbstractArray, d::Integer)
     nd = ndims(A)
     sd = d > nd ? 1 : size(A, d)
@@ -506,12 +625,71 @@ function circshift(a, shiftamts)
     a[I...]::typeof(a)
 end
 
-## Indexing: assign ##
+## Indexing: setindex! ##
 
 # 1-d indexing is assumed defined on subtypes
-assign(t::AbstractArray, x, i::Real) =
-    error("assign not defined for ",typeof(t))
-assign(t::AbstractArray, x) = throw(MethodError(assign, (t, x)))
+setindex!(t::AbstractArray, x, i::Real) =
+    error("setindex! not defined for ",typeof(t))
+setindex!(t::AbstractArray, x) = throw(MethodError(setindex!, (t, x)))
+
+## Indexing: handle more indices than dimensions if "extra" indices are 1
+
+# Don't require vector/matrix subclasses to implement more than 1/2 indices,
+# respectively, by handling the extra dimensions in AbstractMatrix.
+
+function getindex(A::AbstractVector, i1,i2,i3...)
+    if i2*prod(i3) != 1
+        throw(BoundsError())
+    end
+    A[i1]
+end
+function getindex(A::AbstractMatrix, i1,i2,i3,i4...)
+    if i3*prod(i4) != 1
+        throw(BoundsError())
+    end
+    A[i1,i2]
+end
+
+function setindex!(A::AbstractVector, x, i1,i2,i3...)
+    if i2*prod(i3) != 1
+        throw(BoundsError())
+    end
+    A[i1] = x
+end
+function setindex!(A::AbstractMatrix, x, i1,i2,i3,i4...)
+    if i3*prod(i4) != 1
+        throw(BoundsError())
+    end
+    A[i1,i2] = x
+end
+
+## get (getindex with a default value) ##
+
+typealias RangeVecIntList{A<:AbstractVector{Int}} Union((Union(Ranges, AbstractVector{Int})...), AbstractVector{Range1{Int}}, AbstractVector{Range{Int}}, AbstractVector{A})
+
+get(A::AbstractArray, i::Integer, default) = in_bounds(length(A), i) ? A[i] : default
+get(A::AbstractArray, I::(), default) = similar(A, typeof(default), 0)
+get(A::AbstractArray, I::Dims, default) = in_bounds(size(A), I...) ? A[I...] : default
+
+function get!{T}(X::AbstractArray{T}, A::AbstractArray, I::Union(Ranges, AbstractVector{Int}), default::T)
+    ind = findin(I, 1:length(A))
+    X[ind] = A[I[ind]]
+    X[1:first(ind)-1] = default
+    X[last(ind)+1:length(X)] = default
+    X
+end
+
+get(A::AbstractArray, I::Ranges, default) = get!(similar(A, typeof(default), length(I)), A, I, default)
+
+function get!{T}(X::AbstractArray{T}, A::AbstractArray, I::RangeVecIntList, default::T)
+    fill!(X, default)
+    dst, src = indcopy(size(A), I)
+    X[dst...] = A[src...]
+    X
+end
+
+get(A::AbstractArray, I::RangeVecIntList, default) = get!(similar(A, typeof(default), map(length, I)...), A, I, default)
+
 
 ## Concatenation ##
 
@@ -522,8 +700,26 @@ vcat() = Array(None, 0)
 hcat() = Array(None, 0)
 
 ## cat: special cases
-hcat{T}(X::T...) = [ X[j] for i=1, j=1:length(X) ]
-vcat{T}(X::T...) = [ X[i] for i=1:length(X) ]
+hcat{T}(X::T...)         = T[ X[j] for i=1, j=1:length(X) ]
+hcat{T<:Number}(X::T...) = T[ X[j] for i=1, j=1:length(X) ]
+vcat{T}(X::T...)         = T[ X[i] for i=1:length(X) ]
+vcat{T<:Number}(X::T...) = T[ X[i] for i=1:length(X) ]
+
+function vcat(X::Number...)
+    T = None
+    for x in X
+        T = promote_type(T,typeof(x))
+    end
+    hvcat_fill(Array(T,length(X)), X)
+end
+
+function hcat(X::Number...)
+    T = None
+    for x in X
+        T = promote_type(T,typeof(x))
+    end
+    hvcat_fill(Array(T,1,length(X)), X)
+end
 
 function hcat{T}(V::AbstractVector{T}...)
     height = length(V[1])
@@ -606,7 +802,7 @@ function cat(catdim::Integer, X...)
     nargs = length(X)
     dimsX = map((a->isa(a,AbstractArray) ? size(a) : (1,)), X)
     ndimsX = map((a->isa(a,AbstractArray) ? ndims(a) : 1), X)
-    d_max = max(ndimsX)
+    d_max = maximum(ndimsX)
 
     if catdim > d_max + 1
         for i=1:nargs
@@ -674,7 +870,7 @@ function cat_t(catdim::Integer, typeC, A::AbstractArray...)
     nargs = length(A)
     dimsA = map(size, A)
     ndimsA = map(ndims, A)
-    d_max = max(ndimsA)
+    d_max = maximum(ndimsA)
 
     if catdim > d_max + 1
         for i=1:nargs
@@ -731,6 +927,16 @@ hcat(A::AbstractArray...) = cat(2, A...)
 
 # 2d horizontal and vertical concatenation
 
+function hvcat(nbc::Integer, as...)
+    # nbc = # of block columns
+    n = length(as)
+    if mod(n,nbc) != 0
+        error("hvcat: not all rows have the same number of block columns")
+    end
+    nbr = div(n,nbc)
+    hvcat(ntuple(nbr, i->nbc), as...)
+end
+
 function hvcat{T}(rows::(Int...), as::AbstractMatrix{T}...)
     nbr = length(rows)  # number of block rows
 
@@ -796,7 +1002,7 @@ end
 
 function hvcat_fill(a, xs)
     k = 1
-    nr, nc = size(a)
+    nr, nc = size(a,1), size(a,2)
     for i=1:nr
         for j=1:nc
             a[i,j] = xs[k]
@@ -825,6 +1031,7 @@ end
 ## Reductions and scans ##
 
 function isequal(A::AbstractArray, B::AbstractArray)
+    if A === B return true end
     if size(A) != size(B)
         return false
     end
@@ -836,7 +1043,7 @@ function isequal(A::AbstractArray, B::AbstractArray)
     return true
 end
 
-function cmp(A::AbstractArray, B::AbstractArray)
+function lexcmp(A::AbstractArray, B::AbstractArray)
     nA, nB = length(A), length(B)
     for i = 1:min(nA, nB)
         a, b = A[i], B[i]
@@ -847,8 +1054,6 @@ function cmp(A::AbstractArray, B::AbstractArray)
     return cmp(nA, nB)
 end
 
-isless(A::AbstractArray, B::AbstractArray) = cmp(A,B)<0
-
 function (==)(A::AbstractArray, B::AbstractArray)
     if size(A) != size(B)
         return false
@@ -861,29 +1066,31 @@ function (==)(A::AbstractArray, B::AbstractArray)
     return true
 end
 
-function (!=)(A::AbstractArray, B::AbstractArray)
-    if size(A) != size(B)
-        return true
-    end
-    for i = 1:length(A)
-        if A[i]!=B[i]
-            return true
+_cumsum_type{T<:Number}(v::AbstractArray{T}) = typeof(+zero(T))
+_cumsum_type(v) = typeof(v[1]+v[1])
+
+for (f, fp, op) = ((:cumsum, :cumsum_pairwise, :+),
+                   (:cumprod, :cumprod_pairwise, :*) )
+    # in-place cumsum of c = s+v(i1:n), using pairwise summation as for sum
+    @eval function ($fp)(v::AbstractVector, c::AbstractVector, s, i1, n)
+        if n < 128
+            @inbounds c[i1] = ($op)(s, v[i1])
+            for i = i1+1:i1+n-1
+                @inbounds c[i] = $(op)(c[i-1], v[i])
+            end
+        else
+            n2 = div(n,2)
+            ($fp)(v, c, s, i1, n2)
+            ($fp)(v, c, c[(i1+n2)-1], i1+n2, n-n2)
         end
     end
-    return false
-end
 
-for (f, op) = ((:cumsum, :+), (:cumprod, :*) )
     @eval function ($f)(v::AbstractVector)
         n = length(v)
-        c = $(op===:+ ? (:(similar(v,typeof(+zero(eltype(v)))))) :
+        c = $(op===:+ ? (:(similar(v,_cumsum_type(v)))) :
                         (:(similar(v))))
         if n == 0; return c; end
-
-        c[1] = v[1]
-        for i=2:n
-           c[i] = ($op)(v[i], c[i-1])
-        end
+        ($fp)(v, c, $(op==:+ ? :(zero(v[1])) : :(one(v[1]))), 1, n)
         return c
     end
 
@@ -896,7 +1103,7 @@ for (f, op) = ((:cumsum, :+), (:cumprod, :*) )
             axis_stride *= size(A,i)
         end
 
-        B = $(op===:+ ? (:(similar(A,typeof(+zero(eltype(A)))))) :
+        B = $(op===:+ ? (:(similar(A,_cumsum_type(A)))) :
                         (:(similar(A))))
 
         if axis_size < 1
@@ -907,7 +1114,7 @@ for (f, op) = ((:cumsum, :+), (:cumprod, :*) )
             if div(i-1, axis_stride) % axis_size == 0
                B[i] = A[i]
             else
-               B[i] = ($op)(A[i], B[i-axis_stride])
+               B[i] = ($op)(B[i-axis_stride], A[i])
             end
         end
 
@@ -983,8 +1190,8 @@ function hvcat(rows::(Int...), as...)
     vcat(rs...)
 end
 
-function repmat(a::AbstractMatrix, m::Int, n::Int)
-    o,p = size(a)
+function repmat(a::Union(AbstractVector,AbstractMatrix), m::Int, n::Int=1)
+    o, p = size(a,1), size(a,2)
     b = similar(a, o*m, p*n)
     for j=1:n
         d = (j-1)*p+1
@@ -996,7 +1203,16 @@ function repmat(a::AbstractMatrix, m::Int, n::Int)
     end
     return b
 end
-repmat(a::AbstractVector, m::Int, n::Int) = repmat(reshape(a, length(a), 1), m, n)
+
+function repmat(a::AbstractVector, m::Int)
+    o = length(a)
+    b = similar(a, o*m)
+    for i=1:m
+        c = (i-1)*o+1
+        b[c:c+o-1] = a
+    end
+    return b
+end
 
 sub2ind(dims) = 1
 sub2ind(dims, i::Integer) = int(i)
@@ -1021,18 +1237,20 @@ function sub2ind(dims, I::Integer...)
     return index
 end
 
+function sub2ind{T<:Integer}(dims::Array{T}, sub::Array{T})
+    ndims = length(dims)
+    ind = sub[1]
+    stride = 1
+    for k in 2:ndims
+        stride = stride * dims[k - 1]
+        ind += (sub[k] - 1) * stride
+    end
+    return ind
+end
+
 sub2ind{T<:Integer}(dims, I::AbstractVector{T}...) =
     [ sub2ind(dims, map(X->X[i], I)...)::Int for i=1:length(I[1]) ]
 
-ind2sub(dims::(Integer...), ind::Integer) = ind2sub(dims, int(ind))
-ind2sub(dims::(), ind::Integer) = ind==1 ? () : throw(BoundsError())
-ind2sub(dims::(Integer,), ind::Int) = (ind,)
-ind2sub(dims::(Integer,Integer), ind::Int) =
-    (rem(ind-1,dims[1])+1, div(ind-1,dims[1])+1)
-ind2sub(dims::(Integer,Integer,Integer), ind::Int) =
-    (rem(ind-1,dims[1])+1, div(rem(ind-1,dims[1]*dims[2]), dims[1])+1,
-     div(rem(ind-1,dims[1]*dims[2]*dims[3]), dims[1]*dims[2])+1)
-
 function ind2sub(dims::(Integer,Integer...), ind::Int)
     ndims = length(dims)
     stride = dims[1]
@@ -1050,6 +1268,15 @@ function ind2sub(dims::(Integer,Integer...), ind::Int)
     return tuple(ind, sub...)
 end
 
+ind2sub(dims::(Integer...), ind::Integer) = ind2sub(dims, int(ind))
+ind2sub(dims::(), ind::Integer) = ind==1 ? () : throw(BoundsError())
+ind2sub(dims::(Integer,), ind::Int) = (ind,)
+ind2sub(dims::(Integer,Integer), ind::Int) =
+    (rem(ind-1,dims[1])+1, div(ind-1,dims[1])+1)
+ind2sub(dims::(Integer,Integer,Integer), ind::Int) =
+    (rem(ind-1,dims[1])+1, div(rem(ind-1,dims[1]*dims[2]), dims[1])+1,
+     div(rem(ind-1,dims[1]*dims[2]*dims[3]), dims[1]*dims[2])+1)
+
 function ind2sub{T<:Integer}(dims::(Integer,Integer...), ind::AbstractVector{T})
     n = length(dims)
     l = length(ind)
@@ -1063,42 +1290,88 @@ function ind2sub{T<:Integer}(dims::(Integer,Integer...), ind::AbstractVector{T})
     return t
 end
 
+function ind2sub!{T<:Integer}(sub::Array{T}, dims::Array{T}, ind::T)
+    ndims = length(dims)
+    stride = dims[1]
+    for i in 2:(ndims - 1)
+        stride *= dims[i]
+    end
+    for i in (ndims - 1):-1:1
+        rest = rem1(ind, stride)
+        sub[i + 1] = div(ind - rest, stride) + 1
+        ind = rest
+        stride = div(stride, dims[i])
+    end
+    sub[1] = ind
+    return
+end
+
 indices(I) = I
+indices(I::Int) = I
+indices(I::Real) = convert(Int, I)
 indices(I::AbstractArray{Bool,1}) = find(I)
+indices(I::(Any,))            = (indices(I[1]), )
+indices(I::(Any,Any,))        = (indices(I[1]), indices(I[2]))
+indices(I::(Any,Any,Any))     = (indices(I[1]), indices(I[2]), indices(I[3]))
+indices(I::(Any,Any,Any,Any)) = (indices(I[1]), indices(I[2]), indices(I[3]), indices(I[4]))
 indices(I::Tuple) = map(indices, I)
 
+# Generalized repmat
+function repeat{T}(A::Array{T};
+                   inner::Array{Int} = ones(Int, ndims(A)),
+                   outer::Array{Int} = ones(Int, ndims(A)))
+    ndims_in = ndims(A)
+    length_inner = length(inner)
+    length_outer = length(outer)
+    ndims_out = max(ndims_in, length_inner, length_outer)
+
+    if length_inner < ndims_in || length_outer < ndims_in
+        msg = "Inner/outer repetitions must be set for all input dimensions"
+        throw(ArgumentError(msg))
+    end
+
+    size_in = Array(Int, ndims_in)
+    size_out = Array(Int, ndims_out)
+    inner_size_out = Array(Int, ndims_out)
+
+    for i in 1:ndims_in
+        size_in[i] = size(A, i)
+    end
+    for i in 1:ndims_out
+        t1 = ndims_in < i ? 1 : size_in[i]
+        t2 = length_inner < i ? 1 : inner[i]
+        t3 = length_outer < i ? 1 : outer[i]
+        size_out[i] = t1 * t2 * t3
+        inner_size_out[i] = t1 * t2
+    end
+
+    indices_in = Array(Int, ndims_in)
+    indices_out = Array(Int, ndims_out)
+
+    length_out = prod(size_out)
+    R = Array(T, size_out...)
+
+    for index_out in 1:length_out
+        ind2sub!(indices_out, size_out, index_out)
+        for t in 1:ndims_in
+            # "Project" outer repetitions into inner repetitions
+            indices_in[t] = mod1(indices_out[t], inner_size_out[t])
+            # Find inner repetitions using flooring division
+            if inner[t] != 1
+                indices_in[t] = fld1(indices_in[t], inner[t])
+            end
+        end
+        index_in = sub2ind(size_in, indices_in)
+        R[index_out] = A[index_in]
+    end
+
+    return R
+end
+
 ## iteration utilities ##
 
-# TODO: something sensible should happen when each_col et. al. are used with a
-# pure function argument
-function each_col!(f::Function, a::AbstractMatrix)
-    m = size(a,1)
-    for i = 1:m:length(a)
-        f(sub(a, i:(i+m-1)))
-    end
-    return a
-end
-
-function each_row!(f::Function, a::AbstractMatrix)
-    m = size(a,1)
-    for i = 1:m
-        f(sub(a, i:m:length(a)))
-    end
-    return a
-end
-
-function each_vec!(f::Function, a::AbstractMatrix, dim::Integer)
-    if dim == 1; return each_col!(f,a); end
-    if dim == 2; return each_row!(f,a); end
-    error("invalid matrix dimensions: ", dim)
-end
-
-each_col(f::Function, a::AbstractMatrix) = each_col!(f,copy(a))
-each_row(f::Function, a::AbstractMatrix) = each_row!(f,copy(a))
-each_vec(f::Function, a::AbstractMatrix, d::Integer) = each_vec!(f,copy(a),d)
-
 # slow, but useful
-function cartesian_map(body, t::(Int...), it...)
+function cartesianmap(body, t::(Int...), it...)
     idx = length(t)-length(it)
     if idx == 1
         for i = 1:t[1]
@@ -1113,23 +1386,23 @@ function cartesian_map(body, t::(Int...), it...)
     elseif idx > 1
         for j = 1:t[idx]
             for i = 1:t[idx-1]
-                cartesian_map(body, t, i, j, it...)
+                cartesianmap(body, t, i, j, it...)
             end
         end
     else
-        throw(ArgumentError("cartesian_map"))
+        throw(ArgumentError("cartesianmap"))
     end
 end
 
-cartesian_map(body, t::()) = (body(); nothing)
+cartesianmap(body, t::()) = (body(); nothing)
 
-function cartesian_map(body, t::(Int,))
+function cartesianmap(body, t::(Int,))
     for i = 1:t[1]
         body(i)
     end
 end
 
-function cartesian_map(body, t::(Int,Int))
+function cartesianmap(body, t::(Int,Int))
     for j = 1:t[2]
         for i = 1:t[1]
             body(i,j)
@@ -1137,7 +1410,7 @@ function cartesian_map(body, t::(Int,Int))
     end
 end
 
-function cartesian_map(body, t::(Int,Int,Int))
+function cartesianmap(body, t::(Int,Int,Int))
     for k = 1:t[3]
         for j = 1:t[2]
             for i = 1:t[1]
@@ -1147,73 +1420,306 @@ function cartesian_map(body, t::(Int,Int,Int))
     end
 end
 
-function bsxfun(f, a::AbstractArray, b::AbstractArray)
-    nd = max(ndims(a),ndims(b))
-    shp = Array(Int,nd)
-    range = ()
-    xa, xb = false, false
-    for i=1:nd
-        ai, bi = size(a,i), size(b,i)
-        if ai == bi
-            shp[i] = ai
-        elseif ai == 1
-            xa = true
-            shp[i] = bi
-            range = tuple(range..., bi)
-        elseif bi == 1
-            xb = true
-            shp[i] = ai
-            range = tuple(range..., ai)
-        else
-            error("argument dimensions do not match")
-        end
-    end
-    if isempty(range)
-        return f(a, b)
-    end
-    if length(a) == 1
-        return f(a[1], b)
-    elseif length(b) == 1
-        return f(a, b[1])
-    end
-    c = Array(promote_type(eltype(a),eltype(b)), shp...)
+# Basic AbstractArray functions
 
-    aidxs = { 1:size(a,i) for i=1:nd }
-    bidxs = { 1:size(b,i) for i=1:nd }
-    cidxs = { 1:size(c,i) for i=1:nd }
-
-    sliceop = function (idxs::Int...)
-        j = 1
-        for i = 1:nd
-            ai, bi = size(a,i), size(b,i)
-            if ai == bi
-            elseif ai == 1
-                bidxs[i] = idxs[j]
-                cidxs[i] = idxs[j]
-                j+=1
-            else
-                aidxs[i] = idxs[j]
-                cidxs[i] = idxs[j]
-                j+=1
-            end
-        end
-        if xb
-            aa = a[aidxs...]; if length(aa)==1; aa=aa[1]; end
-        else
-            aa = a
-        end
-        if xa
-            bb = b[bidxs...]; if length(bb)==1; bb=bb[1]; end
-        else
-            bb = b
-        end
-        c[cidxs...] = f(aa, bb)
+function nnz{T}(a::AbstractArray{T})
+    n = 0
+    for i = 1:length(a)
+        @inbounds n += (a[i] != zero(T))
     end
-    cartesian_map(sliceop, range)
-    c
+    return n
 end
 
-bsxfun(f, a, b) = f(a, b)
-bsxfun(f, a::AbstractArray, b) = f(a, b)
-bsxfun(f, a, b::AbstractArray) = f(a, b)
-bsxfun(f, a, b, c...) = bsxfun(f, bsxfun(f, a, b), c...)
+# for reductions that expand 0 dims to 1
+reduced_dims(A, region) = ntuple(ndims(A), i->(in(i, region) ? 1 :
+                                               size(A,i)))
+
+# keep 0 dims in place
+reduced_dims0(A, region) = ntuple(ndims(A), i->(size(A,i)==0 ? 0 :
+                                                in(i, region) ? 1 :
+                                                size(A,i)))
+
+reducedim(f::Function, A, region, v0) =
+    reducedim(f, A, region, v0, similar(A, reduced_dims(A, region)))
+
+maximum{T}(A::AbstractArray{T}, region) =
+    isempty(A) ? similar(A,reduced_dims0(A,region)) : reducedim(scalarmax,A,region,typemin(T))
+minimum{T}(A::AbstractArray{T}, region) =
+    isempty(A) ? similar(A,reduced_dims0(A,region)) : reducedim(scalarmin,A,region,typemax(T))
+sum{T}(A::AbstractArray{T}, region)  = reducedim(+,A,region,zero(T))
+prod{T}(A::AbstractArray{T}, region) = reducedim(*,A,region,one(T))
+
+all(A::AbstractArray{Bool}, region) = reducedim(&,A,region,true)
+any(A::AbstractArray{Bool}, region) = reducedim(|,A,region,false)
+sum(A::AbstractArray{Bool}, region) = reducedim(+,A,region,0,similar(A,Int,reduced_dims(A,region)))
+sum(A::AbstractArray{Bool}) = sum(A, [1:ndims(A)])[1]
+prod(A::AbstractArray{Bool}) =
+    error("use all() instead of prod() for boolean arrays")
+prod(A::AbstractArray{Bool}, region) =
+    error("use all() instead of prod() for boolean arrays")
+
+# Pairwise (cascade) summation of A[i1:i1+n-1], which has O(log n) error growth
+# [vs O(n) for a simple loop] with negligible performance cost if
+# the base case is large enough.  See, e.g.:
+#        http://en.wikipedia.org/wiki/Pairwise_summation
+#        Higham, Nicholas J. (1993), "The accuracy of floating point
+#        summation", SIAM Journal on Scientific Computing 14 (4): 783–799.
+# In fact, the root-mean-square error growth, assuming random roundoff
+# errors, is only O(sqrt(log n)), which is nearly indistinguishable from O(1)
+# in practice.  See:
+#        Manfred Tasche and Hansmartin Zeuner, Handbook of
+#        Analytic-Computational Methods in Applied Mathematics (2000).
+function sum_pairwise(A::AbstractArray, i1,n)
+    if n < 128
+        @inbounds s = A[i1]
+        for i = i1+1:i1+n-1
+            @inbounds s += A[i]
+        end
+        return s
+    else
+        n2 = div(n,2)
+        return sum_pairwise(A, i1, n2) + sum_pairwise(A, i1+n2, n-n2)
+    end
+end
+
+function sum{T}(A::AbstractArray{T})
+    n = length(A)
+    n == 0 ? zero(T) : sum_pairwise(A, 1, n)
+end
+
+# Kahan (compensated) summation: O(1) error growth, at the expense
+# of a considerable increase in computational expense.
+function sum_kbn{T<:FloatingPoint}(A::AbstractArray{T})
+    n = length(A)
+    if (n == 0)
+        return zero(T)
+    end
+    s = A[1]
+    c = zero(T)
+    for i in 2:n
+        Ai = A[i]
+        t = s + Ai
+        if abs(s) >= abs(Ai)
+            c += ((s-t) + Ai)
+        else
+            c += ((Ai-t) + s)
+        end
+        s = t
+    end
+
+    s + c
+end
+
+# Uses K-B-N summation
+function cumsum_kbn{T<:FloatingPoint}(v::AbstractVector{T})
+    n = length(v)
+    r = similar(v, n)
+    if n == 0; return r; end
+
+    s = r[1] = v[1]
+    c = zero(T)
+    for i=2:n
+        vi = v[i]
+        t = s + vi
+        if abs(s) >= abs(vi)
+            c += ((s-t) + vi)
+        else
+            c += ((vi-t) + s)
+        end
+        s = t
+        r[i] = s+c
+    end
+    return r
+end
+
+# Uses K-B-N summation
+function cumsum_kbn{T<:FloatingPoint}(A::AbstractArray{T}, axis::Integer)
+    dimsA = size(A)
+    ndimsA = ndims(A)
+    axis_size = dimsA[axis]
+    axis_stride = 1
+    for i = 1:(axis-1)
+        axis_stride *= size(A,i)
+    end
+
+    if axis_size <= 1
+        return A
+    end
+
+    B = similar(A)
+    C = similar(A)
+
+    for i = 1:length(A)
+        if div(i-1, axis_stride) % axis_size == 0
+            B[i] = A[i]
+            C[i] = zero(T)
+        else
+            s = B[i-axis_stride]
+            Ai = A[i]
+            B[i] = t = s + Ai
+            if abs(s) >= abs(Ai)
+                C[i] = C[i-axis_stride] + ((s-t) + Ai)
+            else
+                C[i] = C[i-axis_stride] + ((Ai-t) + s)
+            end
+        end
+    end
+
+    return B + C
+end
+
+function prod{T}(A::AbstractArray{T})
+    if isempty(A)
+        return one(T)
+    end
+    v = A[1]
+    for i=2:length(A)
+        @inbounds v *= A[i]
+    end
+    v
+end
+
+function minimum{T<:Real}(A::AbstractArray{T})
+    if isempty(A); error("minimum: argument is empty"); end
+    v = A[1]
+    for i=2:length(A)
+        @inbounds x = A[i]
+        if x < v || v!=v
+            v = x
+        end
+    end
+    v
+end
+
+function maximum{T<:Real}(A::AbstractArray{T})
+    if isempty(A); error("maximum: argument is empty"); end
+    v = A[1]
+    for i=2:length(A)
+        @inbounds x = A[i]
+        if x > v || v!=v
+            v = x
+        end
+    end
+    v
+end
+
+## map over arrays ##
+
+## transform any set of dimensions
+## dims specifies which dimensions will be transformed. for example
+## dims==1:2 will call f on all slices A[:,:,...]
+mapslices(f::Function, A::AbstractArray, dims) = mapslices(f, A, [dims...])
+function mapslices(f::Function, A::AbstractArray, dims::AbstractVector)
+    if isempty(dims)
+        return A
+    end
+
+    dimsA = [size(A)...]
+    ndimsA = ndims(A)
+    alldims = [1:ndimsA]
+
+    if dims == alldims
+        return f(A)
+    end
+
+    otherdims = setdiff(alldims, dims)
+
+    idx = cell(ndimsA)
+    fill!(idx, 1)
+    Asliceshape = tuple(dimsA[dims]...)
+    itershape   = tuple(dimsA[otherdims]...)
+    for d in dims
+        idx[d] = 1:size(A,d)
+    end
+
+    r1 = f(reshape(A[idx...], Asliceshape))
+
+    # determine result size and allocate
+    Rsize = copy(dimsA)
+    # TODO: maybe support removing dimensions
+    if isempty(size(r1))
+        r1 = [r1]
+    end
+    Rsize[dims] = [size(r1)...; ones(Int,max(0,length(dims)-ndims(r1)))]
+    R = similar(r1, tuple(Rsize...))
+
+    ridx = cell(ndims(R))
+    fill!(ridx, 1)
+    for d in dims
+        ridx[d] = 1:size(R,d)
+    end
+
+    R[ridx...] = r1
+
+    first = true
+    cartesianmap(itershape) do idxs...
+        if first
+            first = false
+        else
+            ia = [idxs...]
+            idx[otherdims] = ia
+            ridx[otherdims] = ia
+            R[ridx...] = f(reshape(A[idx...], Asliceshape))
+        end
+    end
+
+    return R
+end
+
+
+## 1 argument
+function map_to!(f::Callable, first, dest::AbstractArray, A::AbstractArray)
+    dest[1] = first
+    for i=2:length(A)
+        dest[i] = f(A[i])
+    end
+    return dest
+end
+
+function map(f::Callable, A::AbstractArray)
+    if isempty(A); return {}; end
+    first = f(A[1])
+    dest = similar(A, typeof(first))
+    return map_to!(f, first, dest, A)
+end
+
+## 2 argument
+function map_to!(f::Callable, first, dest::AbstractArray, A::AbstractArray, B::AbstractArray)
+    dest[1] = first
+    for i=2:length(A)
+        dest[i] = f(A[i], B[i])
+    end
+    return dest
+end
+
+function map(f::Callable, A::AbstractArray, B::AbstractArray)
+    shp = promote_shape(size(A),size(B))
+    if isempty(A)
+        return similar(A, Any, shp)
+    end
+    first = f(A[1], B[1])
+    dest = similar(A, typeof(first), shp)
+    return map_to!(f, first, dest, A, B)
+end
+
+## N argument
+function map_to!(f::Callable, first, dest::AbstractArray, As::AbstractArray...)
+    n = length(As[1])
+    i = 1
+    ith = a->a[i]
+    dest[1] = first
+    for i=2:n
+        dest[i] = f(map(ith, As)...)
+    end
+    return dest
+end
+
+function map(f::Callable, As::AbstractArray...)
+    shape = mapreduce(size, promote_shape, As)
+    if prod(shape) == 0
+        return similar(As[1], Any, shape)
+    end
+    first = f(map(a->a[1], As)...)
+    dest = similar(As[1], typeof(first), shape)
+    return map_to!(f, first, dest, As...)
+end
