@@ -28,7 +28,7 @@ end
 const FORWARD = int32(-1)
 const BACKWARD = int32(1)
 
-## _jl_FFTW Flags from fftw3.h
+## FFTW Flags from fftw3.h
 
 const MEASURE         = uint32(0)
 const DESTROY_INPUT   = uint32(1 << 0)
@@ -532,7 +532,7 @@ for (Tr,Tc) in ((:Float32,:Complex64),(:Float64,:Complex128))
         function rfft(X::StridedArray{$Tr}, region)
             d1 = region[1]
             osize = [size(X)...]
-            osize[d1] = ifloor(osize[d1]/2) + 1
+            osize[d1] = osize[d1]>>1 + 1
             Y = Array($Tc, osize...)
             p = Plan(X, Y, region, ESTIMATE, NO_TIMELIMIT)
             execute($Tr, p.plan)
@@ -549,7 +549,7 @@ for (Tr,Tc) in ((:Float32,:Complex64),(:Float64,:Complex128))
                            flags::Unsigned, tlim::Real)
             d1 = region[1]
             osize = [size(X)...]
-            osize[d1] = ifloor(osize[d1]/2) + 1
+            osize[d1] = osize[d1]>>1 + 1
             Y = Array($Tc, osize...)
             p = Plan(X, Y, region, flags, tlim)
             return Z::StridedArray{$Tr} -> begin
@@ -566,7 +566,7 @@ for (Tr,Tc) in ((:Float32,:Complex64),(:Float64,:Complex128))
 
         function brfft(X::StridedArray{$Tc}, d::Integer, region::Integer)
             osize = [size(X)...]
-            @assert osize[region] == ifloor(d/2) + 1
+            @assert osize[region] == d>>1 + 1
             osize[region] = d
             Y = Array($Tr, osize...)
             p = Plan(X, Y, region, ESTIMATE | PRESERVE_INPUT, NO_TIMELIMIT)
@@ -579,7 +579,7 @@ for (Tr,Tc) in ((:Float32,:Complex64),(:Float64,:Complex128))
         function brfftd(X::StridedArray{$Tc}, d::Integer, region) 
             d1 = region[1]
             osize = [size(X)...]
-            @assert osize[d1] == ifloor(d/2) + 1
+            @assert osize[d1] == d>>1 + 1
             osize[d1] = d
             Y = Array($Tr, osize...)
             p = Plan(X, Y, region, ESTIMATE, NO_TIMELIMIT)
@@ -604,7 +604,7 @@ for (Tr,Tc) in ((:Float32,:Complex64),(:Float64,:Complex128))
         function plan_brfft(X::StridedArray{$Tc}, d::Integer, region::Integer,
                             flags::Unsigned, tlim::Real)
             osize = [size(X)...]
-            @assert osize[region] == ifloor(d/2) + 1
+            @assert osize[region] == d>>1 + 1
             osize[region] = d
             Y = Array($Tr, osize...)
             p = Plan(X, Y, region, flags | PRESERVE_INPUT, tlim)
@@ -623,7 +623,7 @@ for (Tr,Tc) in ((:Float32,:Complex64),(:Float64,:Complex128))
             end
             d1 = region[1]
             osize = [size(X)...]
-            @assert osize[d1] == ifloor(d/2) + 1
+            @assert osize[d1] == d>>1 + 1
             osize[d1] = d
             Y = Array($Tr, osize...)
             X = copy(X)
@@ -789,41 +789,4 @@ for f in (:plan_r2r, :plan_r2r!)
     end
 end
 
-# FFTW-based transpose routines
-#   TODO: implement in-place transpose! routine (this is supported by FFTW)
-#   TODO: fftw/fftwf can transpose matrices of any type, provided that the
-#         element size is a multiple of sizeof(double) or sizeof(float).
-
-for (libname, fname, elty) in ((:libfftw ,"fftw_plan_guru64_r2r",:Float64),
-                               (:libfftwf,"fftwf_plan_guru64_r2r",:Float32))
-    @eval begin
-        function transpose(X::Matrix{$elty})
-            P = similar(X)
-            (n1, n2) = size(X)
-            plan = ccall(($fname,$libname), Ptr{Void},
-                         (Int32, Ptr{Int}, Int32, Ptr{Int}, Ptr{$elty}, Ptr{$elty}, Ptr{Int32}, Uint32),
-                         0, C_NULL, 2, [n1,n2,1,n2,1,n1], X, P, [HC2R], ESTIMATE | PRESERVE_INPUT)
-            execute($elty, plan)
-            destroy_plan($elty, plan)
-            return P
-        end
-    end
-end
-
-for (libname, fname, celty) in ((:libfftw ,"fftw_plan_guru64_dft",:Complex128),
-                                (:libfftwf,"fftwf_plan_guru64_dft",:Complex64))
-    @eval begin
-        function transpose(X::Matrix{$celty})
-            P = similar(X)
-            (n1, n2) = size(X)
-            plan = ccall(($fname,$libname), Ptr{Void},
-                         (Int32, Ptr{Int}, Int32, Ptr{Int}, Ptr{$celty}, Ptr{$celty}, Int32, Uint32),
-                         0, C_NULL, 2, [n1,n2,1,n2,1,n1], X, P, FORWARD, ESTIMATE | PRESERVE_INPUT)
-            execute($celty, plan)
-            destroy_plan($celty, plan)
-            return P
-        end
-    end
-end
-
 end # module
