@@ -1,18 +1,94 @@
-# formerly built-in methods. can be replaced any time.
+show(x) = show(STDOUT::IO, x)
 
-show(x) = show(OUTPUT_STREAM::IO, x)
+function print(io::IO, s::Symbol)
+    pname = convert(Ptr{Uint8}, s)
+    write(io, pname, int(ccall(:strlen, Csize_t, (Ptr{Uint8},), pname)))
+end
 
-print(io::IO, s::Symbol) = ccall(:jl_print_symbol, Void, (Ptr{Void}, Any,), io, s)
-show(io::IO, x::ANY) = ccall(:jl_show_any, Void, (Any, Any,), io::IO, x)
+function show(io::IO, x::ANY)
+    t = typeof(x)::DataType
+    show(io, t)
+    print(io, '(')
+    if t.names !== () || t.size==0
+        n = length(t.names)
+        for i=1:n
+            f = t.names[i]
+            if !isdefined(x, f)
+                print(io, undef_ref_str)
+            else
+                show(io, x.(f))
+            end
+            if i < n
+                print(io, ',')
+            end
+        end
+    else
+        nb = t.size
+        print(io, "0x")
+        p = pointer_from_objref(x) + sizeof(Ptr{Void})
+        for i=nb-1:-1:0
+            print(io, hex(unsafe_load(convert(Ptr{Uint8}, p+i)), 2))
+        end
+    end
+    print(io,')')
+end
 
+function show(io::IO, f::Function)
+    if isgeneric(f)
+        print(io, f.env.name)
+    elseif isdefined(f, :env) && isa(f.env,Symbol)
+        print(io, f.env)
+    else
+        print(io, "(anonymous function)")
+    end
+end
+
+function show(io::IO, x::IntrinsicFunction)
+    print(io, "(intrinsic function #", box(Int32,unbox(IntrinsicFunction,x)), ")")
+end
+
+function show(io::IO, x::UnionType)
+    if is(x,None)
+        print(io, "None")
+    elseif is(x,Top)
+        print(io, "Top")
+    else
+        print(io, "Union", x.types)
+    end
+end
+
+show(io::IO, x::TypeConstructor) = show(io, x.body)
+
+function show(io::IO, x::DataType)
+    if isvarargtype(x)
+        print(io, x.parameters[1], "...")
+    else
+        print(io, x.name.name)
+        if length(x.parameters) > 0
+            show_comma_array(io, x.parameters, '{', '}')
+        end
+    end
+end
+
+function showcompact{T<:Number}(io::IO, x::Vector{T})
+    print(io, "[")
+    if length(x) > 0
+        showcompact(io, x[1])
+        for j in 2:length(x)
+            print(io, ",")
+            showcompact(io, x[j])
+        end
+    end
+    print(io, "]")
+end
 showcompact(io::IO, x) = show(io, x)
-showcompact(x) = showcompact(OUTPUT_STREAM::IO, x)
+showcompact(x) = showcompact(STDOUT::IO, x)
 
 macro show(exs...)
-    blk = expr(:block)
+    blk = Expr(:block)
     for ex in exs
         push!(blk.args, :(println($(sprint(show_unquoted,ex)*" => "),
-                                 repr(begin value=$(esc(ex)) end))))
+                                  repr(begin value=$(esc(ex)) end))))
     end
     if !isempty(exs); push!(blk.args, :value); end
     return blk
@@ -24,29 +100,22 @@ show(io::IO, ::Nothing) = print(io, "nothing")
 show(io::IO, b::Bool) = print(io, b ? "true" : "false")
 show(io::IO, n::Signed) = (write(io, dec(n)); nothing)
 show(io::IO, n::Unsigned) = print(io, "0x", hex(n,sizeof(n)<<1))
+print(io::IO, n::Unsigned) = print(io, dec(n))
 
 show{T}(io::IO, p::Ptr{T}) =
     print(io, is(T,None) ? "Ptr{Void}" : typeof(p), " @0x$(hex(unsigned(p), WORD_SIZE>>2))")
 
-full_name(m::Module) = m===Main ? () : tuple(full_name(module_parent(m))...,
-                                             module_name(m))
-
 function show(io::IO, m::Module)
     if is(m,Main)
         print(io, "Main")
     else
-        print(io, join(full_name(m),"."))
+        print(io, join(fullname(m),"."))
     end
 end
 
 function show(io::IO, l::LambdaStaticData)
     print(io, "AST(")
-    if isa(l.ast,Expr)
-        show(io, l.ast)
-    else
-        ast = ccall(:jl_uncompress_ast, Any, (Any,Any), l, l.ast)
-        show(io, ast)
-    end
+    show(io, uncompressed_ast(l))
     print(io, ")")
 end
 
@@ -56,19 +125,25 @@ function show_delim_array(io::IO, itr, op, delim, cl, delim_one)
     newline = true
     first = true
     if !done(itr,state)
-	while true
-	    x, state = next(itr,state)
-            multiline = isa(x,AbstractArray) && ndims(x)>1 && length(x)>0
-            if newline
-                if multiline; println(io); end
+        while true
+            if isa(itr,Array) && !isdefined(itr,state)
+                print(io, undef_ref_str)
+                state += 1
+                multiline = false
+            else
+                x, state = next(itr,state)
+                multiline = isa(x,AbstractArray) && ndims(x)>1 && length(x)>0
+                if newline
+                    if multiline; println(io); end
+                end
+                show(io, x)
             end
-	    show(io, x)
-	    if done(itr,state)
+            if done(itr,state)
                 if delim_one && first
                     print(io, delim)
                 end
-		break
-	    end
+                break
+            end
             first = false
             print(io, delim)
             if multiline
@@ -76,7 +151,7 @@ function show_delim_array(io::IO, itr, op, delim, cl, delim_one)
             else
                 newline = true
             end
-	end
+        end
     end
     print(io, cl)
 end
@@ -106,10 +181,10 @@ unquoted(ex::Expr)       = ex.args[1]
 # which shows the contents using show_unquoted(),
 # which shows subexpressions using show_unquoted()
 # ==> AST:s are printed wrapped in a single quotation
-show_indented(x)                 = show_indented(OUTPUT_STREAM, x)
+show_indented(x)                 = show_indented(STDOUT, x)
 show_indented(io::IO, x)         = show_indented(io, x, 0)
 show_indented(io::IO, x, indent) = show(io, x)
-show_unquoted(x)                 = show_unquoted(OUTPUT_STREAM, x)
+show_unquoted(x)                 = show_unquoted(STDOUT, x)
 show_unquoted(io::IO, x)         = show_unquoted(io, x, 0)
 show_unquoted(io::IO, x, indent) = (print(io,"\$("); show(io,x); print(io,')'))
 
@@ -130,7 +205,7 @@ show(io::IO, ex::Expr) = show_indented(io, ex)
 function show_indented(io::IO, ex::Expr, indent::Int)
     if is(ex.head, :block) || is(ex.head, :body)
         show_block(io, "quote", ex, indent); print(io, "end")
-    elseif contains((:tuple, :vcat, :cell1), ex.head)
+    elseif in(ex.head, (:tuple, :vcat, :cell1))
         print(io, ':'); show_unquoted(io, ex, indent + indent_width)        
     else
         default_show_quoted(io, ex, indent)
@@ -138,8 +213,10 @@ function show_indented(io::IO, ex::Expr, indent::Int)
 end
 const paren_quoted_syms = Set{Symbol}(:(:),:(::),:(:=),:(=),:(==),:(===),:(=>))
 function show_indented(io::IO, sym::Symbol, indent::Int)
-    if has(paren_quoted_syms, sym); print(io, ":($sym)")        
-    else                            print(io, ":$sym")        
+    if in(sym, paren_quoted_syms)
+        print(io, ":($sym)")
+    else
+        print(io, ":$sym")
     end
 end
 function default_show_quoted(io::IO, ex, indent::Int)
@@ -225,20 +302,20 @@ function show_unquoted(io::IO, ex::Expr, indent::Int)
             show_unquoted(io, args[2], indent + indent_width)
             print(io, ')')
         end                  
-    elseif (has(_expr_infix, head) && nargs==2) || (is(head,:(:)) && nargs==3)
+    elseif (in(head, _expr_infix) && nargs==2) || (is(head,:(:)) && nargs==3)
         show_list(io, args, head, indent)
-    elseif has(_expr_infix_wide, head) && nargs == 2
+    elseif in(head, _expr_infix_wide) && nargs == 2
         show_list(io, args, " $head ", indent)
     elseif is(head, symbol("::")) && nargs == 1
         print(io, "::")        
         show_unquoted(io, args[1], indent)
-    elseif has(_expr_parens, head)               # :tuple/:vcat/:cell1d
+    elseif haskey(_expr_parens, head)               # :tuple/:vcat/:cell1d
         op, cl = _expr_parens[head]
         print(io, op)
         show_list(io, args, ",", indent)
         if is(head, :tuple) && nargs == 1; print(io, ','); end
         print(io, cl)
-    elseif has(_expr_calls, head) && nargs >= 1  # :call/:ref/:curly
+    elseif haskey(_expr_calls, head) && nargs >= 1  # :call/:ref/:curly
         op, cl = _expr_calls[head]
         show_unquoted(io, args[1], indent)
         show_enclosed_list(io, op, args[2:end], ",", cl, indent)
@@ -247,8 +324,8 @@ function show_unquoted(io::IO, ex::Expr, indent::Int)
     elseif is(head, :(...)) && nargs == 1
         show_unquoted(io, args[1], indent)
         print(io, "...")
-    elseif (nargs == 1 && contains((:return, :abstract, :const), head)) ||
-                          contains((:local,  :global), head)
+    elseif (nargs == 1 && in(head, (:return, :abstract, :const))) ||
+                          in(head, (:local,  :global))
         print(io, head, ' ')
         show_list(io, args, ",", indent)
     elseif is(head, :macrocall) && nargs >= 1
@@ -272,7 +349,7 @@ function show_unquoted(io::IO, ex::Expr, indent::Int)
         show_block(io, "let", args[2:end], args[1], indent); print(io, "end")
     elseif is(head, :block) || is(head, :body)
         show_block(io, "begin", ex, indent); print(io, "end")
-    elseif contains((:for,:while,:function,:if,:type,:module),head) && nargs==2
+    elseif in(head,(:for,:while,:function,:if,:type,:module)) && nargs==2
         show_block(io, head, args[1], args[2], indent); print(io, "end")
     elseif is(head, :quote) && nargs == 1
         show_indented(io, args[1], indent)
@@ -283,8 +360,12 @@ function show_unquoted(io::IO, ex::Expr, indent::Int)
         show(io, args[1])
     elseif is(head, :null)
         print(io, "nothing")
+    elseif is(head, :kw)
+        show_unquoted(io, args[1], indent+indent_width)
+        print(io, '=')
+        show_unquoted(io, args[2], indent+indent_width)
     else
-        print(io, "\$(expr(")
+        print(io, "\$(Expr(")
         show_indented(io, ex.head, indent)
         for arg in args
             print(io, ", ")
@@ -307,6 +388,29 @@ function show_unquoted(io::IO, ex::SymbolNode)
     show_expr_type(io, ex.typ)
 end
 
+function clean_gensym(s::Symbol)
+    s = string(s)
+    i = search(s,'#')
+    if i > 0
+        return s[1:i-1]
+    end
+    return s
+end
+
+function argtype_decl_string(n, t)
+    if isa(n,Expr)
+        n = n.args[1]  # handle n::T in arg list
+    end
+    n = clean_gensym(n)
+    if t === Any && !isempty(n)
+        return n
+    end
+    if t <: Vararg && t !== None && t.parameters[1] === Any
+        return string(n, "...")
+    end
+    return string(n, "::", t)
+end
+
 function show(io::IO, m::Method)
     tv = m.tvars
     if !isa(tv,Tuple)
@@ -315,27 +419,51 @@ function show(io::IO, m::Method)
     if !isempty(tv)
         show_delim_array(io, tv, '{', ',', '}', false)
     end
-    show(io, m.sig)
     li = m.func.code
+    e = uncompressed_ast(li)
+    argnames = e.args[1]
+    if length(argnames) != length(m.sig)
+        s = symbol("?")
+        decls = map(argtype_decl_string, { s for i=1:length(m.sig) }, {m.sig...})
+    else
+        decls = map(argtype_decl_string, argnames, {m.sig...})
+    end
+    print(io, "(")
+    print_joined(io, decls, ",", ",")
+    print(io, ")")
     if li.line > 0
         print(io, " at ", li.file, ":", li.line)
     end
 end
 
-function show(io::IO, mt::MethodTable)
+function show_method_table(io::IO, mt::MethodTable, max::Int=-1, header=true)
     name = mt.name
-    println(io, "# methods for generic function ", name)
+    n = length(mt)
+    if header
+        m = n==1 ? "method" : "methods"
+        print(io, "# $n $m for generic function \"$name\":")
+    end
     d = mt.defs
+    n = rest = 0
     while !is(d,())
-        print(io, name)
-        show(io, d)
-        d = d.next
-        if !is(d,())
+        if max==-1 || n<max || (rest==0 && n==max && d.next === ())
             println(io)
+            print(io, name)
+            show(io, d)
+            n += 1
+        else
+            rest += 1
         end
+        d = d.next
+    end
+    if rest > 0
+        println(io)
+        print(io,"... $rest methods not shown (use methods($name) to see them all)")
     end
 end
 
+show(io::IO, mt::MethodTable) = show_method_table(io, mt)
+
 # dump & xdump - structured tree representation like R's str()
 # - dump is for the user-facing structure
 # - xdump is for the internal structure
@@ -355,13 +483,17 @@ end
 function xdump(fn::Function, io::IO, x, n::Int, indent)
     T = typeof(x)
     print(io, T, " ")
-    if isa(T, CompositeKind)
+    if isa(T, DataType) && length(T.names) > 0
         println(io)
         if n > 0
             for field in T.names
                 if field != symbol("")    # prevents segfault if symbol is blank
                     print(io, indent, "  ", field, ": ")
-                    fn(io, getfield(x, field), n - 1, string(indent, "  "))
+                    if isdefined(x,field)
+                        fn(io, getfield(x, field), n - 1, string(indent, "  "))
+                    else
+                        println(io, undef_ref_str)
+                    end
                 end
             end
         end
@@ -373,35 +505,41 @@ function xdump(fn::Function, io::IO, x::Module, n::Int, indent)
     print(io, Module, " ")
     println(io, x)
 end
+function xdump_elts(fn::Function, io::IO, x::Array{Any}, n::Int, indent, i0, i1)
+    for i in i0:i1
+        print(io, indent, "  ", i, ": ")
+        if !isdefined(x,i)
+            println(undef_ref_str)
+        else
+            fn(io, x[i], n - 1, string(indent, "  "))
+        end
+    end
+end
 function xdump(fn::Function, io::IO, x::Array{Any}, n::Int, indent)
     println("Array($(eltype(x)),$(size(x)))")
     if n > 0
-        for i in 1:(length(x) <= 10 ? length(x) : 5)
-            print(io, indent, "  ", i, ": ")
-            fn(io, x[i], n - 1, string(indent, "  "))
-        end
+        xdump_elts(fn, io, x, n, indent, 1, (length(x) <= 10 ? length(x) : 5))
         if length(x) > 10
             println(io, indent, "  ...")
-            for i in length(x)-4:length(x)
-                print(io, indent, "  ", i, ": ")
-                fn(io, x[i], n - 1, string(indent, "  "))
-            end
+            xdump_elts(fn, io, x, n, indent, length(x)-4, length(x))
         end
     end
 end
 xdump(fn::Function, io::IO, x::Symbol, n::Int, indent) = println(io, typeof(x), " ", x)
 xdump(fn::Function, io::IO, x::Function, n::Int, indent) = println(io, x)
-xdump(fn::Function, io::IO, x::Array, n::Int, indent) = println(io, "Array($(eltype(x)),$(size(x)))", " ", x)
+xdump(fn::Function, io::IO, x::Array, n::Int, indent) =
+               (print(io, "Array($(eltype(x)),$(size(x))) ");
+                show(io, x); println(io))
 
 # Types
-xdump(fn::Function, io::IO, x::UnionKind, n::Int, indent) = println(io, x)
-function xdump(fn::Function, io::IO, x::CompositeKind, n::Int, indent)
+xdump(fn::Function, io::IO, x::UnionType, n::Int, indent) = println(io, x)
+function xdump(fn::Function, io::IO, x::DataType, n::Int, indent)
     println(io, x, "::", typeof(x), " ", " <: ", super(x))
     if n > 0
         for idx in 1:min(10,length(x.names))
             if x.names[idx] != symbol("")    # prevents segfault if symbol is blank
                 print(io, indent, "  ", x.names[idx], "::")
-                if isa(x.types[idx], CompositeKind) 
+                if isa(x.types[idx], DataType)
                     xdump(fn, io, x.types[idx], n - 1, string(indent, "  "))
                 else
                     println(x.types[idx])
@@ -438,17 +576,17 @@ function dumptype(io::IO, x, n::Int, indent)
                                 length(t.parameters) > 0 ? "{$targs}" : "",
                                 " = ", t)
                     end
-                elseif isa(t, UnionKind)
-                    if any(map(tt -> string(x.name) == typargs(tt), t.types))
+                elseif isa(t, UnionType)
+                    if any(tt -> string(x.name) == typargs(tt), t.types)
                         println(io, indent, "  ", s, " = ", t)
                     end
-                elseif isa(t, Type) && super(t).name == x.name
+                elseif isa(t, DataType) && super(t).name == x.name
                     # type aliases
                     if string(s) != string(t.name)
                         println(io, indent, "  ", s, " = ", t.name)
                     elseif t != Any 
                         print(io, indent, "  ")
-                        dumptype(io, t, n - 1, string(indent, "  "))
+                        dump(io, t, n - 1, string(indent, "  "))
                     end
                 end
             end
@@ -458,23 +596,26 @@ end
 
 # For abstract types, use _dumptype only if it's a form that will be called
 # interactively.
-xdump(fn::Function, io::IO, x::AbstractKind) = dumptype(io, x, 5, "")
-xdump(fn::Function, io::IO, x::AbstractKind, n::Int) = dumptype(io, x, n, "")
+xdump(fn::Function, io::IO, x::DataType) = x.abstract ? dumptype(io, x, 5, "") : xdump(fn, io, x, 5, "")
+xdump(fn::Function, io::IO, x::DataType, n::Int) = x.abstract ? dumptype(io, x, n, "") : xdump(fn, io, x, n, "")
 
 # defaults:
 xdump(fn::Function, io::IO, x) = xdump(xdump, io, x, 5, "")  # default is 5 levels
 xdump(fn::Function, io::IO, x, n::Int) = xdump(xdump, io, x, n, "")
-xdump(fn::Function, args...) = xdump(fn, OUTPUT_STREAM::IO, args...)
+xdump(fn::Function, io::IO, args...) = error("invalid arguments to xdump")
+xdump(fn::Function, args...) = xdump(fn, STDOUT::IO, args...)
 xdump(io::IO, args...) = xdump(xdump, io, args...)
-xdump(args...) = xdump(xdump, OUTPUT_STREAM::IO, args...)
-
+xdump(args...) = with_output_limit(()->xdump(xdump, STDOUT::IO, args...), true)
 
 # Here are methods specifically for dump:
 dump(io::IO, x, n::Int) = dump(io, x, n, "")
 dump(io::IO, x) = dump(io, x, 5, "")  # default is 5 levels
-dump(args...) = dump(OUTPUT_STREAM::IO, args...)
-dump(io::IO, x::String, n::Int, indent) = println(io, typeof(x), " \"", x, "\"")
+dump(io::IO, x::String, n::Int, indent) =
+               (print(io, typeof(x), " ");
+                show(io, x); println(io))
 dump(io::IO, x, n::Int, indent) = xdump(dump, io, x, n, indent)
+dump(io::IO, args...) = error("invalid arguments to dump")
+dump(args...) = with_output_limit(()->dump(STDOUT::IO, args...), true)
 
 function dump(io::IO, x::Dict, n::Int, indent)
     println(typeof(x), " len ", length(x))
@@ -493,24 +634,12 @@ function dump(io::IO, x::Dict, n::Int, indent)
 end
 
 # More generic representation for common types:
-dump(io::IO, x::AbstractKind, n::Int, indent) = println(io, x.name)
-dump(io::IO, x::AbstractKind) = dumptype(io, x, 5, "")
-dump(io::IO, x::AbstractKind, n::Int) = dumptype(io, x, n, "")
-dump(io::IO, x::BitsKind, n::Int, indent) = println(io, x.name)
+dump(io::IO, x::DataType, n::Int, indent) = println(io, x.name)
+dump(io::IO, x::DataType, n::Int) = dump(io, x, n, "")
+dump(io::IO, x::DataType) = dump(io, x, 5, "")
 dump(io::IO, x::TypeVar, n::Int, indent) = println(io, x.name)
 
 
-showall(x) = showall(OUTPUT_STREAM::IO, x)
-
-function showall{T}(io::IO, a::AbstractArray{T,1})
-    if is(T,Any)
-        opn = '{'; cls = '}'
-    else
-        opn = '['; cls = ']';
-    end
-    show_comma_array(io, a, opn, cls)
-end
-
 alignment(x::Any) = (0, length(sprint(showcompact, x)))
 alignment(x::Number) = (length(sprint(showcompact, x)), 0)
 alignment(x::Integer) = (length(sprint(showcompact, x)), 0)
@@ -534,7 +663,7 @@ const undef_ref_str = "#undef"
 const undef_ref_alignment = (3,3)
 
 function alignment(
-    X::AbstractMatrix,
+    X::Union(AbstractMatrix,AbstractVector),
     rows::AbstractVector, cols::AbstractVector,
     cols_if_complete::Integer, cols_otherwise::Integer, sep::Integer
 )
@@ -564,8 +693,8 @@ function alignment(
     return a
 end
 
-function print_matrix_row(io,
-    X::AbstractMatrix, A::Vector,
+function print_matrix_row(io::IO,
+    X::Union(AbstractMatrix,AbstractVector), A::Vector,
     i::Integer, cols::AbstractVector, sep::String
 )
     for k = 1:length(A)
@@ -585,7 +714,7 @@ function print_matrix_row(io,
     end
 end
 
-function print_matrix_vdots(io,
+function print_matrix_vdots(io::IO,
     vdots::String, A::Vector, sep::String, M::Integer, m::Integer
 )
     for k = 1:length(A)
@@ -601,8 +730,8 @@ function print_matrix_vdots(io,
     end
 end
 
-function print_matrix(io,
-    X::AbstractMatrix, rows::Integer, cols::Integer,
+function print_matrix(io::IO,
+    X::Union(AbstractMatrix,AbstractVector), rows::Integer, cols::Integer,
     pre::String, sep::String, post::String,
     hdots::String, vdots::String, ddots::String,
     hmod::Integer, vmod::Integer
@@ -612,7 +741,7 @@ function print_matrix(io,
     postsp = ""
     @assert strwidth(hdots) == strwidth(ddots)
     ss = length(sep)
-    m, n = size(X)
+    m, n = size(X,1), size(X,2)
     if m <= rows # rows fit
         A = alignment(X,1:m,1:n,cols,cols,ss)
         if n <= length(A) # rows and cols fit
@@ -676,11 +805,13 @@ function print_matrix(io,
         end
     end
 end
-print_matrix(io, X::AbstractMatrix, rows::Integer, cols::Integer) =
+print_matrix(io::IO, X::Union(AbstractMatrix,AbstractVector),
+             rows::Integer, cols::Integer) =
     print_matrix(io, X, rows, cols, " ", "  ", "",
                  "  \u2026  ", "\u22ee", "  \u22f1  ", 5, 5)
 
-print_matrix(io, X::AbstractMatrix) = print_matrix(io, X, tty_rows()-4, tty_cols())
+print_matrix(io::IO, X::Union(AbstractMatrix,AbstractVector)) =
+               print_matrix(io, X, tty_rows()-4, tty_cols())
 
 summary(x) = string(typeof(x))
 
@@ -688,32 +819,34 @@ dims2string(d) = length(d) == 0 ? "0-dimensional" :
                  length(d) == 1 ? "$(d[1])-element" :
                  join(map(string,d), 'x')
 
-summary{T}(a::AbstractArray{T}) =
-    string(dims2string(size(a)), " ", T, " ", typeof(a).name)
+summary(a::AbstractArray) =
+    string(dims2string(size(a)), " ", typeof(a))
 
-function show_nd(io, a::AbstractArray)
+function show_nd(io::IO, a::AbstractArray, limit, rows, cols)
     if isempty(a)
         return
     end
     tail = size(a)[3:]
     nd = ndims(a)-2
-    function print_slice(io, idxs...)
-        for i = 1:nd
-            ii = idxs[i]
-            if size(a,i+2) > 10
-                if ii == 4 && all(x->x==1,idxs[1:i-1])
-                    for j=i+1:nd
-                        szj = size(a,j+2)
-                        if szj>10 && 3 < idxs[j] <= szj-3
-                            return
+    function print_slice(idxs...)
+        if limit
+            for i = 1:nd
+                ii = idxs[i]
+                if size(a,i+2) > 10
+                    if ii == 4 && all(x->x==1,idxs[1:i-1])
+                        for j=i+1:nd
+                            szj = size(a,j+2)
+                            if szj>10 && 3 < idxs[j] <= szj-3
+                                return
+                            end
                         end
+                        #println(io, idxs)
+                        print(io, "...\n\n")
+                        return
+                    end
+                    if 3 < ii <= size(a,i+2)-3
+                        return
                     end
-                    #println(io, idxs)
-                    print(io, "...\n\n")
-                    return
-                end
-                if 3 < ii <= size(a,i+2)-3
-                    return
                 end
             end
         end
@@ -721,15 +854,15 @@ function show_nd(io, a::AbstractArray)
         for i = 1:(nd-1); print(io, "$(idxs[i]), "); end
         println(io, idxs[end], "] =")
         slice = sub(a, 1:size(a,1), 1:size(a,2), idxs...)
-        print_matrix(io, slice)
+        print_matrix(io, slice, rows, cols)
         print(io, idxs == tail ? "" : "\n\n")
     end
-    cartesian_map((idxs...)->print_slice(io,idxs...), tail)
+    cartesianmap(print_slice, tail)
 end
 
 function whos(m::Module, pattern::Regex)
-    for s in sort!(map(string, names(m)))
-        v = symbol(s)
+    for v in sort(names(m))
+        s = string(v)
         if isdefined(m,v) && ismatch(pattern, s)
             println(rpad(s, 30), summary(eval(m,v)))
         end
@@ -737,7 +870,7 @@ function whos(m::Module, pattern::Regex)
 end
 whos() = whos(r"")
 whos(m::Module) = whos(m, r"")
-whos(pat::Regex) = whos(ccall(:jl_get_current_module, Any, ())::Module, pat)
+whos(pat::Regex) = whos(current_module(), pat)
 
 function show{T}(io::IO, x::AbstractArray{T,0})
     println(io, summary(x),":")
@@ -749,45 +882,79 @@ function show{T}(io::IO, x::AbstractArray{T,0})
     print(io, sx)
 end
 
-function show(io::IO, X::AbstractArray)
-    print(io, summary(X))
+# global flag for limiting output
+# TODO: this should be replaced with a better mechanism. currently it is only
+# for internal use in showing arrays.
+_limit_output = false
+
+# NOTE: this is a possible, so-far-unexported function, providing control of
+# array output. Not sure I want to do it this way.
+showarray(X::AbstractArray; kw...) = showarray(STDOUT, X; kw...)
+function showarray(io::IO, X::AbstractArray;
+                   header::Bool=true, limit::Bool=_limit_output,
+                   rows = tty_rows()-4, cols = tty_cols())
+    header && print(io, summary(X))
     if !isempty(X)
-        println(io, ":")
-        ndims(X)==2 ? print_matrix(io, X) : show_nd(io, X)
+        header && println(io, ":")
+        if ndims(X) == 0
+            return showcompact(io, X[])
+        end
+        if !limit
+            rows = cols = typemax(Int)
+        end
+        if ndims(X)<=2
+            print_matrix(io, X, rows, cols)
+        else
+            show_nd(io, X, limit, rows, cols)
+        end
     end
 end
 
-function showall(io::IO, X::AbstractMatrix)
-    print(io, summary(X))
-    if !isempty(X)
-        println(io, ":")
-        print_matrix(io, X, typemax(Int64), typemax(Int64))
+show(io::IO, X::AbstractArray) = showarray(io, X)
+
+print(io::IO, X::AbstractArray) = writedlm(io, X)
+
+function with_output_limit(thk, lim=true)
+    global _limit_output
+    last = _limit_output
+    _limit_output = lim
+    try
+        thk()
+    finally
+        _limit_output = last
     end
 end
 
-function showall(io::IO, a::AbstractArray)
-    print(io, summary(a))
-    if isempty(a)
-        return
+showall(x) = showall(STDOUT, x)
+function showall(io::IO, x)
+    if _limit_output==false
+        show(io, x)
+    else
+        with_output_limit(false) do
+            show(io, x)
+        end
     end
-    println(io, ":")
-    tail = size(a)[3:]
-    nd = ndims(a)-2
-    function print_slice(io, idxs...)
-        print(io, "[:, :, ")
-        for i = 1:(nd-1); print(io, "$(idxs[i]), "); end
-        println(io, idxs[end], "] =")
-        slice = a[:,:,idxs...]
-        print_matrix(io, reshape(slice, size(slice,1), size(slice,2)),
-                     typemax(Int64), typemax(Int64))
-        print(io, idxs == tail ? "" : "\n\n")
-    end
-    cartesian_map(print_slice, tail)
 end
 
-function show_vector(io, v, opn, cls)
-    X = reshape(v,(1,length(v)))
-    print_matrix(io, X, 1, tty_cols(), opn, ", ", cls, "  \u2026  ", "\u22ee", "  \u22f1  ", 5, 5)
+showlimited(x) = showlimited(STDOUT, x)
+function showlimited(io::IO, x)
+    if _limit_output==true
+        show(io, x)
+    else
+        with_output_limit(true) do
+            show(io, x)
+        end
+    end
+end
+
+function show_vector(io::IO, v, opn, cls)
+    if _limit_output && length(v) > 20
+        show_delim_array(io, v[1:10], opn, ",", "", false)
+        print(io, "  \u2026  ")
+        show_delim_array(io, v[end-9:end], "", ",", cls, false)
+    else
+        show_delim_array(io, v, opn, ",", cls, false)
+    end
 end
 
 show(io::IO, v::AbstractVector{Any}) = show_vector(io, v, "{", "}")
@@ -795,9 +962,6 @@ show(io::IO, v::AbstractVector)      = show_vector(io, v, "[", "]")
 
 # printing BitArrays
 
-summary(a::BitArray) =
-    string(dims2string(size(a)), " ", typeof(a).name)
-
 # (following functions not exported - mainly intended for debug)
 
 function print_bit_chunk(io::IO, c::Uint64, l::Integer)
@@ -812,8 +976,8 @@ end
 
 print_bit_chunk(io::IO, c::Uint64) = print_bit_chunk(io, c, 64)
 
-print_bit_chunk(c::Uint64, l::Integer) = print_bit_chunk(OUTPUT_STREAM, c, l)
-print_bit_chunk(c::Uint64) = print_bit_chunk(OUTPUT_STREAM, c)
+print_bit_chunk(c::Uint64, l::Integer) = print_bit_chunk(STDOUT, c, l)
+print_bit_chunk(c::Uint64) = print_bit_chunk(STDOUT, c)
 
 function bitshow(io::IO, B::BitArray)
     if length(B) == 0
@@ -826,6 +990,6 @@ function bitshow(io::IO, B::BitArray)
     l = (@_mod64 (length(B)-1)) + 1
     print_bit_chunk(io, B.chunks[end], l)
 end
-bitshow(B::BitArray) = bitshow(OUTPUT_STREAM, B)
+bitshow(B::BitArray) = bitshow(STDOUT, B)
 
 bitstring(B::BitArray) = sprint(bitshow, B)
