@@ -45,13 +45,10 @@ extern "C" {
 # define UV_EXTERN /* nothing */
 #endif
 
-
-#define UV_VERSION_MAJOR 0
-#define UV_VERSION_MINOR 9
-
+#include "uv-errno.h"
 
 #if defined(_MSC_VER) && _MSC_VER < 1600
-# include "uv-private/stdint-msvc2008.h"
+# include "stdint-msvc2008.h"
 #else
 # include <stdint.h>
 #endif
@@ -64,80 +61,80 @@ extern "C" {
 
 #if defined(__unix__) || defined(__POSIX__) || \
     defined(__APPLE__) || defined(_AIX)
-# include "uv-private/uv-unix.h"
+# include "uv-unix.h"
 #else
-# include "uv-private/uv-win.h"
+# include "uv-win.h"
 #endif
 
 /* Expand this list if necessary. */
 #define UV_ERRNO_MAP(XX)                                                      \
-  XX( -1, UNKNOWN, "unknown error")                                           \
-  XX(  0, OK, "success")                                                      \
-  XX(  1, EOF, "end of file")                                                 \
-  XX(  2, EADDRINFO, "getaddrinfo error")                                     \
-  XX(  3, EACCES, "permission denied")                                        \
-  XX(  4, EAGAIN, "resource temporarily unavailable")                         \
-  XX(  5, EADDRINUSE, "address already in use")                               \
-  XX(  6, EADDRNOTAVAIL, "address not available")                             \
-  XX(  7, EAFNOSUPPORT, "address family not supported")                       \
-  XX(  8, EALREADY, "connection already in progress")                         \
-  XX(  9, EBADF, "bad file descriptor")                                       \
-  XX( 10, EBUSY, "resource busy or locked")                                   \
-  XX( 11, ECONNABORTED, "software caused connection abort")                   \
-  XX( 12, ECONNREFUSED, "connection refused")                                 \
-  XX( 13, ECONNRESET, "connection reset by peer")                             \
-  XX( 14, EDESTADDRREQ, "destination address required")                       \
-  XX( 15, EFAULT, "bad address in system call argument")                      \
-  XX( 16, EHOSTUNREACH, "host is unreachable")                                \
-  XX( 17, EINTR, "interrupted system call")                                   \
-  XX( 18, EINVAL, "invalid argument")                                         \
-  XX( 19, EISCONN, "socket is already connected")                             \
-  XX( 20, EMFILE, "too many open files")                                      \
-  XX( 21, EMSGSIZE, "message too long")                                       \
-  XX( 22, ENETDOWN, "network is down")                                        \
-  XX( 23, ENETUNREACH, "network is unreachable")                              \
-  XX( 24, ENFILE, "file table overflow")                                      \
-  XX( 25, ENOBUFS, "no buffer space available")                               \
-  XX( 26, ENOMEM, "not enough memory")                                        \
-  XX( 27, ENOTDIR, "not a directory")                                         \
-  XX( 28, EISDIR, "illegal operation on a directory")                         \
-  XX( 29, ENONET, "machine is not on the network")                            \
-  XX( 31, ENOTCONN, "socket is not connected")                                \
-  XX( 32, ENOTSOCK, "socket operation on non-socket")                         \
-  XX( 33, ENOTSUP, "operation not supported on socket")                       \
-  XX( 34, ENOENT, "no such file or directory")                                \
-  XX( 35, ENOSYS, "function not implemented")                                 \
-  XX( 36, EPIPE, "broken pipe")                                               \
-  XX( 37, EPROTO, "protocol error")                                           \
-  XX( 38, EPROTONOSUPPORT, "protocol not supported")                          \
-  XX( 39, EPROTOTYPE, "protocol wrong type for socket")                       \
-  XX( 40, ETIMEDOUT, "connection timed out")                                  \
-  XX( 41, ECHARSET, "invalid Unicode character")                              \
-  XX( 42, EAIFAMNOSUPPORT, "address family for hostname not supported")       \
-  XX( 44, EAISERVICE, "servname not supported for ai_socktype")               \
-  XX( 45, EAISOCKTYPE, "ai_socktype not supported")                           \
-  XX( 46, ESHUTDOWN, "cannot send after transport endpoint shutdown")         \
-  XX( 47, EEXIST, "file already exists")                                      \
-  XX( 48, ESRCH, "no such process")                                           \
-  XX( 49, ENAMETOOLONG, "name too long")                                      \
-  XX( 50, EPERM, "operation not permitted")                                   \
-  XX( 51, ELOOP, "too many symbolic links encountered")                       \
-  XX( 52, EXDEV, "cross-device link not permitted")                           \
-  XX( 53, ENOTEMPTY, "directory not empty")                                   \
-  XX( 54, ENOSPC, "no space left on device")                                  \
-  XX( 55, EIO, "i/o error")                                                   \
-  XX( 56, EROFS, "read-only file system")                                     \
-  XX( 57, ENODEV, "no such device")                                           \
-  XX( 58, ESPIPE, "invalid seek")                                             \
-  XX( 59, ECANCELED, "operation canceled")                                    \
-
-
-#define UV_ERRNO_GEN(val, name, s) UV_##name = val,
-typedef enum {
-  UV_ERRNO_MAP(UV_ERRNO_GEN)
-  UV_MAX_ERRORS
-} uv_err_code;
-#undef UV_ERRNO_GEN
+  XX(E2BIG, "argument list too long")                                         \
+  XX(EACCES, "permission denied")                                             \
+  XX(EADDRINUSE, "address already in use")                                    \
+  XX(EADDRNOTAVAIL, "address not available")                                  \
+  XX(EAFNOSUPPORT, "address family not supported")                            \
+  XX(EAGAIN, "resource temporarily unavailable")                              \
+  XX(EAI_ADDRFAMILY, "address family not supported")                          \
+  XX(EAI_AGAIN, "temporary failure")                                          \
+  XX(EAI_BADFLAGS, "bad ai_flags value")                                      \
+  XX(EAI_CANCELED, "request canceled")                                        \
+  XX(EAI_FAIL, "permanent failure")                                           \
+  XX(EAI_FAMILY, "ai_family not supported")                                   \
+  XX(EAI_MEMORY, "out of memory")                                             \
+  XX(EAI_NODATA, "no address")                                                \
+  XX(EAI_NONAME, "unknown node or service")                                   \
+  XX(EAI_SERVICE, "service not available for socket type")                    \
+  XX(EAI_SOCKTYPE, "socket type not supported")                               \
+  XX(EAI_SYSTEM, "system error")                                              \
+  XX(EALREADY, "connection already in progress")                              \
+  XX(EBADF, "bad file descriptor")                                            \
+  XX(EBUSY, "resource busy or locked")                                        \
+  XX(ECANCELED, "operation canceled")                                         \
+  XX(ECHARSET, "invalid Unicode character")                                   \
+  XX(ECONNABORTED, "software caused connection abort")                        \
+  XX(ECONNREFUSED, "connection refused")                                      \
+  XX(ECONNRESET, "connection reset by peer")                                  \
+  XX(EDESTADDRREQ, "destination address required")                            \
+  XX(EEXIST, "file already exists")                                           \
+  XX(EFAULT, "bad address in system call argument")                           \
+  XX(EHOSTUNREACH, "host is unreachable")                                     \
+  XX(EINTR, "interrupted system call")                                        \
+  XX(EINVAL, "invalid argument")                                              \
+  XX(EIO, "i/o error")                                                        \
+  XX(EISCONN, "socket is already connected")                                  \
+  XX(EISDIR, "illegal operation on a directory")                              \
+  XX(ELOOP, "too many symbolic links encountered")                            \
+  XX(EMFILE, "too many open files")                                           \
+  XX(EMSGSIZE, "message too long")                                            \
+  XX(ENAMETOOLONG, "name too long")                                           \
+  XX(ENETDOWN, "network is down")                                             \
+  XX(ENETUNREACH, "network is unreachable")                                   \
+  XX(ENFILE, "file table overflow")                                           \
+  XX(ENOBUFS, "no buffer space available")                                    \
+  XX(ENODEV, "no such device")                                                \
+  XX(ENOENT, "no such file or directory")                                     \
+  XX(ENOMEM, "not enough memory")                                             \
+  XX(ENONET, "machine is not on the network")                                 \
+  XX(ENOSPC, "no space left on device")                                       \
+  XX(ENOSYS, "function not implemented")                                      \
+  XX(ENOTCONN, "socket is not connected")                                     \
+  XX(ENOTDIR, "not a directory")                                              \
+  XX(ENOTEMPTY, "directory not empty")                                        \
+  XX(ENOTSOCK, "socket operation on non-socket")                              \
+  XX(ENOTSUP, "operation not supported on socket")                            \
+  XX(EPERM, "operation not permitted")                                        \
+  XX(EPIPE, "broken pipe")                                                    \
+  XX(EPROTO, "protocol error")                                                \
+  XX(EPROTONOSUPPORT, "protocol not supported")                               \
+  XX(EPROTOTYPE, "protocol wrong type for socket")                            \
+  XX(EROFS, "read-only file system")                                          \
+  XX(ESHUTDOWN, "cannot send after transport endpoint shutdown")              \
+  XX(ESPIPE, "invalid seek")                                                  \
+  XX(ESRCH, "no such process")                                                \
+  XX(ETIMEDOUT, "connection timed out")                                       \
+  XX(EXDEV, "cross-device link not permitted")                                \
+  XX(UNKNOWN, "unknown error")                                                \
+  XX(EOF, "end of file")                                                      \
 
 #define UV_HANDLE_TYPE_MAP(XX)                                                \
   XX(ASYNC, async)                                                            \
@@ -167,6 +164,13 @@ typedef enum {
   XX(WORK, work)                                                              \
   XX(GETADDRINFO, getaddrinfo)                                                \
 
+typedef enum {
+#define XX(code, _) UV_ ## code = UV__ ## code,
+  UV_ERRNO_MAP(XX)
+#undef XX
+  UV_ERRNO_MAX = UV__EOF - 1
+} uv_errno_t;
+
 typedef enum {
   UV_UNKNOWN_HANDLE = 0,
 #define XX(uc, lc) UV_##uc,
@@ -190,7 +194,6 @@ typedef enum {
 
 /* Handle types. */
 typedef struct uv_loop_s uv_loop_t;
-typedef struct uv_err_s uv_err_t;
 typedef struct uv_handle_s uv_handle_t;
 typedef struct uv_stream_s uv_stream_t;
 typedef struct uv_tcp_s uv_tcp_t;
@@ -230,6 +233,20 @@ typedef enum {
 } uv_run_mode;
 
 
+/*
+ * Returns the libuv version packed into a single integer. 8 bits are used for
+ * each component, with the patch number stored in the 8 least significant
+ * bits. E.g. for libuv 1.2.3 this would return 0x010203.
+ */
+UV_EXTERN unsigned int uv_version(void);
+
+/*
+ * Returns the libuv version number as a string. For non-release versions
+ * "-pre" is appended, so the version number could be "1.2.3-pre".
+ */
+UV_EXTERN const char* uv_version_string(void);
+
+
 /*
  * This function must be called before any other functions in libuv.
  *
@@ -274,16 +291,38 @@ UV_EXTERN void uv_stop(uv_loop_t*);
  */
 UV_EXTERN void uv_ref(uv_handle_t*);
 UV_EXTERN void uv_unref(uv_handle_t*);
+UV_EXTERN int uv_has_ref(const uv_handle_t*);
 
+/*
+ * Update the event loop's concept of "now". Libuv caches the current time
+ * at the start of the event loop tick in order to reduce the number of
+ * time-related system calls.
+ *
+ * You won't normally need to call this function unless you have callbacks
+ * that block the event loop for longer periods of time, where "longer" is
+ * somewhat subjective but probably on the order of a millisecond or more.
+ */
 UV_EXTERN void uv_update_time(uv_loop_t*);
+
+/*
+ * Return the current timestamp in milliseconds. The timestamp is cached at
+ * the start of the event loop tick, see |uv_update_time()| for details and
+ * rationale.
+ *
+ * The timestamp increases monotonically from some arbitrary point in time.
+ * Don't make assumptions about the starting point, you will only get
+ * disappointed.
+ *
+ * Use uv_hrtime() if you need sub-milliseond granularity.
+ */
 UV_EXTERN uint64_t uv_now(uv_loop_t*);
 
 /*
  * Get backend file descriptor. Only kqueue, epoll and event ports are
  * supported.
  *
- * This can be used in conjunction with uv_run_once() to poll in one thread and
- * run the event loop's event callbacks in another.
+ * This can be used in conjunction with `uv_run(loop, UV_RUN_NOWAIT)` to
+ * poll in one thread and run the event loop's event callbacks in another.
  *
  * Useful for embedding libuv's event loop in another event loop.
  * See test/test-embed.c for an example.
@@ -306,20 +345,25 @@ UV_EXTERN int uv_backend_timeout(const uv_loop_t*);
  *
  * `suggested_size` is a hint. Returning a buffer that is smaller is perfectly
  * okay as long as `buf.len > 0`.
+ *
+ * If you return a buffer with `buf.len == 0`, libuv skips the read and calls
+ * your read or recv callback with nread=UV_ENOBUFS.
+ *
+ * Note that returning a zero-length buffer does not stop the handle, call
+ * uv_read_stop() or uv_udp_recv_stop() for that.
  */
 typedef uv_buf_t (*uv_alloc_cb)(uv_handle_t* handle, size_t suggested_size);
 
 /*
- * `nread` is > 0 if there is data available, 0 if libuv is done reading for now
- * or -1 on error.
- *
- * Error details can be obtained by calling uv_last_error(). UV_EOF indicates
- * that the stream has been closed.
+ * `nread` is > 0 if there is data available, 0 if libuv is done reading for
+ * now, or < 0 on error.
  *
  * The callee is responsible for closing the stream when an error happens.
  * Trying to read from the stream again is undefined.
  *
  * The callee is responsible for freeing the buffer, libuv does not reuse it.
+ * The buffer may be a null buffer (where buf.base=NULL and buf.len=0) on EOF
+ * or error.
  */
 typedef void (*uv_read_cb)(uv_stream_t* stream, ssize_t nread, uv_buf_t buf);
 
@@ -352,6 +396,32 @@ typedef void (*uv_getaddrinfo_cb)(uv_getaddrinfo_t* req,
                                   int status,
                                   struct addrinfo* res);
 
+typedef struct {
+  long tv_sec;
+  long tv_nsec;
+} uv_timespec_t;
+
+
+typedef struct {
+  uint64_t st_dev;
+  uint64_t st_mode;
+  uint64_t st_nlink;
+  uint64_t st_uid;
+  uint64_t st_gid;
+  uint64_t st_rdev;
+  uint64_t st_ino;
+  uint64_t st_size;
+  uint64_t st_blksize;
+  uint64_t st_blocks;
+  uint64_t st_flags;
+  uint64_t st_gen;
+  uv_timespec_t st_atim;
+  uv_timespec_t st_mtim;
+  uv_timespec_t st_ctim;
+  uv_timespec_t st_birthtim;
+} uv_stat_t;
+
+
 /*
 * This will be called repeatedly after the uv_fs_event_t is initialized.
 * If uv_fs_event_t was initialized with a directory the filename parameter
@@ -363,8 +433,8 @@ typedef void (*uv_fs_event_cb)(uv_fs_event_t* handle, const char* filename,
 
 typedef void (*uv_fs_poll_cb)(uv_fs_poll_t* handle,
                               int status,
-                              const uv_statbuf_t* prev,
-                              const uv_statbuf_t* curr);
+                              const uv_stat_t* prev,
+                              const uv_stat_t* curr);
 
 typedef void (*uv_signal_cb)(uv_signal_t* handle, int signum);
 
@@ -375,22 +445,11 @@ typedef enum {
 } uv_membership;
 
 
-struct uv_err_s {
-  /* read-only */
-  uv_err_code code;
-  /* private */
-  int sys_errno_;
-};
-
-
 /*
- * Most functions return boolean: 0 for success and -1 for failure.
- * On error the user should then call uv_last_error() to determine
- * the error code.
+ * Most functions return 0 on success or an error code < 0 on failure.
  */
-UV_EXTERN uv_err_t uv_last_error(uv_loop_t*);
-UV_EXTERN const char* uv_strerror(uv_err_t err);
-UV_EXTERN const char* uv_err_name(uv_err_t err);
+UV_EXTERN const char* uv_strerror(int err);
+UV_EXTERN const char* uv_err_name(int err);
 
 
 #define UV_REQ_FIELDS                                                         \
@@ -399,7 +458,7 @@ UV_EXTERN const char* uv_err_name(uv_err_t err);
   /* read-only */                                                             \
   uv_req_type type;                                                           \
   /* private */                                                               \
-  ngx_queue_t active_queue;                                                   \
+  void* active_queue[2];                                                      \
   UV_REQ_PRIVATE_FIELDS                                                       \
 
 /* Abstract base class of all requests. */
@@ -439,7 +498,7 @@ struct uv_shutdown_s {
   uv_close_cb close_cb;                                                       \
   void* data;                                                                 \
   /* private */                                                               \
-  ngx_queue_t handle_queue;                                                   \
+  void* handle_queue[2];                                                      \
   UV_HANDLE_PRIVATE_FIELDS                                                    \
 
 /* The abstract base class of all handles.  */
@@ -480,8 +539,7 @@ UV_EXTERN void uv_walk(uv_loop_t* loop, uv_walk_cb walk_cb, void* arg);
  * It gives you a chance to free up any resources associated with the handle.
  *
  * In-progress requests, like uv_connect_t or uv_write_t, are cancelled and
- * have their callbacks called asynchronously with status=-1 and the error code
- * set to UV_ECANCELED.
+ * have their callbacks called asynchronously with status=UV_ECANCELED.
  */
 UV_EXTERN void uv_close(uv_handle_t* handle, uv_close_cb close_cb);
 
@@ -549,10 +607,12 @@ UV_EXTERN int uv_accept(uv_stream_t* server, uv_stream_t* client);
 
 /*
  * Read data from an incoming stream. The callback will be made several
- * several times until there is no more data to read or uv_read_stop is
- * called. When we've reached EOF nread will be set to -1 and the error is
- * set to UV_EOF. When nread == -1 the buf parameter might not point to a
- * valid buffer; in that case buf.len and buf.base are both set to 0.
+ * times until there is no more data to read or uv_read_stop is called.
+ * When we've reached EOF nread will be set to UV_EOF.
+ *
+ * When nread < 0, the buf parameter might not point to a valid buffer;
+ * in that case buf.len and buf.base are both set to 0.
+ *
  * Note that nread might also be 0, which does *not* indicate an error or
  * eof; it happens when libuv requested a buffer through the alloc callback
  * but then decided that it didn't need that buffer.
@@ -621,6 +681,31 @@ UV_EXTERN int uv_is_readable(const uv_stream_t* handle);
 UV_EXTERN int uv_is_writable(const uv_stream_t* handle);
 
 
+/*
+ * Enable or disable blocking mode for a stream.
+ *
+ * When blocking mode is enabled all writes complete synchronously. The
+ * interface remains unchanged otherwise, e.g. completion or failure of the
+ * operation will still be reported through a callback which is made
+ * asychronously.
+ *
+ * Relying too much on this API is not recommended. It is likely to change
+ * significantly in the future.
+ *
+ * On windows this currently works only for uv_pipe_t instances. On unix it
+ * works for tcp, pipe and tty instances. Be aware that changing the blocking
+ * mode on unix sets or clears the O_NONBLOCK bit. If you are sharing a handle
+ * with another process, the other process is affected by the change too,
+ * which can lead to unexpected results.
+ *
+ * Also libuv currently makes no ordering guarantee when the blocking mode
+ * is changed after write requests have already been submitted. Therefore it is
+ * recommended to set the blocking mode immediately after opening or creating
+ * the stream.
+ */
+UV_EXTERN int uv_stream_set_blocking(uv_stream_t* handle, int blocking);
+
+
 /*
  * Used to determine whether a stream is closing or closed.
  *
@@ -662,12 +747,12 @@ UV_EXTERN int uv_tcp_keepalive(uv_tcp_t* handle,
                                unsigned int delay);
 
 /*
- * This setting applies to Windows only.
  * Enable/disable simultaneous asynchronous accept requests that are
  * queued by the operating system when listening for new tcp connections.
  * This setting is used to tune a tcp server for the desired performance.
  * Having simultaneous accepts can significantly improve the rate of
- * accepting connections (which is why it is enabled by default).
+ * accepting connections (which is why it is enabled by default) but
+ * may lead to uneven load distribution in multi-process setups.
  */
 UV_EXTERN int uv_tcp_simultaneous_accepts(uv_tcp_t* handle, int enable);
 
@@ -725,7 +810,7 @@ typedef void (*uv_udp_send_cb)(uv_udp_send_t* req, int status);
  *  nread   Number of bytes that have been received.
  *          0 if there is no more data to read. You may
  *          discard or repurpose the read buffer.
- *          -1 if a transmission error was detected.
+ *          < 0 if a transmission error was detected.
  *  buf     uv_buf_t with the received data.
  *  addr    struct sockaddr_in or struct sockaddr_in6.
  *          Valid for the duration of the callback only.
@@ -757,6 +842,12 @@ UV_EXTERN int uv_udp_init(uv_loop_t*, uv_udp_t* handle);
 
 /*
  * Opens an existing file descriptor or SOCKET as a udp handle.
+ *
+ * Unix only:
+ *  The only requirement of the sock argument is that it follows the
+ *  datagram contract (works in unconnected mode, supports sendmsg()/recvmsg(),
+ *  etc.). In other words, other datagram-type sockets like raw sockets or
+ *  netlink sockets can also be passed to this function.
  */
 UV_EXTERN int uv_udp_open(uv_udp_t* handle, uv_os_sock_t sock);
 
@@ -769,7 +860,7 @@ UV_EXTERN int uv_udp_open(uv_udp_t* handle, uv_os_sock_t sock);
  *  flags     Unused.
  *
  * Returns:
- *  0 on success, -1 on error.
+ *  0 on success, or an error code < 0 on failure.
  */
 UV_EXTERN int uv_udp_bind(uv_udp_t* handle, struct sockaddr_in addr,
     unsigned flags);
@@ -783,10 +874,11 @@ UV_EXTERN int uv_udp_bind(uv_udp_t* handle, struct sockaddr_in addr,
  *  flags     Should be 0 or UV_UDP_IPV6ONLY.
  *
  * Returns:
- *  0 on success, -1 on error.
+ *  0 on success, or an error code < 0 on failure.
  */
 UV_EXTERN int uv_udp_bind6(uv_udp_t* handle, struct sockaddr_in6 addr,
     unsigned flags);
+
 UV_EXTERN int uv_udp_getsockname(uv_udp_t* handle, struct sockaddr* name,
     int* namelen);
 
@@ -801,7 +893,7 @@ UV_EXTERN int uv_udp_getsockname(uv_udp_t* handle, struct sockaddr* name,
  *  membership          Should be UV_JOIN_GROUP or UV_LEAVE_GROUP
  *
  * Returns:
- *  0 on success, -1 on error.
+ *  0 on success, or an error code < 0 on failure.
  */
 UV_EXTERN int uv_udp_set_membership(uv_udp_t* handle,
     const char* multicast_addr, const char* interface_addr,
@@ -817,7 +909,7 @@ UV_EXTERN int uv_udp_set_membership(uv_udp_t* handle,
  *  on                  1 for on, 0 for off
  *
  * Returns:
- *  0 on success, -1 on error.
+ *  0 on success, or an error code < 0 on failure.
  */
 UV_EXTERN int uv_udp_set_multicast_loop(uv_udp_t* handle, int on);
 
@@ -830,7 +922,7 @@ UV_EXTERN int uv_udp_set_multicast_loop(uv_udp_t* handle, int on);
  *  ttl                 1 through 255
  *
  * Returns:
- *  0 on success, -1 on error.
+ *  0 on success, or an error code < 0 on failure.
  */
 UV_EXTERN int uv_udp_set_multicast_ttl(uv_udp_t* handle, int ttl);
 
@@ -843,7 +935,7 @@ UV_EXTERN int uv_udp_set_multicast_ttl(uv_udp_t* handle, int ttl);
  *  on                  1 for on, 0 for off
  *
  * Returns:
- *  0 on success, -1 on error.
+ *  0 on success, or an error code < 0 on failure.
  */
 UV_EXTERN int uv_udp_set_broadcast(uv_udp_t* handle, int on);
 
@@ -856,7 +948,7 @@ UV_EXTERN int uv_udp_set_broadcast(uv_udp_t* handle, int on);
  *  ttl                 1 through 255
  *
  * Returns:
- *  0 on success, -1 on error.
+ *  0 on success, or an error code < 0 on failure.
  */
 UV_EXTERN int uv_udp_set_ttl(uv_udp_t* handle, int ttl);
 
@@ -874,7 +966,7 @@ UV_EXTERN int uv_udp_set_ttl(uv_udp_t* handle, int ttl);
  *  send_cb   Callback to invoke when the data has been sent out.
  *
  * Returns:
- *  0 on success, -1 on error.
+ *  0 on success, or an error code < 0 on failure.
  */
 UV_EXTERN int uv_udp_send(uv_udp_send_t* req, uv_udp_t* handle,
     uv_buf_t bufs[], int bufcnt, struct sockaddr_in addr,
@@ -893,7 +985,7 @@ UV_EXTERN int uv_udp_send(uv_udp_send_t* req, uv_udp_t* handle,
  *  send_cb   Callback to invoke when the data has been sent out.
  *
  * Returns:
- *  0 on success, -1 on error.
+ *  0 on success, or an error code < 0 on failure.
  */
 UV_EXTERN int uv_udp_send6(uv_udp_send_t* req, uv_udp_t* handle,
     uv_buf_t bufs[], int bufcnt, struct sockaddr_in6 addr,
@@ -910,7 +1002,7 @@ UV_EXTERN int uv_udp_send6(uv_udp_send_t* req, uv_udp_t* handle,
  *  recv_cb   Callback to invoke with received data.
  *
  * Returns:
- *  0 on success, -1 on error.
+ *  0 on success, or an error code < 0 on failure.
  */
 UV_EXTERN int uv_udp_recv_start(uv_udp_t* handle, uv_alloc_cb alloc_cb,
     uv_udp_recv_cb recv_cb);
@@ -922,7 +1014,7 @@ UV_EXTERN int uv_udp_recv_start(uv_udp_t* handle, uv_alloc_cb alloc_cb,
  *  handle    UDP handle. Should have been initialized with `uv_udp_init`.
  *
  * Returns:
- *  0 on success, -1 on error.
+ *  0 on success, or an error code < 0 on failure.
  */
 UV_EXTERN int uv_udp_recv_stop(uv_udp_t* handle);
 
@@ -994,7 +1086,7 @@ enum uv_pipe_flags {
 	UV_PIPE_IPC			 = 0x01,
 	UV_PIPE_SPAWN_SAFE	 = 0x02,
 	UV_PIPE_READABLE	 = 0x04,
-	UV_PIPE_WRITEABLE	 = 0x08
+	UV_PIPE_WRITABLE	 = 0x08
 };
 /*
  * Initialize a pipe. The last argument is a boolean to indicate if
@@ -1086,10 +1178,10 @@ UV_EXTERN int uv_poll_init_socket(uv_loop_t* loop, uv_poll_t* handle,
  * will be called with `status` set to 0, and the detected events set en the
  * `events` field.
  *
- * If an error happens while polling status may be set to -1 and the error
- * code can be retrieved with uv_last_error. The user should not close the
- * socket while uv_poll is active. If the user does that anyway, the callback
- * *may* be called reporting an error status, but this is not guaranteed.
+ * If an error happens while polling status, `status` < 0 and corresponds
+ * with one of the UV_E* error codes. The user should not close the socket
+ * while uv_poll is active. If the user does that anyway, the callback *may*
+ * be called reporting an error status, but this is not guaranteed.
  *
  * Calling uv_poll_start on an uv_poll watcher that is already active is fine.
  * Doing so will update the events mask that is being watched for.
@@ -1210,8 +1302,8 @@ UV_EXTERN int uv_timer_stop(uv_timer_t* handle);
 
 /*
  * Stop the timer, and if it is repeating restart it using the repeat value
- * as the timeout. If the timer has never been started before it returns -1 and
- * sets the error to UV_EINVAL.
+ * as the timeout. If the timer has never been started before it returns
+ * UV_EINVAL.
  */
 UV_EXTERN int uv_timer_again(uv_timer_t* handle);
 
@@ -1247,15 +1339,13 @@ struct uv_getaddrinfo_s {
  * hints is a pointer to a struct addrinfo with additional address type
  * constraints, or NULL. Consult `man -s 3 getaddrinfo` for details.
  *
- * Returns 0 on success, -1 on error. Call uv_last_error() to get the error.
+ * Returns 0 on success or an error code < 0 on failure.
  *
  * If successful, your callback gets called sometime in the future with the
  * lookup result, which is either:
  *
- *  a) status == 0, the res argument points to a valid struct addrinfo, or
- *  b) status == -1, the res argument is NULL.
- *
- * On NXDOMAIN, the status code is -1 and uv_last_error() returns UV_ENOENT.
+ *  a) err == 0, the res argument points to a valid struct addrinfo, or
+ *  b) err < 0, the res argument is NULL. See the UV_EAI_* constants.
  *
  * Call uv_freeaddrinfo() to free the addrinfo structure.
  */
@@ -1271,6 +1361,7 @@ UV_EXTERN int uv_getaddrinfo(uv_loop_t* loop,
  */
 UV_EXTERN void uv_freeaddrinfo(struct addrinfo* ai);
 
+
 typedef struct uv_stdio_container_s {
 	uv_handle_type type;
 	union {
@@ -1361,7 +1452,12 @@ enum uv_process_flags {
    * option is only meaningful on Windows systems. On unix it is silently
    * ignored.
    */
-  UV_PROCESS_WINDOWS_HIDE = (1 << 4)
+  UV_PROCESS_WINDOWS_HIDE = (1 << 4),
+  /*
+   * Reset SIGPIPE to the default handler. Useful for parent processes that ignore 
+   * SIGPIPE, but don't want to make the same assumption for child processes
+   */
+  UV_PROCESS_RESET_SIGPIPE = (1 << 5)
 };
 
 /*
@@ -1387,7 +1483,7 @@ UV_EXTERN int uv_process_kill(uv_process_t*, int signum);
 
 
 /* Kills the process with the specified signal. */
-UV_EXTERN uv_err_t uv_kill(int pid, int signum);
+UV_EXTERN int uv_kill(int pid, int signum);
 
 
 /*
@@ -1408,7 +1504,7 @@ UV_EXTERN int uv_queue_work(uv_loop_t* loop, uv_work_t* req,
 /* Cancel a pending request. Fails if the request is executing or has finished
  * executing.
  *
- * Returns 0 on success, -1 on error. The loop error code is not touched.
+ * Returns 0 on success, or an error code < 0 on failure.
  *
  * Only cancellation of uv_fs_t, uv_getaddrinfo_t and uv_work_t requests is
  * currently supported.
@@ -1419,13 +1515,13 @@ UV_EXTERN int uv_queue_work(uv_loop_t* loop, uv_work_t* req,
  *
  * Here is how cancellation is reported to your callback:
  *
- * - A uv_fs_t request has its req->errorno field set to UV_ECANCELED.
+ * - A uv_fs_t request has its req->result field set to UV_ECANCELED.
  *
  * - A uv_work_t or uv_getaddrinfo_t request has its callback invoked with
- *   status == -1 and uv_last_error(loop).code == UV_ECANCELED.
+ *   status == UV_ECANCELED.
  *
  * This function is currently only implemented on UNIX platforms. On Windows,
- * it always returns -1.
+ * it always returns UV_ENOSYS.
  */
 UV_EXTERN int uv_cancel(uv_req_t* req);
 
@@ -1444,31 +1540,36 @@ struct uv_cpu_info_s {
 
 struct uv_interface_address_s {
   char* name;
+  char phys_addr[6];
   int is_internal;
   union {
     struct sockaddr_in address4;
     struct sockaddr_in6 address6;
   } address;
+  union {
+    struct sockaddr_in netmask4;
+    struct sockaddr_in6 netmask6;
+  } netmask;
 };
 
 UV_EXTERN char** uv_setup_args(int argc, char** argv);
-UV_EXTERN uv_err_t uv_get_process_title(char* buffer, size_t size);
-UV_EXTERN uv_err_t uv_set_process_title(const char* title);
-UV_EXTERN uv_err_t uv_resident_set_memory(size_t* rss);
-UV_EXTERN uv_err_t uv_uptime(double* uptime);
+UV_EXTERN int uv_get_process_title(char* buffer, size_t size);
+UV_EXTERN int uv_set_process_title(const char* title);
+UV_EXTERN int uv_resident_set_memory(size_t* rss);
+UV_EXTERN int uv_uptime(double* uptime);
 
 /*
  * This allocates cpu_infos array, and sets count.  The array
  * is freed using uv_free_cpu_info().
  */
-UV_EXTERN uv_err_t uv_cpu_info(uv_cpu_info_t** cpu_infos, int* count);
+UV_EXTERN int uv_cpu_info(uv_cpu_info_t** cpu_infos, int* count);
 UV_EXTERN void uv_free_cpu_info(uv_cpu_info_t* cpu_infos, int count);
 
 /*
  * This allocates addresses array, and sets count.  The array
  * is freed using uv_free_interface_addresses().
  */
-UV_EXTERN uv_err_t uv_interface_addresses(uv_interface_address_t** addresses,
+UV_EXTERN int uv_interface_addresses(uv_interface_address_t** addresses,
   int* count);
 UV_EXTERN void uv_free_interface_addresses(uv_interface_address_t* addresses,
   int count);
@@ -1526,8 +1627,7 @@ struct uv_fs_s {
   ssize_t result;
   void* ptr;
   const char* path;
-  uv_err_code errorno;
-  uv_statbuf_t statbuf;  /* Stores the result of uv_fs_stat and uv_fs_fstat. */
+  uv_stat_t statbuf;  /* Stores the result of uv_fs_stat and uv_fs_fstat. */
   UV_FS_PRIVATE_FIELDS
 };
 
@@ -1615,10 +1715,10 @@ UV_EXTERN int uv_fs_fchmod(uv_loop_t* loop, uv_fs_t* req, uv_file file,
     int mode, uv_fs_cb cb);
 
 UV_EXTERN int uv_fs_chown(uv_loop_t* loop, uv_fs_t* req, const char* path,
-    int uid, int gid, uv_fs_cb cb);
+    uv_uid_t uid, uv_gid_t gid, uv_fs_cb cb);
 
 UV_EXTERN int uv_fs_fchown(uv_loop_t* loop, uv_fs_t* req, uv_file file,
-    int uid, int gid, uv_fs_cb cb);
+    uv_uid_t uid, uv_gid_t gid, uv_fs_cb cb);
 
 
 enum uv_fs_event {
@@ -1648,13 +1748,13 @@ UV_EXTERN int uv_fs_poll_init(uv_loop_t* loop, uv_fs_poll_t* handle);
 /*
  * Check the file at `path` for changes every `interval` milliseconds.
  *
- * Your callback i invoked with `status == -1` if `path` does not exist
+ * Your callback is invoked with `status < 0` if `path` does not exist
  * or is inaccessible. The watcher is *not* stopped but your callback is
  * not called again until something changes (e.g. when the file is created
  * or the error reason changes).
  *
  * When `status == 0`, your callback receives pointers to the old and new
- * `uv_statbuf_t` structs. They are valid for the duration of the callback
+ * `uv_stat_t` structs. They are valid for the duration of the callback
  * only!
  *
  * For maximum portability, use multi-second intervals. Sub-second intervals
@@ -1720,9 +1820,9 @@ UV_EXTERN int uv_signal_stop(uv_signal_t* handle);
 
 
 /*
- * Gets load avg
+ * Gets load average.
  * See: http://en.wikipedia.org/wiki/Load_(computing)
- * (Returns [0,0,0] for windows and cygwin)
+ * Returns [0,0,0] on Windows.
  */
 UV_EXTERN void uv_loadavg(double avg[3]);
 
@@ -1772,19 +1872,19 @@ UV_EXTERN int uv_ip4_name(struct sockaddr_in* src, char* dst, size_t size);
 UV_EXTERN int uv_ip6_name(struct sockaddr_in6* src, char* dst, size_t size);
 
 /* Cross-platform IPv6-capable implementation of the 'standard' inet_ntop */
-/* and inet_pton functions. On success they return UV_OK. If an error */
+/* and inet_pton functions. On success they return 0. If an error */
 /* the target of the `dst` pointer is unmodified. */
-UV_EXTERN uv_err_t uv_inet_ntop(int af, const void* src, char* dst, size_t size);
-UV_EXTERN uv_err_t uv_inet_pton(int af, const char* src, void* dst);
+UV_EXTERN int uv_inet_ntop(int af, const void* src, char* dst, size_t size);
+UV_EXTERN int uv_inet_pton(int af, const char* src, void* dst);
 
 /* Gets the executable path */
 UV_EXTERN int uv_exepath(char* buffer, size_t* size);
 
 /* Gets the current working directory */
-UV_EXTERN uv_err_t uv_cwd(char* buffer, size_t size);
+UV_EXTERN int uv_cwd(char* buffer, size_t size);
 
 /* Changes the current working directory */
-UV_EXTERN uv_err_t uv_chdir(const char* dir);
+UV_EXTERN int uv_chdir(const char* dir);
 
 /* Gets memory info in bytes */
 UV_EXTERN uint64_t uv_get_free_memory(void);
@@ -1813,8 +1913,6 @@ UV_EXTERN extern uint64_t uv_hrtime(void);
  * Note that this function works on a best-effort basis: there is no guarantee
  * that libuv can discover all file descriptors that were inherited. In general
  * it does a better job on Windows than it does on unix.
- *
- * TODO(bb): insert snarky remark to annoy bnoordhuis and the folks at joyent.
  */
 UV_EXTERN void uv_disable_stdio_inheritance(void);
 
@@ -1841,7 +1939,7 @@ UV_EXTERN int uv_dlsym(uv_lib_t* lib, const char* name, void** ptr);
 UV_EXTERN const char* uv_dlerror(uv_lib_t* lib);
 
 /*
- * The mutex functions return 0 on success, -1 on error
+ * The mutex functions return 0 on success or an error code < 0
  * (unless the return type is void, of course).
  */
 UV_EXTERN int uv_mutex_init(uv_mutex_t* handle);
@@ -1885,7 +1983,8 @@ UV_EXTERN void uv_cond_broadcast(uv_cond_t* cond);
  */
 UV_EXTERN void uv_cond_wait(uv_cond_t* cond, uv_mutex_t* mutex);
 /* Waits on a condition variable with a timeout in nano seconds.
- * Returns 0 for success or -1 on timeout, * aborts when other errors happen.
+ * Returns 0 for success or UV_ETIMEDOUT on timeout, It aborts when other
+ * errors happen.
  *
  * Note:
  * 1. callers should be prepared to deal with spurious wakeups.
@@ -1910,46 +2009,25 @@ UV_EXTERN int uv_thread_create(uv_thread_t *tid,
 UV_EXTERN unsigned long uv_thread_self(void);
 UV_EXTERN int uv_thread_join(uv_thread_t *tid);
 
-/* the presence of these unions force similar struct layout */
+/* The presence of these unions force similar struct layout. */
+#define XX(_, name) uv_ ## name ## _t name;
 union uv_any_handle {
-  uv_handle_t handle;
-  uv_stream_t stream;
-  uv_tcp_t tcp;
-  uv_pipe_t pipe;
-  uv_prepare_t prepare;
-  uv_check_t check;
-  uv_idle_t idle;
-  uv_async_t async;
-  uv_timer_t timer;
-  uv_fs_event_t fs_event;
-  uv_fs_poll_t fs_poll;
-  uv_poll_t poll;
-  uv_process_t process;
-  uv_tty_t tty;
-  uv_udp_t udp;
+  UV_HANDLE_TYPE_MAP(XX)
 };
 
 union uv_any_req {
-  uv_req_t req;
-  uv_write_t write;
-  uv_connect_t connect;
-  uv_shutdown_t shutdown;
-  uv_fs_t fs_req;
-  uv_work_t work_req;
-  uv_udp_send_t udp_send_req;
-  uv_getaddrinfo_t getaddrinfo_req;
+  UV_REQ_TYPE_MAP(XX)
 };
+#undef XX
 
 
 struct uv_loop_s {
   /* User data - use this for whatever. */
   void* data;
-  /* The last error */
-  uv_err_t last_err;
   /* Loop reference counting */
   unsigned int active_handles;
-  ngx_queue_t handle_queue;
-  ngx_queue_t active_reqs;
+  void* handle_queue[2];
+  void* active_reqs[2];
   /* Internal flag to signal loop stop */
   unsigned int stop_flag;
   UV_LOOP_PRIVATE_FIELDS
