@@ -41,7 +41,7 @@ int uv_pipe_init(uv_loop_t* loop, uv_pipe_t* handle, int flags) {
   handle->flags |= ((flags&UV_PIPE_IPC)?UV_HANDLE_PIPE_IPC:0)|
           ((flags&UV_PIPE_SPAWN_SAFE)?UV_HANDLE_PIPE_SPAWN_SAFE:0)|
           ((flags&UV_PIPE_READABLE)?UV_STREAM_READABLE:0)|
-          ((flags&UV_PIPE_WRITEABLE)?UV_STREAM_WRITABLE:0);
+          ((flags&UV_PIPE_WRITABLE)?UV_STREAM_WRITABLE:0);
   /*if (flags == 0 || flags == 1)
       handle->flags |= UV_STREAM_READABLE|UV_STREAM_WRITABLE;*/
   return 0;
@@ -51,44 +51,43 @@ int uv_pipe_init(uv_loop_t* loop, uv_pipe_t* handle, int flags) {
 int uv_pipe_bind(uv_pipe_t* handle, const char* name) {
   struct sockaddr_un saddr;
   const char* pipe_fname;
-  int saved_errno;
   int sockfd;
-  int status;
   int bound;
+  int err;
 
-  saved_errno = errno;
   pipe_fname = NULL;
   sockfd = -1;
-  status = -1;
   bound = 0;
+  err = -EINVAL;
 
   /* Already bound? */
-  if (uv__stream_fd(handle) >= 0) {
-    uv__set_artificial_error(handle->loop, UV_EINVAL);
-    goto out;
-  }
+  if (uv__stream_fd(handle) >= 0)
+    return -EINVAL;
 
   /* Make a copy of the file name, it outlives this function's scope. */
-  if ((pipe_fname = strdup(name)) == NULL) {
-    uv__set_sys_error(handle->loop, ENOMEM);
+  pipe_fname = strdup(name);
+  if (pipe_fname == NULL) {
+    err = -ENOMEM;
     goto out;
   }
 
   /* We've got a copy, don't touch the original any more. */
   name = NULL;
 
-  if ((sockfd = uv__socket(AF_UNIX, SOCK_STREAM, 0)) == -1) {
-    uv__set_sys_error(handle->loop, errno);
+  err = uv__socket(AF_UNIX, SOCK_STREAM, 0);
+  if (err < 0)
     goto out;
-  }
+  sockfd = err;
 
   memset(&saddr, 0, sizeof saddr);
   uv_strlcpy(saddr.sun_path, pipe_fname, sizeof(saddr.sun_path));
   saddr.sun_family = AF_UNIX;
 
   if (bind(sockfd, (struct sockaddr*)&saddr, sizeof saddr)) {
+    err = -errno;
     /* Convert ENOENT to EACCES for compatibility with Windows. */
-    uv__set_sys_error(handle->loop, (errno == ENOENT) ? EACCES : errno);
+    if (err == -ENOENT)
+      err = -EACCES;
     goto out;
   }
   bound = 1;
@@ -96,74 +95,49 @@ int uv_pipe_bind(uv_pipe_t* handle, const char* name) {
   /* Success. */
   handle->pipe_fname = pipe_fname; /* Is a strdup'ed copy. */
   handle->io_watcher.fd = sockfd;
-  status = 0;
+  return 0;
 
 out:
-  /* Clean up on error. */
-  if (status) {
-    if (bound) {
-      /* unlink() before close() to avoid races. */
-      assert(pipe_fname != NULL);
-      unlink(pipe_fname);
-    }
-    close(sockfd);
-
-    free((void*)pipe_fname);
+  if (bound) {
+    /* unlink() before close() to avoid races. */
+    assert(pipe_fname != NULL);
+    unlink(pipe_fname);
   }
-
-  errno = saved_errno;
-  return status;
+  close(sockfd);
+  free((void*)pipe_fname);
+  return err;
 }
 
 int uv_pipe_link(uv_pipe_t *read, uv_pipe_t *write) {
-  int err;
   int fds[2];
+  int err;
+  int flags;
 
   assert(read->loop==write->loop);
   assert(read->flags&UV_STREAM_READABLE);
   assert(write->flags&UV_STREAM_WRITABLE);
   assert(!(write->flags&read->flags&UV_HANDLE_PIPE_IPC));
 
-#ifdef SOCK_NONBLOCK
-  int fl;
+  flags = ((read->flags & UV_PIPE_SPAWN_SAFE)==(write->flags & UV_PIPE_SPAWN_SAFE) && 
+          (write->flags & UV_PIPE_SPAWN_SAFE) != 0) ? UV__F_NONBLOCK : 0;
 
-  fl = SOCK_CLOEXEC;
+  uv__make_pipe(fds,flags);
 
-  if (~((read->flags|write->flags)&UV_PIPE_SPAWN_SAFE)) {
-    if (socketpair(AF_UNIX, SOCK_STREAM|SOCK_NONBLOCK|fl, 0, fds) == 0)
-        goto open_fds;
-
-    if (errno != EINVAL)
-      goto pipe_error;
-    /* errno == EINVAL so maybe the kernel headers lied about
-     * the availability of SOCK_NONBLOCK. This can happen if people
-     * build libuv against newer kernel headers than the kernel
-     * they actually run the software on.
-     */
+  if(flags == 0) {
+    if (~(read->flags & UV_PIPE_SPAWN_SAFE))
+      uv__nonblock(fds[0], 1);
+    if (~(write->flags & UV_PIPE_SPAWN_SAFE))
+      uv__nonblock(fds[1], 1);
   }
-#endif
 
-  if (socketpair(AF_UNIX, SOCK_STREAM, 0, fds))
-    goto pipe_error;
-
-  uv__cloexec(fds[0], 1);
-  uv__cloexec(fds[1], 1);
-
-  if (~(read->flags & UV_PIPE_SPAWN_SAFE))
-    uv__nonblock(fds[0], 1);
-  if (~(write->flags & UV_PIPE_SPAWN_SAFE))
-    uv__nonblock(fds[1], 1);
-
-open_fds: 
-  
-  err = uv__stream_open((uv_stream_t*)read, fds[0], 0);
+  err = uv__stream_open((uv_stream_t*)read, fds[0], UV_STREAM_READABLE);
   if (err) {
       close(fds[0]);
       close(fds[1]);
       goto pipe_error;
   }
 
-  err = uv__stream_open((uv_stream_t*)write, fds[1], 0);
+  err = uv__stream_open((uv_stream_t*)write, fds[1], UV_STREAM_WRITABLE);
   if (err) {
       uv_pipe_close_sync(read);
       close(fds[0]);
@@ -185,29 +159,16 @@ void uv_pipe_close_sync(uv_pipe_t *pipe) {
 }
 
 int uv_pipe_listen(uv_pipe_t* handle, int backlog, uv_connection_cb cb) {
-  int saved_errno;
-  int status;
+  if (uv__stream_fd(handle) == -1)
+    return -EINVAL;
 
-  saved_errno = errno;
-  status = -1;
+  if (listen(uv__stream_fd(handle), backlog))
+    return -errno;
 
-  if (uv__stream_fd(handle) == -1) {
-    uv__set_artificial_error(handle->loop, UV_EINVAL);
-    goto out;
-  }
-  assert(uv__stream_fd(handle) >= 0);
-
-  if ((status = listen(uv__stream_fd(handle), backlog)) == -1) {
-    uv__set_sys_error(handle->loop, errno);
-  } else {
-    handle->connection_cb = cb;
-    handle->io_watcher.cb = uv__pipe_accept;
-    uv__io_start(handle->loop, &handle->io_watcher, UV__POLLIN);
-  }
-
-out:
-  errno = saved_errno;
-  return status;
+  handle->connection_cb = cb;
+  handle->io_watcher.cb = uv__pipe_accept;
+  uv__io_start(handle->loop, &handle->io_watcher, UV__POLLIN);
+  return 0;
 }
 
 
@@ -229,6 +190,14 @@ void uv__pipe_close(uv_pipe_t* handle) {
 
 
 int uv_pipe_open(uv_pipe_t* handle, uv_file fd) {
+#if defined(__APPLE__)
+  int err;
+
+  err = uv__stream_try_select((uv_stream_t*) handle, &fd);
+  if (err)
+    return err;
+#endif /* defined(__APPLE__) */
+
   return uv__stream_open((uv_stream_t*)handle,
                          fd,
                          0);
@@ -240,18 +209,19 @@ void uv_pipe_connect(uv_connect_t* req,
                     const char* name,
                     uv_connect_cb cb) {
   struct sockaddr_un saddr;
-  int saved_errno;
   int new_sock;
   int err;
   int r;
 
-  saved_errno = errno;
   new_sock = (uv__stream_fd(handle) == -1);
-  err = -1;
+  err = -EINVAL;
 
-  if (new_sock)
-    if ((handle->io_watcher.fd = uv__socket(AF_UNIX, SOCK_STREAM, 0)) == -1)
+  if (new_sock) {
+    err = uv__socket(AF_UNIX, SOCK_STREAM, 0);
+    if (err < 0)
       goto out;
+    handle->io_watcher.fd = err;
+  }
 
   memset(&saddr, 0, sizeof saddr);
   uv_strlcpy(saddr.sun_path, name, sizeof(saddr.sun_path));
@@ -263,66 +233,61 @@ void uv_pipe_connect(uv_connect_t* req,
   }
   while (r == -1 && errno == EINTR);
 
-  if (r == -1)
-    if (errno != EINPROGRESS)
-      goto out;
+  if (r == -1 && errno != EINPROGRESS) {
+    err = -errno;
+    goto out;
+  }
 
-  if (new_sock)
-    if (uv__stream_open((uv_stream_t*)handle,
-                        uv__stream_fd(handle),
-                        UV_STREAM_READABLE | UV_STREAM_WRITABLE))
-      goto out;
-
-  uv__io_start(handle->loop, &handle->io_watcher, UV__POLLIN | UV__POLLOUT);
   err = 0;
+  if (new_sock) {
+    err = uv__stream_open((uv_stream_t*)handle,
+                          uv__stream_fd(handle),
+                          UV_STREAM_READABLE | UV_STREAM_WRITABLE);
+  }
+
+  if (err == 0)
+    uv__io_start(handle->loop, &handle->io_watcher, UV__POLLIN | UV__POLLOUT);
 
 out:
-  handle->delayed_error = err ? errno : 0; /* Passed to callback. */
+  handle->delayed_error = err;
   handle->connect_req = req;
 
   uv__req_init(handle->loop, req, UV_CONNECT);
   req->handle = (uv_stream_t*)handle;
   req->cb = cb;
-  ngx_queue_init(&req->queue);
+  QUEUE_INIT(&req->queue);
 
   /* Force callback to run on next tick in case of error. */
-  if (err != 0)
+  if (err)
     uv__io_feed(handle->loop, &handle->io_watcher);
 
   /* Mimic the Windows pipe implementation, always
    * return 0 and let the callback handle errors.
    */
-  errno = saved_errno;
 }
 
 
 /* TODO merge with uv__server_io()? */
 static void uv__pipe_accept(uv_loop_t* loop, uv__io_t* w, unsigned int events) {
   uv_pipe_t* pipe;
-  int saved_errno;
   int sockfd;
 
-  saved_errno = errno;
   pipe = container_of(w, uv_pipe_t, io_watcher);
-
   assert(pipe->type == UV_NAMED_PIPE);
 
   sockfd = uv__accept(uv__stream_fd(pipe));
   if (sockfd == -1) {
-    if (errno != EAGAIN && errno != EWOULDBLOCK) {
-      uv__set_sys_error(pipe->loop, errno);
-      pipe->connection_cb((uv_stream_t*)pipe, -1);
-    }
-  } else {
-    pipe->accepted_fd = sockfd;
-    pipe->connection_cb((uv_stream_t*)pipe, 0);
-    if (pipe->accepted_fd == sockfd) {
-      /* The user hasn't called uv_accept() yet */
-      uv__io_stop(pipe->loop, &pipe->io_watcher, UV__POLLIN);
-    }
+    if (errno != EAGAIN && errno != EWOULDBLOCK)
+      pipe->connection_cb((uv_stream_t*)pipe, -errno);
+    return;
   }
 
-  errno = saved_errno;
+  pipe->accepted_fd = sockfd;
+  pipe->connection_cb((uv_stream_t*)pipe, 0);
+  if (pipe->accepted_fd == sockfd) {
+    /* The user hasn't called uv_accept() yet */
+    uv__io_stop(pipe->loop, &pipe->io_watcher, UV__POLLIN);
+  }
 }
 
 
