@@ -41,7 +41,7 @@ extern char **environ;
 #endif
 
 
-static ngx_queue_t* uv__process_queue(uv_loop_t* loop, int pid) {
+static QUEUE* uv__process_queue(uv_loop_t* loop, int pid) {
   assert(pid > 0);
   return loop->process_handles + pid % ARRAY_SIZE(loop->process_handles);
 }
@@ -49,13 +49,13 @@ static ngx_queue_t* uv__process_queue(uv_loop_t* loop, int pid) {
 
 static uv_process_t* uv__process_find(uv_loop_t* loop, int pid) {
   uv_process_t* handle;
-  ngx_queue_t* h;
-  ngx_queue_t* q;
+  QUEUE* h;
+  QUEUE* q;
 
   h = uv__process_queue(loop, pid);
 
-  ngx_queue_foreach(q, h) {
-    handle = ngx_queue_data(q, uv_process_t, queue);
+  QUEUE_FOREACH(q, h) {
+    handle = QUEUE_DATA(q, uv_process_t, queue);
     if (handle->pid == pid) return handle;
   }
 
@@ -103,10 +103,8 @@ static void uv__chld(uv_signal_t* handle, int signum) {
     if (WIFSIGNALED(status))
       term_signal = WTERMSIG(status);
 
-    if (process->errorno) {
-      uv__set_sys_error(process->loop, process->errorno);
-      exit_status = -1; /* execve() failed */
-    }
+    if (process->errorno)
+      exit_status = process->errorno; /* execve() failed */
 
     process->exit_cb(process, exit_status, term_signal);
   }
@@ -124,7 +122,7 @@ int uv__make_pipe(int fds[2], int flags) {
     return 0;
 
   if (errno != ENOSYS)
-    return -1;
+    return -errno;
 
   no_pipe2 = 1;
 
@@ -132,7 +130,7 @@ skip:
 #endif
 
   if (pipe(fds))
-    return -1;
+    return -errno;
 
   uv__cloexec(fds[0], 1);
   uv__cloexec(fds[1], 1);
@@ -167,8 +165,8 @@ static int uv__process_init_stdio(uv_stdio_container_t* container, int *fd) {
     default:
       assert (0 && "Unexpected flags");
       *fd = -1;
+    return -EINVAL;
       errno = EINVAL;
-      return -1;
   }
   if (*fd == -1) {
     errno = EINVAL;
@@ -215,8 +213,8 @@ static void uv__process_child_init(uv_process_options_t options,
       use_fd = open("/dev/null", fd == 0 ? O_RDONLY : O_RDWR);
 
       if (use_fd == -1) {
-        uv__write_int(error_fd, errno);
-        perror("failed to open stdio");
+        uv__write_int(error_fd, -errno);
+        /* perror("failed to open stdio"); */
         _exit(127);
       }
     }
@@ -233,20 +231,27 @@ static void uv__process_child_init(uv_process_options_t options,
   }
 
   if (options.cwd && chdir(options.cwd)) {
-    uv__write_int(error_fd, errno);
-    perror("chdir()");
+    uv__write_int(error_fd, -errno);
+    /* perror("chdir()"); */
     _exit(127);
   }
 
   if ((options.flags & UV_PROCESS_SETGID) && setgid(options.gid)) {
-    uv__write_int(error_fd, errno);
-    perror("setgid()");
+    uv__write_int(error_fd, -errno);
+    /* perror("setgid()"); */
     _exit(127);
   }
 
   if ((options.flags & UV_PROCESS_SETUID) && setuid(options.uid)) {
+    uv__write_int(error_fd, -errno);
+    /* perror("setuid()"); */
+    _exit(127);
+  }
+
+  if ((options.flags & UV_PROCESS_RESET_SIGPIPE) && signal(SIGPIPE,SIG_DFL) == SIG_ERR)
+  {
     uv__write_int(error_fd, errno);
-    perror("setuid()");
+    /* perror("signal()"); */
     _exit(127);
   }
 
@@ -255,21 +260,22 @@ static void uv__process_child_init(uv_process_options_t options,
   }
 
   execvp(options.file, options.args);
-  uv__write_int(error_fd, errno);
-  perror("execvp()");
+  uv__write_int(error_fd, -errno);
+  /* perror("execvp()"); */
   _exit(127);
 }
 
-
+#include <signal.h>
 int uv_spawn(uv_loop_t* loop,
              uv_process_t* process,
              const uv_process_options_t options) {
   int signal_pipe[2] = { -1, -1 };
   int *pipes;
   int stdio_count;
-  ngx_queue_t* q;
+  QUEUE* q;
   ssize_t r;
   pid_t pid;
+  int err;
   int i;
 
   assert(options.file != NULL);
@@ -277,28 +283,34 @@ int uv_spawn(uv_loop_t* loop,
                              UV_PROCESS_SETGID |
                              UV_PROCESS_SETUID |
                              UV_PROCESS_WINDOWS_HIDE |
-                             UV_PROCESS_WINDOWS_VERBATIM_ARGUMENTS)));
+                             UV_PROCESS_WINDOWS_VERBATIM_ARGUMENTS |
+                             UV_PROCESS_RESET_SIGPIPE)));
+
+  sigset_t sigset, sigoset;
+  sigfillset(&sigset);
+  sigprocmask(SIG_SETMASK, &sigset, &sigoset);
 
   uv__handle_init(loop, (uv_handle_t*)process, UV_PROCESS);
-  ngx_queue_init(&process->queue);
+  QUEUE_INIT(&process->queue);
 
   stdio_count = options.stdio_count;
   if (stdio_count < 3)
     stdio_count = 3;
 
+  err = -ENOMEM;
   pipes = malloc(stdio_count * sizeof(*pipes));
-  if (pipes == NULL) {
-    errno = ENOMEM;
+  if (pipes == NULL)
     goto error;
-  }
 
   for (i = 0; i < stdio_count; i++) {
     pipes[i] = -1;
   }
 
-  for (i = 0; i < options.stdio_count; i++)
-    if (uv__process_init_stdio(options.stdio + i, &pipes[i]))
+  for (i = 0; i < options.stdio_count; i++) {
+    err = uv__process_init_stdio(options.stdio + i, &pipes[i]);
+    if (err)
       goto error;
+  }
 
   /* This pipe is used by the parent to wait until
    * the child has called `execve()`. We need this
@@ -320,20 +332,23 @@ int uv_spawn(uv_loop_t* loop,
    * marked close-on-exec. Then, after the call to `fork()`,
    * the parent polls the read end until it EOFs or errors with EPIPE.
    */
-  if (uv__make_pipe(signal_pipe, 0))
+  err = uv__make_pipe(signal_pipe, 0);
+  if (err)
     goto error;
 
   uv_signal_start(&loop->child_watcher, uv__chld, SIGCHLD);
-
+  
   pid = fork();
 
   if (pid == -1) {
+    err = -errno;
     close(signal_pipe[0]);
     close(signal_pipe[1]);
     goto error;
   }
 
   if (pid == 0) {
+    sigprocmask(SIG_SETMASK, &sigoset, NULL);
     uv__process_child_init(options, stdio_count, pipes, signal_pipe[1]);
     abort();
   }
@@ -357,53 +372,43 @@ int uv_spawn(uv_loop_t* loop,
   close(signal_pipe[0]);
 
   q = uv__process_queue(loop, pid);
-  ngx_queue_insert_tail(q, &process->queue);
+  QUEUE_INSERT_TAIL(q, &process->queue);
 
   process->pid = pid;
   process->exit_cb = options.exit_cb;
   uv__handle_start(process);
 
   free(pipes);
+  sigprocmask(SIG_SETMASK, &sigoset, NULL);
   return 0;
 
 error:
-  uv__set_sys_error(process->loop, errno);
-
   for (i = 0; i < stdio_count; i++) {
     if (options.stdio[i].type == UV_STREAM && options.stdio[i].data.stream == NULL)
       close(pipes[i]);
   }
   free(pipes);
+  sigprocmask(SIG_SETMASK, &sigoset, NULL);
 
-  return -1;
+  return err;
 }
 
 
 int uv_process_kill(uv_process_t* process, int signum) {
-  int r = kill(process->pid, signum);
-
-  if (r) {
-    uv__set_sys_error(process->loop, errno);
-    return -1;
-  } else {
-    return 0;
-  }
+  return uv_kill(process->pid, signum);
 }
 
 
-uv_err_t uv_kill(int pid, int signum) {
-  int r = kill(pid, signum);
-
-  if (r) {
-    return uv__new_sys_error(errno);
-  } else {
-    return uv_ok_;
-  }
+int uv_kill(int pid, int signum) {
+  if (kill(pid, signum))
+    return -errno;
+  else
+    return 0;
 }
 
 
 void uv__process_close(uv_process_t* handle) {
   /* TODO stop signal watcher when this is the last handle */
-  ngx_queue_remove(&handle->queue);
+  QUEUE_REMOVE(&handle->queue);
   uv__handle_stop(handle);
 }
