@@ -35,12 +35,17 @@ static void uv__udp_io(uv_loop_t* loop, uv__io_t* w, unsigned int revents);
 static void uv__udp_recvmsg(uv_loop_t* loop, uv__io_t* w, unsigned int revents);
 static void uv__udp_sendmsg(uv_loop_t* loop, uv__io_t* w, unsigned int revents);
 static int uv__udp_maybe_deferred_bind(uv_udp_t* handle, int domain);
-static int uv__udp_send(uv_udp_send_t* req, uv_udp_t* handle, uv_buf_t bufs[],
-    int bufcnt, struct sockaddr* addr, socklen_t addrlen, uv_udp_send_cb send_cb);
+static int uv__send(uv_udp_send_t* req,
+                    uv_udp_t* handle,
+                    uv_buf_t bufs[],
+                    int bufcnt,
+                    struct sockaddr* addr,
+                    socklen_t addrlen,
+                    uv_udp_send_cb send_cb);
 
 
 void uv__udp_close(uv_udp_t* handle) {
-  uv__io_stop(handle->loop, &handle->io_watcher, UV__POLLIN | UV__POLLOUT);
+  uv__io_close(handle->loop, &handle->io_watcher);
   uv__handle_stop(handle);
   close(handle->io_watcher.fd);
   handle->io_watcher.fd = -1;
@@ -49,32 +54,29 @@ void uv__udp_close(uv_udp_t* handle) {
 
 void uv__udp_finish_close(uv_udp_t* handle) {
   uv_udp_send_t* req;
-  ngx_queue_t* q;
+  QUEUE* q;
 
   assert(!uv__io_active(&handle->io_watcher, UV__POLLIN | UV__POLLOUT));
   assert(handle->io_watcher.fd == -1);
 
   uv__udp_run_completed(handle);
 
-  while (!ngx_queue_empty(&handle->write_queue)) {
-    q = ngx_queue_head(&handle->write_queue);
-    ngx_queue_remove(q);
+  while (!QUEUE_EMPTY(&handle->write_queue)) {
+    q = QUEUE_HEAD(&handle->write_queue);
+    QUEUE_REMOVE(q);
 
-    req = ngx_queue_data(q, uv_udp_send_t, queue);
+    req = QUEUE_DATA(q, uv_udp_send_t, queue);
     uv__req_unregister(handle->loop, req);
 
     if (req->bufs != req->bufsml)
       free(req->bufs);
     req->bufs = NULL;
 
-    if (req->send_cb) {
-      uv__set_artificial_error(handle->loop, UV_ECANCELED);
-      req->send_cb(req, -1);
-    }
+    if (req->send_cb != NULL)
+      req->send_cb(req, -ECANCELED);
   }
 
   /* Now tear down the handle. */
-  handle->flags = 0;
   handle->recv_cb = NULL;
   handle->alloc_cb = NULL;
   /* but _do not_ touch close_cb */
@@ -83,15 +85,15 @@ void uv__udp_finish_close(uv_udp_t* handle) {
 
 static void uv__udp_run_pending(uv_udp_t* handle) {
   uv_udp_send_t* req;
-  ngx_queue_t* q;
+  QUEUE* q;
   struct msghdr h;
   ssize_t size;
 
-  while (!ngx_queue_empty(&handle->write_queue)) {
-    q = ngx_queue_head(&handle->write_queue);
+  while (!QUEUE_EMPTY(&handle->write_queue)) {
+    q = QUEUE_HEAD(&handle->write_queue);
     assert(q != NULL);
 
-    req = ngx_queue_data(q, uv_udp_send_t, queue);
+    req = QUEUE_DATA(q, uv_udp_send_t, queue);
     assert(req != NULL);
 
     memset(&h, 0, sizeof h);
@@ -132,21 +134,21 @@ static void uv__udp_run_pending(uv_udp_t* handle) {
      * why we don't handle partial writes. Just pop the request
      * off the write queue and onto the completed queue, done.
      */
-    ngx_queue_remove(&req->queue);
-    ngx_queue_insert_tail(&handle->write_completed_queue, &req->queue);
+    QUEUE_REMOVE(&req->queue);
+    QUEUE_INSERT_TAIL(&handle->write_completed_queue, &req->queue);
   }
 }
 
 
 static void uv__udp_run_completed(uv_udp_t* handle) {
   uv_udp_send_t* req;
-  ngx_queue_t* q;
+  QUEUE* q;
 
-  while (!ngx_queue_empty(&handle->write_completed_queue)) {
-    q = ngx_queue_head(&handle->write_completed_queue);
-    ngx_queue_remove(q);
+  while (!QUEUE_EMPTY(&handle->write_completed_queue)) {
+    q = QUEUE_HEAD(&handle->write_completed_queue);
+    QUEUE_REMOVE(q);
 
-    req = ngx_queue_data(q, uv_udp_send_t, queue);
+    req = QUEUE_DATA(q, uv_udp_send_t, queue);
     uv__req_unregister(handle->loop, req);
 
     if (req->bufs != req->bufsml)
@@ -159,13 +161,10 @@ static void uv__udp_run_completed(uv_udp_t* handle) {
     /* req->status >= 0 == bytes written
      * req->status <  0 == errno
      */
-    if (req->status >= 0) {
+    if (req->status >= 0)
       req->send_cb(req, 0);
-    }
-    else {
-      uv__set_sys_error(handle->loop, -req->status);
-      req->send_cb(req, -1);
-    }
+    else
+      req->send_cb(req, req->status);
   }
 }
 
@@ -206,12 +205,15 @@ static void uv__udp_recvmsg(uv_loop_t* loop,
   h.msg_name = &peer;
 
   do {
-    buf = handle->alloc_cb((uv_handle_t*)handle, 64 * 1024);
-    assert(buf.len > 0);
+    buf = handle->alloc_cb((uv_handle_t*) handle, 64 * 1024);
+    if (buf.len == 0) {
+      handle->recv_cb(handle, UV_ENOBUFS, buf, NULL, 0);
+      return;
+    }
     assert(buf.base != NULL);
 
     h.msg_namelen = sizeof(peer);
-    h.msg_iov = (struct iovec*)&buf;
+    h.msg_iov = (void*) &buf;
     h.msg_iovlen = 1;
 
     do {
@@ -220,14 +222,10 @@ static void uv__udp_recvmsg(uv_loop_t* loop,
     while (nread == -1 && errno == EINTR);
 
     if (nread == -1) {
-      if (errno == EAGAIN || errno == EWOULDBLOCK) {
-        uv__set_sys_error(handle->loop, EAGAIN);
+      if (errno == EAGAIN || errno == EWOULDBLOCK)
         handle->recv_cb(handle, 0, buf, NULL, 0);
-      }
-      else {
-        uv__set_sys_error(handle->loop, errno);
-        handle->recv_cb(handle, -1, buf, NULL, 0);
-      }
+      else
+        handle->recv_cb(handle, -errno, buf, NULL, 0);
     }
     else {
       flags = 0;
@@ -259,8 +257,8 @@ static void uv__udp_sendmsg(uv_loop_t* loop,
   assert(handle->type == UV_UDP);
   assert(revents & UV__POLLOUT);
 
-  assert(!ngx_queue_empty(&handle->write_queue)
-      || !ngx_queue_empty(&handle->write_completed_queue));
+  assert(!QUEUE_EMPTY(&handle->write_queue)
+      || !QUEUE_EMPTY(&handle->write_completed_queue));
 
   /* Write out pending data first. */
   uv__udp_run_pending(handle);
@@ -268,11 +266,11 @@ static void uv__udp_sendmsg(uv_loop_t* loop,
   /* Drain 'request completed' queue. */
   uv__udp_run_completed(handle);
 
-  if (!ngx_queue_empty(&handle->write_completed_queue)) {
+  if (!QUEUE_EMPTY(&handle->write_completed_queue)) {
     /* Schedule completion callbacks. */
     uv__io_feed(handle->loop, &handle->io_watcher);
   }
-  else if (ngx_queue_empty(&handle->write_queue)) {
+  else if (QUEUE_EMPTY(&handle->write_queue)) {
     /* Pending queue and completion queue empty, stop watcher. */
     uv__io_stop(loop, &handle->io_watcher, UV__POLLOUT);
 
@@ -287,39 +285,32 @@ static int uv__bind(uv_udp_t* handle,
                     struct sockaddr* addr,
                     socklen_t len,
                     unsigned flags) {
-  int saved_errno;
-  int status;
+  int err;
   int yes;
   int fd;
 
-  saved_errno = errno;
-  status = -1;
+  err = -EINVAL;
   fd = -1;
 
   /* Check for bad flags. */
-  if (flags & ~UV_UDP_IPV6ONLY) {
-    uv__set_sys_error(handle->loop, EINVAL);
-    goto out;
-  }
+  if (flags & ~UV_UDP_IPV6ONLY)
+    return -EINVAL;
 
   /* Cannot set IPv6-only mode on non-IPv6 socket. */
-  if ((flags & UV_UDP_IPV6ONLY) && domain != AF_INET6) {
-    uv__set_sys_error(handle->loop, EINVAL);
-    goto out;
-  }
+  if ((flags & UV_UDP_IPV6ONLY) && domain != AF_INET6)
+    return -EINVAL;
 
-  if (handle->io_watcher.fd == -1) {
-    if ((fd = uv__socket(domain, SOCK_DGRAM, 0)) == -1) {
-      uv__set_sys_error(handle->loop, errno);
-      goto out;
-    }
+  fd = handle->io_watcher.fd;
+  if (fd == -1) {
+    fd = uv__socket(domain, SOCK_DGRAM, 0);
+    if (fd == -1)
+      return -errno;
     handle->io_watcher.fd = fd;
   }
 
-  fd = handle->io_watcher.fd;
   yes = 1;
   if (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof yes) == -1) {
-    uv__set_sys_error(handle->loop, errno);
+    err = -errno;
     goto out;
   }
 
@@ -334,7 +325,7 @@ static int uv__bind(uv_udp_t* handle,
 #ifdef SO_REUSEPORT
   yes = 1;
   if (setsockopt(fd, SOL_SOCKET, SO_REUSEPORT, &yes, sizeof yes) == -1) {
-    uv__set_sys_error(handle->loop, errno);
+    err = -errno;
     goto out;
   }
 #endif
@@ -343,31 +334,26 @@ static int uv__bind(uv_udp_t* handle,
 #ifdef IPV6_V6ONLY
     yes = 1;
     if (setsockopt(fd, IPPROTO_IPV6, IPV6_V6ONLY, &yes, sizeof yes) == -1) {
-      uv__set_sys_error(handle->loop, errno);
+      err = -errno;
       goto out;
     }
 #else
-    uv__set_sys_error(handle->loop, ENOTSUP);
+    err = -ENOTSUP;
     goto out;
 #endif
   }
 
   if (bind(fd, addr, len) == -1) {
-    uv__set_sys_error(handle->loop, errno);
+    err = -errno;
     goto out;
   }
 
-  handle->io_watcher.fd = fd;
-  status = 0;
+  return 0;
 
 out:
-  if (status) {
-    close(handle->io_watcher.fd);
-    handle->io_watcher.fd = -1;
-  }
-
-  errno = saved_errno;
-  return status;
+  close(handle->io_watcher.fd);
+  handle->io_watcher.fd = -1;
+  return err;
 }
 
 
@@ -408,17 +394,20 @@ static int uv__udp_maybe_deferred_bind(uv_udp_t* handle, int domain) {
 }
 
 
-static int uv__udp_send(uv_udp_send_t* req,
-                        uv_udp_t* handle,
-                        uv_buf_t bufs[],
-                        int bufcnt,
-                        struct sockaddr* addr,
-                        socklen_t addrlen,
-                        uv_udp_send_cb send_cb) {
+static int uv__send(uv_udp_send_t* req,
+                    uv_udp_t* handle,
+                    uv_buf_t bufs[],
+                    int bufcnt,
+                    struct sockaddr* addr,
+                    socklen_t addrlen,
+                    uv_udp_send_cb send_cb) {
+  int err;
+
   assert(bufcnt > 0);
 
-  if (uv__udp_maybe_deferred_bind(handle, addr->sa_family))
-    return -1;
+  err = uv__udp_maybe_deferred_bind(handle, addr->sa_family);
+  if (err)
+    return err;
 
   uv__req_init(handle->loop, req, UV_UDP_SEND);
 
@@ -428,16 +417,16 @@ static int uv__udp_send(uv_udp_send_t* req,
   req->handle = handle;
   req->bufcnt = bufcnt;
 
-  if (bufcnt <= (int) ARRAY_SIZE(req->bufsml)) {
+  if (bufcnt <= (int) ARRAY_SIZE(req->bufsml))
     req->bufs = req->bufsml;
-  }
-  else if ((req->bufs = malloc(bufcnt * sizeof(bufs[0]))) == NULL) {
-    uv__set_sys_error(handle->loop, ENOMEM);
-    return -1;
-  }
+  else
+    req->bufs = malloc(bufcnt * sizeof(*bufs));
+
+  if (req->bufs == NULL)
+    return -ENOMEM;
 
   memcpy(req->bufs, bufs, bufcnt * sizeof(bufs[0]));
-  ngx_queue_insert_tail(&handle->write_queue, &req->queue);
+  QUEUE_INSERT_TAIL(&handle->write_queue, &req->queue);
   uv__io_start(handle->loop, &handle->io_watcher, UV__POLLOUT);
   uv__handle_start(handle);
 
@@ -446,14 +435,12 @@ static int uv__udp_send(uv_udp_send_t* req,
 
 
 int uv_udp_init(uv_loop_t* loop, uv_udp_t* handle) {
-  memset(handle, 0, sizeof *handle);
-
   uv__handle_init(loop, (uv_handle_t*)handle, UV_UDP);
-  handle->io_watcher.fd = -1;
-  ngx_queue_init(&handle->write_queue);
-  ngx_queue_init(&handle->write_completed_queue);
+  handle->alloc_cb = NULL;
+  handle->recv_cb = NULL;
   uv__io_init(&handle->io_watcher, uv__udp_io, -1);
-
+  QUEUE_INIT(&handle->write_queue);
+  QUEUE_INIT(&handle->write_completed_queue);
   return 0;
 }
 
@@ -477,24 +464,15 @@ int uv__udp_bind6(uv_udp_t* handle, struct sockaddr_in6 addr, unsigned flags) {
 
 
 int uv_udp_open(uv_udp_t* handle, uv_os_sock_t sock) {
-  int saved_errno;
-  int status;
   int yes;
 
-  saved_errno = errno;
-  status = -1;
-
   /* Check for already active socket. */
-  if (handle->io_watcher.fd != -1) {
-    uv__set_artificial_error(handle->loop, UV_EALREADY);
-    goto out;
-  }
+  if (handle->io_watcher.fd != -1)
+    return -EALREADY;  /* FIXME(bnoordhuis) Should be -EBUSY. */
 
   yes = 1;
-  if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof yes) == -1) {
-    uv__set_sys_error(handle->loop, errno);
-    goto out;
-  }
+  if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof yes))
+    return -errno;
 
   /* On the BSDs, SO_REUSEADDR lets you reuse an address that's in the TIME_WAIT
    * state (i.e. was until recently tied to a socket) while SO_REUSEPORT lets
@@ -506,26 +484,22 @@ int uv_udp_open(uv_udp_t* handle, uv_os_sock_t sock) {
    */
 #ifdef SO_REUSEPORT
   yes = 1;
-  if (setsockopt(sock, SOL_SOCKET, SO_REUSEPORT, &yes, sizeof yes) == -1) {
-    uv__set_sys_error(handle->loop, errno);
-    goto out;
-  }
+  if (setsockopt(sock, SOL_SOCKET, SO_REUSEPORT, &yes, sizeof yes))
+    return -errno;
 #endif
 
   handle->io_watcher.fd = sock;
-  status = 0;
-
-out:
-  errno = saved_errno;
-  return status;
+  return 0;
 }
 
 
-int uv_udp_set_membership(uv_udp_t* handle, const char* multicast_addr,
-  const char* interface_addr, uv_membership membership) {
-
-  int optname;
+int uv_udp_set_membership(uv_udp_t* handle,
+                          const char* multicast_addr,
+                          const char* interface_addr,
+                          uv_membership membership) {
   struct ip_mreq mreq;
+  int optname;
+
   memset(&mreq, 0, sizeof mreq);
 
   if (interface_addr) {
@@ -544,13 +518,15 @@ int uv_udp_set_membership(uv_udp_t* handle, const char* multicast_addr,
     optname = IP_DROP_MEMBERSHIP;
     break;
   default:
-    uv__set_sys_error(handle->loop, EFAULT);
-    return -1;
+    return -EINVAL;
   }
 
-  if (setsockopt(handle->io_watcher.fd, IPPROTO_IP, optname, (void*) &mreq, sizeof mreq) == -1) {
-    uv__set_sys_error(handle->loop, errno);
-    return -1;
+  if (setsockopt(handle->io_watcher.fd,
+                 IPPROTO_IP,
+                 optname,
+                 &mreq,
+                 sizeof(mreq))) {
+    return -errno;
   }
 
   return 0;
@@ -558,25 +534,30 @@ int uv_udp_set_membership(uv_udp_t* handle, const char* multicast_addr,
 
 
 static int uv__setsockopt_maybe_char(uv_udp_t* handle, int option, int val) {
-#if __sun
+#if defined(__sun)
   char arg = val;
 #else
   int arg = val;
 #endif
 
   if (val < 0 || val > 255)
-    return uv__set_sys_error(handle->loop, EINVAL);
+    return -EINVAL;
 
   if (setsockopt(handle->io_watcher.fd, IPPROTO_IP, option, &arg, sizeof(arg)))
-    return uv__set_sys_error(handle->loop, errno);
+    return -errno;
 
   return 0;
 }
 
 
 int uv_udp_set_broadcast(uv_udp_t* handle, int on) {
-  if (setsockopt(handle->io_watcher.fd, SOL_SOCKET, SO_BROADCAST, &on, sizeof(on)))
-    return uv__set_sys_error(handle->loop, errno);
+  if (setsockopt(handle->io_watcher.fd,
+                 SOL_SOCKET,
+                 SO_BROADCAST,
+                 &on,
+                 sizeof(on))) {
+    return -errno;
+  }
 
   return 0;
 }
@@ -584,10 +565,10 @@ int uv_udp_set_broadcast(uv_udp_t* handle, int on) {
 
 int uv_udp_set_ttl(uv_udp_t* handle, int ttl) {
   if (ttl < 1 || ttl > 255)
-    return uv__set_sys_error(handle->loop, EINVAL);
+    return -EINVAL;
 
   if (setsockopt(handle->io_watcher.fd, IPPROTO_IP, IP_TTL, &ttl, sizeof(ttl)))
-    return uv__set_sys_error(handle->loop, errno);
+    return -errno;
 
   return 0;
 }
@@ -605,81 +586,67 @@ int uv_udp_set_multicast_loop(uv_udp_t* handle, int on) {
 
 int uv_udp_getsockname(uv_udp_t* handle, struct sockaddr* name, int* namelen) {
   socklen_t socklen;
-  int saved_errno;
-  int rv = 0;
 
-  /* Don't clobber errno. */
-  saved_errno = errno;
-
-  if (handle->io_watcher.fd == -1) {
-    uv__set_sys_error(handle->loop, EINVAL);
-    rv = -1;
-    goto out;
-  }
+  if (handle->io_watcher.fd == -1)
+    return -EINVAL;  /* FIXME(bnoordhuis) -EBADF */
 
   /* sizeof(socklen_t) != sizeof(int) on some systems. */
-  socklen = (socklen_t)*namelen;
+  socklen = (socklen_t) *namelen;
 
-  if (getsockname(handle->io_watcher.fd, name, &socklen) == -1) {
-    uv__set_sys_error(handle->loop, errno);
-    rv = -1;
-  } else {
-    *namelen = (int)socklen;
-  }
+  if (getsockname(handle->io_watcher.fd, name, &socklen))
+    return -errno;
 
-out:
-  errno = saved_errno;
-  return rv;
+  *namelen = (int) socklen;
+  return 0;
 }
 
 
-int uv_udp_send(uv_udp_send_t* req,
-                uv_udp_t* handle,
-                uv_buf_t bufs[],
-                int bufcnt,
-                struct sockaddr_in addr,
-                uv_udp_send_cb send_cb) {
-  return uv__udp_send(req,
-                      handle,
-                      bufs,
-                      bufcnt,
-                      (struct sockaddr*)&addr,
-                      sizeof addr,
-                      send_cb);
-}
-
-
-int uv_udp_send6(uv_udp_send_t* req,
+int uv__udp_send(uv_udp_send_t* req,
                  uv_udp_t* handle,
                  uv_buf_t bufs[],
                  int bufcnt,
-                 struct sockaddr_in6 addr,
+                 struct sockaddr_in addr,
                  uv_udp_send_cb send_cb) {
-  return uv__udp_send(req,
-                      handle,
-                      bufs,
-                      bufcnt,
-                      (struct sockaddr*)&addr,
-                      sizeof addr,
-                      send_cb);
+  return uv__send(req,
+                  handle,
+                  bufs,
+                  bufcnt,
+                  (struct sockaddr*)&addr,
+                  sizeof addr,
+                  send_cb);
 }
 
 
-int uv_udp_recv_start(uv_udp_t* handle,
-                      uv_alloc_cb alloc_cb,
-                      uv_udp_recv_cb recv_cb) {
-  if (alloc_cb == NULL || recv_cb == NULL) {
-    uv__set_artificial_error(handle->loop, UV_EINVAL);
-    return -1;
-  }
+int uv__udp_send6(uv_udp_send_t* req,
+                  uv_udp_t* handle,
+                  uv_buf_t bufs[],
+                  int bufcnt,
+                  struct sockaddr_in6 addr,
+                  uv_udp_send_cb send_cb) {
+  return uv__send(req,
+                  handle,
+                  bufs,
+                  bufcnt,
+                  (struct sockaddr*)&addr,
+                  sizeof addr,
+                  send_cb);
+}
 
-  if (uv__io_active(&handle->io_watcher, UV__POLLIN)) {
-    uv__set_artificial_error(handle->loop, UV_EALREADY);
-    return -1;
-  }
 
-  if (uv__udp_maybe_deferred_bind(handle, AF_INET))
-    return -1;
+int uv__udp_recv_start(uv_udp_t* handle,
+                       uv_alloc_cb alloc_cb,
+                       uv_udp_recv_cb recv_cb) {
+  int err;
+
+  if (alloc_cb == NULL || recv_cb == NULL)
+    return -EINVAL;
+
+  if (uv__io_active(&handle->io_watcher, UV__POLLIN))
+    return -EALREADY;  /* FIXME(bnoordhuis) Should be -EBUSY. */
+
+  err = uv__udp_maybe_deferred_bind(handle, AF_INET);
+  if (err)
+    return err;
 
   handle->alloc_cb = alloc_cb;
   handle->recv_cb = recv_cb;
@@ -691,7 +658,7 @@ int uv_udp_recv_start(uv_udp_t* handle,
 }
 
 
-int uv_udp_recv_stop(uv_udp_t* handle) {
+int uv__udp_recv_stop(uv_udp_t* handle) {
   uv__io_stop(handle->loop, &handle->io_watcher, UV__POLLIN);
 
   if (!uv__io_active(&handle->io_watcher, UV__POLLOUT))
