@@ -58,6 +58,7 @@
 #define UV_HANDLE_SYNC_BYPASS_IOCP              0x00040000
 #define UV_HANDLE_ZERO_READ                     0x00080000
 #define UV_HANDLE_EMULATE_IOCP                  0x00100000
+#define UV_HANDLE_BLOCKING_WRITES               0x00200000
 
 /* Only used by uv_tcp_t handles. */
 #define UV_HANDLE_IPV6                          0x01000000
@@ -138,7 +139,7 @@ void uv_udp_endgame(uv_loop_t* loop, uv_udp_t* handle);
 /*
  * Pipes
  */
-uv_err_t uv_stdio_pipe_server(uv_loop_t* loop, uv_pipe_t* handle, DWORD access,
+int uv_stdio_pipe_server(uv_loop_t* loop, uv_pipe_t* handle, DWORD access,
     char* name, size_t nameSize);
 
 int uv_pipe_listen(uv_pipe_t* handle, int backlog, uv_connection_cb cb);
@@ -200,7 +201,7 @@ void uv_tty_endgame(uv_loop_t* loop, uv_tty_t* handle);
 void uv_process_poll_req(uv_loop_t* loop, uv_poll_t* handle,
     uv_req_t* req);
 
-void uv_poll_close(uv_loop_t* loop, uv_poll_t* handle);
+int uv_poll_close(uv_loop_t* loop, uv_poll_t* handle);
 void uv_poll_endgame(uv_loop_t* loop, uv_poll_t* handle);
 
 
@@ -210,6 +211,7 @@ void uv_poll_endgame(uv_loop_t* loop, uv_poll_t* handle);
 void uv_timer_endgame(uv_loop_t* loop, uv_timer_t* handle);
 
 DWORD uv_get_poll_timeout(uv_loop_t* loop);
+void uv__time_forward(uv_loop_t* loop, uint64_t msecs);
 void uv_process_timers(uv_loop_t* loop);
 
 
@@ -256,6 +258,12 @@ void uv_process_close(uv_loop_t* loop, uv_process_t* handle);
 void uv_process_endgame(uv_loop_t* loop, uv_process_t* handle);
 
 
+/*
+ * Error
+ */
+int uv_translate_sys_error(int sys_errno);
+
+
 /*
  * Getaddrinfo
  */
@@ -297,13 +305,12 @@ void uv__util_init();
 
 int uv_parent_pid();
 void uv_fatal_error(const int errorno, const char* syscall);
-uv_err_code uv_translate_sys_error(int sys_errno);
 
 
 /*
  * Process stdio handles.
  */
-uv_err_t uv__stdio_create(uv_loop_t* loop, uv_process_options_t* options,
+int uv__stdio_create(uv_loop_t* loop, uv_process_options_t* options,
     BYTE** buffer_ptr);
 void uv__stdio_destroy(BYTE* buffer);
 void uv__stdio_noinherit(BYTE* buffer);
