@@ -45,7 +45,52 @@ typedef struct env_var {
 #define E_V(str) { str "=", L##str, sizeof(str), 0, 0 }
 
 
-static uv_err_t uv_utf8_to_utf16_alloc(const char* s, WCHAR** ws_ptr) {
+static HANDLE uv_global_job_handle_;
+static uv_once_t uv_global_job_handle_init_guard_ = UV_ONCE_INIT;
+
+
+static void uv__init_global_job_handle(void) {
+  /* Create a job object and set it up to kill all contained processes when
+   * it's closed. Since this handle is made non-inheritable and we're not
+   * giving it to anyone, we're the only process holding a reference to it.
+   * That means that if this process exits it is closed and all the processes
+   * it contains are killed. All processes created with uv_spawn that are not
+   * spawned with the UV_PROCESS_DETACHED flag are assigned to this job.
+   *
+   * We're setting the JOB_OBJECT_LIMIT_SILENT_BREAKAWAY_OK flag so only the
+   * processes that we explicitly add are affected, and *their* subprocesses
+   * are not. This ensures that our child processes are not limited in their
+   * ability to use job control on Windows versions that don't deal with
+   * nested jobs (prior to Windows 8 / Server 2012). It also lets our child
+   * processes created detached processes without explicitly breaking away
+   * from job control (which uv_spawn doesn't, either).
+   */
+  SECURITY_ATTRIBUTES attr;
+  JOBOBJECT_EXTENDED_LIMIT_INFORMATION info;
+
+  memset(&attr, 0, sizeof attr);
+  attr.bInheritHandle = FALSE;
+
+  memset(&info, 0, sizeof info);
+  info.BasicLimitInformation.LimitFlags =
+      JOB_OBJECT_LIMIT_BREAKAWAY_OK |
+      JOB_OBJECT_LIMIT_SILENT_BREAKAWAY_OK |
+      JOB_OBJECT_LIMIT_DIE_ON_UNHANDLED_EXCEPTION |
+      JOB_OBJECT_LIMIT_KILL_ON_JOB_CLOSE;
+
+  uv_global_job_handle_ = CreateJobObjectW(&attr, NULL);
+  if (uv_global_job_handle_ == NULL)
+    uv_fatal_error(GetLastError(), "CreateJobObjectW");
+
+  if (!SetInformationJobObject(uv_global_job_handle_,
+                               JobObjectExtendedLimitInformation,
+                               &info,
+                               sizeof info))
+    uv_fatal_error(GetLastError(), "SetInformationJobObject");
+}
+
+
+static int uv_utf8_to_utf16_alloc(const char* s, WCHAR** ws_ptr) {
   int ws_len, r;
   WCHAR* ws;
 
@@ -56,12 +101,12 @@ static uv_err_t uv_utf8_to_utf16_alloc(const char* s, WCHAR** ws_ptr) {
                                NULL,
                                0);
   if (ws_len <= 0) {
-    return uv__new_sys_error(GetLastError());
+    return GetLastError();
   }
 
   ws = (WCHAR*) malloc(ws_len * sizeof(WCHAR));
   if (ws == NULL) {
-    return uv__new_artificial_error(UV_ENOMEM);
+    return ERROR_OUTOFMEMORY;
   }
 
   r = MultiByteToWideChar(CP_UTF8,
@@ -73,7 +118,7 @@ static uv_err_t uv_utf8_to_utf16_alloc(const char* s, WCHAR** ws_ptr) {
   assert(r == ws_len);
 
   *ws_ptr = ws;
-  return uv_ok_;
+  return 0;
 }
 
 
@@ -81,7 +126,7 @@ static void uv_process_init(uv_loop_t* loop, uv_process_t* handle) {
   uv__handle_init(loop, (uv_handle_t*) handle, UV_PROCESS);
   handle->exit_cb = NULL;
   handle->pid = 0;
-  handle->spawn_error = uv_ok_;
+  handle->spawn_error = 0;
   handle->exit_signal = 0;
   handle->wait_handle = INVALID_HANDLE_VALUE;
   handle->process_handle = INVALID_HANDLE_VALUE;
@@ -180,7 +225,7 @@ static WCHAR* search_path_join_test(const WCHAR* dir,
   attrs = GetFileAttributesW(result);
 
   if (attrs != INVALID_FILE_ATTRIBUTES &&
-     !(attrs & (FILE_ATTRIBUTE_DIRECTORY | FILE_ATTRIBUTE_REPARSE_POINT))) {
+      !(attrs & FILE_ATTRIBUTE_DIRECTORY)) {
     return result;
   }
 
@@ -453,7 +498,7 @@ WCHAR* quote_cmd_arg(const WCHAR *source, WCHAR *target) {
 }
 
 
-uv_err_t make_program_args(char** args, int verbatim_arguments, WCHAR** dst_ptr) {
+int make_program_args(char** args, int verbatim_arguments, WCHAR** dst_ptr) {
   char** arg;
   WCHAR* dst = NULL;
   WCHAR* temp_buffer = NULL;
@@ -461,7 +506,7 @@ uv_err_t make_program_args(char** args, int verbatim_arguments, WCHAR** dst_ptr)
   size_t temp_buffer_len = 0;
   WCHAR* pos;
   int arg_count = 0;
-  uv_err_t err = uv_ok_;
+  int err = 0;
 
   /* Count the required size. */
   for (arg = args; *arg; arg++) {
@@ -474,7 +519,7 @@ uv_err_t make_program_args(char** args, int verbatim_arguments, WCHAR** dst_ptr)
                                   NULL,
                                   0);
     if (arg_len == 0) {
-      return uv__new_sys_error(GetLastError());
+      return GetLastError();
     }
 
     dst_len += arg_len;
@@ -492,14 +537,14 @@ uv_err_t make_program_args(char** args, int verbatim_arguments, WCHAR** dst_ptr)
   /* Allocate buffer for the final command line. */
   dst = (WCHAR*) malloc(dst_len * sizeof(WCHAR));
   if (dst == NULL) {
-    err = uv__new_artificial_error(UV_ENOMEM);
+    err = ERROR_OUTOFMEMORY;
     goto error;
   }
 
   /* Allocate temporary working buffer. */
   temp_buffer = (WCHAR*) malloc(temp_buffer_len * sizeof(WCHAR));
   if (temp_buffer == NULL) {
-    err = uv__new_artificial_error(UV_ENOMEM);
+    err = ERROR_OUTOFMEMORY;
     goto error;
   }
 
@@ -515,6 +560,7 @@ uv_err_t make_program_args(char** args, int verbatim_arguments, WCHAR** dst_ptr)
                                   temp_buffer,
                                   (int) (dst + dst_len - pos));
     if (arg_len == 0) {
+      err = GetLastError();
       goto error;
     }
 
@@ -533,7 +579,7 @@ uv_err_t make_program_args(char** args, int verbatim_arguments, WCHAR** dst_ptr)
   free(temp_buffer);
 
   *dst_ptr = dst;
-  return uv_ok_;
+  return 0;
 
 error:
   free(dst);
@@ -572,7 +618,7 @@ static void check_required_vars_contains_var(env_var_t* required, int count,
  * these get defined if the input environment block does not contain any
  * values for them.
  */
-uv_err_t make_program_env(char* env_block[], WCHAR** dst_ptr) {
+int make_program_env(char* env_block[], WCHAR** dst_ptr) {
   WCHAR* dst;
   WCHAR* ptr;
   char** env;
@@ -600,7 +646,7 @@ uv_err_t make_program_env(char* env_block[], WCHAR** dst_ptr) {
                               NULL,
                               0);
     if (len <= 0) {
-      return uv__new_sys_error(GetLastError());
+      return GetLastError();
     }
 
     env_len += len;
@@ -611,7 +657,7 @@ uv_err_t make_program_env(char* env_block[], WCHAR** dst_ptr) {
       env_len += required_vars[i].len;
       var_size = GetEnvironmentVariableW(required_vars[i].wide, NULL, 0);
       if (var_size == 0) {
-        return uv__new_sys_error(GetLastError());
+        return GetLastError();
       }
       required_vars[i].value_len = var_size;
       env_len += var_size;
@@ -620,7 +666,7 @@ uv_err_t make_program_env(char* env_block[], WCHAR** dst_ptr) {
 
   dst = malloc(env_len * sizeof(WCHAR));
   if (!dst) {
-    return uv__new_artificial_error(UV_ENOMEM);
+    return ERROR_OUTOFMEMORY;
   }
 
   ptr = dst;
@@ -634,7 +680,7 @@ uv_err_t make_program_env(char* env_block[], WCHAR** dst_ptr) {
                               (int) (env_len - (ptr - dst)));
     if (len <= 0) {
       free(dst);
-      return uv__new_sys_error(GetLastError());
+      return GetLastError();
     }
   }
 
@@ -657,7 +703,7 @@ uv_err_t make_program_env(char* env_block[], WCHAR** dst_ptr) {
   *ptr = L'\0';
 
   *dst_ptr = dst;
-  return uv_ok_;
+  return 0;
 }
 
 
@@ -682,7 +728,8 @@ static void CALLBACK exit_wait_callback(void* data, BOOLEAN didTimeout) {
 
 /* Called on main thread after a child process has exited. */
 void uv_process_proc_exit(uv_loop_t* loop, uv_process_t* handle) {
-  DWORD exit_code;
+  int exit_code;
+  DWORD status;
 
   assert(handle->exit_cb_pending);
   handle->exit_cb_pending = 0;
@@ -704,17 +751,19 @@ void uv_process_proc_exit(uv_loop_t* loop, uv_process_t* handle) {
   /* callback.*/
   uv__handle_stop(handle);
 
-  if (handle->spawn_error.code != UV_OK) {
+  if (handle->spawn_error) {
     /* Spawning failed. */
-    exit_code = (DWORD) -1;
-  } else if (!GetExitCodeProcess(handle->process_handle, &exit_code)) {
+    exit_code = uv_translate_sys_error(handle->spawn_error);
+  } else if (!GetExitCodeProcess(handle->process_handle, &status)) {
     /* Unable to to obtain the exit code. This should never happen. */
-    exit_code = (DWORD) -1;
+    exit_code = uv_translate_sys_error(GetLastError());
+  } else {
+    /* Make sure the exit code is >= 0. */
+    exit_code = status & INT_MAX;
   }
 
   /* Fire the exit callback. */
   if (handle->exit_cb) {
-    loop->last_err = handle->spawn_error;
     handle->exit_cb(handle, exit_code, handle->exit_signal);
   }
 }
@@ -756,7 +805,7 @@ void uv_process_endgame(uv_loop_t* loop, uv_process_t* handle) {
 int uv_spawn(uv_loop_t* loop, uv_process_t* process,
     uv_process_options_t options) {
   int i;
-  uv_err_t err = uv_ok_;
+  int err = 0;
   WCHAR* path = NULL;
   BOOL result;
   WCHAR* application_path = NULL, *application = NULL, *arguments = NULL,
@@ -766,14 +815,12 @@ int uv_spawn(uv_loop_t* loop, uv_process_t* process,
   DWORD process_flags;
 
   if (options.flags & (UV_PROCESS_SETGID | UV_PROCESS_SETUID)) {
-    uv__set_artificial_error(loop, UV_ENOTSUP);
-    return -1;
+    return UV_ENOTSUP;
   }
 
   if (options.file == NULL ||
       options.args == NULL) {
-    uv__set_artificial_error(loop, UV_EINVAL);
-    return -1;
+    return UV_EINVAL;
   }
 
   assert(options.file != NULL);
@@ -787,25 +834,25 @@ int uv_spawn(uv_loop_t* loop, uv_process_t* process,
   process->exit_cb = options.exit_cb;
 
   err = uv_utf8_to_utf16_alloc(options.file, &application);
-  if (err.code != UV_OK)
+  if (err)
     goto done;
 
   err = make_program_args(options.args,
                           options.flags & UV_PROCESS_WINDOWS_VERBATIM_ARGUMENTS,
                           &arguments);
-  if (err.code != UV_OK)
+  if (err)
     goto done;
 
   if (options.env) {
      err = make_program_env(options.env, &env);
-     if (err.code != UV_OK)
+      if (err)
        goto done;
   }
 
   if (options.cwd) {
     /* Explicit cwd */
     err = uv_utf8_to_utf16_alloc(options.cwd, &cwd);
-    if (err.code != UV_OK)
+    if (err)
       goto done;
 
   } else {
@@ -814,19 +861,19 @@ int uv_spawn(uv_loop_t* loop, uv_process_t* process,
 
     cwd_len = GetCurrentDirectoryW(0, NULL);
     if (!cwd_len) {
-      err = uv__new_sys_error(GetLastError());
+      err = GetLastError();
       goto done;
     }
 
     cwd = (WCHAR*) malloc(cwd_len * sizeof(WCHAR));
     if (cwd == NULL) {
-      err = uv__new_artificial_error(UV_ENOMEM);
+      err = ERROR_OUTOFMEMORY;
       goto done;
     }
 
     r = GetCurrentDirectoryW(cwd_len, cwd);
     if (r == 0 || r >= cwd_len) {
-      err = uv__new_sys_error(GetLastError());
+      err = GetLastError();
       goto done;
     }
   }
@@ -837,20 +884,20 @@ int uv_spawn(uv_loop_t* loop, uv_process_t* process,
 
     path_len = GetEnvironmentVariableW(L"PATH", NULL, 0);
     if (path_len == 0) {
-      err = uv__new_sys_error(GetLastError());
+      err = GetLastError();
       goto done;
     }
 
 
     path = (WCHAR*) malloc(path_len * sizeof(WCHAR));
     if (path == NULL) {
-      err = uv__new_artificial_error(UV_ENOMEM);
+      err = ERROR_OUTOFMEMORY;
       goto done;
     }
 
     r = GetEnvironmentVariableW(L"PATH", path, path_len);
     if (r == 0 || r >= path_len) {
-      err = uv__new_sys_error(GetLastError());
+      err = GetLastError();
       goto done;
     }
   }
@@ -860,13 +907,12 @@ int uv_spawn(uv_loop_t* loop, uv_process_t* process,
                                  path);
   if (application_path == NULL) {
     /* Not found. */
-    err = uv__new_artificial_error(UV_ENOENT);
+    err = ERROR_FILE_NOT_FOUND;
     goto done;
   }
 
-
   err = uv__stdio_create(loop, &options, &process->child_stdio_buffer);
-  if (err.code != UV_OK)
+  if (err)
     goto done;
 
   startup.cb = sizeof(startup);
@@ -890,8 +936,19 @@ int uv_spawn(uv_loop_t* loop, uv_process_t* process,
   }
 
   process_flags = CREATE_UNICODE_ENVIRONMENT;
+
   if (options.flags & UV_PROCESS_DETACHED) {
-    process_flags |= DETACHED_PROCESS | CREATE_NEW_PROCESS_GROUP;
+    /* Note that we're not setting the CREATE_BREAKAWAY_FROM_JOB flag. That
+     * means that libuv might not let you create a fully deamonized process
+     * when run under job control. However the type of job control that libuv
+     * itself creates doesn't trickle down to subprocesses so they can still
+     * daemonize.
+     *
+     * A reason to not do this is that CREATE_BREAKAWAY_FROM_JOB makes the
+     * CreateProcess call fail if we're under job control that doesn't allow
+     * breakaway.
+     */
+    process_flags |= CREATE_NEW_PROCESS_GROUP;
   }
 
   if (CreateProcessW(application_path,
@@ -908,6 +965,28 @@ int uv_spawn(uv_loop_t* loop, uv_process_t* process,
     process->process_handle = info.hProcess;
     process->pid = info.dwProcessId;
 
+    /* If the process isn't spawned as detached, assign to the global job */
+    /* object so windows will kill it when the parent process dies. */
+    if (!(options.flags & UV_PROCESS_DETACHED)) {
+      uv_once(&uv_global_job_handle_init_guard_, uv__init_global_job_handle);
+
+      if (!AssignProcessToJobObject(uv_global_job_handle_, info.hProcess)) {
+        /* AssignProcessToJobObject might fail if this process is under job
+         * control and the job doesn't have the
+         * JOB_OBJECT_LIMIT_SILENT_BREAKAWAY_OK flag set, on a Windows version
+         * that doesn't support nested jobs.
+         *
+         * When that happens we just swallow the error and continue without
+         * establishing a kill-child-on-parent-exit relationship, otherwise
+         * there would be no way for libuv applications run under job control
+         * to spawn processes at all.
+         */
+        DWORD err = GetLastError();
+        if (err != ERROR_ACCESS_DENIED)
+          uv_fatal_error(err, "AssignProcessToJobObject");
+      }
+    }
+
     /* Set IPC pid to all IPC pipes. */
     for (i = 0; i < options.stdio_count; i++) {
       const uv_stdio_container_t* fdopt = &options.stdio[i];
@@ -931,7 +1010,7 @@ int uv_spawn(uv_loop_t* loop, uv_process_t* process,
 
   } else {
     /* CreateProcessW failed. */
-    err = uv__new_sys_error(GetLastError());
+    err = GetLastError();
   }
 
 done:
@@ -955,7 +1034,7 @@ done:
   uv__handle_start(process);
 
   /* If an error happened, queue the exit req. */
-  if (err.code != UV_OK) {
+  if (err) {
     process->exit_cb_pending = 1;
     uv_insert_pending_req(loop, (uv_req_t*) &process->exit_req);
   }
@@ -964,28 +1043,29 @@ done:
 }
 
 
-static uv_err_t uv__kill(HANDLE process_handle, int signum) {
+static int uv__kill(HANDLE process_handle, int signum) {
   switch (signum) {
     case SIGTERM:
     case SIGKILL:
     case SIGINT: {
       /* Unconditionally terminate the process. On Windows, killed processes */
       /* normally return 1. */
-      DWORD error, status;
+      DWORD status;
+      int err;
 
       if (TerminateProcess(process_handle, 1))
-        return uv_ok_;
+        return 0;
 
       /* If the process already exited before TerminateProcess was called, */
       /* TerminateProcess will fail with ERROR_ACESS_DENIED. */
-      error = GetLastError();
-      if (error == ERROR_ACCESS_DENIED &&
+      err = GetLastError();
+      if (err == ERROR_ACCESS_DENIED &&
           GetExitCodeProcess(process_handle, &status) &&
           status != STILL_ACTIVE) {
-        return uv__new_artificial_error(UV_ESRCH);
+        return UV_ESRCH;
       }
 
-      return uv__new_sys_error(error);
+      return uv_translate_sys_error(err);
     }
 
     case 0: {
@@ -993,34 +1073,31 @@ static uv_err_t uv__kill(HANDLE process_handle, int signum) {
       DWORD status;
 
       if (!GetExitCodeProcess(process_handle, &status))
-        return uv__new_sys_error(GetLastError());
+        return uv_translate_sys_error(GetLastError());
 
       if (status != STILL_ACTIVE)
-        return uv__new_artificial_error(UV_ESRCH);
+        return UV_ESRCH;
 
-      return uv_ok_;
+      return 0;
     }
 
     default:
       /* Unsupported signal. */
-      return uv__new_artificial_error(UV_ENOSYS);
+      return UV_ENOSYS;
   }
 }
 
 
 int uv_process_kill(uv_process_t* process, int signum) {
-  uv_err_t err;
+  int err;
 
   if (process->process_handle == INVALID_HANDLE_VALUE) {
-    uv__set_artificial_error(process->loop, UV_EINVAL);
-    return -1;
+    return UV_EINVAL;
   }
 
   err = uv__kill(process->process_handle, signum);
-
-  if (err.code != UV_OK) {
-    uv__set_error(process->loop, err.code, err.sys_errno_);
-    return -1;
+  if (err) {
+    return err;  /* err is already translated. */
   }
 
   process->exit_signal = signum;
@@ -1029,21 +1106,22 @@ int uv_process_kill(uv_process_t* process, int signum) {
 }
 
 
-uv_err_t uv_kill(int pid, int signum) {
-  uv_err_t err;
+int uv_kill(int pid, int signum) {
+  int err;
   HANDLE process_handle = OpenProcess(PROCESS_TERMINATE |
     PROCESS_QUERY_INFORMATION, FALSE, pid);
 
   if (process_handle == NULL) {
-    if (GetLastError() == ERROR_INVALID_PARAMETER) {
-      return uv__new_artificial_error(UV_ESRCH);
+    err = GetLastError();
+    if (err == ERROR_INVALID_PARAMETER) {
+      return UV_ESRCH;
     } else {
-      return uv__new_sys_error(GetLastError());
+      return uv_translate_sys_error(err);
     }
   }
 
   err = uv__kill(process_handle, signum);
   CloseHandle(process_handle);
 
-  return err;
+  return err;  /* err is already translated. */
 }
