@@ -31,14 +31,11 @@
 #include "uv.h"
 #include "internal.h"
 
-#pragma push_macro("_WIN32_WINNT")
-#undef _WIN32_WINNT
-#define _WIN32_WINNT 0x0502
+#include <winsock2.h>
 #include <iphlpapi.h>
-#pragma pop_macro("_WIN32_WINNT")
 #include <psapi.h>
 #include <tlhelp32.h>
-
+#include <winperf.h>
 
 /*
  * Max title length; the only thing MSDN tells us about the maximum length
@@ -105,9 +102,10 @@ int uv_utf8_to_utf16(const char* utf8Buffer, WCHAR* utf16Buffer,
 int uv_exepath(char* buffer, size_t* size_ptr) {
   int utf8_len, utf16_buffer_len, utf16_len;
   WCHAR* utf16_buffer;
+  int err;
 
   if (buffer == NULL || size_ptr == NULL || *size_ptr == 0) {
-    return -1;
+    return UV_EINVAL;
   }
 
   if (*size_ptr > 32768) {
@@ -119,12 +117,13 @@ int uv_exepath(char* buffer, size_t* size_ptr) {
 
   utf16_buffer = (WCHAR*) malloc(sizeof(WCHAR) * utf16_buffer_len);
   if (!utf16_buffer) {
-    return -1;
+    return UV_ENOMEM;
   }
 
   /* Get the path as UTF-16. */
   utf16_len = GetModuleFileNameW(NULL, utf16_buffer, utf16_buffer_len);
   if (utf16_len <= 0) {
+    err = GetLastError();
     goto error;
   }
 
@@ -141,6 +140,7 @@ int uv_exepath(char* buffer, size_t* size_ptr) {
                                  NULL,
                                  NULL);
   if (utf8_len == 0) {
+    err = GetLastError();
     goto error;
   }
 
@@ -153,26 +153,26 @@ int uv_exepath(char* buffer, size_t* size_ptr) {
 
  error:
   free(utf16_buffer);
-  return -1;
+  return uv_translate_sys_error(err);
 }
 
 
-uv_err_t uv_cwd(char* buffer, size_t size) {
+int uv_cwd(char* buffer, size_t size) {
   DWORD utf16_len;
   WCHAR utf16_buffer[MAX_PATH];
   int r;
 
   if (buffer == NULL || size == 0) {
-    return uv__new_artificial_error(UV_EINVAL);
+    return UV_EINVAL;
   }
 
   utf16_len = GetCurrentDirectoryW(MAX_PATH, utf16_buffer);
   if (utf16_len == 0) {
-    return uv__new_sys_error(GetLastError());
+    return uv_translate_sys_error(GetLastError());
   } else if (utf16_len > MAX_PATH) {
     /* This should be impossible;  however the CRT has a code path to deal */
     /* with this scenario, so I added a check anyway. */
-    return uv__new_artificial_error(UV_EIO);
+    return UV_EIO;
   }
 
   /* utf16_len contains the length, *not* including the terminating null. */
@@ -196,20 +196,20 @@ uv_err_t uv_cwd(char* buffer, size_t size) {
                           NULL,
                           NULL);
   if (r == 0) {
-    return uv__new_sys_error(GetLastError());
+    return uv_translate_sys_error(GetLastError());
   }
 
-  return uv_ok_;
+  return 0;
 }
 
 
-uv_err_t uv_chdir(const char* dir) {
+int uv_chdir(const char* dir) {
   WCHAR utf16_buffer[MAX_PATH];
   size_t utf16_len;
   WCHAR drive_letter, env_var[4];
 
   if (dir == NULL) {
-    return uv__new_artificial_error(UV_EINVAL);
+    return UV_EINVAL;
   }
 
   if (MultiByteToWideChar(CP_UTF8,
@@ -223,14 +223,14 @@ uv_err_t uv_chdir(const char* dir) {
     /* including terminating null. If it doesn't fit, the path name must be */
     /* too long. */
     if (error == ERROR_INSUFFICIENT_BUFFER) {
-      return uv__new_artificial_error(UV_ENAMETOOLONG);
+      return UV_ENAMETOOLONG;
     } else {
-      return uv__new_sys_error(error);
+      return uv_translate_sys_error(error);
     }
   }
 
   if (!SetCurrentDirectoryW(utf16_buffer)) {
-    return uv__new_sys_error(GetLastError());
+    return uv_translate_sys_error(GetLastError());
   }
 
   /* Windows stores the drive-local path in an "hidden" environment variable, */
@@ -238,9 +238,9 @@ uv_err_t uv_chdir(const char* dir) {
   /* update this, so we'll have to do it. */
   utf16_len = GetCurrentDirectoryW(MAX_PATH, utf16_buffer);
   if (utf16_len == 0) {
-    return uv__new_sys_error(GetLastError());
+    return uv_translate_sys_error(GetLastError());
   } else if (utf16_len > MAX_PATH) {
-    return uv__new_artificial_error(UV_EIO);
+    return UV_EIO;
   }
 
   /* The returned directory should not have a trailing slash, unless it */
@@ -273,11 +273,11 @@ uv_err_t uv_chdir(const char* dir) {
     env_var[3] = L'\0';
 
     if (!SetEnvironmentVariableW(env_var, utf16_buffer)) {
-      return uv__new_sys_error(GetLastError());
+      return uv_translate_sys_error(GetLastError());
     }
   }
 
-  return uv_ok_;
+  return 0;
 }
 
 
@@ -341,8 +341,8 @@ char** uv_setup_args(int argc, char** argv) {
 }
 
 
-uv_err_t uv_set_process_title(const char* title) {
-  uv_err_t err;
+int uv_set_process_title(const char* title) {
+  int err;
   int length;
   WCHAR* title_w = NULL;
 
@@ -351,7 +351,7 @@ uv_err_t uv_set_process_title(const char* title) {
   /* Find out how big the buffer for the wide-char title must be */
   length = uv_utf8_to_utf16(title, NULL, 0);
   if (!length) {
-    err = uv__new_sys_error(GetLastError());
+    err = GetLastError();
     goto done;
   }
 
@@ -363,7 +363,7 @@ uv_err_t uv_set_process_title(const char* title) {
 
   length = uv_utf8_to_utf16(title, title_w, length);
   if (!length) {
-    err = uv__new_sys_error(GetLastError());
+    err = GetLastError();
     goto done;
   };
 
@@ -373,7 +373,7 @@ uv_err_t uv_set_process_title(const char* title) {
   }
 
   if (!SetConsoleTitleW(title_w)) {
-    err = uv__new_sys_error(GetLastError());
+    err = GetLastError();
     goto done;
   }
 
@@ -382,11 +382,11 @@ uv_err_t uv_set_process_title(const char* title) {
   process_title = strdup(title);
   LeaveCriticalSection(&process_title_lock);
 
-  err = uv_ok_;
+  err = 0;
 
 done:
   free(title_w);
-  return err;
+  return uv_translate_sys_error(err);
 }
 
 
@@ -420,7 +420,7 @@ static int uv__get_process_title() {
 }
 
 
-uv_err_t uv_get_process_title(char* buffer, size_t size) {
+int uv_get_process_title(char* buffer, size_t size) {
   uv__once_init();
 
   EnterCriticalSection(&process_title_lock);
@@ -429,14 +429,14 @@ uv_err_t uv_get_process_title(char* buffer, size_t size) {
    * we must query it with getConsoleTitleW
    */
   if (!process_title && uv__get_process_title() == -1) {
-    return uv__new_sys_error(GetLastError());
+    return uv_translate_sys_error(GetLastError());
   }
 
   assert(process_title);
   strncpy(buffer, process_title, size);
   LeaveCriticalSection(&process_title_lock);
 
-  return uv_ok_;
+  return 0;
 }
 
 
@@ -465,23 +465,23 @@ uint64_t uv_hrtime(void) {
 }
 
 
-uv_err_t uv_resident_set_memory(size_t* rss) {
+int uv_resident_set_memory(size_t* rss) {
   HANDLE current_process;
   PROCESS_MEMORY_COUNTERS pmc;
 
   current_process = GetCurrentProcess();
 
   if (!GetProcessMemoryInfo(current_process, &pmc, sizeof(pmc))) {
-    return uv__new_sys_error(GetLastError());
+    return uv_translate_sys_error(GetLastError());
   }
 
   *rss = pmc.WorkingSetSize;
 
-  return uv_ok_;
+  return 0;
 }
 
 
-uv_err_t uv_uptime(double* uptime) {
+int uv_uptime(double* uptime) {
   BYTE stack_buffer[4096];
   BYTE* malloced_buffer = NULL;
   BYTE* buffer = (BYTE*) stack_buffer;
@@ -508,7 +508,7 @@ uv_err_t uv_uptime(double* uptime) {
       break;
     } else if (result != ERROR_MORE_DATA) {
       *uptime = 0;
-      return uv__new_sys_error(result);
+      return uv_translate_sys_error(result);
     }
 
     free(malloced_buffer);
@@ -522,7 +522,7 @@ uv_err_t uv_uptime(double* uptime) {
     buffer = malloced_buffer = (BYTE*) malloc(buffer_size);
     if (malloced_buffer == NULL) {
       *uptime = 0;
-      return uv__new_artificial_error(UV_ENOMEM);
+      return UV_ENOMEM;
     }
   }
 
@@ -562,7 +562,7 @@ uv_err_t uv_uptime(double* uptime) {
         *uptime = (double) (object_type->PerfTime.QuadPart - value) /
                   (double) object_type->PerfFreq.QuadPart;
         free(malloced_buffer);
-        return uv_ok_;
+        return 0;
       }
     }
 
@@ -573,56 +573,59 @@ uv_err_t uv_uptime(double* uptime) {
   /* If we get here, the uptime value was not found. */
   free(malloced_buffer);
   *uptime = 0;
-  return uv__new_artificial_error(UV_ENOSYS);
+  return UV_ENOSYS;
 
  internalError:
   free(malloced_buffer);
   *uptime = 0;
-  return uv__new_artificial_error(UV_EIO);
+  return UV_EIO;
 }
 
 
-uv_err_t uv_cpu_info(uv_cpu_info_t** cpu_infos, int* count) {
+int uv_cpu_info(uv_cpu_info_t** cpu_infos_ptr, int* cpu_count_ptr) {
+  uv_cpu_info_t* cpu_infos;
   SYSTEM_PROCESSOR_PERFORMANCE_INFORMATION* sppi;
   DWORD sppi_size;
   SYSTEM_INFO system_info;
-  DWORD cpu_count, i, r;
+  DWORD cpu_count, r, i;
+  NTSTATUS status;
   ULONG result_size;
-  size_t size;
-  uv_err_t err;
+  int err;
   uv_cpu_info_t* cpu_info;
 
-  *cpu_infos = NULL;
-  *count = 0;
+  cpu_infos = NULL;
+  cpu_count = 0;
+  sppi = NULL;
 
   uv__once_init();
 
   GetSystemInfo(&system_info);
   cpu_count = system_info.dwNumberOfProcessors;
 
-  size = cpu_count * sizeof(uv_cpu_info_t);
-  *cpu_infos = (uv_cpu_info_t*) malloc(size);
-  if (*cpu_infos == NULL) {
-    err = uv__new_artificial_error(UV_ENOMEM);
-    goto out;
-  }
-  memset(*cpu_infos, 0, size);
-
-  sppi_size = sizeof(*sppi) * cpu_count;
-  sppi = (SYSTEM_PROCESSOR_PERFORMANCE_INFORMATION*) malloc(sppi_size);
-  if (!sppi) {
-    uv_fatal_error(ERROR_OUTOFMEMORY, "malloc");
+  cpu_infos = calloc(cpu_count, sizeof *cpu_infos);
+  if (cpu_infos == NULL) {
+    err = ERROR_OUTOFMEMORY;
+    goto error;
   }
 
-  r = pNtQuerySystemInformation(SystemProcessorPerformanceInformation,
-                                sppi,
-                                sppi_size,
-                                &result_size);
-  if (r != ERROR_SUCCESS || result_size != sppi_size) {
-    err = uv__new_sys_error(GetLastError());
-    goto out;
+  sppi_size = cpu_count * sizeof(*sppi);
+  sppi = malloc(sppi_size);
+  if (sppi == NULL) {
+    err = ERROR_OUTOFMEMORY;
+    goto error;
   }
 
+  status = pNtQuerySystemInformation(SystemProcessorPerformanceInformation,
+                                     sppi,
+                                     sppi_size,
+                                     &result_size);
+  if (!NT_SUCCESS(status)) {
+    err = pRtlNtStatusToDosError(status);
+    goto error;
+  }
+
+  assert(result_size == sppi_size);
+
   for (i = 0; i < cpu_count; i++) {
     WCHAR key_name[128];
     HKEY processor_key;
@@ -630,11 +633,14 @@ uv_err_t uv_cpu_info(uv_cpu_info_t** cpu_infos, int* count) {
     DWORD cpu_speed_size = sizeof(cpu_speed);
     WCHAR cpu_brand[256];
     DWORD cpu_brand_size = sizeof(cpu_brand);
+    int len;
 
-    _snwprintf(key_name,
-               ARRAY_SIZE(key_name),
-               L"HARDWARE\\DESCRIPTION\\System\\CentralProcessor\\%d",
-               i);
+    len = _snwprintf(key_name,
+                     ARRAY_SIZE(key_name),
+                     L"HARDWARE\\DESCRIPTION\\System\\CentralProcessor\\%d",
+                     i);
+
+    assert(len > 0 && len < ARRAY_SIZE(key_name));
 
     r = RegOpenKeyExW(HKEY_LOCAL_MACHINE,
                       key_name,
@@ -642,33 +648,35 @@ uv_err_t uv_cpu_info(uv_cpu_info_t** cpu_infos, int* count) {
                       KEY_QUERY_VALUE,
                       &processor_key);
     if (r != ERROR_SUCCESS) {
-      err = uv__new_sys_error(GetLastError());
-      goto out;
+      err = GetLastError();
+      goto error;
     }
 
     if (RegQueryValueExW(processor_key,
                          L"~MHz",
-                         NULL, NULL,
+                         NULL,
+                         NULL,
                          (BYTE*) &cpu_speed,
                          &cpu_speed_size) != ERROR_SUCCESS) {
-      err = uv__new_sys_error(GetLastError());
+      err = GetLastError();
       RegCloseKey(processor_key);
-      goto out;
+      goto error;
     }
 
     if (RegQueryValueExW(processor_key,
                          L"ProcessorNameString",
-                         NULL, NULL,
+                         NULL,
+                         NULL,
                          (BYTE*) &cpu_brand,
                          &cpu_brand_size) != ERROR_SUCCESS) {
-      err = uv__new_sys_error(GetLastError());
+      err = GetLastError();
       RegCloseKey(processor_key);
-      goto out;
+      goto error;
     }
 
     RegCloseKey(processor_key);
 
-    cpu_info = &(*cpu_infos)[i];
+    cpu_info = &cpu_infos[i];
     cpu_info->speed = cpu_speed;
     cpu_info->cpu_times.user = sppi[i].UserTime.QuadPart / 10000;
     cpu_info->cpu_times.sys = (sppi[i].KernelTime.QuadPart -
@@ -677,59 +685,61 @@ uv_err_t uv_cpu_info(uv_cpu_info_t** cpu_infos, int* count) {
     cpu_info->cpu_times.irq = sppi[i].InterruptTime.QuadPart / 10000;
     cpu_info->cpu_times.nice = 0;
 
-    size = uv_utf16_to_utf8(cpu_brand,
-                            cpu_brand_size / sizeof(WCHAR),
-                            NULL,
-                            0);
-    if (size == 0) {
-      err = uv__new_sys_error(GetLastError());
-      goto out;
+
+    len = WideCharToMultiByte(CP_UTF8,
+                              0,
+                              cpu_brand,
+                              cpu_brand_size / sizeof(WCHAR),
+                              NULL,
+                              0,
+                              NULL,
+                              NULL);
+    if (len == 0) {
+      err = GetLastError();
+      goto error;
     }
 
+    assert(len > 0);
+
     /* Allocate 1 extra byte for the null terminator. */
-    cpu_info->model = (char*) malloc(size + 1);
+    cpu_info->model = malloc(len + 1);
     if (cpu_info->model == NULL) {
-      err = uv__new_artificial_error(UV_ENOMEM);
-      goto out;
+      err = ERROR_OUTOFMEMORY;
+      goto error;
     }
 
-    if (uv_utf16_to_utf8(cpu_brand,
-                         cpu_brand_size / sizeof(WCHAR),
-                         cpu_info->model,
-                         size) == 0) {
-      err = uv__new_sys_error(GetLastError());
-      goto out;
+    if (WideCharToMultiByte(CP_UTF8,
+                            0,
+                            cpu_brand,
+                            cpu_brand_size / sizeof(WCHAR),
+                            cpu_info->model,
+                            len,
+                            NULL,
+                            NULL) == 0) {
+      err = GetLastError();
+      goto error;
     }
 
     /* Ensure that cpu_info->model is null terminated. */
-    cpu_info->model[size] = '\0';
-
-    (*count)++;
+    cpu_info->model[len] = '\0';
   }
 
-  err = uv_ok_;
+  free(sppi);
 
- out:
-  if (sppi) {
-    free(sppi);
-  }
+  *cpu_count_ptr = cpu_count;
+  *cpu_infos_ptr = cpu_infos;
 
-  if (err.code != UV_OK &&
-      *cpu_infos != NULL) {
-    int i;
+  return 0;
 
-    for (i = 0; i < *count; i++) {
-      /* This is safe because the cpu_infos memory area is zeroed out */
-      /* immediately after allocating it. */
-      free((*cpu_infos)[i].model);
-    }
-    free(*cpu_infos);
+ error:
+  /* This is safe because the cpu_infos array is zeroed on allocation. */
+  for (i = 0; i < cpu_count; i++)
+    free(cpu_infos[i].model);
 
-    *cpu_infos = NULL;
-    *count = 0;
-  }
+  free(cpu_infos);
+  free(sppi);
 
-  return err;
+  return uv_translate_sys_error(err);
 }
 
 
@@ -744,7 +754,7 @@ void uv_free_cpu_info(uv_cpu_info_t* cpu_infos, int count) {
 }
 
 
-uv_err_t uv_interface_addresses(uv_interface_address_t** addresses_ptr,
+int uv_interface_addresses(uv_interface_address_t** addresses_ptr,
     int* count_ptr) {
   IP_ADAPTER_ADDRESSES* win_address_buf;
   ULONG win_address_buf_size;
@@ -769,7 +779,7 @@ uv_err_t uv_interface_addresses(uv_interface_address_t** addresses_ptr,
     /* ERROR_BUFFER_OVERFLOW, and the required buffer size will be stored in */
     /* win_address_buf_size. */
     r = GetAdaptersAddresses(AF_UNSPEC,
-                             0,
+                             GAA_FLAG_INCLUDE_PREFIX,
                              NULL,
                              win_address_buf,
                              &win_address_buf_size);
@@ -785,7 +795,7 @@ uv_err_t uv_interface_addresses(uv_interface_address_t** addresses_ptr,
         /* all adapters. */
         win_address_buf = malloc(win_address_buf_size);
         if (win_address_buf == NULL)
-          return uv__new_artificial_error(UV_ENOMEM);
+          return UV_ENOMEM;
 
         continue;
 
@@ -793,16 +803,16 @@ uv_err_t uv_interface_addresses(uv_interface_address_t** addresses_ptr,
         /* No adapters were found. */
         uv_address_buf = malloc(1);
         if (uv_address_buf == NULL)
-          return uv__new_artificial_error(UV_ENOMEM);
+          return UV_ENOMEM;
 
         *count_ptr = 0;
         *addresses_ptr = uv_address_buf;
 
-        return uv_ok_;
+        return 0;
       }
 
       case ERROR_ADDRESS_NOT_ASSOCIATED:
-        return uv__new_artificial_error(UV_EAGAIN);
+        return UV_EAGAIN;
 
       case ERROR_INVALID_PARAMETER:
         /* MSDN says:
@@ -813,13 +823,13 @@ uv_err_t uv_interface_addresses(uv_interface_address_t** addresses_ptr,
          * Since the first two conditions are not met, it must be that the
          * adapter data is too big.
          */
-        return uv__new_artificial_error(UV_ENOBUFS);
+        return UV_ENOBUFS;
 
       default:
         /* Other (unspecified) errors can happen, but we don't have any */
         /* special meaning for them. */
         assert(r != ERROR_SUCCESS);
-        return uv__new_sys_error(r);
+        return uv_translate_sys_error(r);
     }
   }
 
@@ -854,7 +864,7 @@ uv_err_t uv_interface_addresses(uv_interface_address_t** addresses_ptr,
                                     FALSE);
     if (name_size <= 0) {
       free(win_address_buf);
-      return uv__new_sys_error(GetLastError());
+      return uv_translate_sys_error(GetLastError());
     }
     uv_address_buf_size += name_size;
 
@@ -873,7 +883,7 @@ uv_err_t uv_interface_addresses(uv_interface_address_t** addresses_ptr,
   uv_address_buf = malloc(uv_address_buf_size);
   if (uv_address_buf == NULL) {
     free(win_address_buf);
-    return uv__new_artificial_error(UV_ENOMEM);
+    return UV_ENOMEM;
   }
 
   /* Compute the start of the uv_interface_address_t array, and the place in */
@@ -886,6 +896,7 @@ uv_err_t uv_interface_addresses(uv_interface_address_t** addresses_ptr,
        win_address != NULL;
        win_address = win_address->Next) {
     IP_ADAPTER_UNICAST_ADDRESS_XP* unicast_address;
+    IP_ADAPTER_PREFIX* prefix;
     int name_size;
     size_t max_name_size;
 
@@ -908,27 +919,52 @@ uv_err_t uv_interface_addresses(uv_interface_address_t** addresses_ptr,
     if (name_size <= 0) {
       free(win_address_buf);
       free(uv_address_buf);
-      return uv__new_sys_error(GetLastError());
+      return uv_translate_sys_error(GetLastError());
     }
 
+    prefix = win_address->FirstPrefix;
+
     /* Add an uv_interface_address_t element for every unicast address. */
+    /* Walk the prefix list in tandem with the address list. */
     for (unicast_address = (IP_ADAPTER_UNICAST_ADDRESS_XP*)
                            win_address->FirstUnicastAddress;
-         unicast_address != NULL;
-         unicast_address = unicast_address->Next) {
+         unicast_address != NULL && prefix != NULL;
+         unicast_address = unicast_address->Next, prefix = prefix->Next) {
       struct sockaddr* sa;
+      ULONG prefix_len;
+
+      sa = unicast_address->Address.lpSockaddr;
+      prefix_len = prefix->PrefixLength;
+
+      memset(uv_address, 0, sizeof *uv_address);
 
       uv_address->name = name_buf;
 
-      sa = unicast_address->Address.lpSockaddr;
-      if (sa->sa_family == AF_INET6)
-        uv_address->address.address6 = *((struct sockaddr_in6 *) sa);
-      else
-        uv_address->address.address4 = *((struct sockaddr_in *) sa);
+      if (win_address->PhysicalAddressLength == sizeof(uv_address->phys_addr)) {
+        memcpy(uv_address->phys_addr,
+               win_address->PhysicalAddress,
+               sizeof(uv_address->phys_addr));
+      }
 
       uv_address->is_internal =
           (win_address->IfType == IF_TYPE_SOFTWARE_LOOPBACK);
 
+      if (sa->sa_family == AF_INET6) {
+        uv_address->address.address6 = *((struct sockaddr_in6 *) sa);
+
+        uv_address->netmask.netmask6.sin6_family = AF_INET6;
+        memset(uv_address->netmask.netmask6.sin6_addr.s6_addr, 0xff, prefix_len >> 3);
+        uv_address->netmask.netmask6.sin6_addr.s6_addr[prefix_len >> 3] =
+            0xff << (8 - prefix_len % 8);
+
+      } else {
+        uv_address->address.address4 = *((struct sockaddr_in *) sa);
+
+        uv_address->netmask.netmask4.sin_family = AF_INET;
+        uv_address->netmask.netmask4.sin_addr.s_addr =
+            htonl(0xffffffff << (32 - prefix_len));
+      }
+
       uv_address++;
     }
 
@@ -940,7 +976,7 @@ uv_err_t uv_interface_addresses(uv_interface_address_t** addresses_ptr,
   *addresses_ptr = uv_address_buf;
   *count_ptr = count;
 
-  return uv_ok_;
+  return 0;
 }
 
 
