@@ -104,16 +104,16 @@ static char buf[32];
 static char test_buf[] = "test-buffer\n";
 
 
-void check_permission(const char* filename, int mode) {
+static void check_permission(const char* filename, unsigned int mode) {
   int r;
   uv_fs_t req;
-  struct stat* s;
+  uv_stat_t* s;
 
   r = uv_fs_stat(uv_default_loop(), &req, filename, NULL);
   ASSERT(r == 0);
   ASSERT(req.result == 0);
 
-  s = req.ptr;
+  s = &req.statbuf;
 #ifdef _WIN32
   /*
    * On Windows, chmod can only modify S_IWUSR (_S_IWRITE) bit,
@@ -196,9 +196,13 @@ static void chown_root_cb(uv_fs_t* req) {
   /* On windows, chown is a no-op and always succeeds. */
   ASSERT(req->result == 0);
 #else
-  /* On unix, chown'ing the root directory is not allowed. */
-  ASSERT(req->result == -1);
-  ASSERT(req->errorno == UV_EPERM);
+  /* On unix, chown'ing the root directory is not allowed -
+   * unless you're root, of course.
+   */
+  if (geteuid() == 0)
+    ASSERT(req->result == 0);
+  else
+    ASSERT(req->result == UV_EPERM);
 #endif
   chown_cb_count++;
   uv_fs_req_cleanup(req);
@@ -207,13 +211,13 @@ static void chown_root_cb(uv_fs_t* req) {
 static void unlink_cb(uv_fs_t* req) {
   ASSERT(req == &unlink_req);
   ASSERT(req->fs_type == UV_FS_UNLINK);
-  ASSERT(req->result != -1);
+  ASSERT(req->result == 0);
   unlink_cb_count++;
   uv_fs_req_cleanup(req);
 }
 
 static void fstat_cb(uv_fs_t* req) {
-  struct stat* s = req->ptr;
+  uv_stat_t* s = req->ptr;
   ASSERT(req->fs_type == UV_FS_FSTAT);
   ASSERT(req->result == 0);
   ASSERT(s->st_size == sizeof(test_buf));
@@ -226,7 +230,7 @@ static void close_cb(uv_fs_t* req) {
   int r;
   ASSERT(req == &close_req);
   ASSERT(req->fs_type == UV_FS_CLOSE);
-  ASSERT(req->result != -1);
+  ASSERT(req->result == 0);
   close_cb_count++;
   uv_fs_req_cleanup(req);
   if (close_cb_count == 3) {
@@ -240,7 +244,7 @@ static void ftruncate_cb(uv_fs_t* req) {
   int r;
   ASSERT(req == &ftruncate_req);
   ASSERT(req->fs_type == UV_FS_FTRUNCATE);
-  ASSERT(req->result != -1);
+  ASSERT(req->result == 0);
   ftruncate_cb_count++;
   uv_fs_req_cleanup(req);
   r = uv_fs_close(loop, &close_req, open_req1.result, close_cb);
@@ -252,7 +256,7 @@ static void read_cb(uv_fs_t* req) {
   int r;
   ASSERT(req == &read_req);
   ASSERT(req->fs_type == UV_FS_READ);
-  ASSERT(req->result != -1);
+  ASSERT(req->result >= 0);  /* FIXME(bnoordhuis) Check if requested size? */
   read_cb_count++;
   uv_fs_req_cleanup(req);
   if (read_cb_count == 1) {
@@ -272,8 +276,7 @@ static void open_cb(uv_fs_t* req) {
   ASSERT(req == &open_req1);
   ASSERT(req->fs_type == UV_FS_OPEN);
   if (req->result < 0) {
-    /* TODO get error with uv_last_error() */
-    fprintf(stderr, "async open error: %d\n", req->errorno);
+    fprintf(stderr, "async open error: %d\n", (int) req->result);
     ASSERT(0);
   }
   open_cb_count++;
@@ -290,8 +293,7 @@ static void open_cb(uv_fs_t* req) {
 static void open_cb_simple(uv_fs_t* req) {
   ASSERT(req->fs_type == UV_FS_OPEN);
   if (req->result < 0) {
-    /* TODO get error with uv_last_error() */
-    fprintf(stderr, "async open error: %d\n", req->errorno);
+    fprintf(stderr, "async open error: %d\n", (int) req->result);
     ASSERT(0);
   }
   open_cb_count++;
@@ -304,7 +306,7 @@ static void fsync_cb(uv_fs_t* req) {
   int r;
   ASSERT(req == &fsync_req);
   ASSERT(req->fs_type == UV_FS_FSYNC);
-  ASSERT(req->result != -1);
+  ASSERT(req->result == 0);
   fsync_cb_count++;
   uv_fs_req_cleanup(req);
   r = uv_fs_close(loop, &close_req, open_req1.result, close_cb);
@@ -316,7 +318,7 @@ static void fdatasync_cb(uv_fs_t* req) {
   int r;
   ASSERT(req == &fdatasync_req);
   ASSERT(req->fs_type == UV_FS_FDATASYNC);
-  ASSERT(req->result != -1);
+  ASSERT(req->result == 0);
   fdatasync_cb_count++;
   uv_fs_req_cleanup(req);
   r = uv_fs_fsync(loop, &fsync_req, open_req1.result, fsync_cb);
@@ -328,7 +330,7 @@ static void write_cb(uv_fs_t* req) {
   int r;
   ASSERT(req == &write_req);
   ASSERT(req->fs_type == UV_FS_WRITE);
-  ASSERT(req->result != -1);
+  ASSERT(req->result >= 0);  /* FIXME(bnoordhuis) Check if requested size? */
   write_cb_count++;
   uv_fs_req_cleanup(req);
   r = uv_fs_fdatasync(loop, &fdatasync_req, open_req1.result, fdatasync_cb);
@@ -340,7 +342,7 @@ static void create_cb(uv_fs_t* req) {
   int r;
   ASSERT(req == &open_req1);
   ASSERT(req->fs_type == UV_FS_OPEN);
-  ASSERT(req->result != -1);
+  ASSERT(req->result >= 0);
   create_cb_count++;
   uv_fs_req_cleanup(req);
   r = uv_fs_write(loop, &write_req, req->result, test_buf, sizeof(test_buf),
@@ -352,7 +354,7 @@ static void create_cb(uv_fs_t* req) {
 static void rename_cb(uv_fs_t* req) {
   ASSERT(req == &rename_req);
   ASSERT(req->fs_type == UV_FS_RENAME);
-  ASSERT(req->result != -1);
+  ASSERT(req->result == 0);
   rename_cb_count++;
   uv_fs_req_cleanup(req);
 }
@@ -361,7 +363,7 @@ static void rename_cb(uv_fs_t* req) {
 static void mkdir_cb(uv_fs_t* req) {
   ASSERT(req == &mkdir_req);
   ASSERT(req->fs_type == UV_FS_MKDIR);
-  ASSERT(req->result != -1);
+  ASSERT(req->result == 0);
   mkdir_cb_count++;
   ASSERT(req->path);
   ASSERT(memcmp(req->path, "test_dir\0", 9) == 0);
@@ -372,7 +374,7 @@ static void mkdir_cb(uv_fs_t* req) {
 static void rmdir_cb(uv_fs_t* req) {
   ASSERT(req == &rmdir_req);
   ASSERT(req->fs_type == UV_FS_RMDIR);
-  ASSERT(req->result != -1);
+  ASSERT(req->result == 0);
   rmdir_cb_count++;
   ASSERT(req->path);
   ASSERT(memcmp(req->path, "test_dir\0", 9) == 0);
@@ -408,9 +410,8 @@ static void empty_readdir_cb(uv_fs_t* req) {
 static void file_readdir_cb(uv_fs_t* req) {
   ASSERT(req == &readdir_req);
   ASSERT(req->fs_type == UV_FS_READDIR);
-  ASSERT(req->result == -1);
+  ASSERT(req->result == UV_ENOTDIR);
   ASSERT(req->ptr == NULL);
-  ASSERT(uv_last_error(req->loop).code == UV_ENOTDIR);
   uv_fs_req_cleanup(req);
   readdir_cb_count++;
 }
@@ -419,7 +420,7 @@ static void file_readdir_cb(uv_fs_t* req) {
 static void stat_cb(uv_fs_t* req) {
   ASSERT(req == &stat_req);
   ASSERT(req->fs_type == UV_FS_STAT || req->fs_type == UV_FS_LSTAT);
-  ASSERT(req->result != -1);
+  ASSERT(req->result == 0);
   ASSERT(req->ptr);
   stat_cb_count++;
   uv_fs_req_cleanup(req);
@@ -438,24 +439,21 @@ static void sendfile_cb(uv_fs_t* req) {
 
 static void open_noent_cb(uv_fs_t* req) {
   ASSERT(req->fs_type == UV_FS_OPEN);
-  ASSERT(req->errorno == UV_ENOENT);
-  ASSERT(req->result == -1);
+  ASSERT(req->result == UV_ENOENT);
   open_cb_count++;
   uv_fs_req_cleanup(req);
 }
 
 static void open_nametoolong_cb(uv_fs_t* req) {
   ASSERT(req->fs_type == UV_FS_OPEN);
-  ASSERT(req->errorno == UV_ENAMETOOLONG);
-  ASSERT(req->result == -1);
+  ASSERT(req->result == UV_ENAMETOOLONG);
   open_cb_count++;
   uv_fs_req_cleanup(req);
 }
 
 static void open_loop_cb(uv_fs_t* req) {
   ASSERT(req->fs_type == UV_FS_OPEN);
-  ASSERT(req->errorno == UV_ELOOP);
-  ASSERT(req->result == -1);
+  ASSERT(req->result == UV_ELOOP);
   open_cb_count++;
   uv_fs_req_cleanup(req);
 }
@@ -468,16 +466,15 @@ TEST_IMPL(fs_file_noent) {
   loop = uv_default_loop();
 
   r = uv_fs_open(loop, &req, "does_not_exist", O_RDONLY, 0, NULL);
-  ASSERT(r == -1);
-  ASSERT(req.result == -1);
-  ASSERT(uv_last_error(loop).code == UV_ENOENT);
+  ASSERT(r == UV_ENOENT);
+  ASSERT(req.result == UV_ENOENT);
   uv_fs_req_cleanup(&req);
 
   r = uv_fs_open(loop, &req, "does_not_exist", O_RDONLY, 0, open_noent_cb);
   ASSERT(r == 0);
 
   ASSERT(open_cb_count == 0);
-  uv_run(loop);
+  uv_run(loop, UV_RUN_DEFAULT);
   ASSERT(open_cb_count == 1);
 
   /* TODO add EACCES test */
@@ -497,16 +494,15 @@ TEST_IMPL(fs_file_nametoolong) {
   name[TOO_LONG_NAME_LENGTH] = 0;
 
   r = uv_fs_open(loop, &req, name, O_RDONLY, 0, NULL);
-  ASSERT(r == -1);
-  ASSERT(req.result == -1);
-  ASSERT(uv_last_error(loop).code == UV_ENAMETOOLONG);
+  ASSERT(r == UV_ENAMETOOLONG);
+  ASSERT(req.result == UV_ENAMETOOLONG);
   uv_fs_req_cleanup(&req);
 
   r = uv_fs_open(loop, &req, name, O_RDONLY, 0, open_nametoolong_cb);
   ASSERT(r == 0);
 
   ASSERT(open_cb_count == 0);
-  uv_run(loop);
+  uv_run(loop, UV_RUN_DEFAULT);
   ASSERT(open_cb_count == 1);
 
   MAKE_VALGRIND_HAPPY();
@@ -524,16 +520,15 @@ TEST_IMPL(fs_file_loop) {
   uv_fs_req_cleanup(&req);
 
   r = uv_fs_open(loop, &req, "test_symlink", O_RDONLY, 0, NULL);
-  ASSERT(r == -1);
-  ASSERT(req.result == -1);
-  ASSERT(uv_last_error(loop).code == UV_ELOOP);
+  ASSERT(r == UV_ELOOP);
+  ASSERT(req.result == UV_ELOOP);
   uv_fs_req_cleanup(&req);
 
   r = uv_fs_open(loop, &req, "test_symlink", O_RDONLY, 0, open_loop_cb);
   ASSERT(r == 0);
 
   ASSERT(open_cb_count == 0);
-  uv_run(loop);
+  uv_run(loop, UV_RUN_DEFAULT);
   ASSERT(open_cb_count == 1);
 
   unlink("test_symlink");
@@ -543,7 +538,7 @@ TEST_IMPL(fs_file_loop) {
 }
 
 static void check_utime(const char* path, double atime, double mtime) {
-  struct stat* s;
+  uv_stat_t* s;
   uv_fs_t req;
   int r;
 
@@ -551,18 +546,10 @@ static void check_utime(const char* path, double atime, double mtime) {
   ASSERT(r == 0);
 
   ASSERT(req.result == 0);
-  s = req.ptr;
+  s = &req.statbuf;
 
-#if defined(_WIN32) || defined(_AIX)
-  ASSERT(s->st_atime == atime);
-  ASSERT(s->st_mtime == mtime);
-#elif !defined(_POSIX_C_SOURCE) || defined(_DARWIN_C_SOURCE)
-  ASSERT(s->st_atimespec.tv_sec  == atime);
-  ASSERT(s->st_mtimespec.tv_sec  == mtime);
-#else
   ASSERT(s->st_atim.tv_sec  == atime);
   ASSERT(s->st_mtim.tv_sec  == mtime);
-#endif
 
   uv_fs_req_cleanup(&req);
 }
@@ -608,9 +595,9 @@ TEST_IMPL(fs_file_async) {
   loop = uv_default_loop();
 
   r = uv_fs_open(loop, &open_req1, "test_file", O_WRONLY | O_CREAT,
-      S_IREAD | S_IWRITE, create_cb);
+      S_IRUSR | S_IWUSR, create_cb);
   ASSERT(r == 0);
-  uv_run(loop);
+  uv_run(loop, UV_RUN_DEFAULT);
 
   ASSERT(create_cb_count == 1);
   ASSERT(write_cb_count == 1);
@@ -621,7 +608,7 @@ TEST_IMPL(fs_file_async) {
   r = uv_fs_rename(loop, &rename_req, "test_file", "test_file2", rename_cb);
   ASSERT(r == 0);
 
-  uv_run(loop);
+  uv_run(loop, UV_RUN_DEFAULT);
   ASSERT(create_cb_count == 1);
   ASSERT(write_cb_count == 1);
   ASSERT(close_cb_count == 1);
@@ -630,7 +617,7 @@ TEST_IMPL(fs_file_async) {
   r = uv_fs_open(loop, &open_req1, "test_file2", O_RDWR, 0, open_cb);
   ASSERT(r == 0);
 
-  uv_run(loop);
+  uv_run(loop, UV_RUN_DEFAULT);
   ASSERT(open_cb_count == 1);
   ASSERT(read_cb_count == 1);
   ASSERT(close_cb_count == 2);
@@ -642,7 +629,7 @@ TEST_IMPL(fs_file_async) {
   r = uv_fs_open(loop, &open_req1, "test_file2", O_RDONLY, 0, open_cb);
   ASSERT(r == 0);
 
-  uv_run(loop);
+  uv_run(loop, UV_RUN_DEFAULT);
   ASSERT(open_cb_count == 2);
   ASSERT(read_cb_count == 2);
   ASSERT(close_cb_count == 3);
@@ -671,70 +658,70 @@ TEST_IMPL(fs_file_sync) {
   loop = uv_default_loop();
 
   r = uv_fs_open(loop, &open_req1, "test_file", O_WRONLY | O_CREAT,
-      S_IWRITE | S_IREAD, NULL);
-  ASSERT(r != -1);
-  ASSERT(open_req1.result != -1);
+      S_IWUSR | S_IRUSR, NULL);
+  ASSERT(r >= 0);
+  ASSERT(open_req1.result >= 0);
   uv_fs_req_cleanup(&open_req1);
 
   r = uv_fs_write(loop, &write_req, open_req1.result, test_buf,
       sizeof(test_buf), -1, NULL);
-  ASSERT(r != -1);
-  ASSERT(write_req.result != -1);
+  ASSERT(r >= 0);
+  ASSERT(write_req.result >= 0);
   uv_fs_req_cleanup(&write_req);
 
   r = uv_fs_close(loop, &close_req, open_req1.result, NULL);
-  ASSERT(r != -1);
-  ASSERT(close_req.result != -1);
+  ASSERT(r == 0);
+  ASSERT(close_req.result == 0);
   uv_fs_req_cleanup(&close_req);
 
   r = uv_fs_open(loop, &open_req1, "test_file", O_RDWR, 0, NULL);
-  ASSERT(r != -1);
-  ASSERT(open_req1.result != -1);
+  ASSERT(r >= 0);
+  ASSERT(open_req1.result >= 0);
   uv_fs_req_cleanup(&open_req1);
 
   r = uv_fs_read(loop, &read_req, open_req1.result, buf, sizeof(buf), -1,
       NULL);
-  ASSERT(r != -1);
-  ASSERT(read_req.result != -1);
+  ASSERT(r >= 0);
+  ASSERT(read_req.result >= 0);
   ASSERT(strcmp(buf, test_buf) == 0);
   uv_fs_req_cleanup(&read_req);
 
   r = uv_fs_ftruncate(loop, &ftruncate_req, open_req1.result, 7, NULL);
-  ASSERT(r != -1);
-  ASSERT(ftruncate_req.result != -1);
+  ASSERT(r == 0);
+  ASSERT(ftruncate_req.result == 0);
   uv_fs_req_cleanup(&ftruncate_req);
 
   r = uv_fs_close(loop, &close_req, open_req1.result, NULL);
-  ASSERT(r != -1);
-  ASSERT(close_req.result != -1);
+  ASSERT(r == 0);
+  ASSERT(close_req.result == 0);
   uv_fs_req_cleanup(&close_req);
 
   r = uv_fs_rename(loop, &rename_req, "test_file", "test_file2", NULL);
-  ASSERT(r != -1);
-  ASSERT(rename_req.result != -1);
+  ASSERT(r == 0);
+  ASSERT(rename_req.result == 0);
   uv_fs_req_cleanup(&rename_req);
 
   r = uv_fs_open(loop, &open_req1, "test_file2", O_RDONLY, 0, NULL);
-  ASSERT(r != -1);
-  ASSERT(open_req1.result != -1);
+  ASSERT(r >= 0);
+  ASSERT(open_req1.result >= 0);
   uv_fs_req_cleanup(&open_req1);
 
   memset(buf, 0, sizeof(buf));
   r = uv_fs_read(loop, &read_req, open_req1.result, buf, sizeof(buf), -1,
       NULL);
-  ASSERT(r != -1);
-  ASSERT(read_req.result != -1);
+  ASSERT(r >= 0);
+  ASSERT(read_req.result >= 0);
   ASSERT(strcmp(buf, "test-bu") == 0);
   uv_fs_req_cleanup(&read_req);
 
   r = uv_fs_close(loop, &close_req, open_req1.result, NULL);
-  ASSERT(r != -1);
-  ASSERT(close_req.result != -1);
+  ASSERT(r == 0);
+  ASSERT(close_req.result == 0);
   uv_fs_req_cleanup(&close_req);
 
   r = uv_fs_unlink(loop, &unlink_req, "test_file2", NULL);
-  ASSERT(r != -1);
-  ASSERT(unlink_req.result != -1);
+  ASSERT(r == 0);
+  ASSERT(unlink_req.result == 0);
   uv_fs_req_cleanup(&unlink_req);
 
   /* Cleanup */
@@ -759,21 +746,21 @@ TEST_IMPL(fs_async_dir) {
   r = uv_fs_mkdir(loop, &mkdir_req, "test_dir", 0755, mkdir_cb);
   ASSERT(r == 0);
 
-  uv_run(loop);
+  uv_run(loop, UV_RUN_DEFAULT);
   ASSERT(mkdir_cb_count == 1);
 
   /* Create 2 files synchronously. */
   r = uv_fs_open(loop, &open_req1, "test_dir/file1", O_WRONLY | O_CREAT,
-      S_IWRITE | S_IREAD, NULL);
-  ASSERT(r != -1);
+      S_IWUSR | S_IRUSR, NULL);
+  ASSERT(r >= 0);
   uv_fs_req_cleanup(&open_req1);
   r = uv_fs_close(loop, &close_req, open_req1.result, NULL);
   ASSERT(r == 0);
   uv_fs_req_cleanup(&close_req);
 
   r = uv_fs_open(loop, &open_req1, "test_dir/file2", O_WRONLY | O_CREAT,
-      S_IWRITE | S_IREAD, NULL);
-  ASSERT(r != -1);
+      S_IWUSR | S_IRUSR, NULL);
+  ASSERT(r >= 0);
   uv_fs_req_cleanup(&open_req1);
   r = uv_fs_close(loop, &close_req, open_req1.result, NULL);
   ASSERT(r == 0);
@@ -782,7 +769,7 @@ TEST_IMPL(fs_async_dir) {
   r = uv_fs_readdir(loop, &readdir_req, "test_dir", 0, readdir_cb);
   ASSERT(r == 0);
 
-  uv_run(loop);
+  uv_run(loop, UV_RUN_DEFAULT);
   ASSERT(readdir_cb_count == 1);
 
   /* sync uv_fs_readdir */
@@ -797,35 +784,35 @@ TEST_IMPL(fs_async_dir) {
 
   r = uv_fs_stat(loop, &stat_req, "test_dir", stat_cb);
   ASSERT(r == 0);
-  uv_run(loop);
+  uv_run(loop, UV_RUN_DEFAULT);
 
   r = uv_fs_stat(loop, &stat_req, "test_dir/", stat_cb);
   ASSERT(r == 0);
-  uv_run(loop);
+  uv_run(loop, UV_RUN_DEFAULT);
 
   r = uv_fs_lstat(loop, &stat_req, "test_dir", stat_cb);
   ASSERT(r == 0);
-  uv_run(loop);
+  uv_run(loop, UV_RUN_DEFAULT);
 
   r = uv_fs_lstat(loop, &stat_req, "test_dir/", stat_cb);
   ASSERT(r == 0);
-  uv_run(loop);
+  uv_run(loop, UV_RUN_DEFAULT);
 
   ASSERT(stat_cb_count == 4);
 
   r = uv_fs_unlink(loop, &unlink_req, "test_dir/file1", unlink_cb);
   ASSERT(r == 0);
-  uv_run(loop);
+  uv_run(loop, UV_RUN_DEFAULT);
   ASSERT(unlink_cb_count == 1);
 
   r = uv_fs_unlink(loop, &unlink_req, "test_dir/file2", unlink_cb);
   ASSERT(r == 0);
-  uv_run(loop);
+  uv_run(loop, UV_RUN_DEFAULT);
   ASSERT(unlink_cb_count == 2);
 
   r = uv_fs_rmdir(loop, &rmdir_req, "test_dir", rmdir_cb);
   ASSERT(r == 0);
-  uv_run(loop);
+  uv_run(loop, UV_RUN_DEFAULT);
   ASSERT(rmdir_cb_count == 1);
 
   /* Cleanup */
@@ -848,7 +835,7 @@ TEST_IMPL(fs_async_sendfile) {
   unlink("test_file");
   unlink("test_file2");
 
-  f = open("test_file", O_WRONLY | O_CREAT, S_IWRITE | S_IREAD);
+  f = open("test_file", O_WRONLY | O_CREAT, S_IWUSR | S_IRUSR);
   ASSERT(f != -1);
 
   r = write(f, "begin\n", 6);
@@ -865,20 +852,20 @@ TEST_IMPL(fs_async_sendfile) {
 
   /* Test starts here. */
   r = uv_fs_open(loop, &open_req1, "test_file", O_RDWR, 0, NULL);
-  ASSERT(r != -1);
-  ASSERT(open_req1.result != -1);
+  ASSERT(r >= 0);
+  ASSERT(open_req1.result >= 0);
   uv_fs_req_cleanup(&open_req1);
 
   r = uv_fs_open(loop, &open_req2, "test_file2", O_WRONLY | O_CREAT,
-      S_IWRITE | S_IREAD, NULL);
-  ASSERT(r != -1);
-  ASSERT(open_req2.result != -1);
+      S_IWUSR | S_IRUSR, NULL);
+  ASSERT(r >= 0);
+  ASSERT(open_req2.result >= 0);
   uv_fs_req_cleanup(&open_req2);
 
   r = uv_fs_sendfile(loop, &sendfile_req, open_req2.result, open_req1.result,
       0, 131072, sendfile_cb);
   ASSERT(r == 0);
-  uv_run(loop);
+  uv_run(loop, UV_RUN_DEFAULT);
 
   ASSERT(sendfile_cb_count == 1);
 
@@ -906,7 +893,10 @@ TEST_IMPL(fs_fstat) {
   int r;
   uv_fs_t req;
   uv_file file;
-  struct stat* s;
+  uv_stat_t* s;
+#ifndef _WIN32
+  struct stat t;
+#endif
 
   /* Setup. */
   unlink("test_file");
@@ -914,9 +904,9 @@ TEST_IMPL(fs_fstat) {
   loop = uv_default_loop();
 
   r = uv_fs_open(loop, &req, "test_file", O_RDWR | O_CREAT,
-      S_IWRITE | S_IREAD, NULL);
-  ASSERT(r != -1);
-  ASSERT(req.result != -1);
+      S_IWUSR | S_IRUSR, NULL);
+  ASSERT(r >= 0);
+  ASSERT(req.result >= 0);
   file = req.result;
   uv_fs_req_cleanup(&req);
 
@@ -930,12 +920,67 @@ TEST_IMPL(fs_fstat) {
   ASSERT(req.result == 0);
   s = req.ptr;
   ASSERT(s->st_size == sizeof(test_buf));
+
+#ifndef _WIN32
+  r = fstat(file, &t);
+  ASSERT(r == 0);
+
+  ASSERT(s->st_dev == (uint64_t) t.st_dev);
+  ASSERT(s->st_mode == (uint64_t) t.st_mode);
+  ASSERT(s->st_nlink == (uint64_t) t.st_nlink);
+  ASSERT(s->st_uid == (uint64_t) t.st_uid);
+  ASSERT(s->st_gid == (uint64_t) t.st_gid);
+  ASSERT(s->st_rdev == (uint64_t) t.st_rdev);
+  ASSERT(s->st_ino == (uint64_t) t.st_ino);
+  ASSERT(s->st_size == (uint64_t) t.st_size);
+  ASSERT(s->st_blksize == (uint64_t) t.st_blksize);
+  ASSERT(s->st_blocks == (uint64_t) t.st_blocks);
+#if defined(__APPLE__)
+  ASSERT(s->st_atim.tv_sec == t.st_atimespec.tv_sec);
+  ASSERT(s->st_atim.tv_nsec == t.st_atimespec.tv_nsec);
+  ASSERT(s->st_mtim.tv_sec == t.st_mtimespec.tv_sec);
+  ASSERT(s->st_mtim.tv_nsec == t.st_mtimespec.tv_nsec);
+  ASSERT(s->st_ctim.tv_sec == t.st_ctimespec.tv_sec);
+  ASSERT(s->st_ctim.tv_nsec == t.st_ctimespec.tv_nsec);
+  ASSERT(s->st_birthtim.tv_sec == t.st_birthtimespec.tv_sec);
+  ASSERT(s->st_birthtim.tv_nsec == t.st_birthtimespec.tv_nsec);
+  ASSERT(s->st_flags == t.st_flags);
+  ASSERT(s->st_gen == t.st_gen);
+#elif defined(__sun) || \
+      defined(_BSD_SOURCE) || \
+      defined(_SVID_SOURCE) || \
+      defined(_XOPEN_SOURCE)
+  ASSERT(s->st_atim.tv_sec == t.st_atim.tv_sec);
+  ASSERT(s->st_atim.tv_nsec == t.st_atim.tv_nsec);
+  ASSERT(s->st_mtim.tv_sec == t.st_mtim.tv_sec);
+  ASSERT(s->st_mtim.tv_nsec == t.st_mtim.tv_nsec);
+  ASSERT(s->st_ctim.tv_sec == t.st_ctim.tv_sec);
+  ASSERT(s->st_ctim.tv_nsec == t.st_ctim.tv_nsec);
+# if defined(__DragonFly__)  || \
+      defined(__FreeBSD__)    || \
+      defined(__OpenBSD__)    || \
+      defined(__NetBSD__)
+  ASSERT(s->st_birthtim.tv_sec == t.st_birthtim.tv_sec);
+  ASSERT(s->st_birthtim.tv_nsec == t.st_birthtim.tv_nsec);
+  ASSERT(s->st_flags == t.st_flags);
+  ASSERT(s->st_gen == t.st_gen);
+# endif
+#else
+  ASSERT(s->st_atim.tv_sec == t.st_atime);
+  ASSERT(s->st_atim.tv_nsec == 0);
+  ASSERT(s->st_mtim.tv_sec == t.st_mtime);
+  ASSERT(s->st_mtim.tv_nsec == 0);
+  ASSERT(s->st_ctim.tv_sec == t.st_ctime);
+  ASSERT(s->st_ctim.tv_nsec == 0);
+#endif
+#endif
+
   uv_fs_req_cleanup(&req);
 
   /* Now do the uv_fs_fstat call asynchronously */
   r = uv_fs_fstat(loop, &req, file, fstat_cb);
   ASSERT(r == 0);
-  uv_run(loop);
+  uv_run(loop, UV_RUN_DEFAULT);
   ASSERT(fstat_cb_count == 1);
 
 
@@ -948,7 +993,7 @@ TEST_IMPL(fs_fstat) {
    * Run the loop just to check we don't have make any extraneous uv_ref()
    * calls. This should drop out immediately.
    */
-  uv_run(loop);
+  uv_run(loop, UV_RUN_DEFAULT);
 
   /* Cleanup. */
   unlink("test_file");
@@ -969,9 +1014,9 @@ TEST_IMPL(fs_chmod) {
   loop = uv_default_loop();
 
   r = uv_fs_open(loop, &req, "test_file", O_RDWR | O_CREAT,
-      S_IWRITE | S_IREAD, NULL);
-  ASSERT(r != -1);
-  ASSERT(req.result != -1);
+      S_IWUSR | S_IRUSR, NULL);
+  ASSERT(r >= 0);
+  ASSERT(req.result >= 0);
   file = req.result;
   uv_fs_req_cleanup(&req);
 
@@ -1014,7 +1059,7 @@ TEST_IMPL(fs_chmod) {
   }
   r = uv_fs_chmod(loop, &req, "test_file", 0200, chmod_cb);
   ASSERT(r == 0);
-  uv_run(loop);
+  uv_run(loop, UV_RUN_DEFAULT);
   ASSERT(chmod_cb_count == 1);
   chmod_cb_count = 0; /* reset for the next test */
 #endif
@@ -1026,7 +1071,7 @@ TEST_IMPL(fs_chmod) {
   }
   r = uv_fs_chmod(loop, &req, "test_file", 0400, chmod_cb);
   ASSERT(r == 0);
-  uv_run(loop);
+  uv_run(loop, UV_RUN_DEFAULT);
   ASSERT(chmod_cb_count == 1);
 
   /* async fchmod */
@@ -1036,7 +1081,7 @@ TEST_IMPL(fs_chmod) {
   }
   r = uv_fs_fchmod(loop, &req, file, 0600, fchmod_cb);
   ASSERT(r == 0);
-  uv_run(loop);
+  uv_run(loop, UV_RUN_DEFAULT);
   ASSERT(fchmod_cb_count == 1);
 
   close(file);
@@ -1045,7 +1090,7 @@ TEST_IMPL(fs_chmod) {
    * Run the loop just to check we don't have make any extraneous uv_ref()
    * calls. This should drop out immediately.
    */
-  uv_run(loop);
+  uv_run(loop, UV_RUN_DEFAULT);
 
   /* Cleanup. */
   unlink("test_file");
@@ -1066,9 +1111,9 @@ TEST_IMPL(fs_chown) {
   loop = uv_default_loop();
 
   r = uv_fs_open(loop, &req, "test_file", O_RDWR | O_CREAT,
-      S_IWRITE | S_IREAD, NULL);
-  ASSERT(r != -1);
-  ASSERT(req.result != -1);
+      S_IWUSR | S_IRUSR, NULL);
+  ASSERT(r >= 0);
+  ASSERT(req.result >= 0);
   file = req.result;
   uv_fs_req_cleanup(&req);
 
@@ -1087,19 +1132,19 @@ TEST_IMPL(fs_chown) {
   /* async chown */
   r = uv_fs_chown(loop, &req, "test_file", -1, -1, chown_cb);
   ASSERT(r == 0);
-  uv_run(loop);
+  uv_run(loop, UV_RUN_DEFAULT);
   ASSERT(chown_cb_count == 1);
 
   /* chown to root (fail) */
   chown_cb_count = 0;
   r = uv_fs_chown(loop, &req, "test_file", 0, 0, chown_root_cb);
-  uv_run(loop);
+  uv_run(loop, UV_RUN_DEFAULT);
   ASSERT(chown_cb_count == 1);
 
   /* async fchown */
   r = uv_fs_fchown(loop, &req, file, -1, -1, fchown_cb);
   ASSERT(r == 0);
-  uv_run(loop);
+  uv_run(loop, UV_RUN_DEFAULT);
   ASSERT(fchown_cb_count == 1);
 
   close(file);
@@ -1108,7 +1153,7 @@ TEST_IMPL(fs_chown) {
    * Run the loop just to check we don't have make any extraneous uv_ref()
    * calls. This should drop out immediately.
    */
-  uv_run(loop);
+  uv_run(loop, UV_RUN_DEFAULT);
 
   /* Cleanup. */
   unlink("test_file");
@@ -1132,9 +1177,9 @@ TEST_IMPL(fs_link) {
   loop = uv_default_loop();
 
   r = uv_fs_open(loop, &req, "test_file", O_RDWR | O_CREAT,
-      S_IWRITE | S_IREAD, NULL);
-  ASSERT(r != -1);
-  ASSERT(req.result != -1);
+      S_IWUSR | S_IRUSR, NULL);
+  ASSERT(r >= 0);
+  ASSERT(req.result >= 0);
   file = req.result;
   uv_fs_req_cleanup(&req);
 
@@ -1152,15 +1197,15 @@ TEST_IMPL(fs_link) {
   uv_fs_req_cleanup(&req);
 
   r = uv_fs_open(loop, &req, "test_file_link", O_RDWR, 0, NULL);
-  ASSERT(r != -1);
-  ASSERT(req.result != -1);
+  ASSERT(r >= 0);
+  ASSERT(req.result >= 0);
   link = req.result;
   uv_fs_req_cleanup(&req);
 
   memset(buf, 0, sizeof(buf));
   r = uv_fs_read(loop, &req, link, buf, sizeof(buf), 0, NULL);
-  ASSERT(r != -1);
-  ASSERT(req.result != -1);
+  ASSERT(r >= 0);
+  ASSERT(req.result >= 0);
   ASSERT(strcmp(buf, test_buf) == 0);
 
   close(link);
@@ -1168,19 +1213,19 @@ TEST_IMPL(fs_link) {
   /* async link */
   r = uv_fs_link(loop, &req, "test_file", "test_file_link2", link_cb);
   ASSERT(r == 0);
-  uv_run(loop);
+  uv_run(loop, UV_RUN_DEFAULT);
   ASSERT(link_cb_count == 1);
 
   r = uv_fs_open(loop, &req, "test_file_link2", O_RDWR, 0, NULL);
-  ASSERT(r != -1);
-  ASSERT(req.result != -1);
+  ASSERT(r >= 0);
+  ASSERT(req.result >= 0);
   link = req.result;
   uv_fs_req_cleanup(&req);
 
   memset(buf, 0, sizeof(buf));
   r = uv_fs_read(loop, &req, link, buf, sizeof(buf), 0, NULL);
-  ASSERT(r != -1);
-  ASSERT(req.result != -1);
+  ASSERT(r >= 0);
+  ASSERT(req.result >= 0);
   ASSERT(strcmp(buf, test_buf) == 0);
 
   close(link);
@@ -1189,7 +1234,7 @@ TEST_IMPL(fs_link) {
    * Run the loop just to check we don't have make any extraneous uv_ref()
    * calls. This should drop out immediately.
    */
-  uv_run(loop);
+  uv_run(loop, UV_RUN_DEFAULT);
 
   /* Cleanup. */
   unlink("test_file");
@@ -1206,17 +1251,15 @@ TEST_IMPL(fs_readlink) {
 
   loop = uv_default_loop();
   ASSERT(0 == uv_fs_readlink(loop, &req, "no_such_file", dummy_cb));
-  ASSERT(0 == uv_run(loop));
+  ASSERT(0 == uv_run(loop, UV_RUN_DEFAULT));
   ASSERT(dummy_cb_count == 1);
   ASSERT(req.ptr == NULL);
-  ASSERT(req.result == -1);
-  ASSERT(req.errorno == UV_ENOENT);
+  ASSERT(req.result == UV_ENOENT);
   uv_fs_req_cleanup(&req);
 
-  ASSERT(-1 == uv_fs_readlink(loop, &req, "no_such_file", NULL));
+  ASSERT(UV_ENOENT == uv_fs_readlink(loop, &req, "no_such_file", NULL));
   ASSERT(req.ptr == NULL);
-  ASSERT(req.result == -1);
-  ASSERT(req.errorno == UV_ENOENT);
+  ASSERT(req.result == UV_ENOENT);
   uv_fs_req_cleanup(&req);
 
   MAKE_VALGRIND_HAPPY();
@@ -1240,9 +1283,9 @@ TEST_IMPL(fs_symlink) {
   loop = uv_default_loop();
 
   r = uv_fs_open(loop, &req, "test_file", O_RDWR | O_CREAT,
-      S_IWRITE | S_IREAD, NULL);
-  ASSERT(r != -1);
-  ASSERT(req.result != -1);
+      S_IWUSR | S_IRUSR, NULL);
+  ASSERT(r >= 0);
+  ASSERT(req.result >= 0);
   file = req.result;
   uv_fs_req_cleanup(&req);
 
@@ -1256,14 +1299,14 @@ TEST_IMPL(fs_symlink) {
   /* sync symlink */
   r = uv_fs_symlink(loop, &req, "test_file", "test_file_symlink", 0, NULL);
 #ifdef _WIN32
-  if (r == -1) {
-    if (uv_last_error(loop).code == UV_ENOTSUP) {
+  if (r < 0) {
+    if (r == UV_ENOTSUP) {
       /*
        * Windows doesn't support symlinks on older versions.
        * We just pass the test and bail out early if we get ENOTSUP.
        */
       return 0;
-    } else if (uv_last_error(loop).code == UV_EPERM) {
+    } else if (r == UV_EPERM) {
       /*
        * Creating a symlink is only allowed when running elevated.
        * We pass the test and bail out early if we get UV_EPERM.
@@ -1277,62 +1320,62 @@ TEST_IMPL(fs_symlink) {
   uv_fs_req_cleanup(&req);
 
   r = uv_fs_open(loop, &req, "test_file_symlink", O_RDWR, 0, NULL);
-  ASSERT(r != -1);
-  ASSERT(req.result != -1);
+  ASSERT(r >= 0);
+  ASSERT(req.result >= 0);
   link = req.result;
   uv_fs_req_cleanup(&req);
 
   memset(buf, 0, sizeof(buf));
   r = uv_fs_read(loop, &req, link, buf, sizeof(buf), 0, NULL);
-  ASSERT(r != -1);
-  ASSERT(req.result != -1);
+  ASSERT(r >= 0);
+  ASSERT(req.result >= 0);
   ASSERT(strcmp(buf, test_buf) == 0);
 
   close(link);
 
   r = uv_fs_symlink(loop, &req, "test_file_symlink", "test_file_symlink_symlink", 0, NULL);
-  ASSERT(r != -1);
+  ASSERT(r == 0);
   uv_fs_req_cleanup(&req);
 
   r = uv_fs_readlink(loop, &req, "test_file_symlink_symlink", NULL);
-  ASSERT(r != -1);
+  ASSERT(r == 0);
   ASSERT(strcmp(req.ptr, "test_file_symlink") == 0);
   uv_fs_req_cleanup(&req);
 
   /* async link */
   r = uv_fs_symlink(loop, &req, "test_file", "test_file_symlink2", 0, symlink_cb);
   ASSERT(r == 0);
-  uv_run(loop);
+  uv_run(loop, UV_RUN_DEFAULT);
   ASSERT(symlink_cb_count == 1);
 
   r = uv_fs_open(loop, &req, "test_file_symlink2", O_RDWR, 0, NULL);
-  ASSERT(r != -1);
-  ASSERT(req.result != -1);
+  ASSERT(r >= 0);
+  ASSERT(req.result >= 0);
   link = req.result;
   uv_fs_req_cleanup(&req);
 
   memset(buf, 0, sizeof(buf));
   r = uv_fs_read(loop, &req, link, buf, sizeof(buf), 0, NULL);
-  ASSERT(r != -1);
-  ASSERT(req.result != -1);
+  ASSERT(r >= 0);
+  ASSERT(req.result >= 0);
   ASSERT(strcmp(buf, test_buf) == 0);
 
   close(link);
 
   r = uv_fs_symlink(loop, &req, "test_file_symlink2", "test_file_symlink2_symlink", 0, NULL);
-  ASSERT(r != -1);
+  ASSERT(r == 0);
   uv_fs_req_cleanup(&req);
 
   r = uv_fs_readlink(loop, &req, "test_file_symlink2_symlink", readlink_cb);
-  ASSERT(r != -1);
-  uv_run(loop);
+  ASSERT(r == 0);
+  uv_run(loop, UV_RUN_DEFAULT);
   ASSERT(readlink_cb_count == 1);
 
   /*
    * Run the loop just to check we don't have make any extraneous uv_ref()
    * calls. This should drop out immediately.
    */
-  uv_run(loop);
+  uv_run(loop, UV_RUN_DEFAULT);
 
   /* Cleanup. */
   unlink("test_file");
@@ -1382,16 +1425,16 @@ TEST_IMPL(fs_symlink_dir) {
 
   r = uv_fs_stat(loop, &req, "test_dir_symlink", NULL);
   ASSERT(r == 0);
-  ASSERT(((struct stat*)req.ptr)->st_mode & S_IFDIR);
+  ASSERT(((uv_stat_t*)req.ptr)->st_mode & S_IFDIR);
   uv_fs_req_cleanup(&req);
 
   r = uv_fs_lstat(loop, &req, "test_dir_symlink", NULL);
   ASSERT(r == 0);
-  ASSERT(((struct stat*)req.ptr)->st_mode & S_IFLNK);
+  ASSERT(((uv_stat_t*)req.ptr)->st_mode & S_IFLNK);
 #ifdef _WIN32
-  ASSERT(((struct stat*)req.ptr)->st_size == strlen(test_dir + 4));
+  ASSERT(((uv_stat_t*)req.ptr)->st_size == strlen(test_dir + 4));
 #else
-  ASSERT(((struct stat*)req.ptr)->st_size == strlen(test_dir));
+  ASSERT(((uv_stat_t*)req.ptr)->st_size == strlen(test_dir));
 #endif
   uv_fs_req_cleanup(&req);
 
@@ -1405,16 +1448,16 @@ TEST_IMPL(fs_symlink_dir) {
   uv_fs_req_cleanup(&req);
 
   r = uv_fs_open(loop, &open_req1, "test_dir/file1", O_WRONLY | O_CREAT,
-      S_IWRITE | S_IREAD, NULL);
-  ASSERT(r != -1);
+      S_IWUSR | S_IRUSR, NULL);
+  ASSERT(r >= 0);
   uv_fs_req_cleanup(&open_req1);
   r = uv_fs_close(loop, &close_req, open_req1.result, NULL);
   ASSERT(r == 0);
   uv_fs_req_cleanup(&close_req);
 
   r = uv_fs_open(loop, &open_req1, "test_dir/file2", O_WRONLY | O_CREAT,
-      S_IWRITE | S_IREAD, NULL);
-  ASSERT(r != -1);
+      S_IWUSR | S_IRUSR, NULL);
+  ASSERT(r >= 0);
   uv_fs_req_cleanup(&open_req1);
   r = uv_fs_close(loop, &close_req, open_req1.result, NULL);
   ASSERT(r == 0);
@@ -1435,7 +1478,7 @@ TEST_IMPL(fs_symlink_dir) {
   uv_fs_req_cleanup(&req);
 
   r = uv_fs_readdir(loop, &readdir_req, "test_dir_symlink", 0, NULL);
-  ASSERT(r == -1);
+  ASSERT(r == UV_ENOENT);
   uv_fs_req_cleanup(&readdir_req);
 
   r = uv_fs_readdir(loop, &readdir_req, "test_dir", 0, NULL);
@@ -1470,9 +1513,9 @@ TEST_IMPL(fs_utime) {
   loop = uv_default_loop();
   unlink(path);
   r = uv_fs_open(loop, &req, path, O_RDWR | O_CREAT,
-      S_IWRITE | S_IREAD, NULL);
-  ASSERT(r != -1);
-  ASSERT(req.result != -1);
+      S_IWUSR | S_IRUSR, NULL);
+  ASSERT(r >= 0);
+  ASSERT(req.result >= 0);
   uv_fs_req_cleanup(&req);
   close(r);
 
@@ -1498,7 +1541,7 @@ TEST_IMPL(fs_utime) {
   utime_req.data = &checkme;
   r = uv_fs_utime(loop, &utime_req, path, atime, mtime, utime_cb);
   ASSERT(r == 0);
-  uv_run(loop);
+  uv_run(loop, UV_RUN_DEFAULT);
   ASSERT(utime_cb_count == 1);
 
   /* Cleanup. */
@@ -1555,17 +1598,17 @@ TEST_IMPL(fs_futime) {
   loop = uv_default_loop();
   unlink(path);
   r = uv_fs_open(loop, &req, path, O_RDWR | O_CREAT,
-      S_IWRITE | S_IREAD, NULL);
-  ASSERT(r != -1);
-  ASSERT(req.result != -1);
+      S_IWUSR | S_IRUSR, NULL);
+  ASSERT(r >= 0);
+  ASSERT(req.result >= 0);
   uv_fs_req_cleanup(&req);
   close(r);
 
   atime = mtime = 400497753; /* 1982-09-10 11:22:33 */
 
   r = uv_fs_open(loop, &req, path, O_RDWR, 0, NULL);
-  ASSERT(r != -1);
-  ASSERT(req.result != -1);
+  ASSERT(r >= 0);
+  ASSERT(req.result >= 0);
   file = req.result; /* FIXME probably not how it's supposed to be used */
   uv_fs_req_cleanup(&req);
 
@@ -1590,7 +1633,7 @@ TEST_IMPL(fs_futime) {
   futime_req.data = &checkme;
   r = uv_fs_futime(loop, &futime_req, file, atime, mtime, futime_cb);
   ASSERT(r == 0);
-  uv_run(loop);
+  uv_run(loop, UV_RUN_DEFAULT);
   ASSERT(futime_cb_count == 1);
 
   /* Cleanup. */
@@ -1608,9 +1651,8 @@ TEST_IMPL(fs_stat_missing_path) {
   loop = uv_default_loop();
 
   r = uv_fs_stat(loop, &req, "non_existent_file", NULL);
-  ASSERT(r == -1);
-  ASSERT(req.result == -1);
-  ASSERT(uv_last_error(loop).code == UV_ENOENT);
+  ASSERT(r == UV_ENOENT);
+  ASSERT(req.result == UV_ENOENT);
   uv_fs_req_cleanup(&req);
 
   MAKE_VALGRIND_HAPPY();
@@ -1639,7 +1681,7 @@ TEST_IMPL(fs_readdir_empty_dir) {
   ASSERT(r == 0);
 
   ASSERT(readdir_cb_count == 0);
-  uv_run(loop);
+  uv_run(loop, UV_RUN_DEFAULT);
   ASSERT(readdir_cb_count == 1);
 
   uv_fs_rmdir(loop, &req, path, NULL);
@@ -1658,15 +1700,14 @@ TEST_IMPL(fs_readdir_file) {
   loop = uv_default_loop();
 
   r = uv_fs_readdir(loop, &readdir_req, path, 0, NULL);
-  ASSERT(r == -1);
-  ASSERT(uv_last_error(loop).code == UV_ENOTDIR);
+  ASSERT(r == UV_ENOTDIR);
   uv_fs_req_cleanup(&readdir_req);
 
   r = uv_fs_readdir(loop, &readdir_req, path, 0, file_readdir_cb);
   ASSERT(r == 0);
 
   ASSERT(readdir_cb_count == 0);
-  uv_run(loop);
+  uv_run(loop, UV_RUN_DEFAULT);
   ASSERT(readdir_cb_count == 1);
 
   MAKE_VALGRIND_HAPPY();
@@ -1683,8 +1724,8 @@ TEST_IMPL(fs_open_dir) {
   loop = uv_default_loop();
 
   r = uv_fs_open(loop, &req, path, O_RDONLY, 0, NULL);
-  ASSERT(r != -1);
-  ASSERT(req.result != -1);
+  ASSERT(r >= 0);
+  ASSERT(req.result >= 0);
   ASSERT(req.ptr == NULL);
   file = r;
   uv_fs_req_cleanup(&req);
@@ -1696,7 +1737,7 @@ TEST_IMPL(fs_open_dir) {
   ASSERT(r == 0);
 
   ASSERT(open_cb_count == 0);
-  uv_run(loop);
+  uv_run(loop, UV_RUN_DEFAULT);
   ASSERT(open_cb_count == 1);
 
   MAKE_VALGRIND_HAPPY();
@@ -1713,41 +1754,41 @@ TEST_IMPL(fs_file_open_append) {
   loop = uv_default_loop();
 
   r = uv_fs_open(loop, &open_req1, "test_file", O_WRONLY | O_CREAT,
-      S_IWRITE | S_IREAD, NULL);
-  ASSERT(r != -1);
-  ASSERT(open_req1.result != -1);
+      S_IWUSR | S_IRUSR, NULL);
+  ASSERT(r >= 0);
+  ASSERT(open_req1.result >= 0);
   uv_fs_req_cleanup(&open_req1);
 
   r = uv_fs_write(loop, &write_req, open_req1.result, test_buf,
       sizeof(test_buf), -1, NULL);
-  ASSERT(r != -1);
-  ASSERT(write_req.result != -1);
+  ASSERT(r >= 0);
+  ASSERT(write_req.result >= 0);
   uv_fs_req_cleanup(&write_req);
 
   r = uv_fs_close(loop, &close_req, open_req1.result, NULL);
-  ASSERT(r != -1);
-  ASSERT(close_req.result != -1);
+  ASSERT(r == 0);
+  ASSERT(close_req.result == 0);
   uv_fs_req_cleanup(&close_req);
 
   r = uv_fs_open(loop, &open_req1, "test_file", O_RDWR | O_APPEND, 0, NULL);
-  ASSERT(r != -1);
-  ASSERT(open_req1.result != -1);
+  ASSERT(r >= 0);
+  ASSERT(open_req1.result >= 0);
   uv_fs_req_cleanup(&open_req1);
 
   r = uv_fs_write(loop, &write_req, open_req1.result, test_buf,
       sizeof(test_buf), -1, NULL);
-  ASSERT(r != -1);
-  ASSERT(write_req.result != -1);
+  ASSERT(r >= 0);
+  ASSERT(write_req.result >= 0);
   uv_fs_req_cleanup(&write_req);
 
   r = uv_fs_close(loop, &close_req, open_req1.result, NULL);
-  ASSERT(r != -1);
-  ASSERT(close_req.result != -1);
+  ASSERT(r == 0);
+  ASSERT(close_req.result == 0);
   uv_fs_req_cleanup(&close_req);
 
-  r = uv_fs_open(loop, &open_req1, "test_file", O_RDONLY, S_IREAD, NULL);
-  ASSERT(r != -1);
-  ASSERT(open_req1.result != -1);
+  r = uv_fs_open(loop, &open_req1, "test_file", O_RDONLY, S_IRUSR, NULL);
+  ASSERT(r >= 0);
+  ASSERT(open_req1.result >= 0);
   uv_fs_req_cleanup(&open_req1);
 
   r = uv_fs_read(loop, &read_req, open_req1.result, buf, sizeof(buf), -1,
@@ -1761,8 +1802,8 @@ TEST_IMPL(fs_file_open_append) {
   uv_fs_req_cleanup(&read_req);
 
   r = uv_fs_close(loop, &close_req, open_req1.result, NULL);
-  ASSERT(r != -1);
-  ASSERT(close_req.result != -1);
+  ASSERT(r == 0);
+  ASSERT(close_req.result == 0);
   uv_fs_req_cleanup(&close_req);
 
   /* Cleanup */
@@ -1783,54 +1824,54 @@ TEST_IMPL(fs_rename_to_existing_file) {
   loop = uv_default_loop();
 
   r = uv_fs_open(loop, &open_req1, "test_file", O_WRONLY | O_CREAT,
-      S_IWRITE | S_IREAD, NULL);
-  ASSERT(r != -1);
-  ASSERT(open_req1.result != -1);
+      S_IWUSR | S_IRUSR, NULL);
+  ASSERT(r >= 0);
+  ASSERT(open_req1.result >= 0);
   uv_fs_req_cleanup(&open_req1);
 
   r = uv_fs_write(loop, &write_req, open_req1.result, test_buf,
       sizeof(test_buf), -1, NULL);
-  ASSERT(r != -1);
-  ASSERT(write_req.result != -1);
+  ASSERT(r >= 0);
+  ASSERT(write_req.result >= 0);
   uv_fs_req_cleanup(&write_req);
 
   r = uv_fs_close(loop, &close_req, open_req1.result, NULL);
-  ASSERT(r != -1);
-  ASSERT(close_req.result != -1);
+  ASSERT(r == 0);
+  ASSERT(close_req.result == 0);
   uv_fs_req_cleanup(&close_req);
 
   r = uv_fs_open(loop, &open_req1, "test_file2", O_WRONLY | O_CREAT,
-      S_IWRITE | S_IREAD, NULL);
-  ASSERT(r != -1);
-  ASSERT(open_req1.result != -1);
+      S_IWUSR | S_IRUSR, NULL);
+  ASSERT(r >= 0);
+  ASSERT(open_req1.result >= 0);
   uv_fs_req_cleanup(&open_req1);
 
   r = uv_fs_close(loop, &close_req, open_req1.result, NULL);
-  ASSERT(r != -1);
-  ASSERT(close_req.result != -1);
+  ASSERT(r == 0);
+  ASSERT(close_req.result == 0);
   uv_fs_req_cleanup(&close_req);
 
   r = uv_fs_rename(loop, &rename_req, "test_file", "test_file2", NULL);
-  ASSERT(r != -1);
-  ASSERT(rename_req.result != -1);
+  ASSERT(r == 0);
+  ASSERT(rename_req.result == 0);
   uv_fs_req_cleanup(&rename_req);
 
   r = uv_fs_open(loop, &open_req1, "test_file2", O_RDONLY, 0, NULL);
-  ASSERT(r != -1);
-  ASSERT(open_req1.result != -1);
+  ASSERT(r >= 0);
+  ASSERT(open_req1.result >= 0);
   uv_fs_req_cleanup(&open_req1);
 
   memset(buf, 0, sizeof(buf));
   r = uv_fs_read(loop, &read_req, open_req1.result, buf, sizeof(buf), -1,
       NULL);
-  ASSERT(r != -1);
-  ASSERT(read_req.result != -1);
+  ASSERT(r >= 0);
+  ASSERT(read_req.result >= 0);
   ASSERT(strcmp(buf, test_buf) == 0);
   uv_fs_req_cleanup(&read_req);
 
   r = uv_fs_close(loop, &close_req, open_req1.result, NULL);
-  ASSERT(r != -1);
-  ASSERT(close_req.result != -1);
+  ASSERT(r == 0);
+  ASSERT(close_req.result == 0);
   uv_fs_req_cleanup(&close_req);
 
   /* Cleanup */
@@ -1851,32 +1892,32 @@ TEST_IMPL(fs_read_file_eof) {
   loop = uv_default_loop();
 
   r = uv_fs_open(loop, &open_req1, "test_file", O_WRONLY | O_CREAT,
-      S_IWRITE | S_IREAD, NULL);
-  ASSERT(r != -1);
-  ASSERT(open_req1.result != -1);
+      S_IWUSR | S_IRUSR, NULL);
+  ASSERT(r >= 0);
+  ASSERT(open_req1.result >= 0);
   uv_fs_req_cleanup(&open_req1);
 
   r = uv_fs_write(loop, &write_req, open_req1.result, test_buf,
       sizeof(test_buf), -1, NULL);
-  ASSERT(r != -1);
-  ASSERT(write_req.result != -1);
+  ASSERT(r >= 0);
+  ASSERT(write_req.result >= 0);
   uv_fs_req_cleanup(&write_req);
 
   r = uv_fs_close(loop, &close_req, open_req1.result, NULL);
-  ASSERT(r != -1);
-  ASSERT(close_req.result != -1);
+  ASSERT(r == 0);
+  ASSERT(close_req.result == 0);
   uv_fs_req_cleanup(&close_req);
 
   r = uv_fs_open(loop, &open_req1, "test_file", O_RDONLY, 0, NULL);
-  ASSERT(r != -1);
-  ASSERT(open_req1.result != -1);
+  ASSERT(r >= 0);
+  ASSERT(open_req1.result >= 0);
   uv_fs_req_cleanup(&open_req1);
 
   memset(buf, 0, sizeof(buf));
   r = uv_fs_read(loop, &read_req, open_req1.result, buf, sizeof(buf), -1,
       NULL);
-  ASSERT(r != -1);
-  ASSERT(read_req.result != -1);
+  ASSERT(r >= 0);
+  ASSERT(read_req.result >= 0);
   ASSERT(strcmp(buf, test_buf) == 0);
   uv_fs_req_cleanup(&read_req);
 
@@ -1887,8 +1928,8 @@ TEST_IMPL(fs_read_file_eof) {
   uv_fs_req_cleanup(&read_req);
 
   r = uv_fs_close(loop, &close_req, open_req1.result, NULL);
-  ASSERT(r != -1);
-  ASSERT(close_req.result != -1);
+  ASSERT(r == 0);
+  ASSERT(close_req.result == 0);
   uv_fs_req_cleanup(&close_req);
 
   /* Cleanup */
