@@ -61,18 +61,34 @@ static void exit_cb(uv_process_t* process, int exit_status, int term_signal) {
 }
 
 
-static void exit_cb_failure_expected(uv_process_t* process, int exit_status,
-    int term_signal) {
+static void expect(uv_process_t* process,
+                   int exit_status,
+                   int term_signal,
+                   int err) {
   printf("exit_cb\n");
   exit_cb_called++;
-  ASSERT(exit_status == -1);
+  ASSERT(exit_status == err);
   ASSERT(term_signal == 0);
   uv_close((uv_handle_t*)process, close_cb);
 }
 
 
+static void exit_cb_expect_enoent(uv_process_t* process,
+                                 int exit_status,
+                                 int term_signal) {
+  expect(process, exit_status, term_signal, UV_ENOENT);
+}
+
+
+static void exit_cb_expect_eperm(uv_process_t* process,
+                                 int exit_status,
+                                 int term_signal) {
+  expect(process, exit_status, term_signal, UV_EPERM);
+}
+
+
 static void kill_cb(uv_process_t* process, int exit_status, int term_signal) {
-  uv_err_t err;
+  int err;
 
   printf("exit_cb\n");
   exit_cb_called++;
@@ -90,7 +106,7 @@ static void kill_cb(uv_process_t* process, int exit_status, int term_signal) {
    * This process should be dead.
    */
   err = uv_kill(process->pid, 0);
-  ASSERT(err.code == UV_ESRCH);
+  ASSERT(err == UV_ESRCH);
 }
 
 static void detach_failure_cb(uv_process_t* process, int exit_status, int term_signal) {
@@ -107,12 +123,10 @@ static uv_buf_t on_alloc(uv_handle_t* handle, size_t suggested_size) {
 
 
 static void on_read(uv_stream_t* tcp, ssize_t nread, uv_buf_t buf) {
-  uv_err_t err = uv_last_error(uv_default_loop());
-
   if (nread > 0) {
     output_used += nread;
   } else if (nread < 0) {
-    ASSERT(err.code == UV_EOF);
+    ASSERT(nread == UV_EOF);
     uv_close((uv_handle_t*)tcp, close_cb);
   }
 }
@@ -146,12 +160,12 @@ static void timer_cb(uv_timer_t* handle, int status) {
 
 
 TEST_IMPL(spawn_fails) {
-  init_process_options("", exit_cb_failure_expected);
+  init_process_options("", exit_cb_expect_enoent);
   options.file = options.args[0] = "program-that-had-better-not-exist";
   ASSERT(0 == uv_spawn(uv_default_loop(), &process, options));
   ASSERT(0 != uv_is_active((uv_handle_t*)&process));
   ASSERT(0 == uv_run(uv_default_loop(), UV_RUN_DEFAULT));
-  ASSERT(uv_last_error(uv_default_loop()).code == UV_ENOENT);
+  ASSERT(1 == exit_cb_called);
 
   MAKE_VALGRIND_HAPPY();
   return 0;
@@ -185,7 +199,7 @@ TEST_IMPL(spawn_stdout) {
   init_process_options("spawn_helper2", exit_cb);
 
   uv_pipe_init(uv_default_loop(), &out, UV_PIPE_READABLE);
-  uv_pipe_init(uv_default_loop(), &child_stdout, UV_PIPE_SPAWN_SAFE|UV_PIPE_WRITEABLE);
+  uv_pipe_init(uv_default_loop(), &child_stdout, UV_PIPE_SPAWN_SAFE|UV_PIPE_WRITABLE);
   r = uv_pipe_link(&out,&child_stdout);
   ASSERT(r == 0);
 
@@ -229,7 +243,7 @@ TEST_IMPL(spawn_stdout_to_file) {
   init_process_options("spawn_helper2", exit_cb);
 
   r = uv_fs_open(uv_default_loop(), &fs_req, "stdout_file", O_CREAT | O_RDWR,
-      S_IREAD | S_IWRITE, NULL);
+      S_IRUSR | S_IWUSR, NULL);
   ASSERT(r != -1);
   uv_fs_req_cleanup(&fs_req);
 
@@ -284,9 +298,9 @@ TEST_IMPL(spawn_stdin) {
   init_process_options("spawn_helper3", exit_cb);
 
   uv_pipe_init(uv_default_loop(), &out, UV_PIPE_READABLE);
-  uv_pipe_init(uv_default_loop(), &in, UV_PIPE_WRITEABLE);
+  uv_pipe_init(uv_default_loop(), &in, UV_PIPE_WRITABLE);
   uv_pipe_init(uv_default_loop(), &child_pipes[0], UV_PIPE_READABLE|UV_PIPE_SPAWN_SAFE);
-  uv_pipe_init(uv_default_loop(), &child_pipes[1], UV_PIPE_WRITEABLE|UV_PIPE_SPAWN_SAFE);
+  uv_pipe_init(uv_default_loop(), &child_pipes[1], UV_PIPE_WRITABLE|UV_PIPE_SPAWN_SAFE);
   r = uv_pipe_link(&child_pipes[0],&in);
   ASSERT(r == 0);
   r = uv_pipe_link(&out,&child_pipes[1]);
@@ -337,7 +351,7 @@ TEST_IMPL(spawn_stdio_greater_than_3) {
   init_process_options("spawn_helper5", exit_cb);
 
   uv_pipe_init(uv_default_loop(), &pipe, UV_PIPE_READABLE);
-  uv_pipe_init(uv_default_loop(), &child_pipe, UV_PIPE_WRITEABLE|UV_PIPE_SPAWN_SAFE);
+  uv_pipe_init(uv_default_loop(), &child_pipe, UV_PIPE_WRITABLE|UV_PIPE_SPAWN_SAFE);
   r = uv_pipe_link(&pipe,&child_pipe);
   ASSERT(r == 0);
   options.stdio = stdio;
@@ -424,7 +438,7 @@ TEST_IMPL(spawn_preserve_env) {
   init_process_options("spawn_helper7", exit_cb);
   
   uv_pipe_init(uv_default_loop(), &out, UV_PIPE_READABLE);
-  uv_pipe_init(uv_default_loop(), &child_out, UV_PIPE_WRITEABLE|UV_PIPE_SPAWN_SAFE);
+  uv_pipe_init(uv_default_loop(), &child_out, UV_PIPE_WRITABLE|UV_PIPE_SPAWN_SAFE);
   r = uv_pipe_link(&out,&child_out);
   ASSERT(r == 0);
 
@@ -465,7 +479,6 @@ TEST_IMPL(spawn_preserve_env) {
 
 TEST_IMPL(spawn_detached) {
   int r;
-  uv_err_t err;
 
   init_process_options("spawn_helper4", detach_failure_cb);
 
@@ -481,11 +494,11 @@ TEST_IMPL(spawn_detached) {
 
   ASSERT(exit_cb_called == 0);
 
-  err = uv_kill(process.pid, 0);
-  ASSERT(err.code == 0);
+  r = uv_kill(process.pid, 0);
+  ASSERT(r == 0);
 
-  err = uv_kill(process.pid, 15);
-  ASSERT(err.code == 0);
+  r = uv_kill(process.pid, 15);
+  ASSERT(r == 0);
 
   MAKE_VALGRIND_HAPPY();
   return 0;
@@ -503,7 +516,7 @@ TEST_IMPL(spawn_and_kill_with_std) {
 
   init_process_options("spawn_helper4", kill_cb);
 
-  r = uv_pipe_init(uv_default_loop(), &in, UV_PIPE_WRITEABLE);
+  r = uv_pipe_init(uv_default_loop(), &in, UV_PIPE_WRITABLE);
   ASSERT(r == 0);
  
   r = uv_pipe_init(uv_default_loop(), &child_pipes[0], UV_PIPE_READABLE|
@@ -512,14 +525,14 @@ TEST_IMPL(spawn_and_kill_with_std) {
   r = uv_pipe_init(uv_default_loop(), &out, UV_PIPE_READABLE);
   ASSERT(r == 0);
 
-  r = uv_pipe_init(uv_default_loop(), &child_pipes[1], UV_PIPE_WRITEABLE|
+  r = uv_pipe_init(uv_default_loop(), &child_pipes[1], UV_PIPE_WRITABLE|
 														UV_PIPE_SPAWN_SAFE);
   ASSERT(r == 0);
 
   r = uv_pipe_init(uv_default_loop(), &err, UV_PIPE_READABLE);
   ASSERT(r == 0);
 
-  r = uv_pipe_init(uv_default_loop(), &child_pipes[2], UV_PIPE_WRITEABLE|
+  r = uv_pipe_init(uv_default_loop(), &child_pipes[2], UV_PIPE_WRITABLE|
 														UV_PIPE_SPAWN_SAFE);
   ASSERT(r == 0);
 
@@ -585,11 +598,11 @@ TEST_IMPL(spawn_and_ping) {
   buf = uv_buf_init("TEST", 4);
 
   uv_pipe_init(uv_default_loop(), &child_pipes[0], UV_PIPE_READABLE|UV_PIPE_SPAWN_SAFE);
-  uv_pipe_init(uv_default_loop(), &in, UV_PIPE_WRITEABLE);
+  uv_pipe_init(uv_default_loop(), &in, UV_PIPE_WRITABLE);
   r = uv_pipe_link(&child_pipes[0],&in);
   ASSERT(r == 0);
   uv_pipe_init(uv_default_loop(), &out, UV_PIPE_READABLE);
-  uv_pipe_init(uv_default_loop(), &child_pipes[1], UV_PIPE_WRITEABLE|UV_PIPE_SPAWN_SAFE);
+  uv_pipe_init(uv_default_loop(), &child_pipes[1], UV_PIPE_WRITABLE|UV_PIPE_SPAWN_SAFE);
   r = uv_pipe_link(&out,&child_pipes[1]);
   ASSERT(r == 0);
   options.stdio = stdio;
@@ -632,7 +645,6 @@ TEST_IMPL(spawn_and_ping) {
 
 TEST_IMPL(kill) {
   int r;
-  uv_err_t err;
 
 #ifdef _WIN32
   no_term_signal = 1;
@@ -646,12 +658,12 @@ TEST_IMPL(kill) {
   /* Sending signum == 0 should check if the
    * child process is still alive, not kill it.
    */
-  err = uv_kill(process.pid, 0);
-  ASSERT(err.code == UV_OK);
+  r = uv_kill(process.pid, 0);
+  ASSERT(r == 0);
 
   /* Kill the process. */
-  err = uv_kill(process.pid, /* SIGTERM */ 15);
-  ASSERT(err.code == UV_OK);
+  r = uv_kill(process.pid, /* SIGTERM */ 15);
+  ASSERT(r == 0);
 
   r = uv_run(uv_default_loop(), UV_RUN_DEFAULT);
   ASSERT(r == 0);
@@ -675,7 +687,7 @@ TEST_IMPL(spawn_detect_pipe_name_collisions_on_windows) {
   init_process_options("spawn_helper2", exit_cb);
 
   uv_pipe_init(uv_default_loop(), &out, UV_PIPE_READABLE);
-  uv_pipe_init(uv_default_loop(), &child_stdout, UV_PIPE_WRITEABLE|UV_PIPE_SPAWN_SAFE);
+  uv_pipe_init(uv_default_loop(), &child_stdout, UV_PIPE_WRITABLE|UV_PIPE_SPAWN_SAFE);
 
   /* Create a pipe that'll cause a collision. */
   _snprintf(name, sizeof(name), "\\\\.\\pipe\\uv\\%p-%d", &out, GetCurrentProcessId());
@@ -719,7 +731,7 @@ TEST_IMPL(spawn_detect_pipe_name_collisions_on_windows) {
 }
 
 
-uv_err_t make_program_args(char** args, int verbatim_arguments, WCHAR** dst_ptr);
+int make_program_args(char** args, int verbatim_arguments, WCHAR** dst_ptr);
 WCHAR* quote_cmd_arg(const WCHAR *source, WCHAR *target);
 
 TEST_IMPL(argument_escaping) {
@@ -741,7 +753,7 @@ TEST_IMPL(argument_escaping) {
   size_t total_size = 0;
   int i;
   int num_args;
-  uv_err_t result;
+  int result;
 
   char* verbatim[] = {
     "cmd.exe",
@@ -781,9 +793,9 @@ TEST_IMPL(argument_escaping) {
   }
 
   result = make_program_args(verbatim, 1, &verbatim_output);
-  ASSERT(result.code == UV_OK);
+  ASSERT(result == 0);
   result = make_program_args(verbatim, 0, &non_verbatim_output);
-  ASSERT(result.code == UV_OK);
+  ASSERT(result == 0);
 
   wprintf(L"    verbatim_output: %s\n", verbatim_output);
   wprintf(L"non_verbatim_output: %s\n", non_verbatim_output);
@@ -797,7 +809,7 @@ TEST_IMPL(argument_escaping) {
   return 0;
 }
 
-uv_err_t make_program_env(char** env_block, WCHAR** dst_ptr);
+int make_program_env(char** env_block, WCHAR** dst_ptr);
 
 TEST_IMPL(environment_creation) {
   int i;
@@ -812,7 +824,7 @@ TEST_IMPL(environment_creation) {
 
   WCHAR expected[512];
   WCHAR* ptr = expected;
-  uv_err_t result;
+  int result;
   WCHAR* str;
   WCHAR* env;
 
@@ -832,7 +844,7 @@ TEST_IMPL(environment_creation) {
   *ptr = '\0';
 
   result = make_program_env(environment, &env);
-  ASSERT(result.code == UV_OK);
+  ASSERT(result == 0);
 
   for (str = env; *str; str += wcslen(str) + 1) {
     wprintf(L"%s\n", str);
@@ -890,14 +902,14 @@ TEST_IMPL(spawn_setuid_fails) {
     struct passwd* pw;
     pw = getpwnam("nobody");
     ASSERT(pw != NULL);
-    r = setuid(pw->pw_uid);
-    ASSERT(r == 0);
+    ASSERT(0 == setgid(pw->pw_gid));
+    ASSERT(0 == setuid(pw->pw_uid));
   }
 
-  init_process_options("spawn_helper1", exit_cb_failure_expected);
+  init_process_options("spawn_helper1", exit_cb_expect_eperm);
 
   options.flags |= UV_PROCESS_SETUID;
-  options.uid = (uv_uid_t) -42424242;
+  options.uid = 0;
 
   r = uv_spawn(uv_default_loop(), &process, options);
   ASSERT(r == 0);
@@ -922,14 +934,14 @@ TEST_IMPL(spawn_setgid_fails) {
     struct passwd* pw;
     pw = getpwnam("nobody");
     ASSERT(pw != NULL);
-    r = setuid(pw->pw_uid);
-    ASSERT(r == 0);
+    ASSERT(0 == setgid(pw->pw_gid));
+    ASSERT(0 == setuid(pw->pw_uid));
   }
 
-  init_process_options("spawn_helper1", exit_cb_failure_expected);
+  init_process_options("spawn_helper1", exit_cb_expect_eperm);
 
   options.flags |= UV_PROCESS_SETGID;
-  options.gid = (uv_gid_t) -42424242;
+  options.gid = 0;
 
   r = uv_spawn(uv_default_loop(), &process, options);
   ASSERT(r == 0);
@@ -964,8 +976,7 @@ TEST_IMPL(spawn_setuid_fails) {
   options.uid = (uv_uid_t) -42424242;
 
   r = uv_spawn(uv_default_loop(), &process, options);
-  ASSERT(r == -1);
-  ASSERT(uv_last_error(uv_default_loop()).code == UV_ENOTSUP);
+  ASSERT(r == UV_ENOTSUP);
 
   r = uv_run(uv_default_loop(), UV_RUN_DEFAULT);
   ASSERT(r == 0);
@@ -986,8 +997,7 @@ TEST_IMPL(spawn_setgid_fails) {
   options.gid = (uv_gid_t) -42424242;
 
   r = uv_spawn(uv_default_loop(), &process, options);
-  ASSERT(r == -1);
-  ASSERT(uv_last_error(uv_default_loop()).code == UV_ENOTSUP);
+  ASSERT(r == UV_ENOTSUP);
 
   r = uv_run(uv_default_loop(), UV_RUN_DEFAULT);
   ASSERT(r == 0);
