@@ -79,8 +79,7 @@ static uv_buf_t on_alloc(uv_handle_t* handle, size_t suggested_size) {
 
 static void after_write(uv_write_t* req, int status) {
   if (status) {
-    uv_err_t err = uv_last_error(loop);
-    fprintf(stderr, "uv_write error: %s\n", uv_strerror(err));
+    fprintf(stderr, "uv_write error: %s\n", uv_strerror(status));
     ASSERT(0);
   }
 
@@ -96,7 +95,7 @@ static void on_read(uv_stream_t* tcp, ssize_t nread, uv_buf_t rdbuf) {
   uv_buf_t wrbuf;
   int r;
 
-  ASSERT(nread > 0 || uv_last_error(uv_default_loop()).code == UV_EOF);
+  ASSERT(nread > 0 || nread == UV_EOF);
 
   if (nread > 0) {
     output_used += nread;
@@ -124,12 +123,12 @@ TEST_IMPL(stdio_over_pipes) {
   init_process_options("stdio_over_pipes_helper", exit_cb);
 
   uv_pipe_init(loop, &child_pipes[0], UV_PIPE_SPAWN_SAFE|UV_PIPE_READABLE);
-  uv_pipe_init(loop, &in, UV_PIPE_WRITEABLE);
+  uv_pipe_init(loop, &in, UV_PIPE_WRITABLE);
   r = uv_pipe_link(&child_pipes[0],&in);
   ASSERT( r == 0 );
   
   uv_pipe_init(loop, &out, UV_PIPE_READABLE);
-  uv_pipe_init(loop, &child_pipes[1], UV_PIPE_SPAWN_SAFE|UV_PIPE_WRITEABLE);
+  uv_pipe_init(loop, &child_pipes[1], UV_PIPE_SPAWN_SAFE|UV_PIPE_WRITABLE);
   r = uv_pipe_link(&out,&child_pipes[1]);
   ASSERT( r == 0 );
 
