@@ -82,7 +82,8 @@
 #define CNAME(csym)		_##csym
 #define HIDENAME(asmsym)	.asmsym
 #define _ENTRY(x)   _START_ENTRY; \
-            .globl CNAME(x); .def CNAME(x); .scl 2; .type 32; .endef; CNAME(x):
+            .globl CNAME(x); .section .drectve; .ascii " -export:" #x; \
+            .section .text; .def CNAME(x); .scl 2; .type 32; .endef; CNAME(x):
 #define END(x) .end
 #endif
 
