@@ -23,7 +23,7 @@ static const float one = 1.0, huge = 1e30;
 
 static const float zero = 0.0;
 
-float
+DLLEXPORT float
 __ieee754_atanhf(float x)
 {
 	float t;
