@@ -25,7 +25,7 @@
 #include "openlibm.h"
 #include "math_private.h"
 
-float
+DLLEXPORT float
 copysignf(float x, float y)
 {
 	u_int32_t ix,iy;
