@@ -31,7 +31,7 @@
 static const double
 two54 =  1.80143985094819840000e+16; /* 0x43500000, 0x00000000 */
 
-double
+DLLEXPORT double
 frexp(double x, int *eptr)
 {
 	int32_t hx, ix, lx;
