@@ -27,7 +27,7 @@
 static const double
 two54 = 1.80143985094819840000e+16;	/* 43500000 00000000 */
 
-double
+DLLEXPORT double
 logb(double x)
 {
 	int32_t lx,ix;
