@@ -20,7 +20,7 @@
 #include "math_private.h"
 
 static const float one=1.0, two=2.0, tiny = 1.0e-30, huge = 1.0e30;
-float
+DLLEXPORT float
 tanhf(float x)
 {
 	float t,z;
