@@ -6,8 +6,9 @@
  */
 
 #include <openlibm.h>
+#include "math_private.h"
 
-double
+DLLEXPORT double
 drem(x, y)
 	double x, y;
 {
