@@ -9,7 +9,7 @@
 #include "openlibm.h"
 #include "math_private.h"
 
-float
+DLLEXPORT float
 dremf(float x, float y)
 {
 	return remainderf(x, y);
