@@ -4,63 +4,93 @@
 #include <stdlib.h>
 #include <string.h>
 #include <assert.h>
-#ifdef __WIN32__
+#ifdef _OS_WINDOWS_
 #include <malloc.h>
 #endif
 #include "julia.h"
 
 // array constructors ---------------------------------------------------------
 
-static jl_array_t *_new_array(jl_type_t *atype,
-                              uint32_t ndims, size_t *dims)
+static inline int store_unboxed(jl_value_t *el_type)
+{
+    return jl_is_datatype(el_type) && jl_is_leaf_type(el_type) && jl_is_immutable(el_type) && jl_is_pointerfree((jl_datatype_t*)el_type);
+}
+
+int jl_array_store_unboxed(jl_value_t *el_type)
+{
+    return store_unboxed(el_type);
+}
+
+// at this size use malloc
+#define MALLOC_THRESH 1048576
+
+#ifdef _P64
+typedef __uint128_t wideint_t;
+#else
+typedef uint64_t wideint_t;
+#endif
+
+#define MAXINTVAL (((size_t)-1)>>1)
+
+static jl_array_t *_new_array_(jl_value_t *atype, uint32_t ndims, size_t *dims,
+                               int isunboxed, int elsz)
 {
     size_t i, tot, nel=1;
-    int isunboxed=0, elsz;
+    wideint_t prod;
     void *data;
     jl_array_t *a;
 
     for(i=0; i < ndims; i++) {
-        if ((long)dims[i] < 0)
-            jl_error("invalid Array dimension size");
-        nel *= dims[i];
+        prod = (wideint_t)nel * (wideint_t)dims[i];
+        if (prod > (wideint_t) MAXINTVAL)
+            jl_error("invalid Array dimensions");
+        nel = prod;
     }
-    jl_type_t *el_type = (jl_type_t*)jl_tparam0(atype);
 
-    isunboxed = jl_is_bits_type(el_type);
     if (isunboxed) {
-        elsz = jl_bitstype_nbits(el_type)/8;
-        tot = elsz * nel;
+        prod = (wideint_t)elsz * (wideint_t)nel;
+        if (prod > (wideint_t) MAXINTVAL)
+            jl_error("invalid Array size");
+        tot = prod;
         if (elsz == 1) {
             // hidden 0 terminator for all byte arrays
             tot++;
         }
     }
     else {
-        elsz = sizeof(void*);
-        tot = sizeof(void*) * nel;
+        prod = (wideint_t)sizeof(void*) * (wideint_t)nel;
+        if (prod > (wideint_t) MAXINTVAL)
+            jl_error("invalid Array size");
+        tot = prod;
     }
 
     int ndimwords = jl_array_ndimwords(ndims);
+    size_t tsz = sizeof(jl_array_t);
+    tsz += ndimwords*sizeof(size_t);
     if (tot <= ARRAY_INLINE_NBYTES) {
-        size_t tsz = tot>sizeof(size_t) ? tot-sizeof(size_t) : tot;
-        a = allocobj((sizeof(jl_array_t)+tsz+ndimwords*sizeof(size_t)+15)&-16);
+        if (isunboxed && elsz >= 4)
+            tsz = (tsz+15)&-16; // align data area 16
+        size_t doffs = tsz;
+        tsz += tot;
+        tsz = (tsz+15)&-16; // align whole object 16
+        a = allocobj(tsz);
         a->type = atype;
-        a->ismalloc = 0;
-        data = (&a->_space[0] + ndimwords*sizeof(size_t));
+        a->how = 0;
+        data = (char*)a + doffs;
         if (tot > 0 && !isunboxed) {
             memset(data, 0, tot);
         }
     }
     else {
-        a = allocobj((sizeof(jl_array_t)+ndimwords*sizeof(size_t)+15)&-16);
-        JL_GC_PUSH(&a);
+        tsz = (tsz+15)&-16; // align whole object size 16
+        a = allocobj(tsz);
+        JL_GC_PUSH1(&a);
         a->type = atype;
-        a->ismalloc = 1;
         // temporarily initialize to make gc-safe
         a->data = NULL;
-        jl_value_t **powner = (jl_value_t**)(&a->_space[0] + ndimwords*sizeof(size_t));
-        *powner = (jl_value_t*)jl_gc_managed_malloc(tot);
-        data = ((jl_mallocptr_t*)*powner)->ptr;
+        a->how = 2;
+        data = jl_gc_managed_malloc(tot);
+        jl_gc_track_malloced_array(a);
         if (!isunboxed)
             memset(data, 0, tot);
         JL_GC_POP();
@@ -68,97 +98,96 @@ static jl_array_t *_new_array(jl_type_t *atype,
 
     a->data = data;
     if (elsz == 1) ((char*)data)[tot-1] = '\0';
+#ifdef STORE_ARRAY_LEN
     a->length = nel;
+#endif
     a->ndims = ndims;
     a->ptrarray = !isunboxed;
     a->elsize = elsz;
+    a->isshared = 0;
+    a->isaligned = 1;
+    a->offset = 0;
     if (ndims == 1) {
         a->nrows = nel;
         a->maxsize = nel;
-        a->offset = 0;
     }
     else {
         size_t *adims = &a->nrows;
         for(i=0; i < ndims; i++)
             adims[i] = dims[i];
     }
-    
+
     return a;
 }
 
-static jl_mallocptr_t *array_new_buffer(jl_array_t *a, size_t newlen);
+static inline jl_array_t *_new_array(jl_value_t *atype, uint32_t ndims, size_t *dims)
+{
+    int isunboxed=0, elsz=sizeof(void*);
+    jl_value_t *el_type = jl_tparam0(atype);
+    isunboxed = store_unboxed(el_type);
+    if (isunboxed)
+        elsz = jl_datatype_size(el_type);
+    return _new_array_(atype, ndims, dims, isunboxed, elsz);
+}
 
-jl_array_t *jl_reshape_array(jl_type_t *atype, jl_array_t *data,
-                             jl_tuple_t *dims)
+jl_array_t *jl_new_array_for_deserialization(jl_value_t *atype, uint32_t ndims, size_t *dims,
+                                             int isunboxed, int elsz)
+{
+    return _new_array_(atype, ndims, dims, isunboxed, elsz);
+}
+
+jl_array_t *jl_reshape_array(jl_value_t *atype, jl_array_t *data, jl_tuple_t *dims)
 {
     size_t i;
     jl_array_t *a;
     size_t ndims = jl_tuple_len(dims);
 
     int ndimwords = jl_array_ndimwords(ndims);
-    a = allocobj((sizeof(jl_array_t) + ndimwords*sizeof(size_t) + 15)&-16);
+    a = allocobj((sizeof(jl_array_t) + sizeof(void*) + ndimwords*sizeof(size_t) + 15)&-16);
     a->type = atype;
     a->ndims = ndims;
+    a->offset = 0;
     a->data = NULL;
-    JL_GC_PUSH(&a);
-
-    char *d = data->data;
-    if (data->ndims == 1) d -= data->offset*data->elsize;
-    if (d == jl_array_inline_data_area(data)) {
-        if (data->ndims == 1) {
-            // data might resize, so switch it to shared representation.
-            // problem: the buffer might be used from C in a way that it's
-            // assumed not to move. for now just hope this doesn't happen.
-            size_t datalen = jl_array_len(data);
-            jl_mallocptr_t *mp = array_new_buffer(data, datalen);
-            memcpy(mp->ptr, data->data, datalen * data->elsize);
-            a->data = mp->ptr;
-            jl_array_data_owner(a) = (jl_value_t*)mp;
-            a->ismalloc = 1;
-
-            data->data = mp->ptr;
-            data->offset = 0;
-            data->maxsize = datalen;
-            jl_array_data_owner(data) = (jl_value_t*)mp;
-            data->ismalloc = 1;
-        }
-        else {
-            a->ismalloc = 0;
-            jl_array_data_owner(a) = (jl_value_t*)data;
-        }
-    }
-    else {
-        a->ismalloc = data->ismalloc;
-        jl_array_data_owner(a) = jl_array_data_owner(data);
-    }
-
-    if (a->data == NULL) a->data = data->data;
-    jl_type_t *el_type = (jl_type_t*)jl_tparam0(atype);
-    if (jl_is_bits_type(el_type)) {
-        a->elsize = jl_bitstype_nbits(el_type)/8;
+    a->isaligned = data->isaligned;
+    jl_value_t *el_type = jl_tparam0(atype);
+    if (store_unboxed(el_type)) {
+        a->elsize = jl_datatype_size(el_type);
         a->ptrarray = 0;
     }
     else {
         a->elsize = sizeof(void*);
         a->ptrarray = 1;
     }
+    JL_GC_PUSH1(&a);
+
+    jl_array_data_owner(a) = (jl_value_t*)data;
+    a->how = 3;
+    a->data = data->data;
+    a->isshared = 1;
+    data->isshared = 1;
 
     if (ndims == 1) {
-        a->length = jl_unbox_long(jl_tupleref(dims,0));
-        a->nrows = a->length;
-        a->maxsize = a->length;
-        a->offset = 0;
+        size_t l = jl_unbox_long(jl_tupleref(dims,0));
+#ifdef STORE_ARRAY_LEN
+        a->length = l;
+#endif
+        a->nrows = l;
+        a->maxsize = l;
     }
     else {
         size_t *adims = &a->nrows;
         size_t l=1;
+        wideint_t prod;
         for(i=0; i < ndims; i++) {
             adims[i] = jl_unbox_long(jl_tupleref(dims, i));
-            if ((long)adims[i] < 0)
-                jl_error("invalid Array dimension size");
-            l *= adims[i];
+            prod = (wideint_t)l * (wideint_t)adims[i];
+            if (prod > (wideint_t) MAXINTVAL)
+                jl_error("invalid Array dimensions");
+            l = prod;
         }
+#ifdef STORE_ARRAY_LEN
         a->length = l;
+#endif
     }
     JL_GC_POP();
 
@@ -166,58 +195,63 @@ jl_array_t *jl_reshape_array(jl_type_t *atype, jl_array_t *data,
 }
 
 // own_buffer != 0 iff GC should call free() on this pointer eventually
-jl_array_t *jl_ptr_to_array_1d(jl_type_t *atype, void *data, size_t nel,
+jl_array_t *jl_ptr_to_array_1d(jl_value_t *atype, void *data, size_t nel,
                                int own_buffer)
 {
     size_t elsz;
     jl_array_t *a;
-    jl_type_t *el_type = (jl_type_t*)jl_tparam0(atype);
+    jl_value_t *el_type = jl_tparam0(atype);
 
-    int isunboxed = jl_is_bits_type(el_type);
+    int isunboxed = store_unboxed(el_type);
     if (isunboxed)
-        elsz = jl_bitstype_nbits(el_type)/8;
+        elsz = jl_datatype_size(el_type);
     else
         elsz = sizeof(void*);
 
     a = allocobj((sizeof(jl_array_t)+jl_array_ndimwords(1)*sizeof(size_t)+15)&-16);
     a->type = atype;
     a->data = data;
+#ifdef STORE_ARRAY_LEN
     a->length = nel;
+#endif
     a->elsize = elsz;
     a->ptrarray = !isunboxed;
     a->ndims = 1;
-
+    a->isshared = 1;
+    a->isaligned = 0;  // TODO: allow passing memalign'd buffers
     if (own_buffer) {
-        a->ismalloc = 1;
-        jl_array_data_owner(a) = (jl_value_t*)jl_gc_acquire_buffer(data,nel*elsz);
+        a->how = 2;
+        jl_gc_track_malloced_array(a);
     }
     else {
-        a->ismalloc = 0;
-        jl_array_data_owner(a) = (jl_value_t*)a;
+        a->how = 0;
     }
 
-    a->nrows = a->length;
-    a->maxsize = a->length;
+    a->nrows = nel;
+    a->maxsize = nel;
     a->offset = 0;
-    
     return a;
 }
 
-jl_array_t *jl_ptr_to_array(jl_type_t *atype, void *data, jl_tuple_t *dims,
+jl_array_t *jl_ptr_to_array(jl_value_t *atype, void *data, jl_tuple_t *dims,
                             int own_buffer)
 {
     size_t i, elsz, nel=1;
     jl_array_t *a;
     size_t ndims = jl_tuple_len(dims);
+    wideint_t prod;
 
     for(i=0; i < ndims; i++) {
-        nel *= jl_unbox_long(jl_tupleref(dims, i));
+        prod = (wideint_t)nel * (wideint_t)jl_unbox_long(jl_tupleref(dims, i));
+        if (prod > (wideint_t) MAXINTVAL)
+            jl_error("invalid Array dimensions");
+        nel = prod;
     }
-    jl_type_t *el_type = (jl_type_t*)jl_tparam0(atype);
+    jl_value_t *el_type = jl_tparam0(atype);
 
-    int isunboxed = jl_is_bits_type(el_type);
+    int isunboxed = store_unboxed(el_type);
     if (isunboxed)
-        elsz = jl_bitstype_nbits(el_type)/8;
+        elsz = jl_datatype_size(el_type);
     else
         elsz = sizeof(void*);
 
@@ -225,24 +259,26 @@ jl_array_t *jl_ptr_to_array(jl_type_t *atype, void *data, jl_tuple_t *dims,
     a = allocobj((sizeof(jl_array_t) + ndimwords*sizeof(size_t)+15)&-16);
     a->type = atype;
     a->data = data;
+#ifdef STORE_ARRAY_LEN
     a->length = nel;
+#endif
     a->elsize = elsz;
     a->ptrarray = !isunboxed;
     a->ndims = ndims;
-
+    a->offset = 0;
+    a->isshared = 1;
+    a->isaligned = 0;
     if (own_buffer) {
-        a->ismalloc = 1;
-        jl_array_data_owner(a) = (jl_value_t*)jl_gc_acquire_buffer(data,nel*elsz);
+        a->how = 2;
+        jl_gc_track_malloced_array(a);
     }
     else {
-        a->ismalloc = 0;
-        jl_array_data_owner(a) = (jl_value_t*)a;
+        a->how = 0;
     }
 
     if (ndims == 1) {
-        a->nrows = a->length;
-        a->maxsize = a->length;
-        a->offset = 0;
+        a->nrows = nel;
+        a->maxsize = nel;
     }
     else {
         size_t *adims = &a->nrows;
@@ -250,16 +286,10 @@ jl_array_t *jl_ptr_to_array(jl_type_t *atype, void *data, jl_tuple_t *dims,
             adims[i] = jl_unbox_long(jl_tupleref(dims, i));
         }
     }
-    
     return a;
 }
 
-jl_array_t *jl_new_array_(jl_type_t *atype, uint32_t ndims, size_t *dims)
-{
-    return _new_array(atype, ndims, dims);
-}
-
-jl_array_t *jl_new_array(jl_type_t *atype, jl_tuple_t *dims)
+jl_array_t *jl_new_array(jl_value_t *atype, jl_tuple_t *dims)
 {
     size_t ndims = jl_tuple_len(dims);
     size_t *adims = alloca(ndims*sizeof(size_t));
@@ -269,18 +299,18 @@ jl_array_t *jl_new_array(jl_type_t *atype, jl_tuple_t *dims)
     return _new_array(atype, ndims, adims);
 }
 
-jl_array_t *jl_alloc_array_1d(jl_type_t *atype, size_t nr)
+jl_array_t *jl_alloc_array_1d(jl_value_t *atype, size_t nr)
 {
     return _new_array(atype, 1, &nr);
 }
 
-jl_array_t *jl_alloc_array_2d(jl_type_t *atype, size_t nr, size_t nc)
+jl_array_t *jl_alloc_array_2d(jl_value_t *atype, size_t nr, size_t nc)
 {
     size_t d[2] = {nr, nc};
     return _new_array(atype, 2, &d[0]);
 }
 
-jl_array_t *jl_alloc_array_3d(jl_type_t *atype, size_t nr, size_t nc, size_t z)
+jl_array_t *jl_alloc_array_3d(jl_value_t *atype, size_t nr, size_t nc, size_t z)
 {
     size_t d[3] = {nr, nc, z};
     return _new_array(atype, 3, &d[0]);
@@ -296,10 +326,10 @@ jl_array_t *jl_pchar_to_array(const char *str, size_t len)
 jl_value_t *jl_array_to_string(jl_array_t *a)
 {
     // TODO: check type of array?
-    jl_struct_type_t* string_type = u8_isvalid(a->data, jl_array_len(a)) == 1 ? // ASCII
+    jl_datatype_t* string_type = u8_isvalid(a->data, jl_array_len(a)) == 1 ? // ASCII
         jl_ascii_string_type : jl_utf8_string_type;
     jl_value_t *s = alloc_2w();
-    s->type = (jl_type_t*)string_type;
+    s->type = (jl_value_t*)string_type;
     jl_set_nth_field(s, 0, (jl_value_t*)a);
     return s;
 }
@@ -307,7 +337,7 @@ jl_value_t *jl_array_to_string(jl_array_t *a)
 jl_value_t *jl_pchar_to_string(const char *str, size_t len)
 {
     jl_array_t *a = jl_pchar_to_array(str, len);
-    JL_GC_PUSH(&a);
+    JL_GC_PUSH1(&a);
     jl_value_t *s = jl_array_to_string(a);
     JL_GC_POP();
     return s;
@@ -325,6 +355,16 @@ jl_array_t *jl_alloc_cell_1d(size_t n)
 
 // array primitives -----------------------------------------------------------
 
+#ifndef STORE_ARRAY_LEN
+DLLEXPORT size_t jl_array_len_(jl_array_t *a)
+{
+    size_t l = 1;
+    for(size_t i=0; i < jl_array_ndims(a); i++)
+        l *= jl_array_dim(a, i);
+    return l;
+}
+#endif
+
 JL_CALLABLE(jl_f_arraylen)
 {
     JL_NARGS(arraylen, 1, 1);
@@ -334,6 +374,9 @@ JL_CALLABLE(jl_f_arraylen)
 
 JL_CALLABLE(jl_f_arraysize)
 {
+    if (nargs != 2) {
+        JL_NARGS(arraysize, 1, 1);
+    }
     JL_TYPECHK(arraysize, array, args[0]);
     jl_array_t *a = (jl_array_t*)args[0];
     size_t nd = jl_array_ndims(a);
@@ -346,11 +389,8 @@ JL_CALLABLE(jl_f_arraysize)
             return jl_box_long(1);
         return jl_box_long((&a->nrows)[dno-1]);
     }
-    else {
-        JL_NARGS(arraysize, 1, 1);
-    }
     jl_tuple_t *d = jl_alloc_tuple(nd);
-    JL_GC_PUSH(&d);
+    JL_GC_PUSH1(&d);
     size_t i;
     for(i=0; i < nd; i++)
         jl_tupleset(d, i, jl_box_long(jl_array_dim(a,i)));
@@ -360,10 +400,10 @@ JL_CALLABLE(jl_f_arraysize)
 
 jl_value_t *jl_arrayref(jl_array_t *a, size_t i)
 {
-    jl_type_t *el_type = (jl_type_t*)jl_tparam0(jl_typeof(a));
+    jl_value_t *el_type = (jl_value_t*)jl_tparam0(jl_typeof(a));
     jl_value_t *elt;
-    if (jl_is_bits_type(el_type)) {
-        elt = jl_new_bits((jl_bits_type_t*)el_type,
+    if (!a->ptrarray) {
+        elt = jl_new_bits((jl_datatype_t*)el_type,
                           &((char*)a->data)[i*a->elsize]);
     }
     else {
@@ -407,11 +447,30 @@ JL_CALLABLE(jl_f_arrayref)
     return jl_arrayref(a, i);
 }
 
-int jl_array_isdefined(jl_value_t **args, int nargs)
+int jl_array_isdefined(jl_value_t **args0, int nargs)
 {
-    assert(jl_is_array(args[0]));
-    jl_array_t *a = (jl_array_t*)args[0];
-    size_t i = array_nd_index(a, &args[1], nargs-1, "isdefined");
+    assert(jl_is_array(args0[0]));
+    jl_array_t *a = (jl_array_t*)args0[0];
+    jl_value_t **args = &args0[1];
+    size_t nidxs = nargs-1;
+    size_t i=0;
+    size_t k, stride=1;
+    size_t nd = jl_array_ndims(a);
+    for(k=0; k < nidxs; k++) {
+        if (!jl_is_long(args[k]))
+            jl_type_error("isdefined", (jl_value_t*)jl_long_type, args[k]);
+        size_t ii = jl_unbox_long(args[k])-1;
+        i += ii * stride;
+        size_t d = k>=nd ? 1 : jl_array_dim(a, k);
+        if (k < nidxs-1 && ii >= d)
+            return 0;
+        stride *= d;
+    }
+    for(; k < nd; k++)
+        stride *= jl_array_dim(a, k);
+    if (i >= stride)
+        return 0;
+
     if (a->ptrarray)
         return ((jl_value_t**)jl_array_data(a))[i] != NULL;
     return 1;
@@ -424,7 +483,7 @@ void jl_arrayset(jl_array_t *a, jl_value_t *rhs, size_t i)
         if (!jl_subtype(rhs, el_type, 1))
             jl_type_error("arrayset", el_type, rhs);
     }
-    if (jl_is_bits_type(el_type)) {
+    if (!a->ptrarray) {
         jl_assign_bits(&((char*)a->data)[i*a->elsize], rhs);
     }
     else {
@@ -451,123 +510,177 @@ void jl_arrayunset(jl_array_t *a, size_t i)
         memset(ptail, 0, a->elsize);
 }
 
-static jl_mallocptr_t *array_new_buffer(jl_array_t *a, size_t newlen)
+// allocate buffer of newlen elements, placing old data at given offset (in #elts)
+static void array_resize_buffer(jl_array_t *a, size_t newlen, size_t oldlen, size_t offs)
 {
-    size_t nbytes = newlen * a->elsize;
-    if (a->elsize == 1) {
+    size_t es = a->elsize;
+    size_t nbytes = newlen * es;
+    size_t offsnb = offs * es;
+    size_t oldnbytes = oldlen * es;
+    size_t oldoffsnb = a->offset * es;
+    if (es == 1)
         nbytes++;
+    assert(!a->isshared || a->how==3);
+    char *newdata;
+    if (a->how == 2) {
+        // already malloc'd - use realloc
+        newdata = jl_gc_managed_realloc((char*)a->data - oldoffsnb, nbytes,
+                                        oldnbytes+oldoffsnb, a->isaligned);
+        if (offs != a->offset) {
+            memmove(&newdata[offsnb], &newdata[oldoffsnb], oldnbytes);
+        }
+    }
+    else {
+        if (nbytes >= MALLOC_THRESH) {
+            newdata = jl_gc_managed_malloc(nbytes);
+            jl_gc_track_malloced_array(a);
+            a->how = 2;
+            a->isaligned = 1;
+        }
+        else {
+            newdata = allocb(nbytes);
+            a->how = 1;
+        }
+        memcpy(newdata + offsnb, (char*)a->data, oldnbytes);
+    }
+
+    a->data = newdata + offsnb;
+    a->isshared = 0;
+    if (a->ptrarray || es==1)
+        memset(newdata+offsnb+oldnbytes, 0, nbytes-oldnbytes-offsnb);
+    a->maxsize = newlen;
+}
+
+static void array_try_unshare(jl_array_t *a)
+{
+    if (a->isshared) {
+        if (a->how != 3)
+            jl_error("cannot resize array with shared data");
+        size_t len = jl_array_nrows(a);
+        array_resize_buffer(a, len, len, a->offset);
     }
-    jl_mallocptr_t *mp = jl_gc_managed_malloc(nbytes);
-    char *newdata = mp->ptr;
-    if (a->ptrarray)
-        memset(newdata, 0, nbytes);
-    if (a->elsize == 1) newdata[nbytes-1] = '\0';
-    return mp;
 }
 
 void jl_array_grow_end(jl_array_t *a, size_t inc)
 {
+    if (a->isshared && a->how!=3) jl_error("cannot resize array with shared data");
     // optimized for the case of only growing and shrinking at the end
-    size_t alen = jl_array_len(a);
+    size_t alen = jl_array_nrows(a);
     if ((alen + inc) > a->maxsize - a->offset) {
         size_t newlen = a->maxsize==0 ? (inc<4?4:inc) : a->maxsize*2;
         while ((alen + inc) > newlen - a->offset)
             newlen *= 2;
-        jl_mallocptr_t *mp = array_new_buffer(a, newlen);
-        char *newdata = mp->ptr;
-        size_t es = a->elsize;
-        newdata += (a->offset*es);
-        size_t anb = alen*es;
-        memcpy(newdata, (char*)a->data, anb);
-        if (es == 1) {
-            memset(newdata + anb, 0, (newlen-a->offset-alen)*es);
-        }
-        a->maxsize = newlen;
-        a->data = newdata;
-        jl_array_data_owner(a) = (jl_value_t*)mp;
-        a->ismalloc = 1;
+        array_resize_buffer(a, newlen, alen, a->offset);
     }
-    a->length += inc; a->nrows += inc;
+#ifdef STORE_ARRAY_LEN
+    a->length += inc;
+#endif
+    a->nrows += inc;
 }
 
 void jl_array_del_end(jl_array_t *a, size_t dec)
 {
-    if (dec > a->length)
+    if (dec == 0) return;
+    if (dec > a->nrows)
         jl_throw(jl_bounds_exception);
-    char *ptail = (char*)a->data + (a->length-dec)*a->elsize;
-    if (a->ptrarray)
-        memset(ptail, 0, dec*a->elsize);
-    else
-        ptail[0] = 0;
-    a->length -= dec; a->nrows -= dec;
+    if (a->isshared) array_try_unshare(a);
+    if (a->elsize > 0) {
+        char *ptail = (char*)a->data + (a->nrows-dec)*a->elsize;
+        assert(ptail < (char*)a->data + (a->length*a->elsize));
+        if (a->ptrarray)
+            memset(ptail, 0, dec*a->elsize);
+        else
+            ptail[0] = 0;
+    }
+#ifdef STORE_ARRAY_LEN
+    a->length -= dec;
+#endif
+    a->nrows -= dec;
+}
+
+void jl_array_sizehint(jl_array_t *a, size_t sz)
+{
+    if (sz <= jl_array_len(a))
+        return;
+    size_t inc = sz - jl_array_len(a);
+    jl_array_grow_end(a, inc);
+#ifdef STORE_ARRAY_LEN
+    a->length -= inc;
+#endif
+    a->nrows -= inc;
 }
 
 void jl_array_grow_beg(jl_array_t *a, size_t inc)
 {
+    if (inc == 0) return;
     // designed to handle the case of growing and shrinking at both ends
-    if (inc == 0)
-        return;
+    if (a->isshared) array_try_unshare(a);
     size_t es = a->elsize;
-    size_t nb = inc*es;
+    size_t incnb = inc*es;
     if (a->offset >= inc) {
-        a->data = (char*)a->data - nb;
+        a->data = (char*)a->data - incnb;
         a->offset -= inc;
     }
     else {
-        size_t alen = a->length;
+        size_t alen = a->nrows;
         size_t anb = alen*es;
-        char *newdata;
-        jl_mallocptr_t *mp = NULL;
         if (inc > (a->maxsize-alen)/2 - (a->maxsize-alen)/20) {
-            size_t newlen = a->maxsize==0 ? 2*inc : a->maxsize*2;
+            size_t newlen = a->maxsize==0 ? inc*2 : a->maxsize*2;
             while (alen+2*inc > newlen-a->offset)
                 newlen *= 2;
-            mp = array_new_buffer(a, newlen);
-            newdata = mp->ptr;
             size_t center = (newlen - (alen + inc))/2;
-            newdata += (center*es);
-            a->maxsize = newlen;
+            array_resize_buffer(a, newlen, alen, center+inc);
+            char *newdata = (char*)a->data - (center+inc)*es;
+            if (a->ptrarray) {
+                memset(newdata, 0, (center+inc)*es);
+            }
             a->offset = center;
+            a->data = newdata + center*es;
         }
         else {
             size_t center = (a->maxsize - (alen + inc))/2;
-            newdata = (char*)a->data - es*a->offset + es*center;
+            char *newdata = (char*)a->data - es*a->offset + es*center;
+            memmove(&newdata[incnb], a->data, anb);
+            a->data = newdata;
             a->offset = center;
         }
-        memmove(&newdata[nb], a->data, anb);
-        a->data = newdata;
-        if (mp) { jl_array_data_owner(a) = (jl_value_t*)mp; a->ismalloc = 1; }
     }
-    a->length += inc; a->nrows += inc;
+#ifdef STORE_ARRAY_LEN
+    a->length += inc;
+#endif
+    a->nrows += inc;
 }
 
 void jl_array_del_beg(jl_array_t *a, size_t dec)
 {
-    if (dec == 0)
-        return;
-    if (dec > a->length)
+    if (dec == 0) return;
+    if (dec > a->nrows)
         jl_throw(jl_bounds_exception);
+    if (a->isshared) array_try_unshare(a);
     size_t es = a->elsize;
     size_t nb = dec*es;
     memset(a->data, 0, nb);
     size_t offset = a->offset;
     offset += dec;
     a->data = (char*)a->data + nb;
-    a->length -= dec; a->nrows -= dec;
+#ifdef STORE_ARRAY_LEN
+    a->length -= dec;
+#endif
+    a->nrows -= dec;
 
     // make sure offset doesn't grow forever due to deleting at beginning
     // and growing at end
     size_t newoffs = offset;
     if (offset >= 13*a->maxsize/20) {
-        newoffs = 17*(a->maxsize - a->length)/100;
+        newoffs = 17*(a->maxsize - a->nrows)/100;
     }
-#ifdef __LP64__
+#ifdef _P64
     while (newoffs > (size_t)((uint32_t)-1)) {
         newoffs = newoffs/2;
     }
 #endif
     if (newoffs != offset) {
-        size_t anb = a->length*es;
+        size_t anb = a->nrows*es;
         size_t delta = (offset - newoffs)*es;
         a->data = (char*)a->data - delta;
         memmove(a->data, (char*)a->data + delta, anb);
@@ -575,9 +688,17 @@ void jl_array_del_beg(jl_array_t *a, size_t dec)
     a->offset = newoffs;
 }
 
-void jl_cell_1d_push(jl_array_t *a, jl_value_t *item)
+DLLEXPORT void jl_cell_1d_push(jl_array_t *a, jl_value_t *item)
 {
     assert(jl_typeis(a, jl_array_any_type));
     jl_array_grow_end(a, 1);
     jl_cellset(a, jl_array_dim(a,0)-1, item);
 }
+
+DLLEXPORT void jl_cell_1d_push2(jl_array_t *a, jl_value_t *b, jl_value_t *c)
+{
+    assert(jl_typeis(a, jl_array_any_type));
+    jl_array_grow_end(a, 2);
+    jl_cellset(a, jl_array_dim(a,0)-2, b);
+    jl_cellset(a, jl_array_dim(a,0)-1, c);
+}
