@@ -6,7 +6,7 @@
 #include <stdio.h>
 #include <string.h>
 #include <assert.h>
-#ifdef __WIN32__
+#ifdef _OS_WINDOWS_
 #include <malloc.h>
 #endif
 #include "julia.h"
@@ -19,7 +19,7 @@ static char flisp_system_image[] = {
 extern fltype_t *iostreamtype;
 static fltype_t *jvtype;
 
-static jl_value_t *scm_to_julia(value_t e);
+static jl_value_t *scm_to_julia(value_t e, int expronly);
 static value_t julia_to_scm(jl_value_t *v);
 
 DLLEXPORT void jl_lisp_prompt(void)
@@ -40,6 +40,13 @@ value_t fl_defined_julia_global(value_t *args, uint32_t nargs)
     return (b != HT_NOTFOUND && b->owner==jl_current_module) ? FL_T : FL_F;
 }
 
+value_t fl_current_julia_module(value_t *args, uint32_t nargs)
+{
+    value_t opaque = cvalue(jvtype, sizeof(void*));
+    *(jl_value_t**)cv_data((cvalue_t*)ptr(opaque)) = (jl_value_t*)jl_current_module;
+    return opaque;
+}
+
 value_t fl_invoke_julia_macro(value_t *args, uint32_t nargs)
 {
     if (nargs < 1)
@@ -49,11 +56,11 @@ value_t fl_invoke_julia_macro(value_t *args, uint32_t nargs)
     JL_GC_PUSHARGS(margs, nargs);
     int i;
     for(i=0; i < nargs; i++) margs[i] = NULL;
-    for(i=1; i < nargs; i++) margs[i] = scm_to_julia(args[i]);
-    jl_value_t *result;
+    for(i=1; i < nargs; i++) margs[i] = scm_to_julia(args[i], 1);
+    jl_value_t *result=NULL;
 
     JL_TRY {
-        margs[0] = scm_to_julia(args[0]);
+        margs[0] = scm_to_julia(args[0], 1);
         f = (jl_function_t*)jl_toplevel_eval(margs[0]);
         result = jl_apply(f, &margs[1], nargs-1);
     }
@@ -68,6 +75,7 @@ value_t fl_invoke_julia_macro(value_t *args, uint32_t nargs)
     // not julia.
     // all calls to invoke-julia-macro happen under a single call to jl_expand,
     // so the preserved value stack is popped there.
+    assert(result != NULL);
     jl_gc_preserve(result);
     value_t scm = julia_to_scm(result);
     fl_gc_handle(&scm);
@@ -90,9 +98,13 @@ value_t fl_invoke_julia_macro(value_t *args, uint32_t nargs)
 static builtinspec_t julia_flisp_ast_ext[] = {
     { "defined-julia-global", fl_defined_julia_global },
     { "invoke-julia-macro", fl_invoke_julia_macro },
+    { "current-julia-module", fl_current_julia_module },
     { NULL, NULL }
 };
 
+static value_t true_sym;
+static value_t false_sym;
+
 DLLEXPORT void jl_init_frontend(void)
 {
     fl_init(2*512*1024);
@@ -111,6 +123,8 @@ DLLEXPORT void jl_init_frontend(void)
                                 NULL, NULL);
 
     assign_global_builtins(julia_flisp_ast_ext);
+    true_sym = symbol("true");
+    false_sym = symbol("false");
 }
 
 static jl_sym_t *scmsym_to_julia(value_t s)
@@ -126,37 +140,37 @@ static jl_sym_t *scmsym_to_julia(value_t s)
     return jl_symbol(symbol_name(s));
 }
 
-static jl_value_t *scm_to_julia_(value_t e);
+static jl_value_t *scm_to_julia_(value_t e, int expronly);
 
-static jl_value_t *full_list(value_t e)
+static jl_value_t *full_list(value_t e, int expronly)
 {
     size_t ln = llength(e);
     if (ln == 0) return jl_an_empty_cell;
     jl_array_t *ar = jl_alloc_cell_1d(ln);
     size_t i=0;
     while (iscons(e)) {
-        jl_cellset(ar, i, scm_to_julia_(car_(e)));
+        jl_cellset(ar, i, scm_to_julia_(car_(e), expronly));
         e = cdr_(e);
         i++;
     }
     return (jl_value_t*)ar;
 }
 
-static jl_value_t *full_list_of_lists(value_t e)
+static jl_value_t *full_list_of_lists(value_t e, int expronly)
 {
     size_t ln = llength(e);
     if (ln == 0) return jl_an_empty_cell;
     jl_array_t *ar = jl_alloc_cell_1d(ln);
     size_t i=0;
     while (iscons(e)) {
-        jl_cellset(ar, i, full_list(car_(e)));
+        jl_cellset(ar, i, full_list(car_(e),expronly));
         e = cdr_(e);
         i++;
     }
     return (jl_value_t*)ar;
 }
 
-static jl_value_t *scm_to_julia(value_t e)
+static jl_value_t *scm_to_julia(value_t e, int expronly)
 {
 #ifdef JL_GC_MARKSWEEP
     int en = jl_gc_is_enabled();
@@ -164,7 +178,7 @@ static jl_value_t *scm_to_julia(value_t e)
 #endif
     jl_value_t *v;
     JL_TRY {
-        v = scm_to_julia_(e);
+        v = scm_to_julia_(e, expronly);
     }
     JL_CATCH {
         // if expression cannot be converted, replace with error expr
@@ -178,57 +192,50 @@ static jl_value_t *scm_to_julia(value_t e)
     return v;
 }
 
-static jl_value_t *scm_to_julia_(value_t e)
+extern int64_t conv_to_int64(void *data, numerictype_t tag);
+
+static jl_value_t *scm_to_julia_(value_t e, int eo)
 {
     if (fl_isnumber(e)) {
-        if (iscprim(e)) {
-            numerictype_t nt = cp_numtype((cprim_t*)ptr(e));
+        int64_t i64;
+        if (isfixnum(e)) {
+            i64 = numval(e);
+        }
+        else {
+            assert(iscprim(e));
+            cprim_t *cp = (cprim_t*)ptr(e);
+            numerictype_t nt = cp_numtype(cp);
             switch (nt) {
             case T_DOUBLE:
-                return (jl_value_t*)jl_box_float64(*(double*)cp_data((cprim_t*)ptr(e)));
+                return (jl_value_t*)jl_box_float64(*(double*)cp_data(cp));
             case T_FLOAT:
-                return (jl_value_t*)jl_box_float32(*(float*)cp_data((cprim_t*)ptr(e)));
-            case T_INT64:
-                return (jl_value_t*)jl_box_int64(*(int64_t*)cp_data((cprim_t*)ptr(e)));
+                return (jl_value_t*)jl_box_float32(*(float*)cp_data(cp));
             case T_UINT8:
-                return (jl_value_t*)jl_box_uint8(*(uint8_t*)cp_data((cprim_t*)ptr(e)));
+                return (jl_value_t*)jl_box_uint8(*(uint8_t*)cp_data(cp));
             case T_UINT16:
-                return (jl_value_t*)jl_box_uint16(*(uint16_t*)cp_data((cprim_t*)ptr(e)));
+                return (jl_value_t*)jl_box_uint16(*(uint16_t*)cp_data(cp));
             case T_UINT32:
-                return (jl_value_t*)jl_box_uint32(*(uint32_t*)cp_data((cprim_t*)ptr(e)));
+                return (jl_value_t*)jl_box_uint32(*(uint32_t*)cp_data(cp));
             case T_UINT64:
-                return (jl_value_t*)jl_box_uint64(*(uint64_t*)cp_data((cprim_t*)ptr(e)));
+                return (jl_value_t*)jl_box_uint64(*(uint64_t*)cp_data(cp));
             default:
                 ;
             }
+            i64 = conv_to_int64(cp_data(cp), nt);
         }
-        if (isfixnum(e)) {
-            int64_t ne = numval(e);
-#ifdef __LP64__
-            return (jl_value_t*)jl_box_int64(ne);
+#ifdef _P64
+        return (jl_value_t*)jl_box_int64(i64);
 #else
-            if (ne > S32_MAX || ne < S32_MIN)
-                return (jl_value_t*)jl_box_int64(ne);
-            return (jl_value_t*)jl_box_int32((int32_t)ne);
-#endif
-        }
-        uint64_t n = toulong(e, "scm_to_julia");
-#ifdef __LP64__
-        return (jl_value_t*)jl_box_int64((int64_t)n);
-#else
-        if (n > S32_MAX)
-            return (jl_value_t*)jl_box_int64((int64_t)n);
-        return (jl_value_t*)jl_box_int32((int32_t)n);
+        if (i64 > (int64_t)S32_MAX || i64 < (int64_t)S32_MIN)
+            return (jl_value_t*)jl_box_int64(i64);
+        return (jl_value_t*)jl_box_int32((int32_t)i64);
 #endif
     }
     if (issymbol(e)) {
-        if (!fl_isgensym(e)) {
-            char *sn = symbol_name(e);
-            if (!strcmp(sn, "true"))
-                return jl_true;
-            else if (!strcmp(sn, "false"))
-                return jl_false;
-        }
+        if (e == true_sym)
+            return jl_true;
+        else if (e == false_sym)
+            return jl_false;
         return (jl_value_t*)scmsym_to_julia(e);
     }
     if (fl_isstring(e)) {
@@ -256,27 +263,29 @@ static jl_value_t *scm_to_julia_(value_t e)
             */
             size_t n = llength(e)-1;
             size_t i;
+            if (sym == null_sym && n == 0)
+                return jl_nothing;
             if (sym == lambda_sym) {
                 jl_expr_t *ex = jl_exprn(lambda_sym, n);
                 e = cdr_(e);
                 value_t largs = car_(e);
-                jl_cellset(ex->args, 0, full_list(largs));
+                jl_cellset(ex->args, 0, full_list(largs,eo));
                 e = cdr_(e);
                 
                 value_t ee = car_(e);
                 jl_array_t *vinf = jl_alloc_cell_1d(3);
-                jl_cellset(vinf, 0, full_list(car_(ee)));
+                jl_cellset(vinf, 0, full_list(car_(ee),eo));
                 ee = cdr_(ee);
-                jl_cellset(vinf, 1, full_list_of_lists(car_(ee)));
+                jl_cellset(vinf, 1, full_list_of_lists(car_(ee),eo));
                 ee = cdr_(ee);
-                jl_cellset(vinf, 2, full_list_of_lists(car_(ee)));
+                jl_cellset(vinf, 2, full_list_of_lists(car_(ee),eo));
                 assert(!iscons(cdr_(ee)));
                 jl_cellset(ex->args, 1, vinf);
                 e = cdr_(e);
                 
                 for(i=2; i < n; i++) {
                     assert(iscons(e));
-                    jl_cellset(ex->args, i, scm_to_julia_(car_(e)));
+                    jl_cellset(ex->args, i, scm_to_julia_(car_(e), eo));
                     e = cdr_(e);
                 }
                 return
@@ -284,30 +293,36 @@ static jl_value_t *scm_to_julia_(value_t e)
             }
 
             e = cdr_(e);
-            if (sym == line_sym && n==1) {
-                return jl_new_struct(jl_linenumbernode_type,
-                                     scm_to_julia_(car_(e)));
-            }
-            if (sym == label_sym) {
-                return jl_new_struct(jl_labelnode_type,
-                                     scm_to_julia_(car_(e)));
-            }
-            if (sym == goto_sym) {
-                return jl_new_struct(jl_gotonode_type,
-                                     scm_to_julia_(car_(e)));
-            }
-            if (sym == quote_sym) {
-                return jl_new_struct(jl_quotenode_type,
-                                     scm_to_julia_(car_(e)));
-            }
-            if (sym == top_sym) {
-                return jl_new_struct(jl_topnode_type,
-                                     scm_to_julia_(car_(e)));
+            if (!eo) {
+                if (sym == line_sym && n==1) {
+                    return jl_new_struct(jl_linenumbernode_type,
+                                         scm_to_julia_(car_(e),0));
+                }
+                if (sym == label_sym) {
+                    return jl_new_struct(jl_labelnode_type,
+                                         scm_to_julia_(car_(e),0));
+                }
+                if (sym == goto_sym) {
+                    return jl_new_struct(jl_gotonode_type,
+                                         scm_to_julia_(car_(e),0));
+                }
+                if (sym == quote_sym) {
+                    return jl_new_struct(jl_quotenode_type,
+                                         scm_to_julia_(car_(e),0));
+                }
+                if (sym == top_sym) {
+                    return jl_new_struct(jl_topnode_type,
+                                         scm_to_julia_(car_(e),0));
+                }
+                if (sym == newvar_sym) {
+                    return jl_new_struct(jl_newvarnode_type,
+                                         scm_to_julia_(car_(e),0));
+                }
             }
             jl_expr_t *ex = jl_exprn(sym, n);
             for(i=0; i < n; i++) {
                 assert(iscons(e));
-                jl_cellset(ex->args, i, scm_to_julia_(car_(e)));
+                jl_cellset(ex->args, i, scm_to_julia_(car_(e),eo));
                 e = cdr_(e);
             }
             return (jl_value_t*)ex;
@@ -344,6 +359,16 @@ static value_t array_to_list(jl_array_t *a)
     return lst;
 }
 
+static value_t julia_to_list2(jl_value_t *a, jl_value_t *b)
+{
+    value_t sa = julia_to_scm(a);
+    fl_gc_handle(&sa);
+    value_t sb = julia_to_scm(b);
+    value_t l = fl_list2(sa, sb);
+    fl_free_gc_handles(1);
+    return l;
+}
+
 static value_t julia_to_scm(jl_value_t *v)
 {
     if (jl_is_symbol(v)) {
@@ -355,6 +380,9 @@ static value_t julia_to_scm(jl_value_t *v)
     if (v == jl_false) {
         return FL_F;
     }
+    if (v == jl_nothing) {
+        return fl_cons(symbol("null"), FL_NIL);
+    }
     if (jl_is_expr(v)) {
         jl_expr_t *ex = (jl_expr_t*)v;
         value_t args = array_to_list(ex->args);
@@ -365,36 +393,26 @@ static value_t julia_to_scm(jl_value_t *v)
         return scmv;
     }
     if (jl_typeis(v, jl_linenumbernode_type)) {
-        return fl_cons(julia_to_scm((jl_value_t*)line_sym),
-                       fl_cons(julia_to_scm(jl_fieldref(v,0)),
-                               FL_NIL));
+        return julia_to_list2((jl_value_t*)line_sym, jl_fieldref(v,0));
     }
     if (jl_typeis(v, jl_labelnode_type)) {
-        return fl_cons(julia_to_scm((jl_value_t*)label_sym),
-                       fl_cons(julia_to_scm(jl_fieldref(v,0)),
-                               FL_NIL));
+        return julia_to_list2((jl_value_t*)label_sym, jl_fieldref(v,0));
     }
     if (jl_typeis(v, jl_gotonode_type)) {
-        return fl_cons(julia_to_scm((jl_value_t*)goto_sym),
-                       fl_cons(julia_to_scm(jl_fieldref(v,0)),
-                               FL_NIL));
+        return julia_to_list2((jl_value_t*)goto_sym, jl_fieldref(v,0));
     }
     if (jl_typeis(v, jl_quotenode_type)) {
-        return fl_cons(julia_to_scm((jl_value_t*)quote_sym),
-                       fl_cons(julia_to_scm(jl_fieldref(v,0)),
-                               FL_NIL));
+        return julia_to_list2((jl_value_t*)quote_sym, jl_fieldref(v,0));
+    }
+    if (jl_typeis(v, jl_newvarnode_type)) {
+        return julia_to_list2((jl_value_t*)newvar_sym, jl_fieldref(v,0));
     }
     if (jl_typeis(v, jl_topnode_type)) {
-        return fl_cons(julia_to_scm((jl_value_t*)top_sym),
-                       fl_cons(julia_to_scm(jl_fieldref(v,0)),
-                               FL_NIL));
+        return julia_to_list2((jl_value_t*)top_sym, jl_fieldref(v,0));
     }
     if (jl_is_long(v) && fits_fixnum(jl_unbox_long(v))) {
         return fixnum(jl_unbox_long(v));
     }
-    if (jl_typeis(v,jl_array_any_type)) {
-        return array_to_list((jl_array_t*)v);
-    }
     value_t opaque = cvalue(jvtype, sizeof(void*));
     *(jl_value_t**)cv_data((cvalue_t*)ptr(opaque)) = v;
     return opaque;
@@ -403,12 +421,12 @@ static value_t julia_to_scm(jl_value_t *v)
 // this is used to parse a line of repl input
 DLLEXPORT jl_value_t *jl_parse_input_line(const char *str)
 {
-    value_t e = fl_applyn(1, symbol_value(symbol("jl-parse-string")),
-                          cvalue_static_cstring(str));
+    value_t s = cvalue_static_cstring(str);
+    value_t e = fl_applyn(1, symbol_value(symbol("jl-parse-string")), s);
     if (e == FL_EOF)
         return jl_nothing;
     
-    return scm_to_julia(e);
+    return scm_to_julia(e,0);
 }
 
 // this is for parsing one expression out of a string, keeping track of
@@ -419,26 +437,28 @@ DLLEXPORT jl_value_t *jl_parse_string(const char *str, int pos0, int greedy)
     value_t p = fl_applyn(3, symbol_value(symbol("jl-parse-one-string")),
                           s, fixnum(pos0), greedy?FL_T:FL_F);
     jl_value_t *expr=NULL, *pos1=NULL;
-    JL_GC_PUSH(&expr, &pos1);
+    JL_GC_PUSH2(&expr, &pos1);
 
     value_t e = car_(p);
     if (e == FL_EOF) {
         expr = (jl_value_t*)jl_null;
     }
     else {
-        expr = scm_to_julia(e);
+        expr = scm_to_julia(e,0);
     }
 
-    pos1 = jl_box_long(toulong(cdr_(p),"parse"));
+    pos1 = jl_box_long(tosize(cdr_(p),"parse"));
     jl_value_t *result = (jl_value_t*)jl_tuple2(expr, pos1);
     JL_GC_POP();
     return result;
 }
 
-void jl_start_parsing_file(const char *fname)
+int jl_start_parsing_file(const char *fname)
 {
-    fl_applyn(1, symbol_value(symbol("jl-parse-file")),
-              cvalue_static_cstring(fname));
+    value_t s = cvalue_static_cstring(fname);
+    if (fl_applyn(1, symbol_value(symbol("jl-parse-file")), s) == FL_F)
+        return 1;
+    return 0;
 }
 
 void jl_stop_parsing()
@@ -458,17 +478,22 @@ jl_value_t *jl_parse_next()
         if (isfixnum(a)) {
             jl_lineno = numval(a);
             //jl_printf(JL_STDERR, "  on line %d\n", jl_lineno);
-            return scm_to_julia(cdr_(c));
+            return scm_to_julia(cdr_(c),0);
         }
     }
-    return scm_to_julia(c);
+    return scm_to_julia(c,0);
 }
 
-void jl_load_file_string(const char *text)
+jl_value_t *jl_load_file_string(const char *text, char *filename)
 {
-    fl_applyn(1, symbol_value(symbol("jl-parse-string-stream")),
-              cvalue_static_cstring(text));
-    jl_parse_eval_all("");
+    value_t t, f;
+    t = cvalue_static_cstring(text);
+    fl_gc_handle(&t);
+    f = cvalue_static_cstring(filename);
+    fl_applyn(2, symbol_value(symbol("jl-parse-string-stream")),
+              t, f);
+    fl_free_gc_handles(1);
+    return jl_parse_eval_all(filename);
 }
 
 // returns either an expression or a thunk
@@ -477,7 +502,7 @@ jl_value_t *jl_expand(jl_value_t *expr)
     int np = jl_gc_n_preserved_values();
     value_t arg = julia_to_scm(expr);
     value_t e = fl_applyn(1, symbol_value(symbol("jl-expand-to-thunk")), arg);
-    jl_value_t *result = scm_to_julia(e);
+    jl_value_t *result = scm_to_julia(e,0);
     while (jl_gc_n_preserved_values() > np) {
         jl_gc_unpreserve();
     }
@@ -490,7 +515,7 @@ DLLEXPORT jl_value_t *jl_macroexpand(jl_value_t *expr)
     value_t arg = julia_to_scm(expr);
     value_t e = fl_applyn(1, symbol_value(symbol("jl-macroexpand")), arg);
     jl_value_t *result;
-    result = scm_to_julia(e);
+    result = scm_to_julia(e,0);
     while (jl_gc_n_preserved_values() > np) {
         jl_gc_unpreserve();
     }
@@ -503,7 +528,7 @@ jl_lambda_info_t *jl_wrap_expr(jl_value_t *expr)
     // `(lambda () (() () ()) ,expr)
     jl_expr_t *le=NULL, *bo=NULL; jl_value_t *vi=NULL;
     jl_value_t *mt = jl_an_empty_cell;
-    JL_GC_PUSH(&le, &vi, &bo);
+    JL_GC_PUSH3(&le, &vi, &bo);
     le = jl_exprn(lambda_sym, 3);
     jl_cellset(le->args, 0, mt);
     vi = (jl_value_t*)jl_alloc_cell_1d(3);
@@ -565,6 +590,17 @@ jl_array_t *jl_lam_capt(jl_expr_t *l)
     return (jl_array_t*)ll;
 }
 
+int jl_lam_vars_captured(jl_expr_t *ast)
+{
+    jl_array_t *vinfos = jl_lam_vinfo(ast);
+    for(int i=0; i < jl_array_len(vinfos); i++) {
+        if (jl_vinfo_capt((jl_array_t*)jl_cellref(vinfos,i))) {
+            return 1;
+        }
+    }
+    return 0;
+}
+
 // get array of body forms
 jl_expr_t *jl_lam_body(jl_expr_t *l)
 {
@@ -618,8 +654,12 @@ static jl_value_t *copy_ast(jl_value_t *expr, jl_tuple_t *sp, int do_sp)
         */
         // TODO: avoid if above condition is true and decls have already
         // been evaluated.
-        JL_GC_PUSH(&li);
+        JL_GC_PUSH1(&li);
         li = jl_add_static_parameters(li, sp);
+        // inner lambda does not need the "def" link. it leads to excess object
+        // retention, for example pointing to the original uncompressed AST
+        // of a top-level thunk that gets type inferred.
+        li->def = li;
         li->ast = jl_prepare_ast(li, li->sparams);
         JL_GC_POP();
         return (jl_value_t*)li;
@@ -627,7 +667,7 @@ static jl_value_t *copy_ast(jl_value_t *expr, jl_tuple_t *sp, int do_sp)
     else if (jl_typeis(expr,jl_array_any_type)) {
         jl_array_t *a = (jl_array_t*)expr;
         jl_array_t *na = jl_alloc_cell_1d(jl_array_len(a));
-        JL_GC_PUSH(&na);
+        JL_GC_PUSH1(&na);
         size_t i;
         for(i=0; i < jl_array_len(a); i++)
             jl_cellset(na, i, copy_ast(jl_cellref(a,i), sp, do_sp));
@@ -637,7 +677,7 @@ static jl_value_t *copy_ast(jl_value_t *expr, jl_tuple_t *sp, int do_sp)
     else if (jl_is_expr(expr)) {
         jl_expr_t *e = (jl_expr_t*)expr;
         jl_expr_t *ne = jl_exprn(e->head, jl_array_len(e->args));
-        JL_GC_PUSH(&ne);
+        JL_GC_PUSH1(&ne);
         if (e->head == lambda_sym) {
             jl_exprarg(ne, 0) = copy_ast(jl_exprarg(e,0), sp, 0);
             jl_exprarg(ne, 1) = copy_ast(jl_exprarg(e,1), sp, 0);
@@ -653,6 +693,48 @@ static jl_value_t *copy_ast(jl_value_t *expr, jl_tuple_t *sp, int do_sp)
     return expr;
 }
 
+DLLEXPORT jl_value_t *jl_copy_ast(jl_value_t *expr)
+{
+    if (jl_is_expr(expr)) {
+        jl_expr_t *e = (jl_expr_t*)expr;
+        size_t i, l = jl_array_len(e->args);
+        jl_expr_t *ne = NULL;
+        JL_GC_PUSH2(&ne, &expr);
+        ne = jl_exprn(e->head, l);
+        if (l == 0) {
+            ne->args = jl_alloc_cell_1d(0);
+        }
+        else {
+            for(i=0; i < l; i++)
+                jl_exprarg(ne, i) = jl_copy_ast(jl_exprarg(e,i));
+        }
+        JL_GC_POP();
+        return (jl_value_t*)ne;
+    }
+    else if (jl_typeis(expr,jl_array_any_type)) {
+        jl_array_t *a = (jl_array_t*)expr;
+        size_t i, l = jl_array_len(a);
+        jl_array_t *na = NULL;
+        JL_GC_PUSH2(&na, &expr);
+        na = jl_alloc_cell_1d(l);
+        for(i=0; i < l; i++)
+            jl_cellset(na, i, jl_copy_ast(jl_cellref(a,i)));
+        JL_GC_POP();
+        return (jl_value_t*)na;
+    }
+    else if (jl_is_quotenode(expr)) {
+        if (jl_is_symbol(jl_fieldref(expr,0)))
+            return expr;
+        jl_value_t *q = NULL;
+        JL_GC_PUSH2(&q, &expr);
+        q = jl_copy_ast(jl_fieldref(expr,0));
+        jl_value_t *v = jl_new_struct(jl_quotenode_type, q);
+        JL_GC_POP();
+        return v;
+    }
+    return expr;
+}
+
 static jl_value_t *dont_copy_ast(jl_value_t *expr, jl_tuple_t *sp, int do_sp)
 {
     if (jl_is_symbol(expr) || jl_is_lambda_info(expr)) {
@@ -681,11 +763,15 @@ static void eval_decl_types(jl_array_t *vi, jl_tuple_t *spenv)
     for(i=0; i < jl_array_len(vi); i++) {
         jl_array_t *v = (jl_array_t*)jl_cellref(vi, i);
         assert(jl_array_len(v) > 1);
-        jl_value_t *ty =
-            jl_interpret_toplevel_expr_with(jl_cellref(v,1),
-                                            &jl_tupleref(spenv,0),
-                                            jl_tuple_len(spenv)/2);
-        jl_cellref(v, 1) = ty;
+        JL_TRY {
+            jl_value_t *ty =
+                jl_interpret_toplevel_expr_with(jl_cellref(v,1),
+                                                &jl_tupleref(spenv,0),
+                                                jl_tuple_len(spenv)/2);
+            jl_cellref(v, 1) = ty;
+        }
+        JL_CATCH {
+        }
     }
 }
 
@@ -712,7 +798,7 @@ jl_value_t *jl_prepare_ast(jl_lambda_info_t *li, jl_tuple_t *sparams)
     jl_tuple_t *spenv = NULL;
     jl_value_t *ast = li->ast;
     if (ast == NULL) return NULL;
-    JL_GC_PUSH(&spenv, &ast);
+    JL_GC_PUSH2(&spenv, &ast);
     spenv = jl_tuple_tvars_to_symbols(sparams);
     if (!jl_is_expr(ast)) {
         ast = jl_uncompress_ast(li, ast);
