@@ -5,6 +5,7 @@
 
 JL_CALLABLE(jl_f_new_expr);
 JL_CALLABLE(jl_f_new_box);
+JL_CALLABLE(jl_f_new_module);
 JL_CALLABLE(jl_f_throw);
 JL_CALLABLE(jl_f_is);
 JL_CALLABLE(jl_f_no_function);
@@ -13,6 +14,7 @@ JL_CALLABLE(jl_f_subtype);
 JL_CALLABLE(jl_f_isa);
 JL_CALLABLE(jl_f_typeassert);
 JL_CALLABLE(jl_f_apply);
+JL_CALLABLE(jl_f_kwcall);
 JL_CALLABLE(jl_f_top_eval);
 JL_CALLABLE(jl_f_isdefined);
 JL_CALLABLE(jl_f_tuple);
@@ -40,6 +42,5 @@ JL_CALLABLE(jl_unprotect_stack);
 JL_CALLABLE(jl_f_task);
 JL_CALLABLE(jl_f_yieldto);
 JL_CALLABLE(jl_f_ctor_trampoline);
-JL_CALLABLE(jl_f_make_callback);
 
 #endif
