@@ -1,6 +1,8 @@
 /*
   implementations of some built-in functions and utilities
 */
+#include "platform.h"
+
 #include <stdlib.h>
 #include <stdio.h>
 #include <string.h>
@@ -10,12 +12,18 @@
 #include <sys/types.h>
 #include <errno.h>
 #include <fcntl.h>
-#include <unistd.h>
-#ifdef __WIN32__
+#if defined(_OS_WINDOWS_)
 #include <malloc.h>
+#if defined(_COMPILER_INTEL_)
+#include <mathimf.h>
+#else
+#include <math.h>
+#endif
+#else
+#include <unistd.h>
+#include <math.h>
 #endif
 #include <ctype.h>
-#include <math.h>
 #include "julia.h"
 #include "builtin_proto.h"
 
@@ -28,7 +36,7 @@ void jl_error(const char *str)
         jl_exit(1);
     }
     jl_value_t *msg = jl_pchar_to_string((char*)str, strlen(str));
-    JL_GC_PUSH(&msg);
+    JL_GC_PUSH1(&msg);
     jl_throw(jl_new_struct(jl_errorexception_type, msg));
 }
 
@@ -44,7 +52,7 @@ void jl_errorf(const char *fmt, ...)
         jl_exit(1);
     }
     jl_value_t *msg = jl_pchar_to_string(buf, nc);
-    JL_GC_PUSH(&msg);
+    JL_GC_PUSH1(&msg);
     jl_throw(jl_new_struct(jl_errorexception_type, msg));
 }
 
@@ -63,7 +71,7 @@ void jl_type_error_rt(const char *fname, const char *context,
                       jl_value_t *ty, jl_value_t *got)
 {
     jl_value_t *ctxt=NULL;
-    JL_GC_PUSH(&ctxt, &got);
+    JL_GC_PUSH2(&ctxt, &got);
     ctxt = jl_pchar_to_string((char*)context, strlen(context));
     jl_value_t *ex = jl_new_struct(jl_typeerror_type, jl_symbol(fname),
                                    ctxt, ty, got);
@@ -105,6 +113,17 @@ void jl_pop_handler(int n)
 
 // primitives -----------------------------------------------------------------
 
+static int bits_equal(void *a, void *b, int sz)
+{
+    switch (sz) {
+    case 1:  return *(int8_t*)a == *(int8_t*)b;
+    case 2:  return *(int16_t*)a == *(int16_t*)b;
+    case 4:  return *(int32_t*)a == *(int32_t*)b;
+    case 8:  return *(int64_t*)a == *(int64_t*)b;
+    default: return memcmp(a, b, sz)==0;
+    }
+}
+
 int jl_egal(jl_value_t *a, jl_value_t *b)
 {
     if (a == b)
@@ -112,21 +131,6 @@ int jl_egal(jl_value_t *a, jl_value_t *b)
     jl_value_t *ta = (jl_value_t*)jl_typeof(a);
     if (ta != (jl_value_t*)jl_typeof(b))
         return 0;
-    if (jl_is_bits_type(ta)) {
-        size_t nb = jl_bitstype_nbits(ta)/8;
-        switch (nb) {
-        case 1:
-            return *(int8_t*)jl_bits_data(a) == *(int8_t*)jl_bits_data(b);
-        case 2:
-            return *(int16_t*)jl_bits_data(a) == *(int16_t*)jl_bits_data(b);
-        case 4:
-            return *(int32_t*)jl_bits_data(a) == *(int32_t*)jl_bits_data(b);
-        case 8:
-            return *(int64_t*)jl_bits_data(a) == *(int64_t*)jl_bits_data(b);
-        default:
-            return memcmp(jl_bits_data(a), jl_bits_data(b), nb)==0;
-        }
-    }
     if (jl_is_tuple(a)) {
         size_t l = jl_tuple_len(a);
         if (l != jl_tuple_len(b))
@@ -137,9 +141,38 @@ int jl_egal(jl_value_t *a, jl_value_t *b)
         }
         return 1;
     }
-    if (ta == (jl_value_t*)jl_union_kind)
-        return jl_egal(jl_fieldref(a,0), jl_fieldref(b,0));
-    return 0;
+    jl_datatype_t *dt = (jl_datatype_t*)ta;
+    if (dt == jl_datatype_type) {
+        jl_datatype_t *dta = (jl_datatype_t*)a;
+        jl_datatype_t *dtb = (jl_datatype_t*)b;
+        return dta->name == dtb->name &&
+            jl_egal((jl_value_t*)dta->parameters, (jl_value_t*)dtb->parameters);
+    }
+    if (dt->mutabl) return 0;
+    size_t sz = dt->size;
+    if (sz == 0) return 1;
+    size_t nf = jl_tuple_len(dt->names);
+    if (nf == 0) {
+        return bits_equal(jl_data_ptr(a), jl_data_ptr(b), sz);
+    }
+    for (size_t f=0; f < nf; f++) {
+        size_t offs = dt->fields[f].offset;
+        char *ao = (char*)jl_data_ptr(a) + offs;
+        char *bo = (char*)jl_data_ptr(b) + offs;
+        int eq;
+        if (dt->fields[f].isptr) {
+            jl_value_t *af = *(jl_value_t**)ao;
+            jl_value_t *bf = *(jl_value_t**)bo;
+            if (af == bf) eq = 1;
+            else if (af==NULL || bf==NULL) eq = 0;
+            else eq = jl_egal(af, bf);
+        }
+        else {
+            eq = bits_equal(ao, bo, dt->fields[f].size);
+        }
+        if (!eq) return 0;
+    }
+    return 1;
 }
 
 JL_CALLABLE(jl_f_is)
@@ -179,6 +212,12 @@ JL_CALLABLE(jl_f_isa)
     return (jl_subtype(args[0],args[1],1) ? jl_true : jl_false);
 }
 
+DLLEXPORT void jl_typeassert(jl_value_t *x, jl_value_t *t)
+{
+    if (!jl_subtype(x,t,1))
+        jl_type_error("typeassert", t, x);
+}
+
 JL_CALLABLE(jl_f_typeassert)
 {
     JL_NARGS(typeassert, 2, 2);
@@ -201,7 +240,7 @@ JL_CALLABLE(jl_f_apply)
             if (jl_is_array(args[1])) {
                 size_t n = jl_array_len(args[1]);
                 jl_tuple_t *t = jl_alloc_tuple(n);
-                JL_GC_PUSH(&t);
+                JL_GC_PUSH1(&t);
                 for(size_t i=0; i < n; i++) {
                     jl_tupleset(t, i, jl_arrayref((jl_array_t*)args[1], i));
                 }
@@ -253,7 +292,7 @@ JL_CALLABLE(jl_f_apply)
 
  fancy_apply: ;
     jl_value_t *argarr = jl_apply(jl_append_any_func, &args[1], nargs-1);
-    JL_GC_PUSH(&argarr);
+    JL_GC_PUSH1(&argarr);
     assert(jl_typeis(argarr, jl_array_any_type));
     jl_value_t *result = jl_apply((jl_function_t*)args[0],
                                   jl_cell_data(argarr), jl_array_len(argarr));
@@ -261,8 +300,42 @@ JL_CALLABLE(jl_f_apply)
     return result;
 }
 
+#include "newobj_internal.h"
+
+void jl_add_constructors(jl_datatype_t *t);
+
+JL_CALLABLE(jl_f_kwcall)
+{
+    if (nargs < 3)
+        jl_error("internal error: malformed keyword argument call");
+    JL_TYPECHK(apply, function, args[0]);
+    jl_function_t *f = (jl_function_t*)args[0];
+    if (f->fptr == jl_f_ctor_trampoline)
+        jl_add_constructors((jl_datatype_t*)f);
+    if (!jl_is_gf(f))
+        jl_error("function does not accept keyword arguments");
+    jl_function_t *sorter = ((jl_methtable_t*)f->env)->kwsorter;
+    if (sorter == NULL) {
+        jl_errorf("function %s does not accept keyword arguments",
+                  jl_gf_name(f)->name);
+    }
+
+    size_t nkeys = jl_unbox_long(args[1]);
+    size_t pa = 3 + 2*nkeys;
+    jl_array_t *container = (jl_array_t*)args[pa-1];
+    assert(jl_array_len(container) > 0);
+    for(size_t i=0; i < nkeys*2; i+=2) {
+        jl_cellset(container, i  , args[2+i]);
+        jl_cellset(container, i+1, args[2+i+1]);
+    }
+
+    return jl_apply(sorter, &args[pa-1], nargs-(pa-1));
+}
+
 // eval -----------------------------------------------------------------------
 
+extern int jl_lineno;
+
 JL_CALLABLE(jl_f_top_eval)
 {
     jl_module_t *m;
@@ -280,19 +353,24 @@ JL_CALLABLE(jl_f_top_eval)
     if (jl_is_symbol(ex)) {
         return jl_eval_global_var(m, (jl_sym_t*)ex);
     }
-    if (m == jl_current_module) {
-        return jl_toplevel_eval(ex);
-    }
     jl_value_t *v=NULL;
+    int last_lineno = jl_lineno;
+    if (m == jl_current_module) {
+        v = jl_toplevel_eval(ex);
+        jl_lineno = last_lineno;
+        return v;
+    }
     jl_module_t *last_m = jl_current_module;
     JL_TRY {
         jl_current_module = m;
         v = jl_toplevel_eval(ex);
     }
     JL_CATCH {
+        jl_lineno = last_lineno;
         jl_current_module = last_m;
         jl_rethrow();
     }
+    jl_lineno = last_lineno;
     jl_current_module = last_m;
     assert(v);
     return v;
@@ -318,10 +396,10 @@ JL_CALLABLE(jl_f_isdefined)
         s = (jl_sym_t*)args[1];
         if (!jl_is_module(args[0])) {
             jl_value_t *vt = (jl_value_t*)jl_typeof(args[0]);
-            if (!jl_is_struct_type(vt)) {
-                jl_type_error("isdefined", (jl_value_t*)jl_struct_kind, args[0]);
+            if (!jl_is_datatype(vt)) {
+                jl_type_error("isdefined", (jl_value_t*)jl_datatype_type, args[0]);
             }
-            return jl_field_isdefined(args[0], s, 1) ? jl_true : jl_false;
+            return jl_field_isdefined(args[0], s, 0) ? jl_true : jl_false;
         }
         JL_TYPECHK(isdefined, module, args[0]);
         m = (jl_module_t*)args[0];
@@ -373,10 +451,10 @@ JL_CALLABLE(jl_f_get_field)
     if (vt == (jl_value_t*)jl_module_type) {
         return jl_eval_global_var((jl_module_t*)v, (jl_sym_t*)args[1]);
     }
-    if (!jl_is_struct_type(vt)) {
-        jl_type_error("getfield", (jl_value_t*)jl_struct_kind, v);
+    if (!jl_is_datatype(vt)) {
+        jl_type_error("getfield", (jl_value_t*)jl_datatype_type, v);
     }
-    jl_struct_type_t *st = (jl_struct_type_t*)vt;
+    jl_datatype_t *st = (jl_datatype_t*)vt;
     jl_sym_t *fld = (jl_sym_t*)args[1];
     jl_value_t *fval = jl_get_nth_field(v, jl_field_index(st, fld, 1));
     if (fval == NULL)
@@ -392,9 +470,11 @@ JL_CALLABLE(jl_f_set_field)
     jl_value_t *vt = (jl_value_t*)jl_typeof(v);
     if (vt == (jl_value_t*)jl_module_type)
         jl_error("cannot assign variables in other modules");
-    if (!jl_is_struct_type(vt))
-        jl_type_error("setfield", (jl_value_t*)jl_struct_kind, v);
-    jl_struct_type_t *st = (jl_struct_type_t*)vt;
+    if (!jl_is_datatype(vt))
+        jl_type_error("setfield", (jl_value_t*)jl_datatype_type, v);
+    jl_datatype_t *st = (jl_datatype_t*)vt;
+    if (!st->mutabl)
+        jl_errorf("type %s is immutable", st->name->name->name);
     jl_sym_t *fld = (jl_sym_t*)args[1];
     size_t i = jl_field_index(st, fld, 1);
     jl_value_t *ft = jl_tupleref(st->types,i);
@@ -413,16 +493,16 @@ JL_CALLABLE(jl_f_field_type)
     jl_value_t *vt = (jl_value_t*)jl_typeof(v);
     if (vt == (jl_value_t*)jl_module_type)
         jl_error("cannot assign variables in other modules");
-    if (!jl_is_struct_type(vt))
-        jl_type_error("fieldtype", (jl_value_t*)jl_struct_kind, v);
-    jl_struct_type_t *st = (jl_struct_type_t*)vt;
+    if (!jl_is_datatype(vt))
+        jl_type_error("fieldtype", (jl_value_t*)jl_datatype_type, v);
+    jl_datatype_t *st = (jl_datatype_t*)vt;
     jl_sym_t *fld = (jl_sym_t*)args[1];
     return jl_tupleref(st->types, jl_field_index(st, fld, 1));
 }
 
 // conversion -----------------------------------------------------------------
 
-static jl_value_t *convert(jl_type_t *to, jl_value_t *x, jl_function_t *conv_f)
+static jl_value_t *convert(jl_value_t *to, jl_value_t *x, jl_function_t *conv_f)
 {
     jl_value_t *args[2];
     if (jl_subtype(x, (jl_value_t*)to, 1))
@@ -439,7 +519,7 @@ JL_CALLABLE(jl_f_convert_tuple)
         return (jl_value_t*)x;
     size_t i, cl=jl_tuple_len(x), pl=jl_tuple_len(to);
     jl_tuple_t *out = jl_alloc_tuple(cl);
-    JL_GC_PUSH(&out);
+    JL_GC_PUSH1(&out);
     jl_value_t *ce, *pe=NULL;
     int pseq=0;
     jl_function_t *f = (jl_function_t*)args[2];
@@ -459,7 +539,7 @@ JL_CALLABLE(jl_f_convert_tuple)
             break;
         }
         assert(pe != NULL);
-        jl_tupleset(out, i, convert((jl_type_t*)pe, ce, f));
+        jl_tupleset(out, i, convert((jl_value_t*)pe, ce, f));
     }
     JL_GC_POP();
     if (out == NULL)
@@ -469,7 +549,7 @@ JL_CALLABLE(jl_f_convert_tuple)
 
 JL_CALLABLE(jl_f_convert_default)
 {
-    jl_type_t *to = (jl_type_t*)args[0];
+    jl_value_t *to = args[0];
     jl_value_t *x = args[1];
     if (!jl_subtype(x, (jl_value_t*)to, 1)) {
         jl_no_method_error((jl_function_t*)args[2], args, 2);
@@ -505,6 +585,18 @@ DLLEXPORT void jl_print_int64(JL_STREAM *s, int64_t i)
     JL_PRINTF(s, "%lld", i);
 }
 
+DLLEXPORT int jl_substrtod(char *str, size_t offset, int len, double *out)
+{
+    char *p;
+    errno = 0;
+    char *bstr = str+offset;
+    *out = strtod(bstr, &p);
+    if ((p == bstr) || (p != (bstr+len)) ||
+        (errno==ERANGE && (*out==0 || *out==HUGE_VAL || *out==-HUGE_VAL)))
+        return 1;
+    return 0;
+}
+
 DLLEXPORT int jl_strtod(char *str, double *out)
 {
     char *p;
@@ -521,11 +613,32 @@ DLLEXPORT int jl_strtod(char *str, double *out)
     return 0;
 }
 
+DLLEXPORT int jl_substrtof(char *str, int offset, int len, float *out)
+{
+    char *p;
+    errno = 0;
+    char *bstr = str+offset;
+#if defined(_OS_WINDOWS_) && !defined(_COMPILER_MINGW_)
+    *out = (float)strtod(bstr, &p);
+#else
+    *out = strtof(bstr, &p);
+#endif
+
+    if ((p == bstr) || (p != (bstr+len)) ||
+        (errno==ERANGE && (*out==0 || *out==HUGE_VALF || *out==-HUGE_VALF)))
+        return 1;
+    return 0;
+}
+
 DLLEXPORT int jl_strtof(char *str, float *out)
 {
     char *p;
     errno = 0;
+#if defined(_OS_WINDOWS_) && !defined(_COMPILER_MINGW_)
+    *out = (float)strtod(str, &p);
+#else
     *out = strtof(str, &p);
+#endif
     if (p == str ||
         (errno==ERANGE && (*out==0 || *out==HUGE_VALF || *out==-HUGE_VALF)))
         return 1;
@@ -539,6 +652,12 @@ DLLEXPORT int jl_strtof(char *str, float *out)
 
 // showing --------------------------------------------------------------------
 
+void jl_flush_cstdio()
+{
+    fflush(stdout);
+    fflush(stderr);
+}
+
 jl_value_t *jl_stdout_obj()
 {
     jl_value_t *stdout_obj = jl_get_global(jl_base_module, jl_symbol("STDOUT"));
@@ -562,9 +681,9 @@ void jl_show(jl_value_t *stream, jl_value_t *v)
             jl_show_gf = (jl_function_t*)jl_get_global(jl_base_module, jl_symbol("show"));
         }
         if (jl_show_gf==NULL || stream==NULL) {
-            JL_PRINTF(JL_STDERR, "could not show value of type %s",
+            JL_PRINTF(JL_STDERR, " could not show value of type %s",
                       jl_is_tuple(v) ? "Tuple" : 
-                      ((jl_tag_type_t*)jl_typeof(v))->name->name->name);
+                      ((jl_datatype_t*)jl_typeof(v))->name->name->name);
             return;
         }
         jl_value_t *args[2] = {stream,v};
@@ -599,7 +718,7 @@ static void show_function(JL_STREAM *s, jl_value_t *v)
 static void show_type(jl_value_t *st, jl_value_t *t)
 {
     uv_stream_t *s =((uv_stream_t**)st)[1];
-    if (jl_is_union_type(t)) {
+    if (jl_is_uniontype(t)) {
         if (t == (jl_value_t*)jl_bottom_type) {
             JL_WRITE(s, "None", 4);
         }
@@ -619,8 +738,8 @@ static void show_type(jl_value_t *st, jl_value_t *t)
         jl_show(st, (jl_value_t*)((jl_typector_t*)t)->body);
     }
     else {
-        assert(jl_is_some_tag_type(t));
-        jl_tag_type_t *tt = (jl_tag_type_t*)t;
+        assert(jl_is_datatype(t));
+        jl_datatype_t *tt = (jl_datatype_t*)t;
         JL_PUTS(tt->name->name->name, s);
         jl_tuple_t *p = tt->parameters;
         if (jl_tuple_len(p) > 0)
@@ -642,21 +761,17 @@ DLLEXPORT void jl_show_any(jl_value_t *str, jl_value_t *v)
         show_function(s, v);
     }
     else if (jl_typeis(v,jl_intrinsic_type)) {
-        JL_PRINTF(s, "# intrinsic function %d", *(uint32_t*)jl_bits_data(v));
+        JL_PRINTF(s, "# intrinsic function %d", *(uint32_t*)jl_data_ptr(v));
     }
     else {
         jl_value_t *t = (jl_value_t*)jl_typeof(v);
-        assert(jl_is_struct_type(t) || jl_is_bits_type(t));
-        jl_tag_type_t *tt = (jl_tag_type_t*)t;
-        JL_PUTS(tt->name->name->name, s);
-        if (tt->parameters != jl_null) {
-            jl_show_tuple(str, tt->parameters, '{', '}', 0);
-        }
+        assert(jl_is_datatype(t));
+        jl_datatype_t *dt = (jl_datatype_t*)t;
+        show_type(str, t);
         JL_PUTC('(', s);
-        if (jl_is_struct_type(tt)) {
-            jl_struct_type_t *st = (jl_struct_type_t*)tt;
+        if (jl_tuple_len(dt->names)>0 || dt->size==0) {
             size_t i;
-            size_t n = jl_tuple_len(st->names);
+            size_t n = jl_tuple_len(dt->names);
             for(i=0; i < n; i++) {
                 jl_value_t *fval = jl_get_nth_field(v, i);
                 if (fval == NULL)
@@ -668,8 +783,8 @@ DLLEXPORT void jl_show_any(jl_value_t *str, jl_value_t *v)
             }
         }
         else {
-            size_t nb = jl_bitstype_nbits(tt)/8;
-            char *data = (char*)jl_bits_data(v);
+            size_t nb = jl_datatype_size(dt);
+            char *data = (char*)jl_data_ptr(v);
             JL_PUTS("0x", s);
             for(int i=nb-1; i >= 0; --i)
                 jl_printf(s, "%02hhx", data[i]);
@@ -681,6 +796,7 @@ DLLEXPORT void jl_show_any(jl_value_t *str, jl_value_t *v)
 // internal functions ---------------------------------------------------------
 
 extern int jl_in_inference;
+extern int jl_boot_file_loaded;
 int jl_eval_with_compiler_p(jl_expr_t *expr, int compileloops);
 
 JL_CALLABLE(jl_trampoline)
@@ -689,6 +805,8 @@ JL_CALLABLE(jl_trampoline)
     jl_function_t *f = (jl_function_t*)F;
     assert(f->linfo != NULL);
     // to run inference on all thunks. slows down loading files.
+    // NOTE: if this call to inference is removed, type_annotate in inference.jl
+    // needs to be updated to infer inner functions.
     if (f->linfo->inferred == 0) {
         if (!jl_in_inference) {
             if (!jl_is_expr(f->linfo->ast)) {
@@ -702,13 +820,16 @@ JL_CALLABLE(jl_trampoline)
     jl_compile(f);
     assert(f->fptr == &jl_trampoline);
     jl_generate_fptr(f);
+    if (jl_boot_file_loaded && jl_is_expr(f->linfo->ast)) {
+        f->linfo->ast = jl_compress_ast(f->linfo, f->linfo->ast);
+    }
     return jl_apply(f, args, nargs);
 }
 
 JL_CALLABLE(jl_f_instantiate_type)
 {
     JL_NARGSV(instantiate_type, 1);
-    if (!jl_is_some_tag_type(args[0]))
+    if (!jl_is_datatype(args[0]))
         JL_TYPECHK(instantiate_type, typector, args[0]);
     return jl_apply_type_(args[0], &args[1], nargs-1);
 }
@@ -720,7 +841,11 @@ JL_CALLABLE(jl_f_new_type_constructor)
     if (!jl_is_type(args[1]))
         jl_type_error("typealias", (jl_value_t*)jl_type_type, args[1]);
     jl_tuple_t *p = (jl_tuple_t*)args[0];
-    return (jl_value_t*)jl_new_type_ctor(p, (jl_type_t*)args[1]);
+    jl_value_t *tc = (jl_value_t*)jl_new_type_ctor(p, args[1]);
+    int i;
+    for(i=0; i < jl_tuple_len(p); i++)
+        ((jl_tvar_t*)jl_tupleref(p,i))->bound = 0;
+    return tc;
 }
 
 JL_CALLABLE(jl_f_typevar)
@@ -729,10 +854,6 @@ JL_CALLABLE(jl_f_typevar)
         JL_NARGS(typevar, 1, 1);
     }
     JL_TYPECHK(typevar, symbol, args[0]);
-    if (jl_boundp(jl_current_module, (jl_sym_t*)args[0]) &&
-        jl_is_type(jl_get_global(jl_current_module, (jl_sym_t*)args[0]))) {
-        jl_printf(JL_STDERR, "Warning: type parameter name %s shadows an identifier.\n", ((jl_sym_t*)args[0])->name);
-    }
     jl_value_t *lb = (jl_value_t*)jl_bottom_type;
     jl_value_t *ub = (jl_value_t*)jl_any_type;
     int b = 0;
@@ -770,7 +891,7 @@ JL_CALLABLE(jl_f_union)
             jl_tupleset(argt, i, args[i]);
         }
     }
-    JL_GC_PUSH(&argt);
+    JL_GC_PUSH1(&argt);
     jl_value_t *u = jl_type_union(argt);
     JL_GC_POP();
     return u;
@@ -788,7 +909,7 @@ JL_CALLABLE(jl_f_methodexists)
     jl_check_type_tuple((jl_tuple_t*)args[1], jl_gf_name(args[0]),
                         "method_exists");
     return jl_method_lookup_by_type(jl_gf_mtable(args[0]),
-                                    (jl_tuple_t*)args[1], 0) != jl_bottom_func ?
+                                    (jl_tuple_t*)args[1],0,0)!=jl_bottom_func ?
         jl_true : jl_false;
 }
 
@@ -799,7 +920,7 @@ JL_CALLABLE(jl_f_applicable)
     if (!jl_is_gf(args[0]))
         jl_error("applicable: not a generic function");
     return jl_method_lookup(jl_gf_mtable(args[0]),
-                            &args[1], nargs-1, 0) != jl_bottom_func ?
+                            &args[1], nargs-1, 1) != jl_bottom_func ?
         jl_true : jl_false;
 }
 
@@ -824,7 +945,7 @@ JL_CALLABLE(jl_f_invoke)
 
 // hashing --------------------------------------------------------------------
 
-#ifdef __LP64__
+#ifdef _P64
 #define bitmix(a,b) int64hash((a)^bswap_64(b))
 #define hash64(a)   int64hash(a)
 #else
@@ -832,45 +953,62 @@ JL_CALLABLE(jl_f_invoke)
 #define hash64(a)   int64to32hash(a)
 #endif
 
+static uptrint_t bits_hash(void *b, size_t sz)
+{
+    switch (sz) {
+    case 1:  return int32hash(*(int8_t*)b);
+    case 2:  return int32hash(*(int16_t*)b);
+    case 4:  return int32hash(*(int32_t*)b);
+    case 8:  return hash64(*(int64_t*)b);
+    default:
+#ifdef _P64
+        return memhash((char*)b, sz);
+#else
+        return memhash32((char*)b, sz);
+#endif
+    }
+}
+
 DLLEXPORT uptrint_t jl_object_id(jl_value_t *v)
 {
     if (jl_is_symbol(v))
         return ((jl_sym_t*)v)->hash;
     jl_value_t *tv = (jl_value_t*)jl_typeof(v);
-    if (jl_is_bits_type(tv)) {
-        size_t nb = jl_bitstype_nbits(tv)/8;
-        uptrint_t h = inthash((uptrint_t)tv);
-        switch (nb) {
-        case 1:
-            return int32hash(*(int8_t*)jl_bits_data(v) ^ h);
-        case 2:
-            return int32hash(*(int16_t*)jl_bits_data(v) ^ h);
-        case 4:
-            return int32hash(*(int32_t*)jl_bits_data(v) ^ h);
-        case 8:
-            return hash64(*(int64_t*)jl_bits_data(v) ^ h);
-        default:
-#ifdef __LP64__
-            return h ^ memhash((char*)jl_bits_data(v), nb);
-#else
-            return h ^ memhash32((char*)jl_bits_data(v), nb);
-#endif
+    if (tv == (jl_value_t*)jl_tuple_type) {
+        uptrint_t h = 0;
+        size_t l = jl_tuple_len(v);
+        for(size_t i = 0; i < l; i++) {
+            uptrint_t u = jl_object_id(jl_tupleref(v,i));
+            h = bitmix(h, u);
         }
+        return h;
     }
-    if (tv == (jl_value_t*)jl_union_kind) {
-#ifdef __LP64__
-        return jl_object_id(jl_fieldref(v,0))^0xA5A5A5A5A5A5A5A5L;
-#else
-        return jl_object_id(jl_fieldref(v,0))^0xA5A5A5A5;
-#endif
+    jl_datatype_t *dt = (jl_datatype_t*)tv;
+    if (dt == jl_datatype_type) {
+        jl_datatype_t *dtv = (jl_datatype_t*)v;
+        uptrint_t h = inthash((uptrint_t)tv);
+        return bitmix(bitmix(h, jl_object_id((jl_value_t*)dtv->name)),
+                      jl_object_id((jl_value_t*)dtv->parameters));
     }
-    if (jl_is_struct_type(tv))
-        return inthash((uptrint_t)v);
-    assert(jl_is_tuple(v));
-    uptrint_t h = 0;
-    size_t l = jl_tuple_len(v);
-    for(size_t i = 0; i < l; i++) {
-        uptrint_t u = jl_object_id(jl_tupleref(v,i));
+    if (dt->mutabl) return inthash((uptrint_t)v);
+    size_t sz = jl_datatype_size(tv);
+    uptrint_t h = inthash((uptrint_t)tv);
+    if (sz == 0) return ~h;
+    size_t nf = jl_tuple_len(dt->names);
+    if (nf == 0) {
+        return bits_hash(jl_data_ptr(v), sz) ^ h;
+    }
+    for (size_t f=0; f < nf; f++) {
+        size_t offs = dt->fields[f].offset;
+        char *vo = (char*)jl_data_ptr(v) + offs;
+        uptrint_t u;
+        if (dt->fields[f].isptr) {
+            jl_value_t *f = *(jl_value_t**)vo;
+            u = f==NULL ? 0 : jl_object_id(f);
+        }
+        else {
+            u = bits_hash(vo, dt->fields[f].size);
+        }
         h = bitmix(h, u);
     }
     return h;
@@ -893,10 +1031,11 @@ void jl_init_primitives(void)
 {
     add_builtin_func("is", jl_f_is);
     add_builtin_func("typeof", jl_f_typeof);
-    add_builtin_func("subtype", jl_f_subtype);
+    add_builtin_func("issubtype", jl_f_subtype);
     add_builtin_func("isa", jl_f_isa);
     add_builtin_func("typeassert", jl_f_typeassert);
     add_builtin_func("apply", jl_f_apply);
+    add_builtin_func("kwcall", jl_f_kwcall);
     add_builtin_func("throw", jl_f_throw);
     add_builtin_func("tuple", jl_f_tuple);
     add_builtin_func("Union", jl_f_union);
@@ -935,11 +1074,8 @@ void jl_init_primitives(void)
     add_builtin("NTuple", (jl_value_t*)jl_ntuple_type);
     add_builtin("Type", (jl_value_t*)jl_type_type);
     add_builtin("Vararg", (jl_value_t*)jl_vararg_type);
-    add_builtin("BitsKind", (jl_value_t*)jl_bits_kind);
-    add_builtin("CompositeKind", (jl_value_t*)jl_struct_kind);
-    add_builtin("AbstractKind", (jl_value_t*)jl_tag_kind);
-    add_builtin("UnionKind", (jl_value_t*)jl_union_kind);
-    // todo: this should only be visible to compiler components
+    add_builtin("DataType", (jl_value_t*)jl_datatype_type);
+    add_builtin("UnionType", (jl_value_t*)jl_uniontype_type);
     add_builtin("Undef", (jl_value_t*)jl_undef_type);
 
     add_builtin("Module", (jl_value_t*)jl_module_type);
@@ -962,8 +1098,9 @@ void jl_init_primitives(void)
     add_builtin("GotoNode", (jl_value_t*)jl_gotonode_type);
     add_builtin("QuoteNode", (jl_value_t*)jl_quotenode_type);
     add_builtin("TopNode", (jl_value_t*)jl_topnode_type);
+    add_builtin("NewvarNode", (jl_value_t*)jl_newvarnode_type);
 
-#ifdef __LP64__
+#ifdef _P64
     add_builtin("Int", (jl_value_t*)jl_int64_type);
 #else
     add_builtin("Int", (jl_value_t*)jl_int32_type);
@@ -971,3 +1108,271 @@ void jl_init_primitives(void)
 
     add_builtin("ANY", jl_ANY_flag);
 }
+
+// toys for debugging -----------------------------------------------------------------------
+
+DLLEXPORT size_t jl_static_show(JL_STREAM *out, jl_value_t *v)
+{
+    // mimic jl_show, but never calling a julia method
+    size_t n = 0;
+    if (v == NULL) {
+        n += JL_PRINTF(out, "<null>");
+    }
+    else if (jl_is_lambda_info(v)) {
+        jl_lambda_info_t *li = (jl_lambda_info_t*)v;
+        n += jl_static_show(out, (jl_value_t*)li->module);
+        n += JL_PRINTF(out, ".%s", li->name->name);
+        if (li->specTypes) {
+            n += jl_static_show(out, (jl_value_t*)li->specTypes);
+        }
+        else {
+            n += JL_PRINTF(out, "(?)");
+        }
+    }
+    else if (jl_is_tuple(v)) {
+        jl_tuple_t *t = (jl_tuple_t*)v;
+        size_t i, len = jl_tuple_len(t);
+        n += JL_PRINTF(out, "(");
+        for (i = 0; i < len; i++) {
+            jl_value_t *v = jl_tupleref(t,i);
+            n += jl_static_show(out, v);
+            if (len == 1)
+                n += JL_PRINTF(out, ",");
+            else if (i != len-1)
+                n += JL_PRINTF(out, ", ");
+        }
+        n += JL_PRINTF(out, ")");
+    }
+    else if (jl_is_vararg_type(v)) {
+        n += jl_static_show(out, jl_tparam0(v));
+        n += JL_PRINTF(out, "...");
+    }
+    else if (jl_is_datatype(v)) {
+        jl_datatype_t *dv = (jl_datatype_t*)v;
+        if (dv->name->module != jl_core_module) {
+            n += jl_static_show(out, (jl_value_t*)dv->name->module);
+            JL_PUTS(".", out); n += 1;
+        }
+        n += JL_PRINTF(out, "%s", dv->name->name->name);
+        if (dv->parameters) {
+            size_t j, tlen = jl_tuple_len(dv->parameters);
+            if (tlen > 0) {
+                n += JL_PRINTF(out, "{");
+                for (j = 0; j < tlen; j++) {
+                    jl_value_t *p = jl_tupleref(dv->parameters,j);
+                    n += jl_static_show(out, p);
+                    if (j != tlen-1)
+                        n += JL_PRINTF(out, ", ");
+                }
+                n += JL_PRINTF(out, "}");
+            }
+        }
+    }
+    else if (jl_is_func(v)) {
+        if (jl_is_gf(v)) {
+            n += JL_PRINTF(out, "%s", jl_gf_name(v)->name);
+        }
+        else {
+            n += JL_PRINTF(out, "<# function>");
+        }
+    }
+    else if (jl_typeis(v, jl_intrinsic_type)) {
+        n += JL_PRINTF(out, "<# intrinsic function %d>", *(uint32_t*)jl_data_ptr(v));
+    }
+    else if (jl_is_int64(v)) {
+        n += JL_PRINTF(out, "%d", jl_unbox_int64(v));
+    }
+    else if (jl_is_int32(v)) {
+        n += JL_PRINTF(out, "%d", jl_unbox_int32(v));
+    }
+    else if (jl_typeis(v,jl_int16_type)) {
+        n += JL_PRINTF(out, "%d", jl_unbox_int16(v));
+    }
+    else if (jl_typeis(v,jl_int8_type)) {
+        n += JL_PRINTF(out, "%d", jl_unbox_int8(v));
+    }
+    else if (jl_is_uint64(v)) {
+        n += JL_PRINTF(out, "0x%016x", jl_unbox_uint64(v));
+    }
+    else if (jl_is_uint32(v)) {
+        n += JL_PRINTF(out, "0x%08x", jl_unbox_uint32(v));
+    }
+    else if (jl_typeis(v,jl_uint16_type)) {
+        n += JL_PRINTF(out, "0x%04x", jl_unbox_uint16(v));
+    }
+    else if (jl_typeis(v,jl_uint8_type)) {
+        n += JL_PRINTF(out, "0x%02x", jl_unbox_uint8(v));
+    }
+    else if (jl_is_cpointer(v)) {
+#ifdef _P64
+        n += JL_PRINTF(out, "0x%016x", jl_unbox_voidpointer(v));
+#else
+        n += JL_PRINTF(out, "0x%08x", jl_unbox_voidpointer(v));
+#endif
+    }
+    else if (jl_is_float32(v)) {
+        n += JL_PRINTF(out, "%g", jl_unbox_float32(v));
+    }
+    else if (jl_is_float64(v)) {
+        n += JL_PRINTF(out, "%g", jl_unbox_float64(v));
+    }
+    else if (v == jl_true) {
+        n += JL_PRINTF(out, "true");
+    }
+    else if (v == jl_false) {
+        n += JL_PRINTF(out, "false");
+    }
+    else if (jl_is_byte_string(v)) {
+        n += JL_PRINTF(out, "\"%s\"", jl_iostr_data(v));
+    }
+    else if (v == jl_bottom_type) {
+        n += JL_PRINTF(out, "Void");
+    }
+    else if (jl_is_uniontype(v)) {
+        n += JL_PRINTF(out, "Union");
+        n += jl_static_show(out, (jl_value_t*)((jl_uniontype_t*)v)->types);
+    }
+    else if (jl_is_typector(v)) {
+        n += jl_static_show(out, ((jl_typector_t*)v)->body);
+    }
+    else if (jl_is_typevar(v)) {
+        n += JL_PRINTF(out, "%s", ((jl_tvar_t*)v)->name->name);
+    }
+    else if (jl_is_module(v)) {
+        jl_module_t *m = (jl_module_t*)v;
+        if (m->parent != m && m->parent != jl_main_module) {
+            n += jl_static_show(out, (jl_value_t*)m->parent);
+            n += JL_PRINTF(out, ".");
+        }
+        n += JL_PRINTF(out, "%s", m->name->name);
+    }
+    else if (jl_is_symbol(v)) {
+        n += JL_PRINTF(out, ":%s", ((jl_sym_t*)v)->name);
+    }
+    else if (jl_is_symbolnode(v)) {
+        n += JL_PRINTF(out, "%s::", jl_symbolnode_sym(v)->name);
+        n += jl_static_show(out, jl_symbolnode_type(v));
+    }
+    else if (jl_is_getfieldnode(v)) {
+        n += jl_static_show(out, jl_getfieldnode_val(v));
+        n += JL_PRINTF(out, ".%s", jl_getfieldnode_name(v)->name);
+        n += JL_PRINTF(out, "::");
+        n += jl_static_show(out, jl_getfieldnode_type(v));
+    }
+    else if (jl_is_labelnode(v)) {
+        n += JL_PRINTF(out, "%d:", jl_labelnode_label(v));
+    }
+    else if (jl_is_gotonode(v)) {
+        n += JL_PRINTF(out, "goto %d", jl_gotonode_label(v));
+    }
+    else if (jl_is_quotenode(v)) {
+        n += JL_PRINTF(out, "quote ");
+        n += jl_static_show(out, jl_fieldref(v,0));
+        n += JL_PRINTF(out, " end");
+    }
+    else if (jl_is_newvarnode(v)) {
+        n += JL_PRINTF(out, "<newvar ");
+        n += jl_static_show(out, jl_fieldref(v,0));
+        n += JL_PRINTF(out, ">");
+    }
+    else if (jl_is_topnode(v)) {
+        n += JL_PRINTF(out, "top(");
+        n += jl_static_show(out, jl_fieldref(v,0));
+        n += JL_PRINTF(out, ")");
+    }
+    else if (jl_is_linenode(v)) {
+        n += JL_PRINTF(out, "# line %d", jl_linenode_line(v));
+    }
+    else if (jl_is_expr(v)) {
+        jl_expr_t *e = (jl_expr_t*)v;
+        if (e->head == assign_sym && jl_array_len(e->args) == 2) {
+            n += jl_static_show(out, jl_exprarg(e,0));
+            n += JL_PRINTF(out, " = ");
+            n += jl_static_show(out, jl_exprarg(e,1));
+        }
+        else {
+            char sep = ' ';
+            if (e->head == body_sym)
+                sep = '\n';
+            n += JL_PRINTF(out, "Expr(:%s", e->head->name);
+            size_t i, len = jl_array_len(e->args);
+            for (i = 0; i < len; i++) {
+                n += JL_PRINTF(out, ",%c", sep);
+                n += jl_static_show(out, jl_exprarg(e,i));
+            }
+            n += JL_PRINTF(out, ")::");
+            n += jl_static_show(out, e->etype);
+        }
+    }
+    else if (jl_is_array(v)) {
+        n += jl_static_show(out, jl_typeof(v));
+        n += JL_PRINTF(out, "[");
+        size_t j, tlen = jl_array_len(v);
+        for (j = 0; j < tlen; j++) {
+            n += jl_static_show(out, jl_arrayref((jl_array_t*)v,j));
+            if (j != tlen-1)
+               n += JL_PRINTF(out, ", ");
+        }
+        n += JL_PRINTF(out, "]");
+    }
+    else if (jl_typeis(v,jl_loaderror_type)) {
+        n += JL_PRINTF(out, "LoadError(at ");
+        n += jl_static_show(out, jl_fieldref(v, 0));
+        n += JL_PRINTF(out, " line ");
+        n += jl_static_show(out, jl_fieldref(v, 1));
+        n += JL_PRINTF(out, ": ");
+        n += jl_static_show(out, jl_fieldref(v, 2));
+        n += JL_PRINTF(out, ")");
+    }
+    else if (jl_typeis(v,jl_errorexception_type)) {
+        n += JL_PRINTF(out, "ErrorException(");
+        n += jl_static_show(out, jl_fieldref(v, 0));
+        n += JL_PRINTF(out, ")");
+    }
+    else if (jl_is_datatype(jl_typeof(v))) {
+        jl_datatype_t *t = (jl_datatype_t*)jl_typeof(v);
+        n += jl_static_show(out, (jl_value_t*)t);
+        n += JL_PRINTF(out, "(");
+        size_t i, tlen = jl_tuple_len(t->names);
+        for (i = 0; i < tlen; i++) {
+            n += JL_PRINTF(out, ((jl_sym_t*)jl_tupleref(t->names, i))->name);
+            jl_fielddesc_t f = t->fields[i];
+            if (f.isptr) {
+                n += JL_PRINTF(out, "=");
+                n += jl_static_show(out, *((jl_value_t**)jl_data_ptr(v)+f.offset/sizeof(void*)));
+            }
+            else {
+                n += JL_PRINTF(out, "::");
+                n += jl_static_show(out, jl_tupleref(t->types, i));
+            }
+            if (i != tlen-1)
+                n += JL_PRINTF(out, ", ");
+        }
+        n += JL_PRINTF(out, ")");
+    }
+    else {
+        n += JL_PRINTF(out, "<?::");
+        n += jl_static_show(out, jl_typeof(v));
+        n += JL_PRINTF(out, ">");
+    }
+    return n;
+}
+
+int in_jl_ = 0;
+DLLEXPORT void jl_(void *jl_value)
+{
+    in_jl_++;
+    JL_TRY {
+        (void)jl_static_show(JL_STDOUT, (jl_value_t*)jl_value);
+        JL_PRINTF(JL_STDOUT,"\n");
+    }
+    JL_CATCH {
+        JL_PRINTF(JL_STDOUT, "\n!!! ERROR in jl_ -- ABORTING !!!\n");
+    }
+    in_jl_--;
+}
+
+DLLEXPORT void jl_breakpoint(jl_value_t* v)
+{
+    // put a breakpoint in you debugger here
+}
