@@ -39,7 +39,7 @@ static GlobalVariable *stringConst(const std::string &txt)
 
 static Value *literal_pointer_val(void *p, Type *t)
 {
-#ifdef __LP64__
+#if defined(_P64)
     return ConstantExpr::getIntToPtr(ConstantInt::get(T_int64, (uint64_t)p),
                                      t);
 #else
@@ -60,11 +60,14 @@ static Value *literal_pointer_val(void *p)
 
 // --- mapping between julia and llvm types ---
 
+static Type *julia_struct_to_llvm(jl_value_t *jt);
+
 static Type *julia_type_to_llvm(jl_value_t *jt)
 {
     if (jt == (jl_value_t*)jl_bool_type) return T_int1;
-    if (jt == (jl_value_t*)jl_float32_type) return T_float32;
-    if (jt == (jl_value_t*)jl_float64_type) return T_float64;
+    if (jt == (jl_value_t*)jl_bottom_type) return T_void;
+    if (!jl_is_leaf_type(jt))
+        return jl_pvalue_llvmt;
     if (jl_is_cpointer_type(jt)) {
         Type *lt = julia_type_to_llvm(jl_tparam0(jt));
         if (lt == NULL)
@@ -73,40 +76,60 @@ static Type *julia_type_to_llvm(jl_value_t *jt)
             lt = T_int8;
         return PointerType::get(lt, 0);
     }
-    if (jl_is_bits_type(jt)) {
-        int nb = jl_bitstype_nbits(jt);
-        if (nb == 8)  return T_int8;
-        if (nb == 16) return T_int16;
-        if (nb == 32) return T_int32;
-        if (nb == 64) return T_int64;
-        else          return Type::getIntNTy(getGlobalContext(), nb);
+    if (jl_is_bitstype(jt)) {
+        int nb = jl_datatype_size(jt);
+        if (jl_is_floattype(jt)) {
+#ifndef DISABLE_FLOAT16
+            if (nb == 2)
+                return Type::getHalfTy(jl_LLVMContext);
+            else 
+#endif
+            if (nb == 4)
+                return Type::getFloatTy(jl_LLVMContext);
+            else if (nb == 8)
+                return Type::getDoubleTy(jl_LLVMContext);
+            else if (nb == 16)
+                return Type::getFP128Ty(jl_LLVMContext);
+        }
+        return Type::getIntNTy(jl_LLVMContext, nb*8);
+    }
+    if (jl_isbits(jt)) {
+        if (((jl_datatype_t*)jt)->size == 0) {
+            // TODO: come up with a representation for a 0-size value,
+            // and make this 0 size everywhere. as an argument, simply
+            // skip passing it.
+            return jl_pvalue_llvmt;
+        }
+        return julia_struct_to_llvm(jt);
     }
-    if (jt == (jl_value_t*)jl_bottom_type) return T_void;
     return jl_pvalue_llvmt;
 }
 
-static Type *julia_struct_to_llvm(jl_value_t *jt) {
-    if (jl_is_struct_type(jt) && !jl_is_array_type(jt)) {
+static Type *julia_struct_to_llvm(jl_value_t *jt)
+{
+    if (jl_is_structtype(jt) && !jl_is_array_type(jt)) {
         if (!jl_is_leaf_type(jt))
-           return NULL;
-        jl_struct_type_t *jst = (jl_struct_type_t*)jt;
+            return NULL;
+        jl_datatype_t *jst = (jl_datatype_t*)jt;
         if (jst->struct_decl == NULL) {
             size_t ntypes = jl_tuple_len(jst->types);
             if (ntypes == 0)
                 return NULL;
+            StructType *structdecl = StructType::create(getGlobalContext(), jst->name->name->name);
+            jst->struct_decl = structdecl;
+
             std::vector<Type *> latypes(0);
             size_t i;
             for(i = 0; i < ntypes; i++) {
                 jl_value_t *ty = jl_tupleref(jst->types, i);
-                Type *lty = julia_type_to_llvm(ty);
-                if (lty == jl_pvalue_llvmt)
-                    return NULL;
+                Type *lty = ty==(jl_value_t*)jl_bool_type ? T_int8 : julia_type_to_llvm(ty);
+                if (jst->fields[i].isptr)
+                    lty = jl_pvalue_llvmt;
                 latypes.push_back(lty);
             }
-            jst->struct_decl = (void*)StructType::create(latypes, jst->name->name->name);
+            structdecl->setBody(latypes);
         }
-        Type *t = (Type*)jst->struct_decl;
-        return t;
+        return (Type*)jst->struct_decl;
     }
     return julia_type_to_llvm(jt);
 }
@@ -199,16 +222,16 @@ static jl_value_t *julia_type_of(Value *v)
     return jl_typeid_to_type(id);
 }
 
-static Value *NoOpCast(Value *v)
+static Value *NoOpInst(Value *v)
 {
-    v = CastInst::Create(Instruction::BitCast, v, v->getType());
+    v = SelectInst::Create(ConstantInt::get(T_int1,1), v, v);
     builder.Insert((Instruction*)v);
     return v;
 }
 
 static Value *mark_julia_type(Value *v, jl_value_t *jt)
 {
-    if (jt == (jl_value_t*)jl_any_type)
+    if (jt == (jl_value_t*)jl_any_type || v->getType()==jl_pvalue_llvmt)
         return v;
     if (has_julia_type(v)) {
         if (julia_type_of(v) == jt)
@@ -218,7 +241,7 @@ static Value *mark_julia_type(Value *v, jl_value_t *jt)
         return v;
     }
     if (dyn_cast<Instruction>(v) == NULL)
-        v = NoOpCast(v);
+        v = NoOpInst(v);
     assert(dyn_cast<Instruction>(v));
     char name[3];
     int id = jl_type_to_typeid(jt);
@@ -232,11 +255,6 @@ static Value *mark_julia_type(Value *v, jl_value_t *jt)
     return v;
 }
 
-static Value *mark_julia_type(Value *v, jl_bits_type_t *jt)
-{
-    return mark_julia_type(v, (jl_value_t*)jt);
-}
-
 // --- generating various error checks ---
 
 static jl_value_t *llvm_type_to_julia(Type *t, bool err=true);
@@ -249,28 +267,43 @@ static Value *emit_typeof(Value *p)
         tt = builder.
             CreateLoad(builder.CreateGEP(tt,ConstantInt::get(T_size,0)),
                        false);
+#ifdef OVERLAP_TUPLE_LEN
+        tt = builder.
+            CreateIntToPtr(builder.
+                           CreateAnd(builder.CreatePtrToInt(tt, T_int64),
+                                     ConstantInt::get(T_int64,0x000ffffffffffffe)),
+                           jl_pvalue_llvmt);
+#endif
         return tt;
     }
-    return literal_pointer_val(llvm_type_to_julia(p->getType()));
+    return literal_pointer_val(julia_type_of(p));
 }
 
-static void emit_error(const std::string &txt, jl_codectx_t *ctx)
+static void just_emit_error(const std::string &txt, jl_codectx_t *ctx)
 {
-    std::string txt2 = "in " + ctx->funcName + ": " + txt;
     Value *zeros[2] = { ConstantInt::get(T_int32, 0),
                         ConstantInt::get(T_int32, 0) };
     builder.CreateCall(jlerror_func,
-                       builder.CreateGEP(stringConst(txt2),
+                       builder.CreateGEP(stringConst(txt),
                                          ArrayRef<Value*>(zeros)));
 }
 
+static Value *emit_error(const std::string &txt, jl_codectx_t *ctx)
+{
+    just_emit_error(txt, ctx);
+    Value *v = builder.CreateUnreachable();
+    BasicBlock *cont = BasicBlock::Create(getGlobalContext(),"after_error",ctx->f);
+    builder.SetInsertPoint(cont);
+    return v;
+}
+
 static void error_unless(Value *cond, const std::string &msg, jl_codectx_t *ctx)
 {
     BasicBlock *failBB = BasicBlock::Create(getGlobalContext(),"fail",ctx->f);
     BasicBlock *passBB = BasicBlock::Create(getGlobalContext(),"pass");
     builder.CreateCondBr(cond, passBB, failBB);
     builder.SetInsertPoint(failBB);
-    emit_error(msg, ctx);
+    just_emit_error(msg, ctx);
     builder.CreateUnreachable();
     ctx->f->getBasicBlockList().push_back(passBB);
     builder.SetInsertPoint(passBB);
@@ -352,8 +385,10 @@ static Value *emit_bounds_check(Value *i, Value *len, jl_codectx_t *ctx)
 {
     Value *im1 = builder.CreateSub(i, ConstantInt::get(T_size, 1));
 #if CHECK_BOUNDS==1
-    Value *ok = builder.CreateICmpULT(im1, len);
-    raise_exception_unless(ok, jlboundserr_var, ctx);
+    if (ctx->boundsCheck.empty() || ctx->boundsCheck.back()==true) {
+        Value *ok = builder.CreateICmpULT(im1, len);
+        raise_exception_unless(ok, jlboundserr_var, ctx);
+    }
 #endif
     return im1;
 }
@@ -368,7 +403,7 @@ static void emit_func_check(Value *x, jl_codectx_t *ctx)
                               literal_pointer_val((jl_value_t*)jl_function_type)),
                  builder.
                  CreateICmpEQ(xty,
-                              literal_pointer_val((jl_value_t*)jl_struct_kind)));
+                              literal_pointer_val((jl_value_t*)jl_datatype_type)));
     BasicBlock *elseBB1 = BasicBlock::Create(getGlobalContext(),"notf", ctx->f);
     BasicBlock *mergeBB1 = BasicBlock::Create(getGlobalContext(),"isf");
     builder.CreateCondBr(isfunc, mergeBB1, elseBB1);
@@ -433,7 +468,7 @@ static Value *typed_store(Value *ptr, Value *idx_0based, Value *rhs,
     Type *elty = julia_type_to_llvm(jltype);
     assert(elty != NULL);
     if (elty==T_int1) { elty = T_int8; }
-    if (jl_is_bits_type(jltype))
+    if (jl_isbits(jltype) && ((jl_datatype_t*)jltype)->size > 0)
         rhs = emit_unbox(elty, PointerType::get(elty,0), rhs);
     else
         rhs = boxed(rhs);
@@ -452,7 +487,8 @@ static Value *julia_bool(Value *cond)
 
 // --- get the inferred type of an AST node ---
 
-static jl_value_t *static_eval(jl_value_t *ex, jl_codectx_t *ctx, bool sparams=true);
+static jl_value_t *static_eval(jl_value_t *ex, jl_codectx_t *ctx, bool sparams=true,
+                               bool allow_alloc=true);
 
 static inline jl_module_t *topmod(jl_codectx_t *ctx)
 {
@@ -514,7 +550,7 @@ static jl_value_t *expr_type(jl_value_t *e, jl_codectx_t *ctx)
             return (jl_value_t*)jl_any_type;
     }
 type_of_constant:
-    if (jl_is_some_tag_type(e))
+    if (jl_is_datatype(e) || jl_is_uniontype(e) || jl_is_typector(e))
         return (jl_value_t*)jl_wrap_Type(e);
     return (jl_value_t*)jl_typeof(e);
 }
@@ -523,11 +559,13 @@ type_of_constant:
 
 static Value *emit_tuplelen(Value *t)
 {
-    Value *lenbits = emit_nthptr(t, 1);
-#ifdef __LP64__
-    return builder.CreatePtrToInt(lenbits, T_int64);
+#ifdef OVERLAP_TUPLE_LEN
+    Value *lenbits = emit_nthptr(t, (size_t)0);
+    return builder.CreateLShr(builder.CreatePtrToInt(lenbits, T_int64),
+                              ConstantInt::get(T_int32, 52));
 #else
-    return builder.CreatePtrToInt(lenbits, T_int32);
+    Value *lenbits = emit_nthptr(t, 1);
+    return builder.CreatePtrToInt(lenbits, T_size);
 #endif
 }
 
@@ -537,7 +575,7 @@ static Value *emit_n_varargs(jl_codectx_t *ctx)
     int nreq = ctx->nReqArgs;
     Value *valen = builder.CreateSub((Value*)ctx->argCount,
                                      ConstantInt::get(T_int32, nreq));
-#ifdef __LP64__
+#ifdef _P64
     return builder.CreateSExt(valen, T_int64);
 #else
     return valen;
@@ -546,10 +584,18 @@ static Value *emit_n_varargs(jl_codectx_t *ctx)
 
 static Value *emit_arraysize(Value *t, Value *dim)
 {
-#ifdef __LP64__
+#ifdef STORE_ARRAY_LEN
+#ifdef _P64
     int o = 3;
 #else
     int o = 4;
+#endif
+#else
+#ifdef _P64
+    int o = 2;
+#else
+    int o = 3;
+#endif
 #endif
     Value *dbits =
         emit_nthptr(t, builder.CreateAdd(dim,
@@ -557,15 +603,57 @@ static Value *emit_arraysize(Value *t, Value *dim)
     return builder.CreatePtrToInt(dbits, T_size);
 }
 
+static jl_arrayvar_t *arrayvar_for(jl_value_t *ex, jl_codectx_t *ctx)
+{
+    if (ex == NULL) return NULL;
+    jl_sym_t *aname=NULL;
+    if (jl_is_symbol(ex))
+        aname = ((jl_sym_t*)ex);
+    else if (jl_is_symbolnode(ex))
+        aname = jl_symbolnode_sym(ex);
+    if (aname && ctx->arrayvars->find(aname) != ctx->arrayvars->end()) {
+        return &(*ctx->arrayvars)[aname];
+    }
+    return NULL;
+}
+
 static Value *emit_arraysize(Value *t, int dim)
 {
     return emit_arraysize(t, ConstantInt::get(T_int32, dim));
 }
 
-static Value *emit_arraylen(Value *t)
+static Value *emit_arraylen_prim(Value *t, jl_value_t *ty)
 {
+#ifdef STORE_ARRAY_LEN
+    (void)ty;
     Value *lenbits = emit_nthptr(t, 2);
     return builder.CreatePtrToInt(lenbits, T_size);
+#else
+    jl_value_t *p1 = jl_tparam1(ty);
+    if (jl_is_long(p1)) {
+        size_t nd = jl_unbox_long(p1);
+        Value *l = ConstantInt::get(T_size, 1);
+        for(size_t i=0; i < nd; i++) {
+            l = builder.CreateMul(l, emit_arraysize(t, (int)(i+1)));
+        }
+        return l;
+    }
+    else {
+        std::vector<Type *> fargt(0);
+        fargt.push_back(jl_pvalue_llvmt);
+        FunctionType *ft = FunctionType::get(T_size, fargt, false);
+        Value *alen = jl_Module->getOrInsertFunction("jl_array_len_", ft);
+        return builder.CreateCall(alen, t);
+    }
+#endif
+}
+
+static Value *emit_arraylen(Value *t, jl_value_t *ex, jl_codectx_t *ctx)
+{
+    jl_arrayvar_t *av = arrayvar_for(ex, ctx);
+    if (av!=NULL)
+        return builder.CreateLoad(av->len);
+    return emit_arraylen_prim(t, expr_type(ex,ctx));
 }
 
 static Value *emit_arrayptr(Value *t)
@@ -573,20 +661,49 @@ static Value *emit_arrayptr(Value *t)
     return emit_nthptr(t, 1);
 }
 
-static Value *bitstype_pointer(Value *x)
+static Value *emit_arrayptr(Value *t, jl_value_t *ex, jl_codectx_t *ctx)
+{
+    jl_arrayvar_t *av = arrayvar_for(ex, ctx);
+    if (av!=NULL)
+        return builder.CreateLoad(av->dataptr);
+    return emit_arrayptr(t);
+}
+
+static Value *emit_arraysize(Value *t, jl_value_t *ex, int dim, jl_codectx_t *ctx)
+{
+    jl_arrayvar_t *av = arrayvar_for(ex, ctx);
+    if (av != NULL && dim <= (int)av->sizes.size())
+        return builder.CreateLoad(av->sizes[dim-1]);
+    return emit_arraysize(t, dim);
+}
+
+static void assign_arrayvar(jl_arrayvar_t &av, Value *ar)
+{
+    builder.CreateStore(builder.CreateBitCast(emit_arrayptr(ar),T_pint8),
+                        av.dataptr);
+    builder.CreateStore(emit_arraylen_prim(ar, av.ty), av.len);
+    for(size_t i=0; i < av.sizes.size(); i++)
+        builder.CreateStore(emit_arraysize(ar,i+1), av.sizes[i]);
+}
+
+static Value *data_pointer(Value *x)
 {
     return builder.CreateGEP(builder.CreateBitCast(x, jl_ppvalue_llvmt),
                              ConstantInt::get(T_size, 1));
 }
 
-static Value *emit_array_nd_index(Value *a, size_t nd, jl_value_t **args,
+static Value *emit_array_nd_index(Value *a, jl_value_t *ex, size_t nd, jl_value_t **args,
                                   size_t nidxs, jl_codectx_t *ctx)
 {
     Value *i = ConstantInt::get(T_size, 0);
     Value *stride = ConstantInt::get(T_size, 1);
+    bool bc = ctx->boundsCheck.empty() || ctx->boundsCheck.back()==true;
 #if CHECK_BOUNDS==1
-    BasicBlock *failBB = BasicBlock::Create(getGlobalContext(), "oob");
-    BasicBlock *endBB = BasicBlock::Create(getGlobalContext(), "idxend");
+    BasicBlock *failBB=NULL, *endBB=NULL;
+    if (bc) {
+        failBB = BasicBlock::Create(getGlobalContext(), "oob");
+        endBB = BasicBlock::Create(getGlobalContext(), "idxend");
+    }
 #endif
     for(size_t k=0; k < nidxs; k++) {
         Value *ii = emit_unbox(T_size, T_psize, emit_unboxed(args[k], ctx));
@@ -594,30 +711,34 @@ static Value *emit_array_nd_index(Value *a, size_t nd, jl_value_t **args,
         i = builder.CreateAdd(i, builder.CreateMul(ii, stride));
         if (k < nidxs-1) {
             Value *d =
-                k >= nd ? ConstantInt::get(T_size, 1) : emit_arraysize(a, k+1);
+                k >= nd ? ConstantInt::get(T_size, 1) : emit_arraysize(a, ex, k+1, ctx);
 #if CHECK_BOUNDS==1
-            BasicBlock *okBB = BasicBlock::Create(getGlobalContext(), "ib");
-            // if !(i < d) goto error
-            builder.CreateCondBr(builder.CreateICmpULT(ii, d), okBB, failBB);
-            ctx->f->getBasicBlockList().push_back(okBB);
-            builder.SetInsertPoint(okBB);
+            if (bc) {
+                BasicBlock *okBB = BasicBlock::Create(getGlobalContext(), "ib");
+                // if !(i < d) goto error
+                builder.CreateCondBr(builder.CreateICmpULT(ii, d), okBB, failBB);
+                ctx->f->getBasicBlockList().push_back(okBB);
+                builder.SetInsertPoint(okBB);
+            }
 #endif
             stride = builder.CreateMul(stride, d);
         }
     }
 #if CHECK_BOUNDS==1
-    Value *alen = emit_arraylen(a);
-    // if !(i < alen) goto error
-    builder.CreateCondBr(builder.CreateICmpULT(i, alen), endBB, failBB);
+    if (bc) {
+        Value *alen = emit_arraylen(a, ex, ctx);
+        // if !(i < alen) goto error
+        builder.CreateCondBr(builder.CreateICmpULT(i, alen), endBB, failBB);
 
-    ctx->f->getBasicBlockList().push_back(failBB);
-    builder.SetInsertPoint(failBB);
-    builder.CreateCall2(jlthrow_line_func, builder.CreateLoad(jlboundserr_var),
-                        ConstantInt::get(T_int32, ctx->lineno));
-    builder.CreateUnreachable();
+        ctx->f->getBasicBlockList().push_back(failBB);
+        builder.SetInsertPoint(failBB);
+        builder.CreateCall2(jlthrow_line_func, builder.CreateLoad(jlboundserr_var),
+                            ConstantInt::get(T_int32, ctx->lineno));
+        builder.CreateUnreachable();
 
-    ctx->f->getBasicBlockList().push_back(endBB);
-    builder.SetInsertPoint(endBB);
+        ctx->f->getBasicBlockList().push_back(endBB);
+        builder.SetInsertPoint(endBB);
+    }
 #endif
 
     return i;
@@ -637,7 +758,7 @@ static Value *tpropagate(Value *a, Value *b)
 static Value *init_bits_value(Value *newv, Value *jt, Type *t, Value *v)
 {
     builder.CreateStore(jt, builder.CreateBitCast(newv, jl_ppvalue_llvmt));
-    builder.CreateStore(v , builder.CreateBitCast(bitstype_pointer(newv),
+    builder.CreateStore(v , builder.CreateBitCast(data_pointer(newv),
                                                   PointerType::get(t,0)));
     return newv;
 }
@@ -648,7 +769,7 @@ static Value *allocate_box_dynamic(Value *jlty, int nb, Value *v)
     if (v->getType()->isPointerTy()) {
         v = builder.CreatePtrToInt(v, T_size);
     }
-    size_t sz = sizeof(void*) + (nb+7)/8;
+    size_t sz = sizeof(void*) + nb;
     Value *newv = builder.CreateCall(jlallocobj_func,
                                      ConstantInt::get(T_size, sz));
     // TODO: make sure this is rooted. I think it is.
@@ -668,7 +789,8 @@ static Value *boxed(Value *v, jl_value_t *jt)
     if (t == T_int1) return julia_bool(v);
     if (jt == NULL)
         jt = julia_type_of(v);
-    jl_bits_type_t *jb = (jl_bits_type_t*)jt;
+    jl_datatype_t *jb = (jl_datatype_t*)jt;
+    assert(jl_is_datatype(jb));
     if (jb == jl_int8_type)
         return builder.CreateCall(box_int8_func,
                                   builder.CreateSExt(v, T_int32));
@@ -679,7 +801,7 @@ static Value *boxed(Value *v, jl_value_t *jt)
     //if (jb == jl_float64_type) return builder.CreateCall(box_float64_func, v);
     if (jb == jl_float64_type) {
         // manually inline alloc & init of Float64 box. cheap, I know.
-#ifdef __LP64__
+#ifdef _P64
         Value *newv = builder.CreateCall(jlalloc2w_func);
 #else
         Value *newv = builder.CreateCall(jlalloc3w_func);
@@ -694,24 +816,28 @@ static Value *boxed(Value *v, jl_value_t *jt)
     if (jb == jl_uint64_type) return builder.CreateCall(box_uint64_func, v);
     if (jb == jl_char_type)   return builder.CreateCall(box_char_func, v);
     // TODO: skip the call for constant arguments
-    if (!jl_is_bits_type(jt)) {
+    if (!jl_isbits(jt)) {
         assert("Don't know how to box this type" && false);
         return NULL;
     }
-    int nb = jl_bitstype_nbits(jt);
-    return allocate_box_dynamic(literal_pointer_val(jt), nb, v);
+    if (!jb->abstract && jb->size == 0) {
+        if (jb->instance == NULL)
+            jl_new_struct_uninit(jb);
+        assert(jb->instance != NULL);
+        return literal_pointer_val(jb->instance);
+    }
+    return allocate_box_dynamic(literal_pointer_val(jt),jl_datatype_size(jt),v);
 }
 
-
 static void emit_cpointercheck(Value *x, const std::string &msg,
-                           jl_codectx_t *ctx)
+                               jl_codectx_t *ctx)
 {
     Value *t = emit_typeof(x);
-    emit_typecheck(t, (jl_value_t*)jl_bits_kind, msg, ctx);
+    emit_typecheck(t, (jl_value_t*)jl_datatype_type, msg, ctx);
 
     Value *istype =
-        builder.CreateICmpEQ(emit_nthptr(t, offsetof(jl_bits_type_t,name)/sizeof(char*)),
-                literal_pointer_val((jl_value_t*)jl_pointer_type->name));
+        builder.CreateICmpEQ(emit_nthptr(t, offsetof(jl_datatype_t,name)/sizeof(char*)),
+                             literal_pointer_val((jl_value_t*)jl_pointer_type->name));
     BasicBlock *failBB = BasicBlock::Create(getGlobalContext(),"fail",ctx->f);
     BasicBlock *passBB = BasicBlock::Create(getGlobalContext(),"pass");
     builder.CreateCondBr(istype, passBB, failBB);
