@@ -4,7 +4,7 @@
 #include <stdlib.h>
 #include <string.h>
 #include <assert.h>
-#ifdef __WIN32__
+#ifdef _OS_WINDOWS_
 #include <malloc.h>
 #endif
 #include "julia.h"
@@ -24,6 +24,8 @@ static const ptrint_t LongExpr_tag   = 25;
 static const ptrint_t LiteralVal_tag = 26;
 static const ptrint_t SmallInt64_tag = 27;
 static const ptrint_t IdTable_tag    = 28;
+static const ptrint_t Int32_tag      = 29;
+static const ptrint_t Array1d_tag    = 30;
 static const ptrint_t Null_tag         = 253;
 static const ptrint_t ShortBackRef_tag = 254;
 static const ptrint_t BackRef_tag      = 255;
@@ -36,7 +38,7 @@ static jl_array_t *tree_literal_values=NULL;
 static jl_value_t *jl_idtable_type=NULL;
 
 // queue of types to cache
-static jl_array_t *tagtype_list=NULL;
+static jl_array_t *datatype_list=NULL;
 
 #define write_uint8(s, n) ios_putc((n), (s))
 #define read_uint8(s) ((uint8_t)ios_getc(s))
@@ -99,51 +101,38 @@ static void jl_serialize_fptr(ios_t *s, void *fptr)
     write_uint16(s, *(ptrint_t*)pbp);
 }
 
-static void jl_serialize_tag_type(ios_t *s, jl_value_t *v)
+static void jl_serialize_datatype(ios_t *s, jl_datatype_t *dt)
 {
-    if (jl_is_struct_type(v)) {
-        writetag(s, (jl_value_t*)jl_struct_kind);
-        jl_serialize_value(s, jl_struct_kind);
-        write_uint16(s, jl_tuple_len(((jl_struct_type_t*)v)->names));
-        write_int32(s, ((jl_struct_type_t*)v)->uid);
-        write_int32(s, ((jl_struct_type_t*)v)->size);
-        write_int32(s, ((jl_struct_type_t*)v)->alignment);
-        size_t nf = jl_tuple_len(((jl_struct_type_t*)v)->names);
-        ios_write(s, (char*)&((jl_struct_type_t*)v)->fields[0], nf*sizeof(jl_fielddesc_t));
-        jl_serialize_value(s, ((jl_struct_type_t*)v)->name);
-        jl_serialize_value(s, ((jl_struct_type_t*)v)->parameters);
-        jl_serialize_value(s, ((jl_struct_type_t*)v)->super);
-        jl_serialize_value(s, ((jl_struct_type_t*)v)->names);
-        jl_serialize_value(s, ((jl_struct_type_t*)v)->types);
-        jl_serialize_value(s, ((jl_struct_type_t*)v)->ctor_factory);
-        jl_serialize_value(s, ((jl_struct_type_t*)v)->env);
-        jl_serialize_value(s, ((jl_struct_type_t*)v)->linfo);
-        jl_serialize_fptr(s, ((jl_struct_type_t*)v)->fptr);
-    }
-    else if (jl_is_bits_type(v)) {
-        writetag(s, jl_struct_kind);
-        jl_serialize_value(s, jl_bits_kind);
-        if (v == (jl_value_t*)jl_int32_type)
-            write_uint8(s, 2);
-        else if (v == (jl_value_t*)jl_bool_type)
-            write_uint8(s, 3);
-        else if (v == (jl_value_t*)jl_int64_type)
-            write_uint8(s, 4);
-        else
-            write_uint8(s, 0);
-        jl_serialize_value(s, ((jl_tag_type_t*)v)->name);
-        jl_serialize_value(s, ((jl_bits_type_t*)v)->parameters);
-        write_int32(s, ((jl_bits_type_t*)v)->nbits);
-        jl_serialize_value(s, ((jl_bits_type_t*)v)->super);
-        write_int32(s, ((jl_bits_type_t*)v)->uid);
-    }
-    else {
-        assert(jl_is_tag_type(v));
-        writetag(s, jl_tag_kind);
-        jl_serialize_value(s, ((jl_tag_type_t*)v)->name);
-        jl_serialize_value(s, ((jl_tag_type_t*)v)->parameters);
-        jl_serialize_value(s, ((jl_tag_type_t*)v)->super);
+    writetag(s, (jl_value_t*)jl_datatype_type);
+    jl_serialize_value(s, (jl_value_t*)jl_datatype_type);
+    int tag = 0;
+    if (dt == jl_int32_type)
+        tag = 2;
+    else if (dt == jl_bool_type)
+        tag = 3;
+    else if (dt == jl_int64_type)
+        tag = 4;
+    write_uint8(s, tag);
+    size_t nf = jl_tuple_len(dt->names);
+    write_uint16(s, nf);
+    write_int32(s, dt->size);
+    if (nf > 0) {
+        write_int32(s, dt->alignment);
+        ios_write(s, (char*)&dt->fields[0], nf*sizeof(jl_fielddesc_t));
+        jl_serialize_value(s, dt->names);
+        jl_serialize_value(s, dt->types);
     }
+    write_uint8(s, dt->abstract | (dt->mutabl<<1) | (dt->pointerfree<<2));
+    if (!dt->abstract)
+        write_int32(s, dt->uid);
+
+    jl_serialize_value(s, dt->parameters);
+    jl_serialize_value(s, dt->name);
+    jl_serialize_value(s, dt->super);
+    jl_serialize_value(s, dt->ctor_factory);
+    jl_serialize_value(s, dt->env);
+    jl_serialize_value(s, dt->linfo);
+    jl_serialize_fptr(s, dt->fptr);
 }
 
 static void jl_serialize_module(ios_t *s, jl_module_t *m)
@@ -179,6 +168,7 @@ static void jl_serialize_module(ios_t *s, jl_module_t *m)
             jl_serialize_value(s, (jl_value_t*)m->usings.items[i]);
         }
     }
+    jl_serialize_value(s, m->constant_table);
 }
 
 static int is_ast_node(jl_value_t *v)
@@ -189,9 +179,9 @@ static int is_ast_node(jl_value_t *v)
             li->ast = jl_compress_ast(li, li->ast);
         return 0;
     }
-    return jl_is_symbol(v) || jl_is_expr(v) ||
+    return jl_is_symbol(v) || jl_is_expr(v) || jl_is_newvarnode(v) ||
         jl_typeis(v, jl_array_any_type) || jl_is_tuple(v) ||
-        jl_is_union_type(v) || jl_is_int32(v) || jl_is_int64(v) ||
+        jl_is_uniontype(v) || jl_is_int32(v) || jl_is_int64(v) ||
         jl_is_symbolnode(v) || jl_is_bool(v) || jl_is_typevar(v) ||
         jl_is_topnode(v) || jl_is_quotenode(v) || jl_is_gotonode(v) ||
         jl_is_labelnode(v) || jl_is_linenode(v) || jl_is_getfieldnode(v);
@@ -273,12 +263,19 @@ static void jl_serialize_value_(ios_t *s, jl_value_t *v)
     }
     else if (jl_is_array(v)) {
         jl_array_t *ar = (jl_array_t*)v;
-        writetag(s, (jl_value_t*)jl_array_type);
+        if (ar->ndims == 1 && ar->elsize < 128) {
+            writetag(s, (jl_value_t*)Array1d_tag);
+            write_uint8(s, (ar->ptrarray<<7) | (ar->elsize & 0x7f));
+        }
+        else {
+            writetag(s, (jl_value_t*)jl_array_type);
+            write_uint16(s, ar->ndims);
+            write_uint16(s, (ar->ptrarray<<15) | (ar->elsize & 0x7fff));
+        }
         jl_serialize_value(s, jl_typeof(ar));
-        jl_value_t *elty = jl_tparam0(jl_typeof(ar));
         for (i=0; i < ar->ndims; i++)
             jl_serialize_value(s, jl_box_long(jl_array_dim(ar,i)));
-        if (jl_is_bits_type(elty)) {
+        if (!ar->ptrarray) {
             size_t tot = jl_array_len(ar) * ar->elsize;
             ios_write(s, jl_array_data(ar), tot);
         }
@@ -305,8 +302,8 @@ static void jl_serialize_value_(ios_t *s, jl_value_t *v)
             jl_serialize_value(s, jl_exprarg(e, i));
         }
     }
-    else if (jl_is_some_tag_type(v)) {
-        jl_serialize_tag_type(s, v);
+    else if (jl_is_datatype(v)) {
+        jl_serialize_datatype(s, (jl_datatype_t*)v);
     }
     else if (jl_is_typevar(v)) {
         writetag(s, jl_tvar_type);
@@ -356,32 +353,41 @@ static void jl_serialize_value_(ios_t *s, jl_value_t *v)
         jl_error("Task cannot be serialized");
     }
     else {
-        jl_value_t *t = (jl_value_t*)jl_typeof(v);
-        if (jl_is_bits_type(t)) {
-            void *data = jl_bits_data(v);
-            if (t == (jl_value_t*)jl_int64_type &&
-                *(int64_t*)data >= S32_MIN && *(int64_t*)data <= S32_MAX) {
-                writetag(s, (jl_value_t*)SmallInt64_tag);
-                write_int32(s, (int32_t)*(int64_t*)data);
+        jl_datatype_t *t = (jl_datatype_t*)jl_typeof(v);
+        void *data = jl_data_ptr(v);
+        if (t == jl_int64_type &&
+            *(int64_t*)data >= S32_MIN && *(int64_t*)data <= S32_MAX) {
+            writetag(s, (jl_value_t*)SmallInt64_tag);
+            write_int32(s, (int32_t)*(int64_t*)data);
+            return;
+        }
+        if (t == jl_int32_type) {
+            writetag(s, (jl_value_t*)Int32_tag);
+            write_int32(s, (int32_t)*(int32_t*)data);
+            return;
+        }
+        if ((jl_value_t*)t == jl_idtable_type)
+            writetag(s, (jl_value_t*)IdTable_tag);
+        else
+            writetag(s, (jl_value_t*)jl_datatype_type);
+        jl_serialize_value(s, t);
+        size_t nf = jl_tuple_len(t->names);
+        if (nf == 0 && jl_datatype_size(t)>0) {
+            if (t->name == jl_pointer_type->name) {
+                write_int32(s, 0);
+#ifdef _P64
+                write_int32(s, 0);
+#endif
             }
             else {
-                int nb = ((jl_bits_type_t*)t)->nbits;
-                writetag(s, jl_bits_kind);
-                jl_serialize_value(s, t);
-                ios_write(s, data, nb/8);
+                ios_write(s, data, jl_datatype_size(t));
             }
         }
-        else if (jl_is_struct_type(t)) {
-            if (t == jl_idtable_type)
-                writetag(s, (jl_value_t*)IdTable_tag);
-            else
-                writetag(s, jl_struct_kind);
-            jl_serialize_value(s, t);
-            if (t == jl_idtable_type) {
+        else {
+            if ((jl_value_t*)t == jl_idtable_type) {
                 jl_array_t *data = (jl_array_t*)jl_get_nth_field(v, 0);
                 jl_value_t **d = (jl_value_t**)data->data;
-                size_t i;
-                for(i=0; i < jl_array_len(data); i+=2) {
+                for(size_t i=0; i < jl_array_len(data); i+=2) {
                     if (d[i+1] != NULL) {
                         jl_serialize_value(s, d[i+1]);
                         jl_serialize_value(s, d[i]);
@@ -390,17 +396,11 @@ static void jl_serialize_value_(ios_t *s, jl_value_t *v)
                 jl_serialize_value(s, NULL);
             }
             else {
-                jl_struct_type_t *st = (jl_struct_type_t*)t;
-                size_t nf = jl_tuple_len(st->names);
-                size_t i;
-                for(i=0; i < nf; i++) {
+                for(size_t i=0; i < nf; i++) {
                     jl_serialize_value(s, jl_get_nth_field(v, i));
                 }
             }
         }
-        else {
-            assert(0);
-        }
     }
 }
 
@@ -419,82 +419,63 @@ static jl_fptr_t jl_deserialize_fptr(ios_t *s)
     return *(jl_fptr_t*)pbp;
 }
 
-static jl_value_t *jl_deserialize_tag_type(ios_t *s, jl_struct_type_t *kind, int pos)
+static jl_value_t *jl_deserialize_datatype(ios_t *s, int pos)
 {
-    if (kind == jl_struct_kind) {
-        uint16_t nf = read_uint16(s);
-        jl_struct_type_t *st = jl_new_uninitialized_struct_type(nf);
-        st->uid = read_int32(s);
-        st->size = read_int32(s);
-        st->alignment = read_int32(s);
-        ios_read(s, (char*)&st->fields[0], nf*sizeof(jl_fielddesc_t));
-        st->instance = NULL;
-        ptrhash_put(&backref_table, (void*)(ptrint_t)pos, st);
-        st->name = (jl_typename_t*)jl_deserialize_value(s);
-        st->parameters = (jl_tuple_t*)jl_deserialize_value(s);
-        st->super = (jl_tag_type_t*)jl_deserialize_value(s);
-        st->names = (jl_tuple_t*)jl_deserialize_value(s);
-        st->types = (jl_tuple_t*)jl_deserialize_value(s);
-        st->ctor_factory = jl_deserialize_value(s);
-        st->env = jl_deserialize_value(s);
-        st->linfo = (jl_lambda_info_t*)jl_deserialize_value(s);
-        st->fptr = jl_deserialize_fptr(s);
-        st->struct_decl = NULL;
-        if (st->name == jl_array_type->name) {
-            // builtin types are not serialized, so their caches aren't
-            // explicitly saved. so we reconstruct the caches of builtin
-            // parametric types here.
-            jl_cell_1d_push(tagtype_list, (jl_value_t*)st);
-        }
-        return (jl_value_t*)st;
-    }
-    else if (kind == jl_bits_kind) {
-        int form = read_uint8(s);
-        jl_bits_type_t *bt;
-        if (form == 2)
-            bt = jl_int32_type;
-        else if (form == 3)
-            bt = jl_bool_type;
-        else if (form == 4)
-            bt = jl_int64_type;
-        else
-            bt = (jl_bits_type_t*)newobj((jl_type_t*)jl_bits_kind,
-                                         BITS_TYPE_NW);
-        ptrhash_put(&backref_table, (void*)(ptrint_t)pos, bt);
-        bt->name = (jl_typename_t*)jl_deserialize_value(s);
-        bt->parameters = (jl_tuple_t*)jl_deserialize_value(s);
+    int tag = read_uint8(s);
+    uint16_t nf = read_uint16(s);
+    size_t size = read_int32(s);
+    jl_datatype_t *dt;
+    if (tag == 2)
+        dt = jl_int32_type;
+    else if (tag == 3)
+        dt = jl_bool_type;
+    else if (tag == 4)
+        dt = jl_int64_type;
+    else
+        dt = jl_new_uninitialized_datatype(nf);
+    dt->size = size;
+    dt->struct_decl = NULL;
+    dt->instance = NULL;
 
-        size_t nbits = read_int32(s);
-        bt->nbits = nbits;
-        bt->fptr = NULL;
-        bt->env = NULL;
-        bt->linfo = NULL;
-        bt->super = (jl_tag_type_t*)jl_deserialize_value(s);
-        bt->uid = read_int32(s);
-        if (bt->name == jl_pointer_type->name) {
-            jl_cell_1d_push(tagtype_list, (jl_value_t*)bt);
-        }
-        return (jl_value_t*)bt;
+    assert(tree_literal_values==NULL);
+    ptrhash_put(&backref_table, (void*)(ptrint_t)pos, dt);
+
+    if (nf > 0) {
+        dt->alignment = read_int32(s);
+        ios_read(s, (char*)&dt->fields[0], nf*sizeof(jl_fielddesc_t));
+        dt->names = (jl_tuple_t*)jl_deserialize_value(s);
+        dt->types = (jl_tuple_t*)jl_deserialize_value(s);
     }
     else {
-        assert(kind == jl_tag_kind);
-        jl_tag_type_t *tt =
-            (jl_tag_type_t*)newobj((jl_type_t*)jl_tag_kind, TAG_TYPE_NW);
-        ptrhash_put(&backref_table, (void*)(ptrint_t)pos, tt);
-        tt->name = (jl_typename_t*)jl_deserialize_value(s);
-        tt->parameters = (jl_tuple_t*)jl_deserialize_value(s);
-        tt->super = (jl_tag_type_t*)jl_deserialize_value(s);
-        tt->fptr = NULL;
-        tt->env = NULL;
-        tt->linfo = NULL;
-        if (tt->name == jl_type_type->name || tt->name == jl_vararg_type->name ||
-            tt->name == jl_abstractarray_type->name) {
-            jl_cell_1d_push(tagtype_list, (jl_value_t*)tt);
-        }
-        return (jl_value_t*)tt;
+        dt->alignment = dt->size;
+        if (dt->alignment > MAX_ALIGN)
+            dt->alignment = MAX_ALIGN;
+        dt->names = dt->types = jl_null;
     }
-    assert(0);
-    return NULL;
+    uint8_t flags = read_uint8(s);
+    dt->abstract = flags&1;
+    dt->mutabl = (flags>>1)&1;
+    dt->pointerfree = (flags>>2)&1;
+    if (!dt->abstract)
+        dt->uid = read_int32(s);
+    else
+        dt->uid = 0;
+    dt->parameters = (jl_tuple_t*)jl_deserialize_value(s);
+    dt->name = (jl_typename_t*)jl_deserialize_value(s);
+    dt->super = (jl_datatype_t*)jl_deserialize_value(s);
+    dt->ctor_factory = jl_deserialize_value(s);
+    dt->env = jl_deserialize_value(s);
+    dt->linfo = (jl_lambda_info_t*)jl_deserialize_value(s);
+    dt->fptr = jl_deserialize_fptr(s);
+    if (dt->name == jl_array_type->name || dt->name == jl_pointer_type->name ||
+        dt->name == jl_type_type->name || dt->name == jl_vararg_type->name ||
+        dt->name == jl_abstractarray_type->name) {
+        // builtin types are not serialized, so their caches aren't
+        // explicitly saved. so we reconstruct the caches of builtin
+        // parametric types here.
+        jl_cell_1d_push(datatype_list, (jl_value_t*)dt);
+    }
+    return (jl_value_t*)dt;
 }
 
 jl_array_t *jl_eqtable_put(jl_array_t *h, void *key, void *val);
@@ -556,17 +537,30 @@ static jl_value_t *jl_deserialize_value(ios_t *s)
             ptrhash_put(&backref_table, (void*)(ptrint_t)pos, s);
         return s;
     }
-    else if (vtag == (jl_value_t*)jl_array_type) {
+    else if (vtag == (jl_value_t*)jl_array_type ||
+             vtag == (jl_value_t*)Array1d_tag) {
+        int16_t ndims;
+        int isunboxed, elsize;
+        if (vtag == (jl_value_t*)Array1d_tag) {
+            ndims = 1;
+            elsize = read_uint8(s);
+            isunboxed = !(elsize>>7);
+            elsize = elsize&0x7f;
+        }
+        else {
+            ndims = read_uint16(s);
+            elsize = read_uint16(s);
+            isunboxed = !(elsize>>15);
+            elsize = elsize&0x7fff;
+        }
         jl_value_t *aty = jl_deserialize_value(s);
-        jl_value_t *elty = jl_tparam0(aty);
-        int16_t ndims = jl_unbox_long(jl_tparam1(aty));
         size_t *dims = alloca(ndims*sizeof(size_t));
         for(i=0; i < ndims; i++)
             dims[i] = jl_unbox_long(jl_deserialize_value(s));
-        jl_array_t *a = jl_new_array_((jl_type_t*)aty, ndims, dims);
+        jl_array_t *a = jl_new_array_for_deserialization((jl_value_t*)aty, ndims, dims, isunboxed, elsize);
         if (usetable)
             ptrhash_put(&backref_table, (void*)(ptrint_t)pos, (jl_value_t*)a);
-        if (jl_is_bits_type(elty)) {
+        if (!a->ptrarray) {
             size_t tot = jl_array_len(a) * a->elsize;
             ios_read(s, jl_array_data(a), tot);
         }
@@ -597,7 +591,7 @@ static jl_value_t *jl_deserialize_value(ios_t *s)
         return jl_cellref(tree_literal_values, read_uint16(s));
     }
     else if (vtag == (jl_value_t*)jl_tvar_type) {
-        jl_tvar_t *tv = (jl_tvar_t*)newobj((jl_type_t*)jl_tvar_type, 4);
+        jl_tvar_t *tv = (jl_tvar_t*)newobj((jl_value_t*)jl_tvar_type, 4);
         if (usetable)
             ptrhash_put(&backref_table, (void*)(ptrint_t)pos, tv);
         tv->name = (jl_sym_t*)jl_deserialize_value(s);
@@ -608,7 +602,7 @@ static jl_value_t *jl_deserialize_value(ios_t *s)
     }
     else if (vtag == (jl_value_t*)jl_function_type) {
         jl_function_t *f =
-            (jl_function_t*)newobj((jl_type_t*)jl_function_type, 3);
+            (jl_function_t*)newobj((jl_value_t*)jl_function_type, 3);
         if (usetable)
             ptrhash_put(&backref_table, (void*)(ptrint_t)pos, f);
         f->linfo = (jl_lambda_info_t*)jl_deserialize_value(s);
@@ -618,7 +612,7 @@ static jl_value_t *jl_deserialize_value(ios_t *s)
     }
     else if (vtag == (jl_value_t*)jl_lambda_info_type) {
         jl_lambda_info_t *li =
-            (jl_lambda_info_t*)newobj((jl_type_t*)jl_lambda_info_type,
+            (jl_lambda_info_t*)newobj((jl_value_t*)jl_lambda_info_type,
                                       LAMBDA_INFO_NW);
         if (usetable)
             ptrhash_put(&backref_table, (void*)(ptrint_t)pos, li);
@@ -656,7 +650,7 @@ static jl_value_t *jl_deserialize_value(ios_t *s)
                 break;
             jl_binding_t *b = jl_get_binding_wr(m, (jl_sym_t*)name);
             b->value = jl_deserialize_value(s);
-            b->type = (jl_type_t*)jl_deserialize_value(s);
+            b->type = (jl_value_t*)jl_deserialize_value(s);
             b->owner = (jl_module_t*)jl_deserialize_value(s);
             int8_t flags = read_int8(s);
             b->constp = (flags>>2) & 1;
@@ -667,6 +661,7 @@ static jl_value_t *jl_deserialize_value(ios_t *s)
         for(size_t i=0; i < ni; i++) {
             arraylist_push(&m->usings, jl_deserialize_value(s));
         }
+        m->constant_table = (jl_array_t*)jl_deserialize_value(s);
         return (jl_value_t*)m;
     }
     else if (vtag == (jl_value_t*)SmallInt64_tag) {
@@ -675,64 +670,68 @@ static jl_value_t *jl_deserialize_value(ios_t *s)
             ptrhash_put(&backref_table, (void*)(ptrint_t)pos, v);
         return v;
     }
-    else if (vtag == (jl_value_t*)jl_bits_kind) {
-        jl_bits_type_t *bt = (jl_bits_type_t*)jl_deserialize_value(s);
-        int nby = bt->nbits/8;
-        char *data = alloca(nby);
-        ios_read(s, data, nby);
-        jl_value_t *v=NULL;
-        if (bt == jl_int32_type)
-            v = jl_box_int32(*(int32_t*)data);
-        else if (bt == jl_int64_type)
-            v = jl_box_int64(*(int64_t*)data);
-        else if (bt == jl_bool_type)
-            v = jl_box_bool(*(int8_t*)data);
-        else {
-            switch (bt->nbits) {
-            case  8: v = jl_box8 (bt, *(int8_t*) data); break;
-            case 16: v = jl_box16(bt, *(int16_t*)data); break;
-            case 32: v = jl_box32(bt, *(int32_t*)data); break;
-            case 64: v = jl_box64(bt, *(int64_t*)data); break;
-            default:
-                v = (jl_value_t*)allocobj(sizeof(void*)+nby);
-                v->type = (jl_type_t*)bt;
-                memcpy(jl_bits_data(v), data, nby);
-            }
-        }
+    else if (vtag == (jl_value_t*)Int32_tag) {
+        jl_value_t *v = jl_box_int32(read_int32(s));
         if (usetable)
             ptrhash_put(&backref_table, (void*)(ptrint_t)pos, v);
         return v;
     }
-    else if (vtag == (jl_value_t*)jl_struct_kind || vtag == (jl_value_t*)IdTable_tag) {
-        jl_struct_type_t *typ = (jl_struct_type_t*)jl_deserialize_value(s);
-        if (typ == jl_struct_kind || typ == jl_bits_kind)
-            return jl_deserialize_tag_type(s, typ, pos);
-        size_t nf = jl_tuple_len(typ->names);
-        jl_value_t *v = jl_new_struct_uninit(typ);
-        if (usetable)
-            ptrhash_put(&backref_table, (void*)(ptrint_t)pos, v);
-        if (vtag == (jl_value_t*)IdTable_tag) {
-            jl_array_t *a = jl_alloc_cell_1d(32);
-            while (1) {
-                jl_value_t *val = jl_deserialize_value(s);
-                if (val == NULL)
-                    break;
-                jl_value_t *key = jl_deserialize_value(s);
-                a = jl_eqtable_put(a, key, val);
+    else if (vtag == (jl_value_t*)jl_datatype_type || vtag == (jl_value_t*)IdTable_tag) {
+        jl_datatype_t *dt = (jl_datatype_t*)jl_deserialize_value(s);
+        if (dt == jl_datatype_type)
+            return jl_deserialize_datatype(s, pos);
+        size_t nf = jl_tuple_len(dt->names);
+        jl_value_t *v;
+        if (nf == 0 && jl_datatype_size(dt)>0) {
+            int nby = jl_datatype_size(dt);
+            char *data = alloca(nby);
+            ios_read(s, data, nby);
+            v = NULL;
+            if (dt == jl_int32_type)
+                v = jl_box_int32(*(int32_t*)data);
+            else if (dt == jl_int64_type)
+                v = jl_box_int64(*(int64_t*)data);
+            else if (dt == jl_bool_type)
+                v = jl_box_bool(*(int8_t*)data);
+            else {
+                switch (nby) {
+                case 1: v = jl_box8 (dt, *(int8_t*) data); break;
+                case 2: v = jl_box16(dt, *(int16_t*)data); break;
+                case 4: v = jl_box32(dt, *(int32_t*)data); break;
+                case 8: v = jl_box64(dt, *(int64_t*)data); break;
+                default:
+                    v = (jl_value_t*)allocobj(sizeof(void*)+nby);
+                    v->type = (jl_value_t*)dt;
+                    memcpy(jl_data_ptr(v), data, nby);
+                }
             }
-            jl_set_nth_field(v, 0, (jl_value_t*)a);
+            if (usetable)
+                ptrhash_put(&backref_table, (void*)(ptrint_t)pos, v);
         }
         else {
-            for(i=0; i < nf; i++) {
-                jl_set_nth_field(v, i, jl_deserialize_value(s));
+            v = jl_new_struct_uninit(dt);
+            if (usetable)
+                ptrhash_put(&backref_table, (void*)(ptrint_t)pos, v);
+            if (vtag == (jl_value_t*)IdTable_tag) {
+                jl_array_t *a = jl_alloc_cell_1d(32);
+                while (1) {
+                    jl_value_t *val = jl_deserialize_value(s);
+                    if (val == NULL)
+                        break;
+                    jl_value_t *key = jl_deserialize_value(s);
+                    a = jl_eqtable_put(a, key, val);
+                }
+                jl_set_nth_field(v, 0, (jl_value_t*)a);
+            }
+            else {
+                for(i=0; i < nf; i++) {
+                    jl_set_nth_field(v, i, jl_deserialize_value(s));
+                }
             }
         }
         // TODO: put WeakRefs on the weak_refs list
         return v;
     }
-    else if (vtag == (jl_value_t*)jl_tag_kind) {
-        return jl_deserialize_tag_type(s, jl_tag_kind, pos);
-    }
     assert(0);
     return NULL;
 }
@@ -748,7 +747,10 @@ void jl_save_system_image(char *fname)
     jl_gc_disable();
     htable_reset(&backref_table, 50000);
     ios_t f;
-    ios_file(&f, fname, 1, 1, 1, 1);
+    if (ios_file(&f, fname, 1, 1, 1, 1) == NULL) {
+        JL_PRINTF(JL_STDERR, "Cannot open system image file \"%s\" for writing.\n", fname);
+        exit(1);
+    }
 
     // orphan old Base module if present
     jl_base_module = (jl_module_t*)jl_get_global(jl_main_module, jl_symbol("Base"));
@@ -756,7 +758,7 @@ void jl_save_system_image(char *fname)
     // delete cached slow ASCIIString constructor if present
     jl_methtable_t *mt = jl_gf_mtable((jl_function_t*)jl_ascii_string_type);
     jl_array_t *spec = mt->defs->func->linfo->specializations;
-    if (jl_array_len(spec) > 0 &&
+    if (spec != NULL && jl_array_len(spec) > 0 &&
         ((jl_lambda_info_t*)jl_cellref(spec,0))->inferred == 0) {
         mt->cache = JL_NULL;
         mt->cache_arg1 = JL_NULL;
@@ -790,15 +792,15 @@ void jl_restore_system_image(char *fname)
     ios_t f;
     char *fpath = fname;
     if (ios_file(&f, fpath, 1, 0, 0, 0) == NULL) {
-        JL_PRINTF(JL_STDERR, "system image file not found\n");
-        jl_exit(1);
+        JL_PRINTF(JL_STDERR, "System image file \"%s\" not found\n", fname);
+        exit(1);
     }
 #ifdef JL_GC_MARKSWEEP
     int en = jl_gc_is_enabled();
     jl_gc_disable();
 #endif
 
-    tagtype_list = jl_alloc_cell_1d(0);
+    datatype_list = jl_alloc_cell_1d(0);
 
     jl_array_type->env = jl_deserialize_value(&f);
     
@@ -810,18 +812,11 @@ void jl_restore_system_image(char *fname)
     jl_current_module = jl_base_module; // run start_image in Base
 
     // cache builtin parametric types
-    for(int i=0; i < jl_array_len(tagtype_list); i++) {
-        jl_value_t *v = jl_cellref(tagtype_list, i);
-        uint32_t uid=0;
-        if (jl_is_struct_type(v))
-            uid = ((jl_struct_type_t*)v)->uid;
-        else if (jl_is_bits_type(v))
-            uid = ((jl_bits_type_t*)v)->uid;
-        jl_cache_type_((jl_tag_type_t*)v);
-        if (jl_is_struct_type(v))
-            ((jl_struct_type_t*)v)->uid = uid;
-        else if (jl_is_bits_type(v))
-            ((jl_bits_type_t*)v)->uid = uid;
+    for(int i=0; i < jl_array_len(datatype_list); i++) {
+        jl_value_t *v = jl_cellref(datatype_list, i);
+        uint32_t uid = ((jl_datatype_t*)v)->uid;
+        jl_cache_type_((jl_datatype_t*)v);
+        ((jl_datatype_t*)v)->uid = uid;
     }
 
     jl_get_builtin_hooks();
@@ -849,13 +844,17 @@ jl_value_t *jl_ast_rettype(jl_lambda_info_t *li, jl_value_t *ast)
 {
     if (jl_is_expr(ast))
         return jl_lam_body((jl_expr_t*)ast)->etype;
-    tree_literal_values = li->def->roots;
+    tree_literal_values = li->module->constant_table;
     ios_t src;
     jl_array_t *bytes = (jl_array_t*)ast;
     ios_mem(&src, 0);
     ios_setbuf(&src, bytes->data, jl_array_len(bytes), 0);
     src.size = jl_array_len(bytes);
+    int en = jl_gc_is_enabled();
+    jl_gc_disable();
     jl_value_t *rt = jl_deserialize_value(&src);
+    if (en)
+        jl_gc_enable();
     tree_literal_values = NULL;
     return rt;
 }
@@ -869,11 +868,9 @@ jl_value_t *jl_compress_ast(jl_lambda_info_t *li, jl_value_t *ast)
     int en = jl_gc_is_enabled();
     jl_gc_disable();
 
-    jl_lambda_info_t *def = li->def;
-    if (def->roots == NULL) {
-        def->roots = jl_alloc_cell_1d(0);
-    }
-    tree_literal_values = def->roots;
+    if (li->module->constant_table == NULL)
+        li->module->constant_table = jl_alloc_cell_1d(0);
+    tree_literal_values = li->module->constant_table;
     li->capt = (jl_value_t*)jl_lam_capt((jl_expr_t*)ast);
     if (jl_array_len(li->capt) == 0)
         li->capt = NULL;
@@ -883,8 +880,8 @@ jl_value_t *jl_compress_ast(jl_lambda_info_t *li, jl_value_t *ast)
     //JL_PRINTF(JL_STDERR, "%d bytes, %d values\n", dest.size, vals->length);
 
     jl_value_t *v = (jl_value_t*)jl_takebuf_array(&dest);
-    if (jl_array_len(tree_literal_values) == 0) {
-        def->roots = NULL;
+    if (jl_array_len(tree_literal_values) == 0 && last_tlv == NULL) {
+        li->module->constant_table = NULL;
     }
     tree_literal_values = last_tlv;
     if (en)
@@ -896,7 +893,7 @@ DLLEXPORT
 jl_value_t *jl_uncompress_ast(jl_lambda_info_t *li, jl_value_t *data)
 {
     jl_array_t *bytes = (jl_array_t*)data;
-    tree_literal_values = li->def->roots;
+    tree_literal_values = li->module->constant_table;
     ios_t src;
     ios_mem(&src, 0);
     ios_setbuf(&src, bytes->data, jl_array_len(bytes), 0);
@@ -923,17 +920,17 @@ void jl_init_serializer(void)
     htable_new(&id_to_fptr, 0);
     htable_new(&backref_table, 50000);
 
-    void *tags[] = { jl_symbol_type, jl_tag_kind, jl_bits_kind, jl_struct_kind,
+    void *tags[] = { jl_symbol_type, jl_datatype_type,
                      jl_function_type, jl_tuple_type, jl_array_type,
                      jl_expr_type, (void*)LongSymbol_tag, (void*)LongTuple_tag,
                      (void*)LongExpr_tag, (void*)LiteralVal_tag,
                      (void*)SmallInt64_tag, (void*)IdTable_tag,
+                     (void*)Int32_tag, (void*)Array1d_tag,
                      jl_module_type, jl_tvar_type, jl_lambda_info_type,
 
                      jl_null, jl_false, jl_true, jl_any_type, jl_symbol("Any"),
                      jl_symbol("Array"), jl_symbol("TypeVar"),
-                     jl_symbol("FuncKind"), jl_symbol("Box"),
-                     jl_symbol("apply"),
+                     jl_symbol("Box"), jl_symbol("apply"),
                      lambda_sym, body_sym, return_sym, call_sym, colons_sym,
                      null_sym, goto_ifnot_sym, assign_sym,
 
@@ -958,7 +955,7 @@ void jl_init_serializer(void)
                      jl_symbol("eq_int"), jl_symbol("slt_int"),
                      jl_symbol("sle_int"), jl_symbol("ne_int"),
                      jl_symbol("arrayset"), jl_symbol("arrayref"),
-                     jl_symbol("arraylen"),
+                     jl_symbol("arraylen"), jl_symbol("boundscheck"),
                      jl_symbol("convert"), jl_symbol("typeassert"),
                      jl_symbol("getfield"), jl_symbol("setfield"),
                      jl_symbol("tupleref"), jl_symbol("tuplelen"),
@@ -975,7 +972,7 @@ void jl_init_serializer(void)
                      jl_box_int32(24), jl_box_int32(25), jl_box_int32(26),
                      jl_box_int32(27), jl_box_int32(28), jl_box_int32(29),
                      jl_box_int32(30), jl_box_int32(31), jl_box_int32(32),
-#ifndef __LP64__
+#ifndef _P64
                      jl_box_int32(33), jl_box_int32(34), jl_box_int32(35),
                      jl_box_int32(36), jl_box_int32(37), jl_box_int32(38),
                      jl_box_int32(39), jl_box_int32(40), jl_box_int32(41),
@@ -999,7 +996,7 @@ void jl_init_serializer(void)
                      jl_box_int64(24), jl_box_int64(25), jl_box_int64(26),
                      jl_box_int64(27), jl_box_int64(28), jl_box_int64(29),
                      jl_box_int64(30), jl_box_int64(31), jl_box_int64(32),
-#ifdef __LP64__
+#ifdef _P64
                      jl_box_int64(33), jl_box_int64(34), jl_box_int64(35),
                      jl_box_int64(36), jl_box_int64(37), jl_box_int64(38),
                      jl_box_int64(39), jl_box_int64(40), jl_box_int64(41),
@@ -1017,14 +1014,14 @@ void jl_init_serializer(void)
                      jl_type_type, jl_bottom_type, jl_pointer_type,
                      jl_vararg_type, jl_ntuple_type, jl_abstractarray_type,
                      jl_box_type, jl_typector_type, jl_undef_type, jl_top_type,
-                     jl_typename_type, jl_task_type, jl_union_kind,
+                     jl_typename_type, jl_task_type, jl_uniontype_type,
                      jl_typetype_type, jl_typetype_tvar, jl_ANY_flag,
                      jl_array_any_type, jl_intrinsic_type, jl_method_type,
-                     jl_methtable_type, jl_voidpointer_type,
+                     jl_methtable_type, jl_voidpointer_type, jl_newvarnode_type,
                      jl_array_symbol_type, jl_tupleref(jl_tuple_type,0),
 
                      jl_symbol_type->name, jl_pointer_type->name,
-                     jl_tag_kind->name, jl_union_kind->name, jl_bits_kind->name, jl_struct_kind->name,
+                     jl_datatype_type->name, jl_uniontype_type->name,
                      jl_array_type->name, jl_expr_type->name,
                      jl_typename_type->name, jl_type_type->name, jl_methtable_type->name,
                      jl_method_type->name, jl_tvar_type->name,
@@ -1060,7 +1057,7 @@ void jl_init_serializer(void)
                       jl_f_set_field, jl_f_field_type, 
                       jl_f_arraylen, jl_f_arrayref, 
                       jl_f_arrayset, jl_f_arraysize, 
-                      jl_f_instantiate_type,
+                      jl_f_instantiate_type, jl_f_kwcall,
                       jl_f_convert_default, jl_f_convert_tuple,
                       jl_trampoline, jl_f_new_type_constructor, 
                       jl_f_typevar, jl_f_union, 
@@ -1068,6 +1065,7 @@ void jl_init_serializer(void)
                       jl_f_invoke, jl_apply_generic, 
                       jl_unprotect_stack, jl_f_task, 
                       jl_f_yieldto, jl_f_ctor_trampoline,
+                      jl_f_new_module,
                       NULL };
     i=2;
     while (fptrs[i-2] != NULL) {
