@@ -42,8 +42,18 @@
 #include <locale.h>
 #include <limits.h>
 #include <errno.h>
-#include <math.h>
+
+#include "platform.h"
+
+#if defined(_OS_WINDOWS_) && !defined(_COMPILER_MINGW_)
+char * basename(char *);
+char * dirname(char *);
+#else
 #include <libgen.h>
+#endif
+
+//#define MEMDEBUG
+
 #include "libsupport.h"
 #include "flisp.h"
 #include "opcodes.h"
@@ -101,7 +111,7 @@ static value_t NIL, LAMBDA, IF, TRYCATCH;
 static value_t BACKQUOTE, COMMA, COMMAAT, COMMADOT, FUNCTION;
 
 static value_t pairsym, symbolsym, fixnumsym, vectorsym, builtinsym, vu8sym;
-static value_t definesym, defmacrosym, forsym, labelsym, setqsym;
+static value_t definesym, defmacrosym, forsym, setqsym;
 static value_t tsym, Tsym, fsym, Fsym, booleansym, nullsym, evalsym, fnsym;
 // for reading characters
 static value_t nulsym, alarmsym, backspacesym, tabsym, linefeedsym, newlinesym;
@@ -246,7 +256,7 @@ symbol_t *symtab = NULL;
 
 int fl_is_keyword_name(char *str, size_t len)
 {
-    return ((str[0] == ':' || str[len-1] == ':') && str[1] != '\0');
+    return len>1 && ((str[0] == ':' || str[len-1] == ':') && str[1] != '\0');
 }
 
 static symbol_t *mk_symbol(char *str)
@@ -254,7 +264,7 @@ static symbol_t *mk_symbol(char *str)
     symbol_t *sym;
     size_t len = strlen(str);
 
-    sym = (symbol_t*)malloc(sizeof(symbol_t)-sizeof(void*) + len + 1);
+    sym = (symbol_t*)malloc((sizeof(symbol_t)-sizeof(void*)+len+1+7)&-8);
     assert(((uptrint_t)sym & 0x7) == 0); // make sure malloc aligns 8
     sym->left = sym->right = NULL;
     sym->flags = 0;
@@ -535,12 +545,12 @@ void gc(int mustgrow)
     void *temp;
     uint32_t i, f, top;
     fl_readstate_t *rs;
-
+    size_t hsz = grew ? heapsize*2 : heapsize;
+#ifdef MEMDEBUG
+    tospace = LLT_ALLOC(hsz);
+#endif
     curheap = tospace;
-    if (grew)
-        lim = curheap+heapsize*2-sizeof(cons_t);
-    else
-        lim = curheap+heapsize-sizeof(cons_t);
+    lim = curheap + hsz - sizeof(cons_t);
 
     if (fl_throwing_frame > curr_frame) {
         top = fl_throwing_frame - 4;
@@ -587,7 +597,14 @@ void gc(int mustgrow)
     // if we're using > 80% of the space, resize tospace so we have
     // more space to fill next time. if we grew tospace last time,
     // grow the other half of the heap this time to catch up.
-    if (grew || ((lim-curheap) < (int)(heapsize/5)) || mustgrow) {
+    if (grew || mustgrow
+#ifdef MEMDEBUG
+        // GC more often
+        || ((lim-curheap) < (int)(heapsize/128))
+#else
+        || ((lim-curheap) < (int)(heapsize/5))
+#endif
+        ) {
         temp = LLT_REALLOC(tospace, heapsize*2);
         if (temp == NULL)
             fl_raise(memory_exception_value);
@@ -601,6 +618,9 @@ void gc(int mustgrow)
         }
         grew = !grew;
     }
+#ifdef MEMDEBUG
+    LLT_FREE(tospace);
+#endif
     if (curheap > lim)  // all data was live
         gc(0);
 }
@@ -1370,7 +1390,7 @@ static value_t apply_cl(uint32_t nargs)
                 if (fits_fixnum(s))
                     v = fixnum(s);
                 else
-                    v = mk_long(s);
+                    v = mk_ptrdiff(s);
             }
             else {
                 v = fl_add_any(&Stack[SP-2], 2, 0);
@@ -1408,7 +1428,7 @@ static value_t apply_cl(uint32_t nargs)
                 if (fits_fixnum(s))
                     v = fixnum(s);
                 else
-                    v = mk_long(s);
+                    v = mk_ptrdiff(s);
             }
             else {
                 Stack[SP-1] = fl_neg(Stack[SP-1]);
@@ -1513,7 +1533,7 @@ static value_t apply_cl(uint32_t nargs)
                 if (isfixnum(e))
                     i = numval(e);
                 else
-                    i = (uint32_t)toulong(e, "aref");
+                    i = (uint32_t)tosize(e, "aref");
                 if ((unsigned)i >= vector_size(v))
                     bounds_error("aref", v, e);
                 v = vector_elt(v, i);
@@ -2147,34 +2167,36 @@ value_t fl_map1(value_t *args, u_int32_t nargs)
     if (nargs < 2)
         lerror(ArgError, "map: too few arguments");
     if (!iscons(args[1])) return NIL;
-    value_t first, last, v;
+    value_t *first, *last, v;
     int64_t argSP = args-Stack;
     assert(argSP >= 0 && argSP < N_STACK);
     if (nargs == 2) {
-        if (SP+3 > N_STACK) grow_stack();
+        if (SP+4 > N_STACK) grow_stack();
         PUSH(Stack[argSP]);
         PUSH(car_(Stack[argSP+1]));
         v = _applyn(1);
+        POPN(2);
         PUSH(v);
         v = mk_cons();
         car_(v) = POP(); cdr_(v) = NIL;
-        last = first = v;
+        PUSH(v);
+        PUSH(v);
+        first = &Stack[SP-2];
+        last = &Stack[SP-1];
         Stack[argSP+1] = cdr_(Stack[argSP+1]);
-        fl_gc_handle(&first);
-        fl_gc_handle(&last);
         while (iscons(Stack[argSP+1])) {
-            Stack[SP-2] = Stack[argSP];
-            Stack[SP-1] = car_(Stack[argSP+1]);
+            PUSH(Stack[argSP]);
+            PUSH(car_(Stack[argSP+1]));
             v = _applyn(1);
+            POPN(2);
             PUSH(v);
             v = mk_cons();
             car_(v) = POP(); cdr_(v) = NIL;
-            cdr_(last) = v;
-            last = v;
+            cdr_(*last) = v;
+            *last = v;
             Stack[argSP+1] = cdr_(Stack[argSP+1]);
         }
         POPN(2);
-        fl_free_gc_handles(2);
     }
     else {
         size_t i;
@@ -2185,29 +2207,31 @@ value_t fl_map1(value_t *args, u_int32_t nargs)
             Stack[argSP+i] = cdr_(Stack[argSP+i]);
         }
         v = _applyn(nargs-1);
+        POPN(nargs);
         PUSH(v);
         v = mk_cons();
         car_(v) = POP(); cdr_(v) = NIL;
-        last = first = v;
-        fl_gc_handle(&first);
-        fl_gc_handle(&last);
+        PUSH(v);
+        PUSH(v);
+        first = &Stack[SP-2];
+        last = &Stack[SP-1];
         while (iscons(Stack[argSP+1])) {
-            Stack[SP-nargs] = Stack[argSP];
+            PUSH(Stack[argSP]);
             for(i=1; i < nargs; i++) {
-                Stack[SP-nargs+i] = car(Stack[argSP+i]);
+                PUSH(car(Stack[argSP+i]));
                 Stack[argSP+i] = cdr_(Stack[argSP+i]);
             }
             v = _applyn(nargs-1);
+            POPN(nargs);
             PUSH(v);
             v = mk_cons();
             car_(v) = POP(); cdr_(v) = NIL;
-            cdr_(last) = v;
-            last = v;
+            cdr_(*last) = v;
+            *last = v;
         }
-        POPN(nargs);
-        fl_free_gc_handles(2);
+        POPN(2);
     }
-    return first;
+    return *first;
 }
 
 static builtinspec_t core_builtin_info[] = {
@@ -2241,7 +2265,11 @@ static void lisp_init(size_t initial_heapsize)
     heapsize = initial_heapsize;
 
     fromspace = LLT_ALLOC(heapsize);
+#ifdef MEMDEBUG
+    tospace   = NULL;
+#else
     tospace   = LLT_ALLOC(heapsize);
+#endif
     curheap = fromspace;
     lim = curheap+heapsize-sizeof(cons_t);
     consflags = bitvector_new(heapsize/sizeof(cons_t), 1);
@@ -2270,7 +2298,7 @@ static void lisp_init(size_t initial_heapsize)
     vectorsym = symbol("vector");     builtinsym = symbol("builtin");
     booleansym = symbol("boolean");   nullsym = symbol("null");
     definesym = symbol("define");     defmacrosym = symbol("define-macro");
-    forsym = symbol("for");           labelsym = symbol("label");
+    forsym = symbol("for");
     setqsym = symbol("set!");         evalsym = symbol("eval");
     vu8sym = symbol("vu8");           fnsym = symbol("fn");
     nulsym = symbol("nul");           alarmsym = symbol("alarm");
@@ -2296,11 +2324,11 @@ static void lisp_init(size_t initial_heapsize)
     setc(symbol("procedure?"), builtin(OP_FUNCTIONP));
     setc(symbol("top-level-bound?"), builtin(OP_BOUNDP));
 
-#ifdef __linux__
+#if defined(_OS_LINUX_)
     set(symbol("*os-name*"), symbol("linux"));
-#elif defined(WIN32) || defined(WIN64)
+#elif defined(_OS_WINDOWS_)
     set(symbol("*os-name*"), symbol("win32"));
-#elif defined(__APPLE__)
+#elif defined(_OS_DARWIN_)
     set(symbol("*os-name*"), symbol("macos"));
 #else
     set(symbol("*os-name*"), symbol("unknown"));
