@@ -52,6 +52,22 @@ int isnumtok_base(char *tok, value_t *pval, int base)
             return 1;
         }
     }
+    // hexadecimal float literals
+    else if (((tok[0]=='0' && tok[1]=='x') || (base == 16)) &&
+        strpbrk(tok, "pP")) {
+        d = strtod(tok, &end);
+        if (*end == '\0') {
+            if (pval) *pval = mk_double(d);
+            return 1;
+        }
+        // floats can end in f or f0
+        if (end > tok && end[0] == 'f' &&
+            (end[1] == '\0' ||
+             (end[1] == '0' && end[2] == '\0'))) {
+            if (pval) *pval = mk_float((float)d);
+            return 1;
+        }
+    }
 
     if (tok[0] == '+') {
         if (!strcmp(tok,"+NaN") || !strcasecmp(tok,"+nan.0")) {
@@ -76,15 +92,17 @@ int isnumtok_base(char *tok, value_t *pval, int base)
         i64 = strtoll(tok, &end, base);
         if (errno)
             return 0;
+        int done = (*end == '\0');  // must access *end before alloc
         if (pval) *pval = return_from_int64(i64);
-        return (*end == '\0');
+        return done;
     }
     errno = 0;
     ui64 = strtoull_0b0o(tok, &end, base);
     if (errno)
         return 0;
+    int done = (*end == '\0');  // must access *end before alloc
     if (pval) *pval = return_from_uint64(ui64);
-    return (*end == '\0');
+    return done;
 }
 
 static int isnumtok(char *tok, value_t *pval)
@@ -419,6 +437,7 @@ static value_t read_vector(value_t label, u_int32_t closer)
     while (peek() != closer) {
         if (ios_eof(F))
             lerror(ParseError, "read: unexpected end of input");
+        v = Stack[SP-1]; // reload after possible alloc in peek()
         if (i >= vector_size(v)) {
             v = Stack[SP-1] = vector_grow(v);
             if (label != UNBOUND)
