@@ -1,7 +1,7 @@
 #include <stdlib.h>
 #include <setjmp.h>
 #include <assert.h>
-#ifdef __WIN32__
+#ifdef _OS_WINDOWS_
 #include <malloc.h>
 #endif
 #include "julia.h"
@@ -11,8 +11,9 @@ extern int jl_lineno;
 
 static jl_value_t *eval(jl_value_t *e, jl_value_t **locals, size_t nl);
 static jl_value_t *eval_body(jl_array_t *stmts, jl_value_t **locals, size_t nl,
-                             int start);
+                             int start, int toplevel);
 jl_value_t *jl_eval_module_expr(jl_expr_t *ex);
+int jl_is_toplevel_only_expr(jl_value_t *e);
 
 jl_value_t *jl_interpret_toplevel_expr(jl_value_t *e)
 {
@@ -66,6 +67,8 @@ jl_value_t *jl_eval_global_var(jl_module_t *m, jl_sym_t *e)
     return v;
 }
 
+int jl_has_intrinsics(jl_expr_t *e);
+
 extern int jl_boot_file_loaded;
 extern int inside_typedef;
 
@@ -108,22 +111,68 @@ static jl_value_t *eval(jl_value_t *e, jl_value_t **locals, size_t nl)
             return jl_f_get_field(NULL, gfargs, 2);
         }
         if (jl_is_lambda_info(e)) {
-            return (jl_value_t*)jl_new_closure(NULL, (jl_value_t*)jl_null,
-                                               (jl_lambda_info_t*)e);
+            jl_lambda_info_t *li = (jl_lambda_info_t*)e;
+            if (jl_boot_file_loaded && li->ast && jl_is_expr(li->ast)) {
+                li->ast = jl_compress_ast(li, li->ast);
+            }
+            return (jl_value_t*)jl_new_closure(NULL, (jl_value_t*)jl_null, li);
         }
         if (jl_is_linenode(e)) {
             jl_lineno = jl_linenode_line(e);
         }
+        if (jl_is_newvarnode(e)) {
+            jl_value_t *var = jl_fieldref(e,0);
+            assert(jl_is_symbol(var));
+            for(size_t i=0; i < nl; i++) {
+                if (locals[i*2] == var) {
+                    locals[i*2+1] = NULL;
+                    break;
+                }
+            }
+            return (jl_value_t*)jl_nothing;
+        }
         return e;
     }
     jl_expr_t *ex = (jl_expr_t*)e;
     jl_value_t **args = &jl_cellref(ex->args,0);
+    size_t nargs = jl_array_len(ex->args);
     if (ex->head == call_sym ||  ex->head == call1_sym) {
+        if (jl_is_lambda_info(args[0])) {
+            // directly calling an inner function ("let")
+            jl_lambda_info_t *li = (jl_lambda_info_t*)args[0];
+            if (jl_is_expr(li->ast) && !jl_lam_vars_captured((jl_expr_t*)li->ast) &&
+                !jl_has_intrinsics((jl_expr_t*)li->ast)) {
+                size_t na = nargs-1;
+                if (na == 0)
+                    return jl_interpret_toplevel_thunk(li);
+                jl_array_t *formals = jl_lam_args((jl_expr_t*)li->ast);
+                size_t nreq = jl_array_len(formals);
+                if (nreq==0 || !jl_is_rest_arg(jl_cellref(formals,nreq-1))) {
+                    jl_value_t **ar;
+                    JL_GC_PUSHARGS(ar, na*2);
+                    for(int i=0; i < na*2; i++) {
+                        ar[i] = NULL;
+                    }
+                    for(int i=0; i < na; i++) {
+                        ar[i*2+1] = eval(args[i+1], locals, nl);
+                    }
+                    if (na != nreq) {
+                        jl_error("wrong number of arguments");
+                    }
+                    for(int i=0; i < na; i++) {
+                        ar[i*2] = (jl_value_t*)jl_decl_var(jl_cellref(formals,i));
+                    }
+                    jl_value_t *ret = jl_interpret_toplevel_thunk_with(li, ar, na);
+                    JL_GC_POP();
+                    return ret;
+                }
+            }
+        }
         jl_function_t *f = (jl_function_t*)eval(args[0], locals, nl);
         if (!jl_is_func(f))
             jl_type_error("apply", (jl_value_t*)jl_function_type,
                           (jl_value_t*)f);
-        return do_call(f, &args[1], jl_array_len(ex->args)-1, locals, nl);
+        return do_call(f, &args[1], nargs-1, locals, nl);
     }
     else if (ex->head == assign_sym) {
         jl_value_t *sym = args[0];
@@ -140,9 +189,13 @@ static jl_value_t *eval(jl_value_t *e, jl_value_t **locals, size_t nl)
     }
     else if (ex->head == new_sym) {
         jl_value_t *thetype = eval(args[0], locals, nl);
-        JL_GC_PUSH(&thetype);
-        assert(jl_is_struct_type(thetype));
-        jl_value_t *v = jl_new_struct_uninit((jl_struct_type_t*)thetype);
+        jl_value_t *v=NULL;
+        JL_GC_PUSH2(&thetype, &v);
+        assert(jl_is_structtype(thetype));
+        v = jl_new_struct_uninit((jl_datatype_t*)thetype);
+        for(size_t i=1; i < nargs; i++) {
+            jl_set_nth_field(v, i-1, eval(args[i], locals, nl));
+        }
         JL_GC_POP();
         return v;
     }
@@ -150,7 +203,7 @@ static jl_value_t *eval(jl_value_t *e, jl_value_t **locals, size_t nl)
         return (jl_value_t*)jl_nothing;
     }
     else if (ex->head == body_sym) {
-        return eval_body(ex->args, locals, nl, 0);
+        return eval_body(ex->args, locals, nl, 0, 0);
     }
     else if (ex->head == exc_sym) {
         return jl_exception_in_transit;
@@ -162,26 +215,47 @@ static jl_value_t *eval(jl_value_t *e, jl_value_t **locals, size_t nl)
         jl_sym_t *fname = (jl_sym_t*)args[0];
         jl_value_t **bp=NULL;
         jl_binding_t *b=NULL;
-        for (size_t i=0; i < nl; i++) {
-            if (locals[i*2] == (jl_value_t*)fname) {
-                bp = &locals[i*2+1];
-                break;
+        jl_value_t *gf=NULL;
+        int kw=0;
+        if (jl_is_expr(fname)) {
+            if (((jl_expr_t*)fname)->head == kw_sym) {
+                kw = 1;
+                fname = (jl_sym_t*)jl_exprarg(fname, 0);
+            }
+            gf = eval((jl_value_t*)fname, locals, nl);
+            assert(jl_is_function(gf));
+            assert(jl_is_gf(gf));
+            if (jl_is_expr(fname))
+                fname = (jl_sym_t*)jl_fieldref(jl_exprarg(fname, 2), 0);
+            if (!kw)
+                bp = &gf;
+            else
+                bp = (jl_value_t**)&((jl_methtable_t*)((jl_function_t*)gf)->env)->kwsorter;
+            assert(jl_is_symbol(fname));
+        }
+        else {
+            for (size_t i=0; i < nl; i++) {
+                if (locals[i*2] == (jl_value_t*)fname) {
+                    bp = &locals[i*2+1];
+                    break;
+                }
+            }
+            if (bp == NULL) {
+                b = jl_get_binding_for_method_def(jl_current_module, fname);
+                bp = &b->value;
             }
         }
-        if (bp == NULL) {
-            b = jl_get_binding_for_method_def(jl_current_module, fname);
-            bp = &b->value;
-        }
-        jl_value_t *atypes=NULL, *meth=NULL, *tvars=NULL;
-        JL_GC_PUSH(&atypes, &meth, &tvars);
+        jl_value_t *atypes=NULL, *meth=NULL;
+        JL_GC_PUSH2(&atypes, &meth);
         atypes = eval(args[1], locals, nl);
         meth = eval(args[2], locals, nl);
-        tvars = eval(args[3], locals, nl);
-        jl_method_def(fname, bp, b, (jl_tuple_t*)atypes,
-                      (jl_function_t*)meth, (jl_tuple_t*)tvars);
+        jl_method_def(fname, bp, b, (jl_tuple_t*)atypes, (jl_function_t*)meth);
         JL_GC_POP();
         return *bp;
     }
+    else if (ex->head == copyast_sym) {
+        return jl_copy_ast(eval(args[0], locals, nl));
+    }
     else if (ex->head == const_sym) {
         jl_value_t *sym = args[0];
         for (size_t i=0; i < nl; i++) {
@@ -206,19 +280,20 @@ static jl_value_t *eval(jl_value_t *e, jl_value_t **locals, size_t nl)
         jl_value_t *name = args[0];
         jl_value_t *para = eval(args[1], locals, nl);
         jl_value_t *super = NULL;
-        JL_GC_PUSH(&para, &super);
-        jl_tag_type_t *tt=jl_new_tagtype(name, jl_any_type, (jl_tuple_t*)para);
+        JL_GC_PUSH2(&para, &super);
+        jl_datatype_t *dt =
+            jl_new_abstracttype(name, jl_any_type, (jl_tuple_t*)para);
         jl_binding_t *b = jl_get_binding_wr(jl_current_module, (jl_sym_t*)name);
-        jl_checked_assignment(b, (jl_value_t*)tt);
+        jl_checked_assignment(b, (jl_value_t*)dt);
         super = eval(args[2], locals, nl);
-        jl_set_tag_type_super(tt, super);
+        jl_set_datatype_super(dt, super);
         JL_GC_POP();
         return (jl_value_t*)jl_nothing;
     }
     else if (ex->head == bitstype_sym) {
         jl_value_t *name = args[0];
         jl_value_t *super = NULL, *para = NULL, *vnb = NULL;
-        JL_GC_PUSH(&para, &super, &vnb);
+        JL_GC_PUSH3(&para, &super, &vnb);
         para = eval(args[1], locals, nl);
         vnb  = eval(args[2], locals, nl);
         if (!jl_is_long(vnb))
@@ -227,36 +302,42 @@ static jl_value_t *eval(jl_value_t *e, jl_value_t **locals, size_t nl)
         if (nb < 1 || nb>=(1<<23) || (nb&7) != 0)
             jl_errorf("invalid number of bits in type %s",
                       ((jl_sym_t*)name)->name);
-        jl_bits_type_t *bt = jl_new_bits_type(name, jl_any_type, (jl_tuple_t*)para, nb);
+        jl_datatype_t *dt =
+            jl_new_bitstype(name, jl_any_type, (jl_tuple_t*)para, nb);
         jl_binding_t *b = jl_get_binding_wr(jl_current_module, (jl_sym_t*)name);
-        jl_checked_assignment(b, (jl_value_t*)bt);
+        jl_checked_assignment(b, (jl_value_t*)dt);
         super = eval(args[3], locals, nl);
-        jl_set_tag_type_super((jl_tag_type_t*)bt, super);
+        jl_set_datatype_super(dt, super);
         JL_GC_POP();
         return (jl_value_t*)jl_nothing;
     }
     else if (ex->head == compositetype_sym) {
-        void jl_add_constructors(jl_struct_type_t *t);
+        void jl_add_constructors(jl_datatype_t *t);
         jl_value_t *name = args[0];
         jl_value_t *para = eval(args[1], locals, nl);
         jl_value_t *fnames = NULL;
         jl_value_t *super = NULL;
-        jl_struct_type_t *st = NULL;
-        JL_GC_PUSH(&para, &super, &fnames, &st);
+        jl_datatype_t *dt = NULL;
+        JL_GC_PUSH4(&para, &super, &fnames, &dt);
         fnames = eval(args[2], locals, nl);
-        st = jl_new_struct_type((jl_sym_t*)name, jl_any_type, (jl_tuple_t*)para,
-                                (jl_tuple_t*)fnames, NULL);
-        st->ctor_factory = eval(args[3], locals, nl);
+        dt = jl_new_datatype((jl_sym_t*)name, jl_any_type, (jl_tuple_t*)para,
+                             (jl_tuple_t*)fnames, NULL,
+                             0, args[6]==jl_true ? 1 : 0);
+        dt->fptr = jl_f_ctor_trampoline;
+        dt->ctor_factory = eval(args[3], locals, nl);
         jl_binding_t *b = jl_get_binding_wr(jl_current_module, (jl_sym_t*)name);
-        jl_checked_assignment(b, (jl_value_t*)st);
+        jl_checked_assignment(b, (jl_value_t*)dt);
         inside_typedef = 1;
-        st->types = (jl_tuple_t*)eval(args[5], locals, nl);
+        dt->types = (jl_tuple_t*)eval(args[5], locals, nl);
         inside_typedef = 0;
-        jl_check_type_tuple(st->types, st->name->name, "type definition");
+        jl_check_type_tuple(dt->types, dt->name->name, "type definition");
         super = eval(args[4], locals, nl);
-        jl_set_tag_type_super((jl_tag_type_t*)st, super);
-        jl_compute_struct_offsets(st);
-        jl_add_constructors(st);
+        jl_set_datatype_super(dt, super);
+        for(size_t i=0; i < jl_tuple_len(para); i++) {
+            ((jl_tvar_t*)jl_tupleref(para,i))->bound = 0;
+        }
+        jl_compute_field_offsets(dt);
+        jl_add_constructors(dt);
         JL_GC_POP();
         return (jl_value_t*)jl_nothing;
     }
@@ -286,6 +367,9 @@ static jl_value_t *eval(jl_value_t *e, jl_value_t **locals, size_t nl)
             jl_errorf("syntax: %s", jl_string_data(args[0]));
         jl_throw(args[0]);
     }
+    else if (ex->head == boundscheck_sym) {
+        return (jl_value_t*)jl_nothing;
+    }
     jl_errorf("unsupported or misplaced expression %s", ex->head->name);
     return (jl_value_t*)jl_nothing;
 }
@@ -303,11 +387,17 @@ static int label_idx(jl_value_t *tgt, jl_array_t *stmts)
     return j;
 }
 
+jl_value_t *jl_toplevel_eval_body(jl_array_t *stmts)
+{
+    return eval_body(stmts, NULL, 0, 0, 1);
+}
+
 static jl_value_t *eval_body(jl_array_t *stmts, jl_value_t **locals, size_t nl,
-                             int start)
+                             int start, int toplevel)
 {
     jl_handler_t __eh;
     size_t i=start;
+
     while (1) {
         jl_value_t *stmt = jl_cellref(stmts,i);
         if (jl_is_gotonode(stmt)) {
@@ -328,12 +418,16 @@ static jl_value_t *eval_body(jl_array_t *stmts, jl_value_t **locals, size_t nl,
                 }
             }
             else if (head == return_sym) {
-                return eval(jl_exprarg(stmt,0), locals, nl);
+                jl_value_t *ex = jl_exprarg(stmt,0);
+                if (toplevel && jl_is_toplevel_only_expr(ex))
+                    return jl_toplevel_eval(ex);
+                else
+                    return eval(ex, locals, nl);
             }
             else if (head == enter_sym) {
                 jl_enter_handler(&__eh);
                 if (!jl_setjmp(__eh.eh_ctx,1)) {
-                    return eval_body(stmts, locals, nl, i+1);
+                    return eval_body(stmts, locals, nl, i+1, toplevel);
                 }
                 else {
                     i = label_idx(jl_exprarg(stmt,0), stmts);
@@ -345,11 +439,17 @@ static jl_value_t *eval_body(jl_array_t *stmts, jl_value_t **locals, size_t nl,
                 jl_pop_handler(hand_n_leave);
             }
             else {
-                eval(stmt, locals, nl);
+                if (toplevel && jl_is_toplevel_only_expr(stmt))
+                    jl_toplevel_eval(stmt);
+                else
+                    eval(stmt, locals, nl);
             }
         }
         else {
-            eval(stmt, locals, nl);
+            if (toplevel && jl_is_toplevel_only_expr(stmt))
+                jl_toplevel_eval(stmt);
+            else
+                eval(stmt, locals, nl);
         }
         i++;
     }
@@ -378,7 +478,7 @@ jl_value_t *jl_interpret_toplevel_thunk_with(jl_lambda_info_t *lam,
         locals[i*2]   = loc[(i-llength)*2];
         locals[i*2+1] = loc[(i-llength)*2+1];
     }
-    r = eval_body(stmts, locals, nl, 0);
+    r = eval_body(stmts, locals, nl, 0, 1);
     JL_GC_POP();
     return r;
 }
