@@ -31,7 +31,10 @@
 	((quoted? e) '())
 	(else (case (car e)
 		((=)            (list (decl-var (cadr e))))
-		((method)       (list (cadr e)))
+		((method)       (let ((n (method-expr-name e)))
+				  (if (symbol? n)
+				      (list n)
+				      '())))
 		((lambda)       '())
 		((local local!) '())
 		((break-block)  (find-possible-globals (caddr e)))
@@ -52,6 +55,8 @@
    (append
     ;; vars assigned at the outer level
     (filter (lambda (x) (not (some-gensym? x))) (find-assigned-vars e '()))
+    ;; vars declared const outside any scope block
+    (find-decls 'const e '())
     ;; vars assigned anywhere, if they have been defined as global
     (filter defined-julia-global (find-possible-globals e)))))
 
@@ -133,16 +138,30 @@
 
 (define (jl-parse-string s)
   (parser-wrap (lambda ()
-		 (let* ((inp  (make-token-stream (open-input-string s)))
-			(expr (julia-parse inp)))
-		   (expand-toplevel-expr expr)))))
+		 (let ((inp  (make-token-stream (open-input-string s))))
+		   ;; parse all exprs into a (toplevel ...) form
+		   (let loop ((exprs '()))
+		     ;; delay expansion so macros run in the Task executing
+		     ;; the input, not the task parsing it (issue #2378)
+		     ;; used to be (expand-toplevel-expr expr)
+		     (let ((expr (julia-parse inp)))
+		       (if (eof-object? expr)
+			   (cond ((null? exprs)     expr)
+				 ((length= exprs 1) (car exprs))
+				 (else (cons 'toplevel (reverse! exprs))))
+			   (if (and (pair? expr) (eq? (car expr) 'toplevel))
+			       (loop (nreconc (cdr expr) exprs))
+			       (loop (cons expr exprs))))))))))
 
 ;; parse file-in-a-string
-(define (jl-parse-string-stream str)
-  (jl-parser-set-stream "string" (open-input-string str)))
+(define (jl-parse-string-stream str filename)
+  (jl-parser-set-stream filename (open-input-string str)))
 
 (define (jl-parse-file s)
-  (jl-parser-set-stream s (open-input-file s)))
+  (trycatch
+   (begin (jl-parser-set-stream s (open-input-file s))
+	  #t)
+   (lambda (e) #f)))
 
 (define *filename-stack* '())
 (define *ts-stack* '())
@@ -167,11 +186,11 @@
 			  (julia-parse current-token-stream)))))
     (if (eof-object? e)
 	e
-	(cons (+ (input-port-line (ts:port current-token-stream))
-		 (if (eqv? (peek-token current-token-stream) #\newline)
-		     -1 0))
-	      (parser-wrap
-	       (lambda ()
+	(parser-wrap
+	 (lambda ()
+	   (cons (+ (input-port-line (ts:port current-token-stream))
+		    (if (eqv? (peek-token current-token-stream) #\newline)
+			-1 0))
 		 (if (and (pair? e) (or (eq? (car e) 'error)
 					(eq? (car e) 'continue)))
 		     e
