@@ -22,7 +22,7 @@
 
 #include "dtypes.h"
 
-#ifdef WIN32
+#ifdef _OS_WINDOWS_
 #include <malloc.h>
 #define snprintf _snprintf
 #else
@@ -257,7 +257,7 @@ size_t u8_strlen(const char *s)
     return count;
 }
 
-#if defined(__WIN32__) || defined(__linux__)
+#if defined(_OS_WINDOWS_) || defined(_OS_LINUX_)
 extern int wcwidth(uint32_t ch);
 #endif
 
@@ -288,7 +288,6 @@ size_t u8_strwidth(const char *s)
             }
             ch -= offsetsFromUTF8[nb];
             w = wcwidth(ch);  // might return -1
-            w=0;
             if (w > 0) tot += w;
         }
     }
@@ -358,7 +357,7 @@ char read_escape_control_char(char c)
     else if (c == 'r')
         return '\r';
     else if (c == 'e')
-        return '\e';
+        return '\x1B';
     else if (c == 'b')
         return '\b';
     else if (c == 'f')
@@ -454,7 +453,7 @@ int u8_escape_wchar(char *buf, size_t sz, uint32_t ch)
         return buf_put2c(buf, "\\t");
     else if (ch == L'\r')
         return buf_put2c(buf, "\\r");
-    else if (ch == L'\e')
+    else if (ch == L'\x1B')
         return buf_put2c(buf, "\\e");
     else if (ch == L'\b')
         return buf_put2c(buf, "\\b");
@@ -614,7 +613,7 @@ size_t u8_vprintf(const char *fmt, va_list ap)
     sz = 512;
     buf = (char*)alloca(sz);
     cnt = vsnprintf(buf, sz, fmt, ap);
-    if ((ssize_t)cnt < 0)
+    if ((intptr_t)cnt < 0)
         return 0;
     if (cnt >= sz) {
         buf = (char*)malloc(cnt + 1);
@@ -709,7 +708,7 @@ int u8_isvalid(const char *str, size_t length)
     return ret;
 }
 
-int u8_reverse(char *dest, char * src, size_t len)
+int u8_reverse(char *dest, char *src, size_t len)
 {
     size_t si=0, di=len;
     unsigned char c;
