@@ -1,23 +1,20 @@
 /*
   evaluating top-level expressions, loading source files
 */
+#include "platform.h"
+
 #include <stdlib.h>
-#include <stdio.h>
 #include <string.h>
-#include <stdarg.h>
 #include <setjmp.h>
 #include <assert.h>
 #include <sys/types.h>
 #include <errno.h>
-#include <fcntl.h>
-#include <unistd.h>
-#ifdef __WIN32__
+#if defined(_OS_WINDOWS_)
 #include <malloc.h>
+#else
+#include <unistd.h>
 #endif
-#include <ctype.h>
-#include <math.h>
 #include "julia.h"
-#include <sys/stat.h>
 #include "builtin_proto.h"
 
 DLLEXPORT char *julia_home = NULL;
@@ -42,6 +39,9 @@ jl_value_t *jl_eval_module_expr(jl_expr_t *ex)
 {
     assert(ex->head == module_sym);
     jl_module_t *last_module = jl_current_module;
+    if (jl_array_len(ex->args) != 3 || !jl_is_expr(jl_exprarg(ex,2))) {
+        jl_error("syntax: malformed module expression");
+    }
     int std_imports = (jl_exprarg(ex,0)==jl_true);
     jl_sym_t *name = (jl_sym_t*)jl_exprarg(ex, 1);
     if (!jl_is_symbol(name)) {
@@ -57,11 +57,10 @@ jl_value_t *jl_eval_module_expr(jl_expr_t *ex)
     newm->parent = parent_module;
     b->value = (jl_value_t*)newm;
     if (parent_module == jl_main_module && name == jl_symbol("Base")) {
+        base_module_conflict = (jl_base_module != NULL);
         jl_old_base_module = jl_base_module;
         // pick up Base module during bootstrap
         jl_base_module = newm;
-    } else {
-        base_module_conflict = 1;
     }
     // export all modules from Main
     if (parent_module == jl_main_module)
@@ -74,7 +73,7 @@ jl_value_t *jl_eval_module_expr(jl_expr_t *ex)
         }
     }
 
-    JL_GC_PUSH(&last_module);
+    JL_GC_PUSH1(&last_module);
     jl_current_module = newm;
 
     jl_array_t *exprs = ((jl_expr_t*)jl_exprarg(ex, 2))->args;
@@ -92,6 +91,8 @@ jl_value_t *jl_eval_module_expr(jl_expr_t *ex)
     JL_GC_POP();
     jl_current_module = last_module;
 
+#if 0
+    // some optional post-processing steps
     size_t i;
     void **table = newm->bindings.table;
     for(i=1; i < newm->bindings.size; i+=2) {
@@ -108,13 +109,14 @@ jl_value_t *jl_eval_module_expr(jl_expr_t *ex)
             */
         }
     }
+#endif
     return jl_nothing;
 }
 
 static int is_intrinsic(jl_module_t *m, jl_sym_t *s)
 {
     jl_value_t *v = jl_get_global(m, s);
-    return (v != NULL && jl_typeof(v)==(jl_type_t*)jl_intrinsic_type);
+    return (v != NULL && jl_typeof(v)==(jl_value_t*)jl_intrinsic_type);
 }
 
 // module referenced by TopNode from within m
@@ -126,7 +128,7 @@ jl_module_t *jl_base_relative_to(jl_module_t *m)
     return (m==jl_core_module||m==jl_old_base_module||jl_base_module==NULL) ? m : jl_base_module;
 }
 
-static int has_intrinsics(jl_expr_t *e)
+int jl_has_intrinsics(jl_expr_t *e)
 {
     if (jl_array_len(e->args) == 0)
         return 0;
@@ -139,7 +141,7 @@ static int has_intrinsics(jl_expr_t *e)
     int i;
     for(i=0; i < jl_array_len(e->args); i++) {
         jl_value_t *a = jl_exprarg(e,i);
-        if (jl_is_expr(a) && has_intrinsics((jl_expr_t*)a))
+        if (jl_is_expr(a) && jl_has_intrinsics((jl_expr_t*)a))
             return 1;
     }
     return 0;
@@ -191,7 +193,7 @@ int jl_eval_with_compiler_p(jl_expr_t *expr, int compileloops)
             }
         }
     }
-    if (has_intrinsics(expr)) return 1;
+    if (jl_has_intrinsics(expr)) return 1;
     return 0;
 }
 
@@ -201,37 +203,66 @@ static jl_value_t *require_func=NULL;
 
 static jl_module_t *eval_import_path_(jl_array_t *args, int retrying)
 {
-    // in A.B.C, first find a binding for A in the chain of module scopes
+    // in .A.B.C, first find a binding for A in the chain of module scopes
     // following parent links. then evaluate the rest of the path from there.
+    // in A.B, look for A in Main first.
     jl_sym_t *var = (jl_sym_t*)jl_cellref(args,0);
+    size_t i=1;
     assert(jl_is_symbol(var));
-    jl_module_t *m = jl_current_module;
+    jl_module_t *m;
+
+    if (var != dot_sym) {
+        m = jl_main_module;
+    }
+    else {
+        m = jl_current_module;
+        while (1) {
+            var = (jl_sym_t*)jl_cellref(args,i);
+            i++;
+            if (var != dot_sym) {
+                if (i == jl_array_len(args))
+                    return m;
+                else
+                    break;
+            }
+            m = m->parent;
+        }
+    }
+
     while (1) {
-        jl_binding_t *mb = jl_get_binding(m, var);
-        if (mb != NULL) {
-            if (mb->value == NULL || !jl_is_module(mb->value))
-                jl_errorf("invalid module path");
-            m = (jl_module_t*)mb->value;
-            break;
+        if (jl_binding_resolved_p(m, var)) {
+            jl_binding_t *mb = jl_get_binding(m, var);
+            assert(mb != NULL);
+            if (mb->owner == m || mb->imported) {
+                m = (jl_module_t*)mb->value;
+                if (m == NULL || !jl_is_module(m))
+                    jl_errorf("invalid module path (%s does not name a module)", var->name);
+                break;
+            }
         }
         if (m == jl_main_module) {
-            if (!retrying) {
+            if (!retrying && i==1) { // (i==1) => no require() for relative imports
                 if (require_func == NULL && jl_base_module != NULL)
                     require_func = jl_get_global(jl_base_module, jl_symbol("require"));
                 if (require_func != NULL) {
                     jl_value_t *str = jl_cstr_to_string(var->name);
-                    JL_GC_PUSH(&str);
+                    JL_GC_PUSH1(&str);
                     jl_apply((jl_function_t*)require_func, &str, 1);
                     JL_GC_POP();
                     return eval_import_path_(args, 1);
                 }
             }
+        }
+        if (retrying && require_func) {
+            JL_PRINTF(JL_STDERR, "Warning: requiring \"%s\" did not define a corresponding module.\n", var->name);
+            return NULL;
+        }
+        else {
             jl_errorf("in module path: %s not defined", var->name);
         }
-        m = m->parent;
     }
 
-    for(size_t i=1; i < jl_array_len(args)-1; i++) {
+    for(; i < jl_array_len(args)-1; i++) {
         jl_value_t *s = jl_cellref(args,i);
         assert(jl_is_symbol(s));
         m = (jl_module_t*)jl_eval_global_var(m, (jl_sym_t*)s);
@@ -246,6 +277,19 @@ static jl_module_t *eval_import_path(jl_array_t *args)
     return eval_import_path_(args, 0);
 }
 
+jl_value_t *jl_toplevel_eval_body(jl_array_t *stmts);
+
+int jl_is_toplevel_only_expr(jl_value_t *e)
+{
+    return jl_is_expr(e) &&
+        (((jl_expr_t*)e)->head == module_sym ||
+         ((jl_expr_t*)e)->head == importall_sym ||
+         ((jl_expr_t*)e)->head == import_sym ||
+         ((jl_expr_t*)e)->head == using_sym ||
+         ((jl_expr_t*)e)->head == export_sym ||
+         ((jl_expr_t*)e)->head == toplevel_sym);
+}
+
 jl_value_t *jl_toplevel_eval_flex(jl_value_t *e, int fast)
 {
     //jl_show(ex);
@@ -264,34 +308,51 @@ jl_value_t *jl_toplevel_eval_flex(jl_value_t *e, int fast)
     }
 
     // handle import, using, importall, export toplevel-only forms
-    if (ex->head == using_sym || ex->head == importall_sym) {
+    if (ex->head == importall_sym) {
         jl_module_t *m = eval_import_path(ex->args);
+        if (m==NULL) return jl_nothing;
         jl_sym_t *name = (jl_sym_t*)jl_cellref(ex->args, jl_array_len(ex->args)-1);
-        assert(jl_is_symbol(name));
+        if (!jl_is_symbol(name))
+            jl_error("syntax: malformed \"importall\" statement");
         m = (jl_module_t*)jl_eval_global_var(m, name);
         if (!jl_is_module(m))
 	    jl_errorf("invalid %s statement: name exists but does not refer to a module", ex->head->name);
-	if (ex->head == using_sym) {
-	    jl_module_using(jl_current_module, m);
-	}
-	else {
-            jl_module_importall(jl_current_module, m);
+        jl_module_importall(jl_current_module, m);
+        return jl_nothing;
+    }
+
+    if (ex->head == using_sym) {
+        jl_module_t *m = eval_import_path(ex->args);
+        if (m==NULL) return jl_nothing;
+        jl_sym_t *name = (jl_sym_t*)jl_cellref(ex->args, jl_array_len(ex->args)-1);
+        if (!jl_is_symbol(name))
+            jl_error("syntax: malformed \"using\" statement");
+        jl_module_t *u = (jl_module_t*)jl_eval_global_var(m, name);
+        if (jl_is_module(u)) {
+            jl_module_using(jl_current_module, u);
+        }
+        else {
+            jl_module_use(jl_current_module, m, name);
         }
         return jl_nothing;
     }
 
     if (ex->head == import_sym) {
         jl_module_t *m = eval_import_path(ex->args);
+        if (m==NULL) return jl_nothing;
         jl_sym_t *name = (jl_sym_t*)jl_cellref(ex->args, jl_array_len(ex->args)-1);
-        assert(jl_is_symbol(name));
+        if (!jl_is_symbol(name))
+            jl_error("syntax: malformed \"import\" statement");
         jl_module_import(jl_current_module, m, name);
         return jl_nothing;
     }
 
     if (ex->head == export_sym) {
         for(size_t i=0; i < jl_array_len(ex->args); i++) {
-            jl_module_export(jl_current_module,
-                             (jl_sym_t*)jl_cellref(ex->args, i));
+            jl_sym_t *name = (jl_sym_t*)jl_cellref(ex->args, i);
+            if (!jl_is_symbol(name))
+                jl_error("syntax: malformed \"export\" statement");
+            jl_module_export(jl_current_module, name);
         }
         return jl_nothing;
     }
@@ -308,7 +369,7 @@ jl_value_t *jl_toplevel_eval_flex(jl_value_t *e, int fast)
     jl_value_t *result;
     jl_lambda_info_t *thk=NULL;
     int ewc = 0;
-    JL_GC_PUSH(&thunk, &thk, &ex);
+    JL_GC_PUSH3(&thunk, &thk, &ex);
 
     if (ex->head != body_sym && ex->head != thunk_sym) {
         // not yet expanded
@@ -320,6 +381,7 @@ jl_value_t *jl_toplevel_eval_flex(jl_value_t *e, int fast)
         for(i=0; i < jl_array_len(ex->args); i++) {
             res = jl_toplevel_eval_flex(jl_cellref(ex->args, i), fast);
         }
+        JL_GC_POP();
         return res;
     }
 
@@ -328,14 +390,9 @@ jl_value_t *jl_toplevel_eval_flex(jl_value_t *e, int fast)
         assert(jl_is_lambda_info(thk));
         ewc = jl_eval_with_compiler_p(jl_lam_body((jl_expr_t*)thk->ast), fast);
         if (!ewc) {
-            jl_array_t *vinfos = jl_lam_vinfo((jl_expr_t*)thk->ast);
-            int i;
-            for(i=0; i < jl_array_len(vinfos); i++) {
-                if (jl_vinfo_capt((jl_array_t*)jl_cellref(vinfos,i))) {
-                    // interpreter doesn't handle closure environment
-                    ewc = 1;
-                    break;
-                }
+            if (jl_lam_vars_captured((jl_expr_t*)thk->ast)) {
+                // interpreter doesn't handle closure environment
+                ewc = 1;
             }
         }
     }
@@ -345,7 +402,15 @@ jl_value_t *jl_toplevel_eval_flex(jl_value_t *e, int fast)
             ewc = 1;
         }
         else {
-            result = jl_interpret_toplevel_expr((jl_value_t*)ex);
+            if (ex->head == body_sym) {
+                result = jl_toplevel_eval_body(ex->args);
+            }
+            else if (jl_is_toplevel_only_expr((jl_value_t*)ex)) {
+                result = jl_toplevel_eval((jl_value_t*)ex);
+            }
+            else {
+                result = jl_interpret_toplevel_expr((jl_value_t*)ex);
+            }
             JL_GC_POP();
             return result;
         }
@@ -371,13 +436,13 @@ jl_value_t *jl_toplevel_eval(jl_value_t *v)
 }
 
 // repeatedly call jl_parse_next and eval everything
-void jl_parse_eval_all(char *fname)
+jl_value_t *jl_parse_eval_all(char *fname)
 {
     //jl_printf(JL_STDERR, "***** loading %s\n", fname);
     int last_lineno = jl_lineno;
     jl_lineno=0;
-    jl_value_t *fn=NULL, *ln=NULL, *form=NULL;
-    JL_GC_PUSH(&fn, &ln, &form);
+    jl_value_t *fn=NULL, *ln=NULL, *form=NULL, *result=jl_nothing;
+    JL_GC_PUSH4(&fn, &ln, &form, &result);
     JL_TRY {
         // handle syntax error
         while (1) {
@@ -392,7 +457,7 @@ void jl_parse_eval_all(char *fname)
                     jl_interpret_toplevel_expr(form);
                 }
             }
-            (void)jl_toplevel_eval_flex(form, 1);
+            result = jl_toplevel_eval_flex(form, 1);
         }
     }
     JL_CATCH {
@@ -406,38 +471,42 @@ void jl_parse_eval_all(char *fname)
     jl_stop_parsing();
     jl_lineno = last_lineno;
     JL_GC_POP();
+    return result;
 }
 
 int asprintf(char **strp, const char *fmt, ...);
 
-void jl_load(const char *fname)
+jl_value_t *jl_load(const char *fname)
 {
     if (jl_current_module == jl_base_module) {
         //This deliberatly uses ios, because stdio initialization has been moved to Julia
         jl_printf(JL_STDOUT, "%s\n", fname);
     }
     char *fpath = (char*)fname;
-    uv_statbuf_t stbuf;
+    uv_stat_t stbuf;
     if (jl_stat(fpath, (char*)&stbuf) != 0 || (stbuf.st_mode & S_IFMT) != S_IFREG) {
         jl_errorf("could not open file %s", fpath);
     }
-    jl_start_parsing_file(fpath);
-    jl_parse_eval_all(fpath);
+    if (jl_start_parsing_file(fpath) != 0) {
+        jl_errorf("could not open file %s", fpath);
+    }
+    jl_value_t *result = jl_parse_eval_all(fpath);
     if (fpath != fname) free(fpath);
     if (jl_current_module == jl_base_module) {
-        jl_printf(JL_STDOUT, "\e[1F\e[2K");
+        jl_printf(JL_STDOUT, "\x1B[1F\x1B[2K");
     }
+    return result;
 }
 
 // load from filename given as a ByteString object
-DLLEXPORT void jl_load_(jl_value_t *str)
+DLLEXPORT jl_value_t *jl_load_(jl_value_t *str)
 {
-    jl_load(jl_string_data(str));
+    return jl_load(jl_string_data(str));
 }
 
 // type definition ------------------------------------------------------------
 
-void jl_reinstantiate_inner_types(jl_tag_type_t *t);
+void jl_reinstantiate_inner_types(jl_datatype_t *t);
 
 void jl_check_type_tuple(jl_tuple_t *t, jl_sym_t *name, const char *ctx)
 {
@@ -449,27 +518,52 @@ void jl_check_type_tuple(jl_tuple_t *t, jl_sym_t *name, const char *ctx)
     }
 }
 
-void jl_set_tag_type_super(jl_tag_type_t *tt, jl_value_t *super)
+void jl_set_datatype_super(jl_datatype_t *tt, jl_value_t *super)
 {
-    if (!jl_is_tag_type(super) || super == (jl_value_t*)jl_undef_type ||
+    if (!jl_is_datatype(super) || super == (jl_value_t*)jl_undef_type ||
+        !jl_is_abstracttype(super) ||
+        jl_subtype(super,(jl_value_t*)jl_vararg_type,0) ||
         jl_subtype(super,(jl_value_t*)jl_type_type,0)) {
         jl_errorf("invalid subtyping in definition of %s",tt->name->name->name);
     }
-    tt->super = (jl_tag_type_t*)super;
+    tt->super = (jl_datatype_t*)super;
     if (jl_tuple_len(tt->parameters) > 0) {
         tt->name->cache = (jl_value_t*)jl_null;
-        jl_reinstantiate_inner_types((jl_tag_type_t*)tt);
+        jl_reinstantiate_inner_types(tt);
     }
 }
 
 // method definition ----------------------------------------------------------
 
 extern int jl_boot_file_loaded;
-void jl_add_constructors(jl_struct_type_t *t);
+void jl_add_constructors(jl_datatype_t *t);
+
+static int type_contains(jl_value_t *ty, jl_value_t *x)
+{
+    if (ty == x) return 1;
+    if (jl_is_tuple(ty)) {
+        size_t i, l=jl_tuple_len(ty);
+        for(i=0; i < l; i++) {
+            jl_value_t *e = jl_tupleref(ty,i);
+            if (e==x || type_contains(e, x))
+                return 1;
+        }
+    }
+    if (jl_is_uniontype(ty))
+        return type_contains(jl_fieldref(ty,0), x);
+    if (jl_is_datatype(ty))
+        return type_contains((jl_value_t*)((jl_datatype_t*)ty)->parameters, x);
+    return 0;
+}
+
+void print_func_loc(JL_STREAM *s, jl_lambda_info_t *li);
 
 jl_value_t *jl_method_def(jl_sym_t *name, jl_value_t **bp, jl_binding_t *bnd,
-                          jl_tuple_t *argtypes, jl_function_t *f, jl_tuple_t *t)
+                          jl_tuple_t *argtypes, jl_function_t *f)
 {
+    // argtypes is a tuple ((types...), (typevars...))
+    jl_tuple_t *t = (jl_tuple_t*)jl_t1(argtypes);
+    argtypes = (jl_tuple_t*)jl_t0(argtypes);
     jl_value_t *gf;
     if (bnd) {
         //jl_declare_constant(bnd);
@@ -486,16 +580,16 @@ jl_value_t *jl_method_def(jl_sym_t *name, jl_value_t **bp, jl_binding_t *bnd,
     else {
         gf = *bp;
         if (!jl_is_gf(gf)) {
-            if (jl_is_struct_type(gf) &&
+            if (jl_is_datatype(gf) &&
                 ((jl_function_t*)gf)->fptr == jl_f_ctor_trampoline) {
-                jl_add_constructors((jl_struct_type_t*)gf);
+                jl_add_constructors((jl_datatype_t*)gf);
             }
             if (!jl_is_gf(gf)) {
                 jl_error("invalid method definition: not a generic function");
             }
         }
     }
-    JL_GC_PUSH(&gf);
+    JL_GC_PUSH1(&gf);
     assert(jl_is_function(f));
     assert(jl_is_tuple(argtypes));
     assert(jl_is_tuple(t));
@@ -505,15 +599,24 @@ jl_value_t *jl_method_def(jl_sym_t *name, jl_value_t **bp, jl_binding_t *bnd,
         if (!jl_is_type(elt) && !jl_is_typevar(elt)) {
             jl_lambda_info_t *li = f->linfo;
             jl_errorf("invalid type for argument %s in method definition for %s at %s:%d",
-                      ((jl_sym_t*)jl_arrayref(jl_lam_args((jl_expr_t*)li->ast),i))->name,
+                      jl_is_expr(li->ast) ?
+                      ((jl_sym_t*)jl_arrayref(jl_lam_args((jl_expr_t*)li->ast),i))->name :
+                      "?",
                       name->name, li->file->name, li->line);
         }
     }
 
+    int ishidden = !!strchr(name->name, '#');
     for(size_t i=0; i < jl_tuple_len(t); i++) {
-        if (!jl_is_typevar(jl_tupleref(t,i)))
-            jl_type_error_rt(name->name, "method definition",
-                             (jl_value_t*)jl_tvar_type, jl_tupleref(t,i));
+        jl_value_t *tv = jl_tupleref(t,i);
+        if (!jl_is_typevar(tv))
+            jl_type_error_rt(name->name, "method definition", (jl_value_t*)jl_tvar_type, tv);
+        if (!ishidden && !type_contains((jl_value_t*)argtypes, tv)) {
+            JL_PRINTF(JL_STDERR, "Warning: static parameter %s does not occur in signature for %s",
+                      ((jl_tvar_t*)tv)->name->name, name->name);
+            print_func_loc(JL_STDERR, f->linfo);
+            JL_PRINTF(JL_STDERR, ".\nThe method will not be callable.\n");
+        }
     }
     jl_add_method((jl_function_t*)gf, argtypes, f, t);
     if (jl_boot_file_loaded &&
