@@ -1,26 +1,25 @@
-macro check_bit_operation(func, RetT, args, opts...)
-    quote
-        if length($opts) > 0 && $opts[1] == :skiptc
-            tc(r1,r2) = true
-        else
-            function tc(r1, r2)
-                if isa(r1, Tuple) && isa(r2, Tuple) && length(r1) == length(r2)
-                    return all(map(x->tc(x...), [zip(r1,r2)...]))
-                elseif isa(r1,BitArray)
-                    return isa(r2, Union(BitArray,Array{Bool}))
-                else
-                    return typeof(r1) == typeof(r2)
-                end
-            end
-        end
-        r1 = ($func)($args...)
-        r2 = ($func)(map(x->(isa(x, BitArray) ? bitunpack(x) : x), $args)...)
-        @test isa(r1, $RetT)
-        @test tc(r1, r2)
-        @test isequal(r1, convert($RetT, r2))
+function tc(r1, r2)
+    if isa(r1, Tuple) && isa(r2, Tuple) && length(r1) == length(r2)
+        return all(map(x->tc(x...), [zip(r1,r2)...]))
+    elseif isa(r1,BitArray)
+        return isa(r2, Union(BitArray,Array{Bool}))
+    else
+        return typeof(r1) == typeof(r2)
     end
 end
 
+function check_bitop(func, RetT, args)
+    r1 = func(args...)
+    r2 = func(map(x->(isa(x, BitArray) ? bitunpack(x) : x), args)...)
+    @test isa(r1, RetT)
+    @test tc(r1, r2)
+    @test isequal(r1, convert(RetT, r2))
+end
+
+macro check_bit_operation(func, RetT, args)
+    :(check_bitop($func, $RetT, $args))
+end
+
 let t0 = time()
     global timesofar
     function timesofar(str)
@@ -46,7 +45,7 @@ s4 = 4
 
 b1 = randbool(n1, n2)
 @test isequal(bitpack(bitunpack(b1)), b1)
-i1 = rand(1:2, n1, n2) - 1
+i1 = rand(false:true, n1, n2)
 @test isequal(bitunpack(bitpack(i1)), i1)
 
 timesofar("conversions")
@@ -72,17 +71,17 @@ m1 = rand(1:n1)
 m2 = rand(1:n2)
 b2 = randbool(m1, m2)
 @check_bit_operation copy! BitMatrix (b1, b2)
-@check_bit_operation ref BitMatrix (b1, 1:m1, m2:n2)
-@check_bit_operation ref BitVector (b1, 1:m1, m2)
-@check_bit_operation ref BitMatrix (b1, 1:m1, [n2,m2,1])
+@check_bit_operation getindex BitMatrix (b1, 1:m1, m2:n2)
+@check_bit_operation getindex BitVector (b1, 1:m1, m2)
+@check_bit_operation getindex BitMatrix (b1, 1:m1, [n2,m2,1])
 b2 = randbool(m1, m2)
-@check_bit_operation assign BitMatrix (b1, b2, 1:m1, n2-m2+1:n2)
+@check_bit_operation setindex! BitMatrix (b1, b2, 1:m1, n2-m2+1:n2)
 k1 = randperm(m1)
 k2 = randperm(m2)
-@check_bit_operation assign BitMatrix (b1, b2, 1:m1, k2)
-@check_bit_operation assign BitMatrix (b1, b2, k1, k2)
+@check_bit_operation setindex! BitMatrix (b1, b2, 1:m1, k2)
+@check_bit_operation setindex! BitMatrix (b1, b2, k1, k2)
 b2 = randbool(m1)
-@check_bit_operation assign BitMatrix (b1, b2, 1:m1, m2)
+@check_bit_operation setindex! BitMatrix (b1, b2, 1:m1, m2)
 
 for p1 = [rand(1:v1) 1 63 64 65 191 192 193]
     for p2 = [rand(1:v1) 1 63 64 65 191 192 193]
@@ -108,7 +107,7 @@ timesofar("indexing")
 
 ## Dequeue functionality ##
 
-b1 = BitArray()
+b1 = BitArray(0)
 i1 = Bool[]
 for m = 1 : v1
     x = randbool()
@@ -127,6 +126,16 @@ for m1 = 0 : v1
     end
 end
 
+for m1 = 0 : v1
+    for m2 = [0, 1, 63, 64, 65, 127, 128, 129]
+        b1 = randbool(m1)
+        b2 = randbool(m2)
+        i1 = bitunpack(b1)
+        i2 = bitunpack(b2)
+        @test isequal(bitunpack(prepend!(b1, b2)), prepend!(i1, i2))
+    end
+end
+
 b1 = randbool(v1)
 i1 = bitunpack(b1)
 for m = 1 : v1
@@ -138,7 +147,7 @@ end
 @test length(b1) == 0
 
 
-b1 = BitArray()
+b1 = BitArray(0)
 i1 = Bool[]
 for m = 1 : v1
     x = randbool()
@@ -158,7 +167,7 @@ for m = 1 : v1
 end
 @test length(b1) == 0
 
-b1 = BitArray()
+b1 = BitArray(0)
 i1 = bitunpack(b1)
 for m = 1 : v1
     j = rand(1:m)
@@ -181,8 +190,8 @@ b1 = randbool(v1)
 i1 = bitunpack(b1)
 for m = v1 : -1 : 1
     j = rand(1:m)
-    delete!(b1, j)
-    delete!(i1, j)
+    splice!(b1, j)
+    splice!(i1, j)
     @test isequal(bitunpack(b1), i1)
 end
 @test length(b1) == 0
@@ -190,8 +199,8 @@ end
 b1 = randbool(v1)
 i1 = bitunpack(b1)
 for j in [63, 64, 65, 127, 128, 129, 191, 192, 193]
-    delete!(b1, j)
-    delete!(i1, j)
+    splice!(b1, j)
+    splice!(i1, j)
     @test isequal(bitunpack(b1), i1)
 end
 
@@ -201,12 +210,28 @@ for m1 = 1 : v1
     for m2 = m1 : v1
         b2 = copy(b1)
         i2 = copy(i1)
-        delete!(b2, m1:m2)
-        delete!(i2, m1:m2)
+        splice!(b2, m1:m2)
+        splice!(i2, m1:m2)
         @test isequal(bitunpack(b2), i2)
     end
 end
 
+b1 = randbool(v1)
+i1 = bitunpack(b1)
+for m1 = 1 : v1 + 1
+    for m2 = m1 - 1 : v1
+        for v2 = [0, 1, 63, 64, 65, 127, 128, 129, 191, 192, 193, rand(1:v1)]
+            b2 = copy(b1)
+            i2 = copy(i1)
+            b3 = randbool(v2)
+            i3 = bitunpack(b3)
+            splice!(b2, m1:m2, b3)
+            splice!(i2, m1:m2, i3)
+            @test isequal(bitunpack(b2), i2)
+        end
+    end
+end
+
 b1 = randbool(v1)
 i1 = bitunpack(b1)
 empty!(b1)
@@ -246,7 +271,7 @@ b2 = randbool(n1, n2)
 @check_bit_operation (+) Matrix{Int} (b1, b2)
 @check_bit_operation (-) Matrix{Int} (b1, b2)
 @check_bit_operation (.*) BitMatrix (b1, b2)
-@check_bit_operation (./) Matrix{Float64} (b1, b2)
+@check_bit_operation (./) Matrix{Float32} (b1, b2)
 @check_bit_operation (.^) BitMatrix (b1, b2)
 
 b2 = trues(n1, n2)
@@ -262,7 +287,7 @@ while true
 end
 b2 = randbool(n1, n1)
 
-@check_bit_operation (*) BitMatrix (b1, b2)
+@check_bit_operation (*) Matrix{Int} (b1, b2)
 @check_bit_operation (/) Matrix{Float64} (b1, b1)
 @check_bit_operation (\) Matrix{Float64} (b1, b1)
 
@@ -271,7 +296,7 @@ b0 = falses(0)
 @check_bit_operation (|) BitVector (b0, b0)
 @check_bit_operation ($) BitVector (b0, b0)
 @check_bit_operation (.*) BitVector (b0, b0)
-@check_bit_operation (*) BitMatrix (b0, b0')
+@check_bit_operation (*) Matrix{Int} (b0, b0')
 
 # Matrix{Bool}/Matrix{Int}
 b1 = randbool(n1, n2)
@@ -320,8 +345,8 @@ cf1 = complex(f1)
 @check_bit_operation (.*) Matrix{Uint8} (u1, b2)
 
 for (x1,t1) = {(f1, Float64),
-               (ci1, ComplexPair{Int}),
-               (cu1, ComplexPair{Uint8}),
+               (ci1, Complex{Int}),
+               (cu1, Complex{Uint8}),
                (cf1, Complex128)}
     @check_bit_operation (+) Matrix{t1} (x1, b2)
     @check_bit_operation (-) Matrix{t1} (x1, b2)
@@ -329,10 +354,10 @@ for (x1,t1) = {(f1, Float64),
 end
 
 b2 = trues(n1, n2)
-@check_bit_operation (./) Matrix{Float64} (true, b2)
+@check_bit_operation (./) Matrix{Float32} (true, b2)
 @check_bit_operation div BitMatrix (true, b2)
 @check_bit_operation mod BitMatrix (true, b2)
-@check_bit_operation (./) Matrix{Float64} (false, b2)
+@check_bit_operation (./) Matrix{Float32} (false, b2)
 @check_bit_operation div BitMatrix (false, b2)
 @check_bit_operation mod BitMatrix (false, b2)
 
@@ -340,7 +365,7 @@ b2 = trues(n1, n2)
 @check_bit_operation div Matrix{Int} (i1, b2)
 @check_bit_operation mod Matrix{Int} (i1, b2)
 
-@check_bit_operation (./) Matrix{Float64} (u1, b2)
+@check_bit_operation (./) Matrix{Float32} (u1, b2)
 @check_bit_operation div Matrix{Uint8} (u1, b2)
 @check_bit_operation mod Matrix{Uint8} (u1, b2)
 
@@ -348,15 +373,15 @@ b2 = trues(n1, n2)
 @check_bit_operation div Matrix{Float64} (f1, b2)
 @check_bit_operation mod Matrix{Float64} (f1, b2)
 
-for x1 = {ci1, cu1, cf1}
-    @check_bit_operation (./) Matrix{Complex128} (x1, b2)
-end
+@check_bit_operation (./) Matrix{Complex128} (ci1, b2)
+@check_bit_operation (./) Matrix{Complex64 } (cu1, b2)
+@check_bit_operation (./) Matrix{Complex128} (cf1, b2)
 
 b2 = randbool(n1, n2)
 @check_bit_operation (.^) BitMatrix (false, b2)
 @check_bit_operation (.^) BitMatrix (true, b2)
-@check_bit_operation (.^) Matrix{Uint8} (0x0, b2) skiptc
-@check_bit_operation (.^) Matrix{Uint8} (0x1, b2) skiptc
+@check_bit_operation (.^) Matrix{Uint8} (0x0, b2)
+@check_bit_operation (.^) Matrix{Uint8} (0x1, b2)
 @check_bit_operation (.^) Matrix{Int} (-1, b2)
 @check_bit_operation (.^) Matrix{Int} (0, b2)
 @check_bit_operation (.^) Matrix{Int} (1, b2)
@@ -364,10 +389,10 @@ b2 = randbool(n1, n2)
 @check_bit_operation (.^) Matrix{Float64} (1.0, b2)
 @check_bit_operation (.^) Matrix{Complex128} (0.0im, b2)
 @check_bit_operation (.^) Matrix{Complex128} (1.0im, b2)
-@check_bit_operation (.^) Matrix{ComplexPair{Int}} (0im, b2)
-@check_bit_operation (.^) Matrix{ComplexPair{Int}} (1im, b2)
-@check_bit_operation (.^) Matrix{ComplexPair{Uint8}} (0x0im, b2) skiptc
-@check_bit_operation (.^) Matrix{ComplexPair{Uint8}} (0x1im, b2) skiptc
+@check_bit_operation (.^) Matrix{Complex{Int}} (0im, b2)
+@check_bit_operation (.^) Matrix{Complex{Int}} (1im, b2)
+@check_bit_operation (.^) Matrix{Complex{Uint8}} (0x0im, b2)
+@check_bit_operation (.^) Matrix{Complex{Uint8}} (0x1im, b2)
 
 # Matrix/Number
 b1 = randbool(n1, n2)
@@ -390,8 +415,8 @@ cf2 = complex(f2)
 @check_bit_operation (-) Matrix{Int} (b1, false)
 @check_bit_operation (.*) BitMatrix (b1, true)
 @check_bit_operation (.*) BitMatrix (b1, false)
-@check_bit_operation (./) Matrix{Float64} (b1, true)
-@check_bit_operation (./) Matrix{Float64} (b1, false)
+@check_bit_operation (./) Matrix{Float32} (b1, true)
+@check_bit_operation (./) Matrix{Float32} (b1, false)
 @check_bit_operation div BitMatrix (b1, true)
 @check_bit_operation mod BitMatrix (b1, true)
 
@@ -411,7 +436,7 @@ cf2 = complex(f2)
 @check_bit_operation (+) Matrix{Uint8} (b1, u2)
 @check_bit_operation (-) Matrix{Uint8} (b1, u2)
 @check_bit_operation (.*) Matrix{Uint8} (b1, u2)
-@check_bit_operation (./) Matrix{Float64} (b1, u2)
+@check_bit_operation (./) Matrix{Float32} (b1, u2)
 @check_bit_operation div Matrix{Uint8} (b1, u2)
 @check_bit_operation mod Matrix{Uint8} (b1, u2)
 
@@ -422,15 +447,15 @@ cf2 = complex(f2)
 @check_bit_operation div Matrix{Float64} (b1, f2)
 @check_bit_operation mod Matrix{Float64} (b1, f2)
 
-@check_bit_operation (+) Matrix{ComplexPair{Int}} (b1, ci2)
-@check_bit_operation (-) Matrix{ComplexPair{Int}} (b1, ci2)
-@check_bit_operation (.*) Matrix{ComplexPair{Int}} (b1, ci2)
+@check_bit_operation (+) Matrix{Complex{Int}} (b1, ci2)
+@check_bit_operation (-) Matrix{Complex{Int}} (b1, ci2)
+@check_bit_operation (.*) Matrix{Complex{Int}} (b1, ci2)
 @check_bit_operation (./) Matrix{Complex128} (b1, ci2)
 
-@check_bit_operation (+) Matrix{ComplexPair{Uint8}} (b1, cu2)
-@check_bit_operation (-) Matrix{ComplexPair{Uint8}} (b1, cu2)
-@check_bit_operation (.*) Matrix{ComplexPair{Uint8}} (b1, cu2)
-@check_bit_operation (./) Matrix{Complex128} (b1, cu2)
+@check_bit_operation (+) Matrix{Complex{Uint8}} (b1, cu2)
+@check_bit_operation (-) Matrix{Complex{Uint8}} (b1, cu2)
+@check_bit_operation (.*) Matrix{Complex{Uint8}} (b1, cu2)
+@check_bit_operation (./) Matrix{Complex64} (b1, cu2)
 
 @check_bit_operation (+) Matrix{Complex128} (b1, cf2)
 @check_bit_operation (-) Matrix{Complex128} (b1, cf2)
@@ -447,15 +472,15 @@ cf2 = complex(f2)
 @check_bit_operation (.^) Matrix{Float64} (b1, 0.0)
 @check_bit_operation (.^) Matrix{Float64} (b1, 1.0)
 @check_bit_operation (.^) Matrix{Complex128} (b1, 0.0im)
-@check_bit_operation (.^) Matrix{ComplexPair{Uint8}} (b1, 0x0im) skiptc
-@check_bit_operation (.^) Matrix{ComplexPair{Int}} (b1, 0im) skiptc
+@check_bit_operation (.^) Matrix{Complex64 } (b1, 0x0im)
+@check_bit_operation (.^) Matrix{Complex128} (b1, 0im)
 
 b1 = trues(n1, n2)
 @check_bit_operation (.^) Matrix{Complex128} (b1, -1.0im)
 @check_bit_operation (.^) Matrix{Complex128} (b1, 1.0im)
-@check_bit_operation (.^) Matrix{Complex128} (b1, -1im) skiptc
-@check_bit_operation (.^) Matrix{Complex128} (b1, 1im) skiptc
-@check_bit_operation (.^) Matrix{Complex128} (b1, 0x1im) skiptc
+@check_bit_operation (.^) Matrix{Complex128} (b1, -1im)
+@check_bit_operation (.^) Matrix{Complex128} (b1, 1im)
+@check_bit_operation (.^) Matrix{Complex64 } (b1, 0x1im)
 
 timesofar("binary arithmetic")
 
@@ -536,7 +561,7 @@ for i = 3:v1-1
 end
 
 b1 = randbool(n1, n2)
-@check_bit_operation findn_nzs (Vector{Int}, Vector{Int}, BitArray) (b1,)
+@check_bit_operation findnz (Vector{Int}, Vector{Int}, BitArray) (b1,)
 
 timesofar("nnz&find")
 
@@ -545,19 +570,17 @@ timesofar("nnz&find")
 b1 = randbool(s1, s2, s3, s4)
 m1 = 1
 m2 = 3
-@check_bit_operation max BitArray{4} (b1, (), (m1, m2))
-@check_bit_operation min BitArray{4} (b1, (), (m1, m2))
+@check_bit_operation maximum BitArray{4} (b1, (m1, m2))
+@check_bit_operation minimum BitArray{4} (b1, (m1, m2))
 @check_bit_operation sum Array{Int,4} (b1, (m1, m2))
 
-@check_bit_operation max Bool (b1,)
-@check_bit_operation min Bool (b1,)
+@check_bit_operation maximum Bool (b1,)
+@check_bit_operation minimum Bool (b1,)
 @check_bit_operation any Bool (b1,)
 @check_bit_operation all Bool (b1,)
 @check_bit_operation sum Int (b1,)
 
 b0 = falses(0)
-@check_bit_operation max Bool (b0,)
-@check_bit_operation min Bool (b0,)
 @check_bit_operation any Bool (b0,)
 @check_bit_operation all Bool (b0,)
 @check_bit_operation sum Int (b0,)
@@ -643,6 +666,42 @@ for k = -max(n1,n2) : max(n1,n2)
     @check_bit_operation triu BitMatrix (b1, k)
 end
 
+b1 = randbool(n1, n1)
+@check_bit_operation istril Bool (b1,)
+b1 = randbool(n1, n2)
+@check_bit_operation istril Bool (b1,)
+b1 = randbool(n2, n1)
+@check_bit_operation istril Bool (b1,)
+
+b1 = tril(randbool(n1, n1))
+@check_bit_operation istril Bool (b1,)
+b1 = tril(randbool(n1, n2))
+@check_bit_operation istril Bool (b1,)
+b1 = tril(randbool(n2, n1))
+@check_bit_operation istril Bool (b1,)
+
+b1 = randbool(n1, n1)
+@check_bit_operation istriu Bool (b1,)
+b1 = randbool(n1, n2)
+@check_bit_operation istriu Bool (b1,)
+b1 = randbool(n2, n1)
+@check_bit_operation istriu Bool (b1,)
+
+b1 = triu(randbool(n1, n1))
+@check_bit_operation istriu Bool (b1,)
+b1 = triu(randbool(n1, n2))
+@check_bit_operation istriu Bool (b1,)
+b1 = triu(randbool(n2, n1))
+@check_bit_operation istriu Bool (b1,)
+
+b1 = randbool(n1)
+b2 = randbool(n2)
+@check_bit_operation kron BitVector (b1, b2)
+
+b1 = randbool(s1, s2)
+b2 = randbool(s3, s4)
+@check_bit_operation kron BitMatrix (b1, b2)
+
 #b1 = randbool(v1)
 #@check_bit_operation diff Vector{Int} (b1,)
 #b1 = randbool(n1, n2)
