@@ -1,3 +1,3 @@
 ccall_test_func(x) = ccall((:testUcharX, "./libccalltest"), Int32, (Uint8,), x)
-@assert ccall_test_func(3) == 1
-@assert ccall_test_func(259) == 1
+@test ccall_test_func(3) == 1
+@test ccall_test_func(259) == 1
