@@ -1,17 +1,23 @@
 # Check that serializer hasn't gone out-of-frame
 @test Base.ser_tag[Symbol] == 2
-@test Base.ser_tag[()] == 49
-@test Base.ser_tag[false] == 125
+@test Base.ser_tag[()] == 47
+@test Base.ser_tag[false] == 123
 
 # issue #1770
 let
     a = ['T', 'e', 's', 't']
-    f = memio()
+    f = IOBuffer()
     serialize(f, a)
     seek(f, 0)
     @test deserialize(f) == a
-    f = IOString()
+    f = IOBuffer()
     serialize(f, a)
     seek(f, 0)
     @test deserialize(f) == a
+
+    # issue #4414
+    seek(f,0)
+    serialize(f, :β)
+    seek(f,0)
+    @test deserialize(f) === :β
 end
